// add local coefficients for a particle with charge q at relative position (x,y,z)
// to the (non-negative triangle) local coefficient matrix upto order p
template <typename Real, typename CoefficientMatrix>
inline __attribute__((flatten)) void P2L(
        const Real x,
        const Real y,
        const Real z,
        const Real q,
        CoefficientMatrix & mu,
        const size_t p)
{
    typedef Real real_type;
    typedef typename CoefficientMatrix::value_type complex_type;
    Vec_traits<Real> VT;

    real_type dist_squared = x * x + y * y + z * z;
    real_type recip_dist = rsqrt(dist_squared);
    real_type recip_dist_squared = reciprocal(dist_squared);
    real_type twice_recip_dist_squared = recip_dist_squared + recip_dist_squared;
    real_type z_recip_dist_squared = z * recip_dist_squared;
    real_type twice_z_recip_dist_squared = z_recip_dist_squared + z_recip_dist_squared;

    complex_type complex_x_y(x, y);
    complex_type complex_x_y_m(VT.same(1));     // iterative computation of ((x + iy)^m)

    // mu_0_0  (for the first iteration)
    real_type mu_mplus1_mplus1 = q * recip_dist;
    real_type e_m = recip_dist_squared;         // iterative computation of ((2*m + 1)/R^2)

    // mu_0_0 upto mu_p_p-1
    for (size_t m = 0; m < p; ++m) {
        // mu_m_m  (from previous iteration)
        real_type mu_m_m = mu_mplus1_mplus1;
        mu(m, m) += complex_x_y_m * mu_m_m;

        // mu_m+1_m+1  (for the next iteration)
        mu_mplus1_mplus1 = e_m * mu_m_m;

        // mu_m+1_m
        real_type mu_mplus1_m = mu_mplus1_mplus1 * z;
        mu(m + 1, m) += complex_x_y_m * mu_mplus1_m;

        // mu_m+2_m upto mu_p_m
        real_type e_mplus1 = e_m + twice_recip_dist_squared;
        real_type mu_lminus2_m = mu_m_m;
        real_type mu_lminus1_m = mu_mplus1_m;
        real_type f_l = e_mplus1 * z;           // iterative computation of ((2*l - 1)*z/R^2)
        real_type h_l = e_m;                    // iterative computation of ((2*l - 3)/R^2)
        real_type g_l = h_l;                    // iterative computation of (((l - 1)^2 - m^2)/R^2)
        for (size_t l = m + 2; l <= p; ++l) {
            // mu_l_m
            real_type mu_l_m = f_l * mu_lminus1_m - g_l * mu_lminus2_m;
            mu(l, m) += complex_x_y_m * mu_l_m;

            // (for the next iteration)
            mu_lminus2_m = mu_lminus1_m;
            mu_lminus1_m = mu_l_m;
            f_l += twice_z_recip_dist_squared;
            h_l += twice_recip_dist_squared;
            g_l += h_l;
        }

        // (for the next iteration)
        complex_x_y_m *= complex_x_y;
        e_m = e_mplus1;
    }

    // mu_p_p
    mu(p, p) += complex_x_y_m * mu_mplus1_mplus1;
}
