/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */
#ifndef _AnBe_FMM_bulk_ops_hpp_
#define _AnBe_FMM_bulk_ops_hpp_

#include <vector>
#include <cstddef>

namespace fmsolvr {

namespace bulk {

template <typename InputCoefficents,
          typename RotationOperator,
          typename ShiftOperator,
          typename OutputCoefficients>
class Bulk4Ptr {
    const InputCoefficents * in_;
    const RotationOperator * R_;
    const ShiftOperator * Op_;
    OutputCoefficients * out_;

public:
    typedef InputCoefficents in_type;
    typedef RotationOperator R_type;
    typedef ShiftOperator Op_type;
    typedef OutputCoefficients out_type;

    template <typename T>
    Bulk4Ptr(const T & x) :
        in_(&x.in()), R_(&x.R()), Op_(&x.Op()), out_(&x.out())
    { }

    const InputCoefficents & in() const
    { return *in_; }

    const RotationOperator & R() const
    { return *R_; }

    const ShiftOperator & Op() const
    { return *Op_; }

    OutputCoefficients & out() const
    { return *out_; }
};

template <typename InputCoefficents,
          typename RotationOperator,
          typename ShiftOperator,
          typename OutputCoefficients>
class BulkRotOp {
    typedef typename RotationOperator::packed_pyramid_type packed_pyramid_type;
    typedef typename RotationOperator::e_ikphi_table_type e_ikphi_table_type;

    const InputCoefficents & in_;
    const packed_pyramid_type & d_f_forward_;
    const packed_pyramid_type & d_f_backward_;
    const packed_pyramid_type & d_g_forward_;
    const packed_pyramid_type & d_g_backward_;
    const e_ikphi_table_type & eikphitable_;
    const ShiftOperator & Op_;
    OutputCoefficients & out_;

public:
    typedef InputCoefficents in_type;
    typedef ShiftOperator Op_type;
    typedef OutputCoefficients out_type;

    template <typename T>
    BulkRotOp(const T & x)
        : in_(x.in()),
          d_f_forward_(x.R().FoldedCoefficientsRotateMultipoleForward()),
          d_f_backward_(x.R().FoldedCoefficientsRotateMultipoleBackward()),
          d_g_forward_(x.R().FoldedCoefficientsRotateLocalForward()),
          d_g_backward_(x.R().FoldedCoefficientsRotateLocalBackward()),
          eikphitable_(x.R().e_ikphi()),
          Op_(x.Op()),
          out_(x.out())
    { }

    const InputCoefficents & in() const
    { return in_; }

    const packed_pyramid_type & FoldedCoefficientsRotateMultipoleForward() const
    { return d_f_forward_; }

    const packed_pyramid_type & FoldedCoefficientsRotateMultipoleBackward() const
    { return d_f_backward_; }

    const packed_pyramid_type & FoldedCoefficientsRotateLocalForward() const
    { return d_g_forward_; }

    const packed_pyramid_type & FoldedCoefficientsRotateLocalBackward() const
    { return d_g_backward_; }

    const e_ikphi_table_type & e_ikphi() const
    { return eikphitable_; }

    const BulkRotOp & R() const
    { return *this; }

    const ShiftOperator & Op() const
    { return Op_; }

    OutputCoefficients & out() const
    { return out_; }
};

template <typename RAIterator, typename Scratch>
void M2Lp3(RAIterator first, RAIterator last, size_t p, Scratch &scratch);

}  // namespace bulk

}  // namespace fmsolvr

//#include "bulk_ops_impl.hpp"

#endif
// vim: et:ts=4:sw=4
