/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <cassert>
#include <initializer_list>
#include <algorithm>

#include "ipowlog.hpp"
#include "MultidimensionalCoordinates.hpp"


template <typename Tp>
Tp wraparound(Tp val, Tp dim)
{
    while (val < 0)
        val += dim;
    while (val >= dim)
        val -= dim;

    return val;
}

#if 0
// little endian: coord[0] = least significant dim
template <unsigned NDims, typename Integral = ssize_t>
struct MultidimensionalCoordinates
{
    typedef Integral integral_type;

    MultidimensionalCoordinates()
    {
        std::fill(coord, coord + ndims(), 0);
    }

    MultidimensionalCoordinates(std::initializer_list<integral_type> il)
    {
        assert(il.size() == ndims());
        std::copy(il.begin(), il.end(), coord);
    }

    static size_t ndims()
    {
        return NDims;
    }

    const integral_type & operator [] (size_t index) const
    {
        return coord[index];
    }

    integral_type & operator [] (size_t index)
    {
        return coord[index];
    }

    // linearize coordinates inside a ndims()-dimensional cube with side length 'length'
    integral_type linearize(integral_type length) const
    {
        integral_type res = 0;
        for (ssize_t i = ndims() - 1; i >= 0; --i)
            res = res * length + coord[i];
        return res;
    }

    // de-linearize coordinates inside a ndims()-dimensional cube with side length 'length'
    static MultidimensionalCoordinates xdelinearize(integral_type lin, integral_type length)
    {
        MultidimensionalCoordinates a;
        for (size_t i = 0; i < ndims(); ++i)
            a[i] = lin % length;
            lin /= length;
        assert(lin == 0);
        return a;
    }

    MultidimensionalCoordinates & operator += (const MultidimensionalCoordinates & a)
    {
        for (size_t i = 0; i < ndims(); ++i)
            coord[i] += a[i];
        return *this;
    }

    MultidimensionalCoordinates & operator -= (const MultidimensionalCoordinates & a)
    {
        for (size_t i = 0; i < ndims(); ++i)
            coord[i] -= a[i];
        return *this;
    }

    MultidimensionalCoordinates operator + (const MultidimensionalCoordinates & a) const
    {
        MultidimensionalCoordinates b(*this);
        b += a;
        return b;
    }

    MultidimensionalCoordinates operator - (const MultidimensionalCoordinates & a) const
    {
        MultidimensionalCoordinates b(*this);
        b -= a;
        return b;
    }

    bool operator == (const MultidimensionalCoordinates & a) const
    {
        return !(*this != a);
    }

    bool operator != (const MultidimensionalCoordinates & a) const
    {
        for (size_t i = 0; i < ndims(); ++i)
            if (coord[i] != a[i])
                return true;
        return false;
    }

private:
    integral_type coord[NDims];
};

template <unsigned NDims, typename Integral>
std::ostream & operator << (std::ostream & os, const MultidimensionalCoordinates<NDims, Integral> & a)
{
    os << "(";
    for (size_t i = 0; i < a.ndims(); ++i) {
        if (i > 0)
            os << ",";
        os << a[i];
    }
    os << ")";
    return os;
}
#endif


template <typename Coords>
Coords get_mother(const Coords & child)
{
    Coords mother;
    for (size_t i = 0; i < mother.ndims(); ++i)
        mother[i] = child[i] >> 1;
    return mother;
}

template <typename Coords>
Coords get_first_child(const Coords & mother)
{
    Coords child;
    for (size_t i = 0; i < child.ndims(); ++i)
        child[i] = mother[i] << 1;
    return child;
}

// first enumerate children of a mother, thereafter move to next mother
// one bit per dimension as child index, other bits linearized mother
template <typename Coords>
typename Coords::integral_type linearize_neighbor(const Coords & neighbor, size_t length)
{
    typename Coords::integral_type mother_length = length >> 1;
    typename Coords::integral_type mother = 0, child = 0;
    for (ssize_t i = neighbor.ndims() - 1; i >= 0; --i) {
        child |= (neighbor[i] & 1) << i;
        mother = mother * mother_length + (neighbor[i] >> 1);
    }
    return (mother << neighbor.ndims()) + child;
}

// one bit per dimension as child index, other bits linearized mother
template <typename Coords>
Coords delinearize_neighbor(typename Coords::integral_type lin_neighbor, size_t length)
{
    typename Coords::integral_type lin_mother = lin_neighbor >> Coords::ndims();
    typename Coords::integral_type mother_length = length >> 1;
    Coords neighbor;
    for (size_t i = 0; i < neighbor.ndims(); ++i) {
        neighbor[i] = ((lin_mother % mother_length) << 1) | ((lin_neighbor >> i) & 1);
        lin_mother /= mother_length;
    }
    return neighbor;
}


template <typename Coords>
void ff_communicate(const Coords & rank, const ssize_t ws, const ssize_t c, const ssize_t depth)
{
    const size_t ndims = rank.ndims();
    const ssize_t log2c = ilog2(c);
    assert(c == (1 << log2c));
    const ssize_t children_per_mother = ipow<ssize_t>(2, ndims);
    assert(c <= children_per_mother);
    const ssize_t dim = 1 << depth;
    const ssize_t my_team = linearize_neighbor(rank, dim) >> log2c;
    const ssize_t neighbor_mother_dim = 2 * ws + 1;
    const ssize_t neighbor_dim = 2 * neighbor_mother_dim;
    assert(neighbor_dim <= dim);  // don't overlap within neighbordim
    const ssize_t neighbors = ipow(neighbor_dim, ssize_t(ndims));
    //const ssize_t neighbor_teams = (neighbors + c - 1) / c;

    const ssize_t rounds = (neighbors + c * c - 1) / (c * c);

    Coords mother = get_mother(rank);
    Coords first_neighbor;
    for (size_t i = 0; i < ndims; ++i)
        first_neighbor[i] = 2 * (mother[i] - ws);
    Coords first_sibling = get_first_child(mother);

    ssize_t position_in_mother = (rank - first_sibling).linearize(2);
    ssize_t position_in_team = position_in_mother % c;
    ssize_t team_in_mother = position_in_mother >> log2c;

    ssize_t skew_team = position_in_team + team_in_mother;
    ssize_t skew_in_team = position_in_team;

    ssize_t teams_per_mother = children_per_mother / c;
    ssize_t recv_skew = -1;
    ssize_t send_skew = -1;
    for (ssize_t tim = 0; tim < teams_per_mother; ++tim) {
        for (ssize_t pit = 0; pit < c; ++pit) {
            ssize_t _recv_skew = c * tim + pit;
            ssize_t _send_skew = c * tim + c * pit + pit;
            if (_send_skew % children_per_mother == position_in_mother) {
                recv_skew = _recv_skew;
                send_skew = _send_skew;
            }
        }
    }
    assert(0 <= send_skew);
    assert(0 <= recv_skew && recv_skew < children_per_mother);

    std::cout
        << rank
        << " send_skew=" << send_skew
        << " recv_skew=" << recv_skew
        << std::endl;

    ssize_t neighbor_team_idx = 0;

    for (ssize_t r = 0; r < rounds; ++r) {
        ssize_t lin_receive_from = c * (neighbor_team_idx + skew_team)
                                 + skew_in_team;  // only talk to other team members at the same position
        lin_receive_from %= neighbors;
        ssize_t receive_nitems = std::max(ssize_t(0), std::min(c, neighbors - c * (neighbor_team_idx + skew_team)));
        Coords receive_from = delinearize_neighbor<Coords>(lin_receive_from, neighbor_dim);

        Coords receive_from_abs = first_neighbor + receive_from;
        for (size_t i = 0; i < ndims; ++i)
            receive_from_abs[i] = wraparound(receive_from_abs[i], dim);

        // find a reference point: the first-sibling-neighbor sending to the first-sibling-receiver in round 0
        ssize_t reference_sender_offset = r * c * c + send_skew;
        ssize_t reference_receiver_offset = (neighbors - children_per_mother) / 2 + recv_skew;
        Coords send_to_abs =
            rank
            - delinearize_neighbor<Coords>(reference_sender_offset, neighbor_dim)
            + delinearize_neighbor<Coords>(reference_receiver_offset, neighbor_dim);
        for (size_t i = 0; i < ndims; ++i)
            send_to_abs[i] = wraparound(send_to_abs[i], dim);
        ssize_t send_nitems = std::max(ssize_t(0), std::min(c, neighbors - reference_sender_offset));

        bool self_comm = rank == send_to_abs && rank == send_to_abs;
        assert((rank != receive_from_abs && rank != send_to_abs) || (rank == receive_from_abs && rank == send_to_abs));
        assert((rank == receive_from_abs) == (rank == send_to_abs));

        ssize_t receive_from_team = linearize_neighbor(receive_from_abs, dim) >> log2c;
        ssize_t send_to_team = linearize_neighbor(send_to_abs, dim) >> log2c;
        assert(self_comm || (my_team != receive_from_team && my_team != send_to_team));

        std::cout << r << ": "
            << rank
            << " "
            << (receive_nitems > 0 ? "" : "!")
            << (receive_from_abs != rank ? "" : "self")
            << "recv" << receive_from_abs
            << "#" << receive_nitems
            << " " << receive_from
            << " " << neighbor_team_idx
            << " "
            << (send_nitems > 0 ? "" : "!")
            << (send_to_abs != rank ? "" : "self")
            << "send" << send_to_abs
            << "#" << send_nitems
            << " rso=" << reference_sender_offset
            << " rro=" << reference_receiver_offset
            << std::endl;

        // advance by C teams
        neighbor_team_idx += c;
    }
}

int main()
{
    ff_communicate(MultidimensionalCoordinates<3> {2, 2, 2}, 1, 2, 5);
    ff_communicate(MultidimensionalCoordinates<3> {3, 2, 2}, 1, 2, 5);
    ff_communicate(MultidimensionalCoordinates<3> {2, 3, 2}, 1, 2, 5);
    ff_communicate(MultidimensionalCoordinates<3> {3, 3, 3}, 1, 2, 5);
    ff_communicate(MultidimensionalCoordinates<3> {2, 2, 2}, 1, 4, 5);
    ff_communicate(MultidimensionalCoordinates<3> {2, 2, 2}, 1, 8, 3);
    std::cout << std::endl;
    for (int i = 2; i < 6; ++i)
    for (int j = 2; j < 4; ++j)
    for (int k = 2; k < 4; ++k)
        ff_communicate(MultidimensionalCoordinates<3> {i, j, k}, 1, 8, 3);
    std::cout << std::endl;
    ff_communicate(MultidimensionalCoordinates<2> {2, 2}, 1, 1, 3);
    std::cout << std::endl;
    ff_communicate(MultidimensionalCoordinates<2> {2, 2}, 1, 2, 3);
    ff_communicate(MultidimensionalCoordinates<1> {2}, 1, 2, 3);
    std::cout << std::endl;
    for (int i = 0; i < 8; ++i)
        ff_communicate(MultidimensionalCoordinates<1> {i}, 1, 1, 3);
    std::cout << "0:" << std::endl;
    for (int j = 2; j < 4; ++j)
        for (int i = 2; i < 4; ++i)
            ff_communicate(MultidimensionalCoordinates<2> {i, j}, 1, 2, 3);
    std::cout << "0:" << std::endl;
    for (int i = 2; i < 4; ++i)
        for (int j = 2; j < 4; ++j)
            for (int k = 2; k < 4; ++k)
                ff_communicate(MultidimensionalCoordinates<3> {i, j, k}, 1, 4, 3);
    if (1) {
        std::cout << std::endl;
        for (int j = 0; j < 8; ++j)
            for (int i = 0; i < 8; ++i)
                ff_communicate(MultidimensionalCoordinates<2> {i, j}, 1, 4, 3);
    }
}

// vim: et:ts=4:sw=4
