/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "fmsolvr/abc.hpp"
#include "fmsolvr/aligned_alloc.hpp"
#include "fmsolvr/dipole_compensation.hpp"
#include "fmsolvr/math/change_of_basis.hpp"
#include "fmsolvr/math/rsqrt.hpp"
#include "fmsolvr/multipole.hpp"
#include "fmsolvr/operator/lattice/latticeoperator.hpp"
#include "fmsolvr/operator/p4/local2local.hpp"
#include "fmsolvr/operator/p4/local2local_derivative.hpp"
#include "fmsolvr/operator/p4/multipole2local.hpp"
#include "fmsolvr/operator/p4/multipole2multipole_derivative.hpp"
#include "fmsolvr/operator/particle2multipole.hpp"
#include "fmsolvr/tree/simulationbox.hpp"
#include "fmsolvr/util/buildinfo.hpp"
#include "fmsolvr/util/float128.hpp"
#include "fmsolvr/xyz.hpp"
#include "fmsolvr/xyzs.hpp"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <cmath>
#include <cassert>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <sys/time.h>
#ifdef WITH_OPENMP
#include <omp.h>
#endif
#ifdef WITH_STD_THREAD
#include <thread>
#endif
#ifndef __NVCC__
#if defined(__SSE__)
#include <immintrin.h>
#endif
#endif
#ifdef WITH_BOOST
#include <boost/thread/thread.hpp>
#endif

#include "read_qxyz.hpp"
#include "read_fphi.hpp"
#include "write_fphi.hpp"

#include "Mlm400_64_iter64_1D.h"
#include "Mlm400_64_iter64_2D.h"
#include "Mlm400_64_iter64_3D.h"
#include "compare_with_reference.hpp"

#ifndef FP_TYPE
#define FP_TYPE double
#endif

#define _my_STRINGIFY(x) # x
#define STRINGIFY(x) _my_STRINGIFY(x)

namespace fmsolvr
{

#ifdef WITH_FLOAT4
typedef float float4 __attribute__ ((vector_size (16)));

template <>
struct XYZ<float> {
	typedef float value_type;
	typedef float4 value4_type;
	union {
		struct {
			value_type x,y,z,_unused;
		};
		value4_type f4;
	};

	XYZ(value_type x, value_type y, value_type z): x(x), y(y), z(z) {}

	XYZ(value4_type f4) : f4(f4) {}

	XYZ operator - (const XYZ & a) const
	{
		return XYZ(f4 - a.f4);
	}

	XYZ operator * (const value_type s) const
	{
		value4_type s4 = {s, s, s, s};
		return XYZ(f4 * s4);
	}

	XYZ & operator += (const XYZ & a)
	{
		f4 += a.f4;
		return *this;
	}

	XYZ & operator -= (const XYZ & a)
	{
		f4 -= a.f4;
		return *this;
	}
};
#endif

#ifdef WITH_DOUBLE4
typedef double double4 __attribute__ ((vector_size (32)));

template <>
struct XYZ<double> {
	typedef double value_type;
	typedef double4 value4_type;
	union {
		struct {
			value_type x,y,z,_unused;
		};
		value4_type f4 __attribute__ ((aligned (32)));
	};

	XYZ(value_type x, value_type y, value_type z): x(x), y(y), z(z) {}

	XYZ(value4_type f4) : f4(f4) {}

	XYZ operator - (const XYZ & a) const
	{
		return XYZ(f4 - a.f4);
	}

	XYZ operator * (const value_type s) const
	{
		value4_type s4 = {s, s, s, s};
		return XYZ(f4 * s4);
	}

	XYZ & operator += (const XYZ & a)
	{
		f4 += a.f4;
		return *this;
	}

	XYZ & operator -= (const XYZ & a)
	{
		f4 -= a.f4;
		return *this;
	}
};
#endif


template <typename Real>
struct coulomb_types
{
	typedef Real real_type;
	typedef XYZ<Real> real3_type;
	typedef std::vector<real_type> vector1_type;
#ifndef WITH_DOUBLE4
	typedef std::vector<real3_type> vector3_type;
#else
	typedef std::vector<real3_type, aligned_allocator<real3_type> > vector3_type;
#endif
};

template <typename Real>
struct coulomb_data
{
	typedef typename coulomb_types<Real>::real_type real_type;
	typedef typename coulomb_types<Real>::real3_type real3_type;
	typedef typename coulomb_types<Real>::vector1_type vector1_type;
	typedef typename coulomb_types<Real>::vector3_type vector3_type;

	const vector1_type & vq;
	const vector3_type & vx;
	vector3_type & vF;
	vector1_type & vPhi;

	coulomb_data(
			const vector1_type & vq,
			const vector3_type & vx,
			vector3_type & vF,
			vector1_type & vPhi
			) :
		vq(vq), vx(vx), vF(vF), vPhi(vPhi)
	{ }
};


typedef coulomb_data<FP_TYPE> coulomb_data_type;


void read_input(
		const char* f,
		coulomb_data_type::vector1_type & vq,
		coulomb_data_type::vector3_type & vx
		)
{
	vq.clear();
	vx.clear();
	fmsolvr::read_qxyz<FP_TYPE>(f, vq, vx);
	std::cout << "Read " << vq.size() << " particles from " << f << std::endl;
}

void write_output(
		const char* f,
		const coulomb_data_type::vector3_type & vF,
		const coulomb_data_type::vector1_type & vPhi
		)
{
	fmsolvr::write_fphi<FP_TYPE>(f, vF, vPhi);
	std::cout << "Wrote information for " << vF.size() << " particles to " << f << std::endl;
}


// Returns number of seconds since the epoch, high resolution.
inline double
timestamp()
{
    struct timeval tp;
    gettimeofday(&tp, NULL);
    return double(tp.tv_sec) + tp.tv_usec / 1000000.;
}

#ifdef WITH_FLOAT4
template <>
float length<float>(const XYZ<float> & a)
{
	XYZ<float> aa(a.f4 * a.f4);
	return std::sqrt(aa.x() + aa.y() + aa.z());
}
#endif

#ifdef WITH_DOUBLE4
template <>
double length<double>(const XYZ<double> & a)
{
	XYZ<double> aa(a.f4 * a.f4);
	return std::sqrt(aa.x() + aa.y() + aa.z());
}
#endif

// 1 / |a|
template <typename Real>
Real rcplength(const XYZ<Real> & a)
{
	return rsqrt(a.x() * a.x() + a.y() * a.y() + a.z() * a.z());
}

#ifdef WITH_FLOAT4
template <>
float rcplength<float>(const XYZ<float> & a)
{
	XYZ<float> aa(a.f4 * a.f4);
	return rsqrt(aa.x() + aa.y() + aa.z());
}
#endif

#ifdef WITH_DOUBLE4
template <>
double rcplength<double>(const XYZ<double> & a)
{
	XYZ<double> aa(a.f4 * a.f4);
	return rsqrt(aa.x() + aa.y() + aa.z());
}
#endif

template <typename Real>
Real sumdistancehalfmatrix(const typename coulomb_types<Real>::vector3_type & vx)
{
	Real sum = 0.;
	size_t n = vx.size();
	for (size_t i = 0; i < n; ++i)
		for (size_t j = 0; j < i; ++j)
		{
			sum += length(vx[i] - vx[j]);
		}
	return sum;
}


int NTHREADS = 1;

// sequential
template <typename CoulombData>
void coulompute(CoulombData & cd, typename CoulombData::real_type & Ec)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector3_type & vx(cd.vx);
	vector3_type & vF(cd.vF);
	size_t n = vx.size();

	vF.clear();
	vF.resize(n, Real3(0., 0., 0.));
#ifdef COULOMB_POTENTIAL
	vector1_type & vPhi(cd.vPhi);
	vPhi.clear();
	vPhi.resize(n, 0.);
#endif
	Real Ec2 = 0.;

	coulompute_slice(cd, Ec2, 0, n);
	Ec = Ec2 / 2.;
}


// sequential
template <bool periodic_a, bool periodic_b, bool periodic_c,
	 typename CoulombData, typename LatticeOperator, typename Real33>
void coulompute_periodic(CoulombData & cd, typename CoulombData::real_type & Ec,
		const int p,
		const LatticeOperator & L,
		const Real33 & abc,
		ssize_t ws)
{
	const bool compute_field = true;
	const bool disable_dipole_compensation = false;
	const bool negative_gradient = true;

	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector1_type & vq(cd.vq);
	const vector3_type & vx(cd.vx);
	vector3_type & vF(cd.vF);
	size_t n = vx.size();

	// ref_point is the reference point ("Aufpunkt") (in global coordinates)
	// corresponding to the centre of the (to be generated) simulation box in R^3
	Real3 ref_point(0., 0., 0.);
	const char * refpointstring = 0;
	if ((refpointstring = getenv("REFERENCE_POINT")) || (refpointstring = getenv("CENTER")))
	{
		if (refpointstring[0] == 'c' || refpointstring[0] == 'C') {
			ref_point = Real3(abc.center());
		} else {
			double args[3];
			int k = sscanf(refpointstring, "%lf %lf %lf", args + 0, args + 1, args + 2);
			if (k == 1) {
				ref_point = Real3(args[0], args[0], args[0]);
			} else if (k == 3) {
				ref_point = Real3(args[0], args[1], args[2]);
			} else {
				std::cerr << "Error: Cannot parse CENTER." << std::endl;
				exit(1);
			}
		}
	}
	std::cout << "reference point: " << ref_point << std::endl;

	vF.clear();
	vF.resize(n, Real3(0., 0., 0.));
#ifdef COULOMB_POTENTIAL
	vector1_type & vPhi(cd.vPhi);
	vPhi.clear();
	vPhi.resize(n, 0.);
#endif
	Real Ec2 = 0.;

	Real33 convert_to_abc = change_of_basis_from_standard_basis(abc);
	Real33 convert_from_abc = change_of_basis_to_standard_basis(abc);

	vector3_type periodics;
	{
		const ssize_t range_i = periodic_a ? ws : 0;
		const ssize_t range_j = periodic_b ? ws : 0;
		const ssize_t range_k = periodic_c ? ws : 0;

		periodics.reserve((2 * range_i + 1) * (2 * range_j + 1) * (2 * range_k + 1) - 1);
		for (int i = -range_i; i <= +range_i; ++i)
			for (int j = -range_j; j <= +range_j; ++j)
				for (int k = -range_k; k <= +range_k; ++k)
					if (i || j || k)
						periodics.push_back(abc * Real3(i, j, k));
	}

	// dipole
	// NOTE: this works only for box_charge == 0.
	Real3 dp;
	Real totalq = 0.;
	for (size_t i = 0; i < n; ++i) {
		dp += Real3(vx[i]) * vq[i];
		totalq += vq[i];
	}
	Real3 dp_abc = convert_to_abc * dp;
	Real3 pdp_abc(periodic_a ? dp_abc.x() : 0., periodic_b ? dp_abc.y() : 0., periodic_c ? dp_abc.z() : 0.);
	if (disable_dipole_compensation)
		pdp_abc = Real3();
	Real3 pdp = convert_from_abc * pdp_abc;
	if (periodic_a && periodic_b && periodic_c) {
		//assert(pdp == dp);
		//pdp = dp;
	}
	std::cout << "dipole: " << dp << "_xyz  " << dp_abc << "_abc  total charge: " << totalq << std::endl;
	std::cout << "periodicity: " << Real3(periodic_a, periodic_b, periodic_c) << "  ws: " << ws << std::endl;
	std::cout << "dipole to be compensated: " << pdp << "_xyz  " << pdp_abc << "_abc" << std::endl;

	Real3 origin(0., 0., 0.);
	Real q0, qa, qb, qc;
	fmsolvr::compute_dipole_compensating_charges(pdp, abc, q0, qa, qb, qc);
	std::cout << "fictive charges: " << q0 << " " << qa << " " << qb << " " << qc << std::endl;
	Real3 rdp = dp + origin * q0 + abc.a() * qa + abc.b() * qb + abc.c() * qc;
	std::cout << "remaining dipole: " << rdp << "_xyz  " << convert_to_abc * rdp << "_abc" << std::endl;

	vector3_type fake_particles;
	vector1_type fake_charges;
	{
		fmsolvr::SimulationBox<Real33> simulationbox(abc, ref_point, periodic_a, periodic_b, periodic_c);
            auto lambda = [&](Real3 xyz, Real q) {
                fake_particles.push_back(xyz);
                fake_charges.push_back(q);
            };
            fmsolvr::generate_dipole_compensating_particles(
                    simulationbox, ws, q0, qa, qb, qc, lambda);
	}

	// NEAR FIELD
	if (compute_field) {
		coulompute_field_slice(cd, 0, n);
		coulompute_field_slice_periodic(cd, 0, n, periodics.size(), periodics);
	} else {
	coulompute_slice(cd, Ec2, 0, n);
	coulompute_slice_periodic(cd, Ec2, 0, n, periodics.size(), periodics);
	}

	// FAR FIELD
	typedef LatticeOperator CoeffMatrix;
	typedef typename CoeffMatrix::complex_type complex_type;

	P2M_divtable<Real> div_lut_p(p);
	CoeffMatrix omega(p);
	// omega and mu use the expansion centre (0,0,0) in local coordinate
	// systems, therefore convert coordinates:
	// - particle data is given in global coordinates, the box is centered
	//   at ref_point
	for (size_t particle = 0; particle < n; ++particle) {
		Real3 xyz = vx[particle];
		Real q = vq[particle];
		P2M(q, xyz, omega, ref_point, div_lut_p);
	}
	// - the 4 fictive charges are placed on the box spanning vectors with
	//   origin (0,0,0) at a corner
	Real3 m = abc.center();
	P2M(q0, origin, omega, m, div_lut_p);
	P2M(qa, abc.a(), omega, m, div_lut_p);
	P2M(qb, abc.b(), omega, m, div_lut_p);
	P2M(qc, abc.c(), omega, m, div_lut_p);
	omega.populate_lower();

	CoeffMatrix mu(p);
	single_M2L(omega, L, mu, p);

	const bool fake_particles_in_far_field = true;
	const bool far_field_via_dmu = true;  // takes precedence over far_field_via_domega
	const bool far_field_via_domega = true;

	if (fake_particles_in_far_field) {
		for (size_t particle = 0; particle < fake_particles.size(); ++particle) {
			Real3 xyz = fake_particles[particle];
			Real q = fake_charges[particle];
			P2L(q, xyz, mu, ref_point, p);
		}
	}

	mu.populate_lower();

	CoeffMatrix dmux(p);
	CoeffMatrix dmuy(p);
	CoeffMatrix dmuz(p);

	if (far_field_via_dmu) {
		DxL2L_reference(mu, dmux, p);
		DyL2L_reference(mu, dmuy, p);
		DzL2L_reference(mu, dmuz, p);
		dmux.populate_lower();
		dmuy.populate_lower();
		dmuz.populate_lower();
	}

	// far field Field
	for (size_t i = 0; i < n; ++i) {
		Real3 xyz = vx[i];
		Real q = (compute_field ? -1. : negative_gradient ? -vq[i] : vq[i]);

		if (far_field_via_dmu) {
			CoeffMatrix omega(p);

			// expand the particle at (x,y,z) into a multipole at ref_point
			P2M(q, xyz, omega, ref_point, div_lut_p);
			omega.populate_lower();

			// FIXME: optimize multiplication: (re, 0) = (re, im) * (re, im)
			complex_type Fx(0.);
			complex_type Fy(0.);
			complex_type Fz(0.);

			for (ssize_t l = 0; l <= p; ++l) {
				for (ssize_t m = -l; m <= l; ++m) {
					complex_type omega_lm = omega.get(l, m);
					Fx += dmux.get(l, m) * omega_lm;
					Fy += dmuy.get(l, m) * omega_lm;
					Fz += dmuz.get(l, m) * omega_lm;
				}
			}

			// FIXME: verify imag() part is (reasonably) zero
			vF[i] -= Real3(Fx.real(), Fy.real(), Fz.real());
		} else if (far_field_via_domega) {
			CoeffMatrix omega(p);
			CoeffMatrix domegax(p);
			CoeffMatrix domegay(p);
			CoeffMatrix domegaz(p);

			// expand the particle at (x,y,z) into a multipole at ref_point
			P2M(q, xyz, omega, ref_point, div_lut_p);
			omega.populate_lower();

			DxM2M_reference(omega, domegax, p);
			DyM2M_reference(omega, domegay, p);
			DzM2M_reference(omega, domegaz, p);
			domegax.populate_lower();
			domegay.populate_lower();
			domegaz.populate_lower();

			// FIXME: optimize multiplication: (re, 0) = (re, im) * (re, im)
			complex_type Fx(0.);
			complex_type Fy(0.);
			complex_type Fz(0.);

			for (ssize_t l = 0; l <= p; ++l) {
				for (ssize_t m = -l; m <= l; ++m) {
					complex_type mu_lm = mu.get(l, m);
					Fx += domegax.get(l, m) * mu_lm;
					Fy += domegay.get(l, m) * mu_lm;
					Fz += domegaz.get(l, m) * mu_lm;
				}
			}

			// FIXME: verify imag() part is (reasonably) zero
			vF[i] += Real3(Fx.real(), Fy.real(), Fz.real());
		} else {
			// shift mu from ref_point (e.g. (0,0,0)) to the particle at (x,y,z)
			CoeffMatrix mu_shift(1);
			L2L(mu, ref_point, mu_shift, xyz, p, 1);

			Real Fx = mu_shift.upper(1, 1).real();
			Real Fy = mu_shift.upper(1, 1).imag();
			Real Fz = mu_shift.upper(1, 0).real();
			vF[i] += Real3(Fx, Fy, Fz) * q;
		}
	}

	// far field Potential
	for (size_t i = 0; i < n; ++i) {
		Real3 xyz = vx[i];
#if 0
		// shift mu from ref_point (e.g. (0,0,0)) to the particle at (x,y,z)
		CoeffMatrix mu_shift(0);
		L2L(mu, ref_point, mu_shift, xyz, p, 0);
		vPhi[i] += mu_shift.upper(0, 0).real();
#else
		// shift mu from ref_point (e.g. (0,0,0)) to the particle at (x,y,z)
		L2L_Operator<CoeffMatrix> O(ref_point, xyz, p);

		complex_type Phi_i(0.);
		// FIXME: optimize multiplication: (re, 0) = (re, im) * (re, im)
		for (ssize_t l = 0; l <= p; ++l) {
			for (ssize_t m = -l; m <= l; ++m) {
				Phi_i += mu.get(l, m) * O.get(l, m);
			}
		}
		// FIXME: verify imag() part is (reasonably) zero
		vPhi[i] += Phi_i.real();
#endif
	}

	// dipole correction
	// NOTE: this works only for box_charge == 0.
	Real3 adp;
	{
	    auto it = fake_particles.begin();
	    auto itq = fake_charges.begin();
	    for ( ; it != fake_particles.end(); ++it, ++itq) {
		adp += Real3(*it) * (*itq);
		//std::cout << it->x << " " << it->y << " " << it->z << " " << it->q << std::endl;
	    }
	}
	std::cout << "antdipole: " << adp << std::endl;

	if (!fake_particles_in_far_field) {
		if (compute_field) {
			coulompute_field_slice_box2box(cd, 0, n, fake_particles, fake_charges);
		} else {
		coulompute_slice_box2box(cd, Ec2, 0, n, fake_particles, fake_charges);
		}
	}

	// Forces and Energy
	Ec2 = 0.;
	for (size_t i = 0; i < n; ++i) {
		Real q = -(!compute_field ? -1. : negative_gradient ? -vq[i] : vq[i]);
		vF[i] *= q;
#ifdef COULOMB_ENERGY
		Ec2 += vq[i] * vPhi[i];
#endif
	}
	Ec = Ec2 / 2.;

	// TODO

	// ewald correction
	// TODO
}


template <typename CoulombData>
void coulompute_slice(CoulombData cd, typename CoulombData::real_type & Ec2, size_t begin, size_t end)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector1_type & vq(cd.vq);
	const vector3_type & vx(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);
	Real Ec = Ec2;
	size_t n = vx.size();

	for (size_t i = begin; i < end; ++i) {
		Real3 F_i(0., 0., 0.);
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
		Real Phi_i = 0.;
#endif

		for (size_t j = 0; j < n; ++j) {
			if (i != j) {
				Real3 diff = vx[i] - vx[j];		// (x_i - x_j)
				Real rlen = rcplength(diff);		// 1 / |x_i - x_j|
				Real q_j_rlen = vq[j] * rlen;		// q_j / |x_i - x_j|
				Real rlen2 = rlen * rlen;		// 1 / |x_i - x_j|^2
				Real q_j_rlen3 = q_j_rlen * rlen2;	// q_j / |x_i - x_j|^3
				F_i += diff * q_j_rlen3;		// (x_i - x_j) * q_j / |x_i - x_j|^3
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
				Phi_i += q_j_rlen;
#endif
			}
		}
		Real q_i = vq[i];
		vF[i] += F_i * q_i;
#ifdef COULOMB_POTENTIAL
		vPhi[i] += Phi_i;
#endif
#ifdef COULOMB_ENERGY
		Ec += q_i * Phi_i;
#endif
	}
	Ec2 = Ec;
}


template <typename CoulombData>
void coulompute_field_slice(CoulombData cd, size_t begin, size_t end)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector1_type & vq(cd.vq);
	const vector3_type & vx(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);
	size_t n = vx.size();

	for (size_t i = begin; i < end; ++i) {
		Real3 F_i(0., 0., 0.);
#if defined(COULOMB_POTENTIAL)
		Real Phi_i = 0.;
#endif

		for (size_t j = 0; j < n; ++j) {
			if (i != j) {
				Real3 diff = vx[i] - vx[j];		// (x_i - x_j)
				Real rlen = rcplength(diff);		// 1 / |x_i - x_j|
				Real q_j_rlen = vq[j] * rlen;		// q_j / |x_i - x_j|
				Real rlen2 = rlen * rlen;		// 1 / |x_i - x_j|^2
				Real q_j_rlen3 = q_j_rlen * rlen2;	// q_j / |x_i - x_j|^3
				F_i += diff * q_j_rlen3;		// (x_i - x_j) * q_j / |x_i - x_j|^3
#if defined(COULOMB_POTENTIAL)
				Phi_i += q_j_rlen;
#endif
			}
		}
		vF[i] += F_i;
#ifdef COULOMB_POTENTIAL
		vPhi[i] += Phi_i;
#endif
	}
}


template <typename CoulombData>
void coulompute_slice_box2box(CoulombData cd, typename CoulombData::real_type & Ec2, size_t begin, size_t end)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector1_type & vqi(cd.vq);
	const vector3_type & vxi(cd.vx);
	const vector1_type & vqj(cd.vq);
	const vector3_type & vxj(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);
	Real Ec = Ec2;
	size_t n = vxj.size();

	size_t i_begin = begin;
	size_t i_end = end;

	size_t j_begin = 0;
	size_t j_end = n;

	for (size_t i = i_begin; i < i_end; ++i) {
		Real3 F_i(0., 0., 0.);
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
		Real Phi_i = 0.;
#endif

		for (size_t j = j_begin; j < j_end; ++j) {
			if (1) {
				Real3 diff = vxi[i] - vxj[j];		// (x_i - x_j)
				Real rlen = rcplength(diff);		// 1 / |x_i - x_j|
				Real q_j_rlen = vqj[j] * rlen;		// q_j / |x_i - x_j|
				Real rlen2 = rlen * rlen;		// 1 / |x_i - x_j|^2
				Real q_j_rlen3 = q_j_rlen * rlen2;	// q_j / |x_i - x_j|^3
				F_i += diff * q_j_rlen3;		// (x_i - x_j) * q_j / |x_i - x_j|^3
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
				Phi_i += q_j_rlen;
#endif
			}
		}
		Real q_i = vqi[i];
		vF[i] += F_i * q_i;
#ifdef COULOMB_POTENTIAL
		vPhi[i] += Phi_i;
#endif
#ifdef COULOMB_ENERGY
		Ec += q_i * Phi_i;
#endif
	}
	Ec2 = Ec;
}


template <typename CoulombData>
void coulompute_slice_box2box(CoulombData cd, typename CoulombData::real_type & Ec2, size_t begin, size_t end,
		const typename CoulombData::vector3_type & vxj, const typename CoulombData::vector1_type & vqj)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector1_type & vqi(cd.vq);
	const vector3_type & vxi(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);
	Real Ec = Ec2;
	size_t n = vxj.size();

	size_t i_begin = begin;
	size_t i_end = end;

	size_t j_begin = 0;
	size_t j_end = n;

	for (size_t i = i_begin; i < i_end; ++i) {
		Real3 F_i(0., 0., 0.);
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
		Real Phi_i = 0.;
#endif

		for (size_t j = j_begin; j < j_end; ++j) {
			if (1) {
				Real3 diff = vxi[i] - vxj[j];		// (x_i - x_j)
				Real rlen = rcplength(diff);		// 1 / |x_i - x_j|
				Real q_j_rlen = vqj[j] * rlen;		// q_j / |x_i - x_j|
				Real rlen2 = rlen * rlen;		// 1 / |x_i - x_j|^2
				Real q_j_rlen3 = q_j_rlen * rlen2;	// q_j / |x_i - x_j|^3
				F_i += diff * q_j_rlen3;		// (x_i - x_j) * q_j / |x_i - x_j|^3
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
				Phi_i += q_j_rlen;
#endif
			}
		}
		Real q_i = vqi[i];
		vF[i] += F_i * q_i;
#ifdef COULOMB_POTENTIAL
		vPhi[i] += Phi_i;
#endif
#ifdef COULOMB_ENERGY
		Ec += q_i * Phi_i;
#endif
	}
	Ec2 = Ec;
}


template <typename CoulombData>
void coulompute_field_slice_box2box(CoulombData cd, size_t begin, size_t end,
		const typename CoulombData::vector3_type & vxj, const typename CoulombData::vector1_type & vqj)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	//const vector1_type & vqi(cd.vq);
	const vector3_type & vxi(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);
	size_t n = vxj.size();

	size_t i_begin = begin;
	size_t i_end = end;

	size_t j_begin = 0;
	size_t j_end = n;

	for (size_t i = i_begin; i < i_end; ++i) {
		Real3 F_i(0., 0., 0.);
#if defined(COULOMB_POTENTIAL)
		Real Phi_i = 0.;
#endif

		for (size_t j = j_begin; j < j_end; ++j) {
			if (1) {
				Real3 diff = vxi[i] - vxj[j];		// (x_i - x_j)
				Real rlen = rcplength(diff);		// 1 / |x_i - x_j|
				Real q_j_rlen = vqj[j] * rlen;		// q_j / |x_i - x_j|
				Real rlen2 = rlen * rlen;		// 1 / |x_i - x_j|^2
				Real q_j_rlen3 = q_j_rlen * rlen2;	// q_j / |x_i - x_j|^3
				F_i += diff * q_j_rlen3;		// (x_i - x_j) * q_j / |x_i - x_j|^3
#if defined(COULOMB_POTENTIAL)
				Phi_i += q_j_rlen;
#endif
			}
		}
		vF[i] += F_i;
#ifdef COULOMB_POTENTIAL
		vPhi[i] += Phi_i;
#endif
	}
}


template <typename CoulombData, typename Periodics>
void coulompute_slice_periodic(CoulombData cd, typename CoulombData::real_type & Ec2, size_t begin, size_t end, size_t periodic_neighbors, const Periodics & periodicstep)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector1_type & vqi(cd.vq);
	const vector3_type & vxi(cd.vx);
	const vector1_type & vqj(cd.vq);
	const vector3_type & vxj(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);
	Real Ec = Ec2;
	size_t n = vxj.size();

	size_t i_begin = begin;
	size_t i_end = end;

	size_t j_begin = 0;
	size_t j_end = n;

	for (size_t i = i_begin; i < i_end; ++i) {
		Real3 F_i(0., 0., 0.);
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
		Real Phi_i = 0.;
#endif

		for (size_t j = j_begin; j < j_end; ++j) {
			Real3 basediff = vxi[i] - vxj[j];		// (x_i - x_j)
			Real q_j = vqj[j];
			for (size_t k = 0; k < periodic_neighbors; ++k) {
				Real3 diff = basediff;
			        diff += periodicstep[k];
				Real rlen = rcplength(diff);		// 1 / |x_i - x_j|
				Real q_j_rlen = q_j * rlen;		// q_j / |x_i - x_j|
				Real rlen2 = rlen * rlen;		// 1 / |x_i - x_j|^2
				Real q_j_rlen3 = q_j_rlen * rlen2;	// q_j / |x_i - x_j|^3
				F_i += diff * q_j_rlen3;		// (x_i - x_j) * q_j / |x_i - x_j|^3
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
				Phi_i += q_j_rlen;
#endif
			}
		}
		Real q_i = vqi[i];
		vF[i] += F_i * q_i;
#ifdef COULOMB_POTENTIAL
		vPhi[i] += Phi_i;
#endif
#ifdef COULOMB_ENERGY
		Ec += q_i * Phi_i;
#endif
	}
	Ec2 = Ec;
}


template <typename CoulombData, typename Periodics>
void coulompute_field_slice_periodic(CoulombData cd, size_t begin, size_t end, size_t periodic_neighbors, const Periodics & periodicstep)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	//const vector1_type & vqi(cd.vq);
	const vector3_type & vxi(cd.vx);
	const vector1_type & vqj(cd.vq);
	const vector3_type & vxj(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);
	size_t n = vxj.size();

	size_t i_begin = begin;
	size_t i_end = end;

	size_t j_begin = 0;
	size_t j_end = n;

	for (size_t i = i_begin; i < i_end; ++i) {
		Real3 F_i(0., 0., 0.);
#if defined(COULOMB_POTENTIAL)
		Real Phi_i = 0.;
#endif

		for (size_t j = j_begin; j < j_end; ++j) {
			Real3 basediff = vxi[i] - vxj[j];		// (x_i - x_j)
			Real q_j = vqj[j];
			for (size_t k = 0; k < periodic_neighbors; ++k) {
				Real3 diff = basediff;
			        diff += periodicstep[k];
				Real rlen = rcplength(diff);		// 1 / |x_i - x_j|
				Real q_j_rlen = q_j * rlen;		// q_j / |x_i - x_j|
				Real rlen2 = rlen * rlen;		// 1 / |x_i - x_j|^2
				Real q_j_rlen3 = q_j_rlen * rlen2;	// q_j / |x_i - x_j|^3
				F_i += diff * q_j_rlen3;		// (x_i - x_j) * q_j / |x_i - x_j|^3
#if defined(COULOMB_POTENTIAL)
				Phi_i += q_j_rlen;
#endif
			}
		}
		vF[i] += F_i;
#ifdef COULOMB_POTENTIAL
		vPhi[i] += Phi_i;
#endif
	}
}


#ifdef WITH_OPENMP
// sequential algorithm with parallel outer loop
template <typename CoulombData>
void coulompute_omp(CoulombData & cd, typename CoulombData::real_type & Ec)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector1_type & vq(cd.vq);
	const vector3_type & vx(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);

	size_t n = vx.size();
	vF.clear();
	vF.resize(n, Real3(0., 0., 0.));
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
	vPhi.clear();
	vPhi.resize(n, 0.);
#endif
	Real Ec2 = 0.;

	NTHREADS = omp_get_max_threads();
	std::cout << "threads: " << omp_get_max_threads() << std::endl;
#pragma omp parallel for
	for (size_t i = 0; i < n; ++i) {
		Real3 F_i(0., 0., 0.);
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
		Real Phi_i = 0.;
#endif

		for (size_t j = 0; j < n; ++j) {
			if (i != j) {
				Real3 diff = vx[i] - vx[j];		// (x_i - x_j)
				Real rlen = rcplength(diff);		// 1 / |x_i - x_j|
				Real q_j_rlen = vq[j] * rlen;		// q_j / |x_i - x_j|
				Real rlen2 = rlen * rlen;		// 1 / |x_i - x_j|^2
				Real q_j_rlen3 = q_j_rlen * rlen2;	// q_j / |x_i - x_j|^3
				F_i += diff * q_j_rlen3;		// (x_i - x_j) * q_j / |x_i - x_j|^3
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
				Phi_i += q_j_rlen;
#endif
			}
		}
		Real q_i = vq[i];
		vF[i] += F_i * q_i;
#ifdef COULOMB_POTENTIAL
		vPhi[i] += Phi_i;
#endif
	}
	for (size_t i = 0; i < n; ++i) {
#ifdef COULOMB_ENERGY
		Ec2 += vq[i] * vPhi[i];
#endif
	}
	Ec = Ec2 / 2.;
}
#endif

#ifdef WITH_STD_THREAD
// sequential algorithm with parallel outer loop: std::thread
template <typename CoulombData>
void coulompute_threaded(CoulombData & cd, typename CoulombData::real_type & Ec)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector3_type & vx(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);

	size_t n = vx.size();
	vF.clear();
	vF.resize(n, Real3(0., 0., 0.));
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
	vPhi.clear();
	vPhi.resize(n, 0.);
#endif
	Real Ec2 = 0.;

	int nthreads = std::thread::hardware_concurrency();
	if (getenv("OMP_NUM_THREADS"))
		nthreads = atoi(getenv("OMP_NUM_THREADS"));
	if (nthreads < 1)
		nthreads = 1;
	NTHREADS = nthreads;
	std::cout << "threads: " << nthreads << std::endl;
	{
		std::vector<std::thread> threads;
		std::vector<Real> vEc(nthreads, 0.);
		size_t current = 0;
		for (int t = 0; t < nthreads; ++t)
		{
			size_t slice = (n - current) / (nthreads - t);
			threads.push_back(std::thread(coulompute_slice<CoulombData>, cd, std::ref(vEc[t]), current, current + slice));
			current += slice;
		}
		assert(current == n);
		for (size_t t = 0; t < threads.size(); ++t) {
			threads[t].join();
			Ec2 += vEc[t];
		}
	}
	Ec = Ec2 / 2.;
}
#endif


// sequential algorithm with parallel outer loop: pthreads
template <typename CoulombData>
struct coulompute_pthread_params
{
	typedef typename CoulombData::real_type Real;

	CoulombData * cd;
	Real Ec;
	size_t begin, end;

	coulompute_pthread_params(CoulombData & cd, size_t begin, size_t end) :
		cd(&cd), Ec(0.), begin(begin), end(end)
	{ }
};

template <typename CoulombData>
void* coulompute_worker_pthreads(void * data)
{
	typedef coulompute_pthread_params<CoulombData> threaddata_type;
	threaddata_type * param = reinterpret_cast<threaddata_type *>(data);
	coulompute_slice(*(param->cd), param->Ec, param->begin, param->end);
	return 0;
}

template <typename CoulombData>
void coulompute_pthreads(CoulombData & cd, typename CoulombData::real_type & Ec)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector3_type & vx(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);

	size_t n = vx.size();
	vF.clear();
	vF.resize(n, Real3(0., 0., 0.));
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
	vPhi.clear();
	vPhi.resize(n, 0.);
#endif
	Real Ec2 = 0.;

	int nthreads = 1;
	if (getenv("OMP_NUM_THREADS"))
		nthreads = atoi(getenv("OMP_NUM_THREADS"));
	if (nthreads < 1)
		nthreads = 1;
	NTHREADS = nthreads;
	std::cout << "threads: " << nthreads << std::endl;
	{
		typedef coulompute_pthread_params<CoulombData> threaddata_type;
		std::vector<threaddata_type> threaddata;
		std::vector<pthread_t> threads;

		threaddata.reserve(nthreads);
		threads.reserve(nthreads);
		size_t current = 0;
		for (int t = 0; t < nthreads; ++t)
		{
			size_t slice = (n - current) / (nthreads - t);
			threaddata.push_back(threaddata_type(cd, current, current + slice));
			pthread_t thr;
			if (pthread_create(&thr, 0, coulompute_worker_pthreads<CoulombData>, &threaddata[t]) == 0)
				threads.push_back(thr);
			else
				throw "pthread_create";
			current += slice;
		}
		assert(current == n);
		for (size_t t = 0; t < threads.size(); ++t) {
			pthread_join(threads[t], 0);
			Ec2 += threaddata[t].Ec;
		}
	}
	Ec = Ec2 / 2.;
}


#ifdef WITH_BOOST
// sequential algorithm with parallel outer loop: boost::thread
template <typename CoulombData>
void coulompute_boost_thread(CoulombData & cd, typename CoulombData::real_type & Ec)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector3_type & vx(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);

	size_t n = vx.size();
	vF.clear();
	vF.resize(n, Real3(0., 0., 0.));
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
	vPhi.clear();
	vPhi.resize(n, 0.);
#endif
	Real Ec2 = 0.;

	int nthreads = 0;
	if (getenv("OMP_NUM_THREADS"))
		nthreads = atoi(getenv("OMP_NUM_THREADS"));
	if (nthreads < 1)
		nthreads = 1;
	NTHREADS = nthreads;
	std::cout << "threads: " << nthreads << std::endl;
	{
		std::vector<boost::thread> threads(nthreads);
		std::vector<Real> vEc(nthreads, 0.);
		size_t current = 0;
		for (int t = 0; t < nthreads; ++t)
		{
			size_t slice = (n - current) / (nthreads - t);
			threads[t] = boost::thread(coulompute_slice<CoulombData>, cd, std::ref(vEc[t]), current, current + slice);
			current += slice;
		}
		assert(current == n);
		for (size_t t = 0; t < threads.size(); ++t) {
			threads[t].join();
			Ec2 += vEc[t];
		}
	}
	Ec = Ec2 / 2.;
}
#endif


// compute F(i,j) and F(j,i) at the same time
// reduces floating point ops (e.g. |x_i - x_j| can be shared) by a constant factor
// at the expense of more memory updates (linear -> quadratic)
// (sequential algo has quadratic number of register updates)
template <typename CoulombData>
void coulompute2(CoulombData & cd, typename CoulombData::real_type & Ec)
{
	typedef typename CoulombData::vector1_type vector1_type;
	typedef typename CoulombData::vector3_type vector3_type;
	typedef typename CoulombData::real_type Real;
	typedef typename vector3_type::value_type Real3;

	const vector1_type & vq(cd.vq);
	const vector3_type & vx(cd.vx);
	vector3_type & vF(cd.vF);
	vector1_type & vPhi(cd.vPhi);

	size_t n = vx.size();
	vF.clear();
	vF.resize(n, Real3(0., 0., 0.));
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
	vPhi.clear();
	vPhi.resize(n, 0.);
#endif
	Real Ec2 = 0.;

	// vF is not accumulating with previous entries !
	for (size_t i = 0; i < n; ++i) {
		for (size_t j = 0; j < i; ++j) {
				Real3 diff = vx[i] - vx[j];		// (x_i - x_j)
				Real rlen = rcplength(diff);		// 1 / |x_i - x_j|
				Real q_j_rlen = vq[j] * rlen;		// q_j / |x_i - x_j|
				Real q_i_rlen = vq[i] * rlen;		// q_i / |x_j - x_i|
				Real rlen2 = rlen * rlen;		// 1 / |x_i - x_j|^2
				Real q_j_rlen3 = q_j_rlen * rlen2;	// q_j / |x_i - x_j|^3
				Real q_i_rlen3 = q_i_rlen * rlen2;	// q_i / |x_j - x_i|^3
				vF[i] += diff * q_j_rlen3;		// (x_i - x_j) * q_j / |x_i - x_j|^3
				vF[j] += diff * (- q_i_rlen3);		// (x_j - x_i) * q_i / |x_j - x_i|^3
#if defined(COULOMB_POTENTIAL) || defined(COULOMB_ENERGY)
				vPhi[i] += q_j_rlen;
				vPhi[j] += q_i_rlen;
#endif
		}
	}
	for (size_t i = 0; i < n; ++i) {
		Real q_i = vq[i];
		vF[i] = vF[i] * q_i;
#ifdef COULOMB_ENERGY
		Ec2 += q_i * vPhi[i];
#endif
	}
	Ec = Ec2 / 2.;
}


template <typename CoulombData>
void verify_result(
		const CoulombData & cd,
		typename CoulombData::real_type Ec,
		const char* f,
		long double & reldeltaEc
		)
{
#if defined(HAVE_FLOAT128)
    typedef __float128 RefReal;
#else
    typedef long double RefReal;
#endif
    reldeltaEc = fmsolvr::compare_with_reference<RefReal>(cd.vq, cd.vF, cd.vPhi, Ec, f);
}

int main(int argc, char ** argv)
{
	typedef coulomb_data_type::real_type Real;
	typedef coulomb_data_type::real3_type Real3;

	if (argc < 4 || argc > 5) {
		std::cerr << "Usage: $0 input output algo [referencesolution]" << std::endl;
		return 1;
	}

	int algo = atoi(argv[3]);

	std::cout << "# coulomb.cc " << buildinfo<Real>() << std::endl;
	std::cout << "# " << compilerinfo() << std::endl;

	std::cout << "using " << FP_NAME << " and algorithm " << algo << std::endl;

	double t;
	coulomb_data_type::vector1_type vq;
	coulomb_data_type::vector3_type vx;

	read_input(argv[1], vq, vx);
	assert(vq.size() == vx.size());

	if (algo < 0) {
		t = -timestamp();
		Real dhmsum = sumdistancehalfmatrix<Real>(vx);
		t += timestamp();
		std::cout << "Computed + summed half distance matrix in " << t << " seconds, result = " << dhmsum << std::endl;
		// 182 seconds, result = 1.2982e+09
		return 0;
	}

	if (vx.size())
	{
		Real xmin = vx[0].x(), ymin = vx[0].y(), zmin = vx[0].z(), xmax = vx[0].x(), ymax = vx[0].y(), zmax = vx[0].z();
		for (auto particle : vx) {
			xmin = std::min(xmin, particle.x());
			ymin = std::min(ymin, particle.y());
			zmin = std::min(zmin, particle.z());
			xmax = std::max(xmax, particle.x());
			ymax = std::max(ymax, particle.y());
			zmax = std::max(zmax, particle.z());
		}
		Real3 mincorner(xmin, ymin, zmin);
		Real3 maxcorner(xmax, ymax, zmax);
		std::cout << "# orthonormal bounding box: " << mincorner << ", " << maxcorner << " center: " << ((maxcorner + mincorner) * 0.5) << std::endl;
	}

	typedef MultipoleCoefficientsUpperLower<Real> CoeffMatrix;
	typedef ABC<Real3> Real33;
	int ws_periodic = 1;
	Real3 a(1, 0, 0);
	Real3 b(0, 1, 0);
	Real3 c(0, 0, 1);
	const Real33 standard_basis(a, b, c);
	if (const char * unitbox = getenv("UNITBOX")) {
		double args[9];
		int ws_;
		int k = sscanf(unitbox, "%lf %lf %lf %lf %lf %lf %lf %lf %lf %d",
					args + 0, args + 1, args + 2,
					args + 3, args + 4, args + 5,
					args + 6, args + 7, args + 8,
					&ws_);
		if (k == 1) {
			a = Real3(args[0], 0, 0);
			b = Real3(0, args[0], 0);
			c = Real3(0, 0, args[0]);
		} else if (k == 3) {
			a = Real3(args[0], 0, 0);
			b = Real3(0, args[1], 0);
			c = Real3(0, 0, args[2]);
		} else if (k == 9 || k == 10) {
			a = Real3(args[0], args[1], args[2]);
			b = Real3(args[3], args[4], args[5]);
			c = Real3(args[6], args[7], args[8]);
			if (k == 10) {
				ws_periodic = ws_;
			}
		} else {
			std::cerr << "Error: Cannot parse UNITBOX." << std::endl;
			exit(1);
		}
	}
	int p = 25;
	if (const char * multipoleorder = getenv("MULTIPOLEORDER"))
		p = atoi(multipoleorder);
        if (algo >= 100)
        {
            std::cout << "# unitbox: a=" << a << " b=" << b << " c=" << c;
            std::cout << " ws=" << ws_periodic << " p=" << p << std::endl;
        }
	const Real33 abc(a, b, c);
	coulomb_data_type::vector3_type vF;
	coulomb_data_type::vector1_type vPhi;
	Real Ec;
	coulomb_data_type couldata(vq, vx, vF, vPhi);
	t = -timestamp();
	switch (algo) {
		case 0:	coulompute(couldata, Ec); break;
#ifdef WITH_OPENMP
		case 1:	coulompute_omp(couldata, Ec); break;
#endif
#ifdef WITH_STD_THREAD
		case 3:	coulompute_threaded(couldata, Ec); break;
#endif
		case 4:	coulompute_pthreads(couldata, Ec); break;
#ifdef WITH_BOOST
		case 5:	coulompute_boost_thread(couldata, Ec); break;
#endif
		case 2:	coulompute2(couldata, Ec); break;
		case 100: {
				CoeffMatrix L(2 * p, LatticeSums_1D);
				L.populate_lower();
				coulompute_periodic<false, false, true>(couldata, Ec, p, L, standard_basis, 1);
			} break;
		case 200: {
				CoeffMatrix L(2 * p, LatticeSums_2D);
				L.populate_lower();
				coulompute_periodic<true, true, false>(couldata, Ec, p, L, standard_basis, 1);
			} break;
		case 300: {
				CoeffMatrix L(2 * p, LatticeSums_3D);
				L.populate_lower();
				coulompute_periodic<true, true, true>(couldata, Ec, p, L, standard_basis, 1);
			} break;
		case 111: {
				CoeffMatrix L(2 * p);
				size_t its = LatticeOperator1D(abc, ws_periodic, L, p, 1., 42);
				L.populate_lower();
				std::cout << "LOp1Diterations " << its << std::endl;
				coulompute_periodic<false, false, true>(couldata, Ec, p, L, abc, ws_periodic);
			} break;
		case 112: {
				CoeffMatrix L(2 * p);
				size_t its = LatticeOperator1D2D<true, false, false>(abc, ws_periodic, L, p, 1., 42);
				L.populate_lower();
				std::cout << "LOp1D2Diterations " << its << std::endl;
				coulompute_periodic<true, false, false>(couldata, Ec, p, L, abc, ws_periodic);
			} break;
		case 113: {
				CoeffMatrix L(2 * p);
				size_t its = LatticeOperator1D2D<false, true, false>(abc, ws_periodic, L, p, 1., 42);
				L.populate_lower();
				std::cout << "LOp1D2Diterations " << its << std::endl;
				coulompute_periodic<false, true, false>(couldata, Ec, p, L, abc, ws_periodic);
			} break;
		case 222: {
				CoeffMatrix L(2 * p);
				size_t its = LatticeOperator2D(abc, ws_periodic, L, p, 1., 42);
				L.populate_lower();
				std::cout << "LOp2Diterations " << its << std::endl;
				coulompute_periodic<true, true, false>(couldata, Ec, p, L, abc, ws_periodic);
			} break;
		case 223: {
				CoeffMatrix L(2 * p);
				size_t its = LatticeOperator1D2D<false, true, true>(abc, ws_periodic, L, p, 1., 42);
				L.populate_lower();
				std::cout << "LOp1D2Diterations " << its << std::endl;
				coulompute_periodic<false, true, true>(couldata, Ec, p, L, abc, ws_periodic);
			} break;
		case 224: {
				CoeffMatrix L(2 * p);
				size_t its = LatticeOperator1D2D<true, false, true>(abc, ws_periodic, L, p, 1., 42);
				L.populate_lower();
				std::cout << "LOp1D2Diterations " << its << std::endl;
				coulompute_periodic<true, false, true>(couldata, Ec, p, L, abc, ws_periodic);
			} break;
		case 333: {
				CoeffMatrix L(2 * p);
				size_t its = LatticeOperator3D(abc, ws_periodic, L, p, 1., 42);
				L.populate_lower();
				std::cout << "LOp3Diterations " << its << std::endl;
				coulompute_periodic<true, true, true>(couldata, Ec, p, L, abc, ws_periodic);
			} break;
		case 399: {
				CoeffMatrix L(2 * p);
				L.populate_lower();
				coulompute_periodic<true, true, true>(couldata, Ec, p, L, abc, ws_periodic);
			} break;
		default: std::cout << "Unsupported algorithm " << algo << std::endl;
			 return 0;
	}
	t += timestamp();
	std::cout << "Computed Coulomb forces etc. in " << t << " seconds" << std::endl;

#ifdef COULOMB_ENERGY
	std::cout << "Coulomb Energy Ec = " << Ec << std::endl;
#endif

	long double reldeltaEc = -0.;
	if (argc == 5)
		verify_result(couldata, Ec, argv[4], reldeltaEc);

	std::cout << "COULOMB"
		<< "\t"	<< GITBRANCH << ":" << GITREVISION
		<< "\t" << vx.size()
		<< "\t" << FP_NAME
		<< "\t" << algo
		<< "\t" << NTHREADS
		<< "\t" << t
		<< std::setprecision(16)
	       	<< "\t" << Ec
		<< "\t" << reldeltaEc
		<< std::endl;

	write_output(argv[2], vF, vPhi);

	return 0;
}

}  // namespace fmsolvr

int main(int argc, char **argv)
{
    return fmsolvr::main(argc, argv);
}
