/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <vector>
#include <algorithm>
#include <cassert>
#include <fenv.h>

#include "fmsolvr.hpp"

#include "read_qxyz.hpp"
#include "write_qxyz.hpp"
#include "write_fphi.hpp"
#include "compare_with_reference.hpp"

#include "fmsolvr_tasking_config.hpp"

#include "tools/stat.hpp"

#include "parallelization/ThreadIdentifier.cpp"

namespace fmsolvr
{

const XYZ<double> e1(1, 0, 0);
const XYZ<double> e2(0, 1, 0);
const XYZ<double> e3(0, 0, 1);

namespace testdata {
#if 1
#include "qxyz_NaCl444.hpp"

    ABC<XYZ<double> > abc(e1 * boxlength, e2 * boxlength, e3 * boxlength);
    XYZ<double> reference_point(0., 0., 0.);
#elif 0
#include "qxyz_NaCl_25_25_25.hpp"

    ABC<XYZ<double> > abc(e1 * boxlength, e2 * boxlength, e3 * boxlength);
    XYZ<double> reference_point(0., 0., 0.);
#else
#include "qxyz_silicamelt.hpp"

    ABC<XYZ<double> > abc(e1 * boxlength, e2 * boxlength, e3 * boxlength);
    XYZ<double> reference_point(abc.center());
#endif
}


template <typename RealOut, typename RealIn = RealOut>
struct fake_aos4
{
    typedef RealOut Real;
    typedef size_t size_type;
    typedef const RealIn * const_pointer;
    typedef typename fmsolvr::StackedMultiReal<Real>::stacked_real4_type Real4;
    typedef Real4 value_type;

    const_pointer x, y, z, s;
    size_type n;

    fake_aos4(const_pointer x, const_pointer y, const_pointer z, const_pointer s, size_type n)
        : x(x), y(y), z(z), s(s), n(n)
    { }

    size_type size() const
    {
        return n;
    }

    Real4 operator [] (size_type i) const
    {
        return Real4(x[i], y[i], z[i], s[i]);
    }
};


template <typename Vector3, typename Vector1>
struct fake_aos31
{
    typedef typename Vector1::value_type Real;
    typedef size_t size_type;
    typedef typename fmsolvr::StackedMultiReal<Real>::stacked_real4_type Real4;
    typedef Real4 value_type;

    const Vector3 & xyz;
    const Vector1 & s;

    fake_aos31(const Vector3 & xyz, const Vector1 & s)
        : xyz(xyz), s(s)
    { }

    size_type size() const
    {
        return xyz.size();
    }

    Real4 operator [] (size_type i) const
    {
        return Real4(xyz[i], s[i]);
    }
};


inline size_t boxes_above_depth(size_t d)
{
    // sum(0, d-1, 8^i)
    return ((size_t(1) << (3 * d)) - 1 ) / 7;
}

template <template <typename> class SIMDengine>
int fmmtest(int argc, char ** argv)
{
#ifdef USE_FLOAT
    typedef float Real;
#elif defined(USE_LONGDOUBLE)
    typedef long double Real;
#elif defined(USE_FLOAT128)
    typedef __float128 Real;
#else
    typedef double Real;
#endif

    std::cout << "# FMM test"
        << " " << buildinfo<Real>()
        << std::endl;
    std::cout << "# " << SIMDinfo<typename SIMDengine<Real>::type>() << std::endl;
    std::cout << "# " << compilerinfo() << std::endl;
    std::cout << std::scientific << std::setprecision(8);

    feenableexcept(FE_DIVBYZERO | FE_INVALID | FE_OVERFLOW);

    papiwrapper<> pw;

    const char * particles_from_file = (argc > 1) ? argv[1] : NULL;
    const size_t p = get_env_int("MULTIPOLEORDER", 20);
    const ssize_t ws = get_env_int("WS", 1);
    const size_t depth = get_env_int("DEPTH", 4);
    const bool open_boundary_conditions = get_env_int("OPENBOUNDARY", 0) != 0;
    const bool periodic_a = !open_boundary_conditions;
    const bool periodic_b = !open_boundary_conditions;
    const bool periodic_c = !open_boundary_conditions;

    typedef typename fmsolvr::default_allocator<Real>::type Alloc;
    typedef fmsolvr::BoxID<fmsolvr::BoxID4D<uint32_t> > BoxID;  // with a 64-bit size_t this is sufficient upto depth 21
    typedef fmsolvr::FMMBaseCoulomb<Real, BoxID, SIMDengine, Alloc> FMMBase;
    typedef typename FMMBase::simulationbox_type SimulationBox;
    typedef typename FMMBase::real3_type Real3;
    typedef typename FMMBase::real33_type Real33;
    typedef fmsolvr::FMMHandle<FMMBase> FMMHandle;
    typedef typename FMMHandle::PrecomputedDataHandle PrecomputedDataHandle;
    typedef typename FMMHandle::PrecomputedRotationDataHandle PrecomputedRotationDataHandle;
    typedef typename FMMHandle::TreeHandle TreeHandle;
    typedef typename FMMHandle::ParticleHandle ParticleHandle;
    typedef typename FMMHandle::ParticleInteractionHandle ParticleInteractionHandle;
    typedef typename FMMHandle::P2P_type P2P_type;
    typedef typename ParticleHandle::potential_vector_type potential_vector_type;
    typedef typename ParticleHandle::field_vector_type field_vector_type;

    typedef IntraNodeContainer<FMMHandle,
                               MutexContainer<MCSLock, FMMHandle>,
                               EventDispatcher>
            IntraNodeContainer;


    typedef std::vector<Real> strength_vector_type;
    typedef std::vector<Real3> position_vector_type;


    strength_vector_type * vq = NULL;
    position_vector_type * vxyz = NULL;
    if (particles_from_file) {
        vq = new strength_vector_type;
        vxyz = new position_vector_type;
        fmsolvr::read_qxyz<Real>(particles_from_file, *vq, *vxyz);
    }

    size_t n = particles_from_file ? vxyz->size() : testdata::n;
    Real33 abc(e1, e2, e3);
    Real3 reference_center(0., 0., 0.);
    if (particles_from_file) {
        if (const char * unitbox = getenv("UNITBOX")) {
            double args[9];
            int k = sscanf(unitbox, "%lf %lf %lf %lf %lf %lf %lf %lf %lf",
                           args + 0, args + 1, args + 2,
                           args + 3, args + 4, args + 5,
                           args + 6, args + 7, args + 8);
            if (k == 1) {
                abc = Real33(Real3(args[0], 0, 0), Real3(0, args[0], 0), Real3(0, 0, args[0]));
            } else if (k == 3) {
                abc = Real33(Real3(args[0], 0, 0), Real3(0, args[1], 0), Real3(0, 0, args[2]));
            } else if (k == 9) {
                abc = Real33(Real3(args[0], args[1], args[2]),
                             Real3(args[3], args[4], args[5]),
                             Real3(args[6], args[7], args[8]));
            } else {
                std::cerr << "Error: Cannot parse UNITBOX." << std::endl;
                exit(1);
            }
        }
        if (const char * center = getenv("CENTER")) {
            if (center[0] == 'c' || center[0] == 'C') {
                reference_center = abc.center();
            } else {
                double args[3];
                int k = sscanf(center, "%lf %lf %lf", args + 0, args + 1, args + 2);
                if (k == 1) {
                    reference_center = Real3(args[0], args[0], args[0]);
                } else if (k == 3) {
                    reference_center = Real3(args[0], args[1], args[2]);
                } else {
                    std::cerr << "Error: Cannot parse CENTER." << std::endl;
                    exit(1);
                }
            }
        }
    } else {
        abc = testdata::abc;
        reference_center = Real3(testdata::reference_point);
    }
    size_t num_threads = get_env_int("NUM_THREADS", 1);

    // TODO: get the number of hardware threads

    SimulationBox simulationbox(abc, reference_center, periodic_a, periodic_b, periodic_c);
    ParticleHandle particle_handle(simulationbox, n, depth);
    ParticleInteractionHandle particle_interaction_handle(particle_handle);
    PrecomputedDataHandle fmm_precomputed(simulationbox, depth, ws, p);
    PrecomputedRotationDataHandle fmm_precomputed_p3(simulationbox, depth, ws, p);
    TreeHandle fmm_tree(depth, p);
    bool valid = fmm_precomputed.validate(depth, ws, p);
    assert(valid);


    valid = fmm_precomputed_p3.validate(depth, ws, p);
    if (!valid)
        assert(valid);
    FMMHandle fmm_handle(
            particle_interaction_handle,
            fmm_precomputed,
            fmm_precomputed_p3,
            fmm_tree,
            depth,
            ws,
            p,
            true,
            std::cout);

    IntraNodeContainer intra_node_container(fmm_handle, num_threads);

    double t = -timestamp();

    std::cout << "N=" << n << "  ws=" << ws << "  p=" << p << "  depth=" << depth;
    if (simulationbox.open_boundary_conditions())
    {
        std::cout << "  open";
    }
    else
    {
        int pdim = simulationbox.periodic_a() + simulationbox.periodic_b() + simulationbox.periodic_c();
        std::cout << "  " << pdim << "D-periodic ("
            << (simulationbox.periodic_a() ? "A" : "")
            << (simulationbox.periodic_b() ? "B" : "")
            << (simulationbox.periodic_c() ? "C" : "")
            << ")";
    }
    std::cout << " boundaries" << std::endl;
    std::cout << "box=" << simulationbox.abc() << "  ref_point=" << simulationbox.center() << std::endl;
    std::cout << "scaling factor: omega=" << fmm_handle.omega_scaling_factor()
              << " m2l_operator=" << fmm_handle.m2l_operator_scaling_factor() << std::endl;

    {
        std::cout << "# PASS 0: reorder\n";
        pw.start();
        if (particles_from_file) {
            fake_aos31<position_vector_type, strength_vector_type> particle_data(*vxyz, *vq);
            fmsolvr::distribute_particles(particle_handle, particle_data);
        } else {
            fake_aos4<Real, double> particle_data(testdata::x, testdata::y, testdata::z, testdata::q, testdata::n);
            fmsolvr::distribute_particles(particle_handle, particle_data);
        }
        pw.stop();
        pw.report(n);
    }
    if (const char *opfile = get_env_string("WRITE_PARTICLES_FILE")) {
        auto operm_particles = fmsolvr::make_vector_permutation_adaptor(particle_handle.ordered_particles(), particle_handle.particle_permutation());
        fmsolvr::write_qxyz<Real>(opfile, operm_particles);
        std::cout << "Wrote " << particle_handle.ordered_particles().size() << " ordered particles to " << opfile << std::endl;
    }
    std::cout << "boxes@depth(" << depth << ")=" << particle_handle.num_boxes_lowest()
              << "  empty_boxes=" << particle_handle.count_empty_boxes()
              << "  inactive_particles=" << particle_handle.count_inactive_particles()
              << std::endl;

    if (get_env_int("N2REF", 0) != 0)
    {
        std::cout << "# N^2:\n";
        pw.start();
        potential_vector_type potential(n, 0.);
        field_vector_type efield(n, Real3(0., 0., 0.));
        field_vector_type force(n, Real3(0., 0., 0.));
        auto particles = particle_handle.particle_adapter();
        typename ParticleHandle::outputadapter_type result(potential, efield);

        P2P_type P2P(result, particles, particles);
        P2P.overlapping_interactions(0, n, 0, n);

        Real Ec2 = 0;
        for (size_t i = 0; i < n; ++i) {
            Real q_i = particles.load_xyzs(i).s();
            force[i] = efield[i] * q_i;
            Ec2 += q_i * potential[i];
        }
        Real Ec = Ec2 / 2.;

        pw.stop();
        pw.report(n * (n - 1));

        std::cout << "Ec = " << Ec << std::endl;

        if (argc > 2) {
            fmsolvr::write_fphi<Real>(argv[2],
                                      fmsolvr::make_vector_permutation_adaptor(force, particle_handle.particle_permutation()),
                                      fmsolvr::make_vector_permutation_adaptor(potential, particle_handle.particle_permutation()));
        }
    }


    const int total_loops = get_env_int("LOOPS", 1);
    std::vector<double> timer;
    timer.reserve(total_loops);

    auto & tree_omega = fmm_handle.tree_omega();
    auto & tree_mu = fmm_handle.tree_mu();
    for (int act_loop = 0; act_loop < total_loops; ++act_loop)
    {
        // RESET
        intra_node_container.prepareForComputation();
        fmm_handle.particle_handle().zero();
        for (unsigned d = 0; d <= fmm_handle.depth(); ++d)
        {
            for (auto &mu : tree_mu.LevelRange(d))
            {
                mu.zero(p);
                mu.clearSourceCounters();
            }
            for (auto &omega : tree_omega.LevelRange(d))
            {
                omega.zero(p);
                omega.clearSourceCounters();
            }
        }

        double start_tasks = -timestamp();

        fmsolvr::pass12345(pw, fmm_handle, intra_node_container);
        timer.push_back(timestamp() + start_tasks);
        std::cout << "time: " << timer.back() << std::endl;
    }

    std::cout << "Execution statistics (#total iterations: " << total_loops << ")" << std::endl;
    printf("#STATS threads loops %-14s %-14s %-14s %-14s %-14s\n", "Min", "Q1(25%)", "Median", "Q3(75%)", "Max");
    printf("STATS  %7zd %5d %-15.6e%-15.6e%-15.6e%-15.6e%-15.6e\n",
           num_threads,
           total_loops,
           quantile(0, timer),
           quantile(0.25, timer),
           quantile(0.5, timer),
           quantile(0.75, timer),
           quantile(1, timer));

    potential_vector_type potential(n, 0.);
    field_vector_type force(n, Real3(0., 0., 0.));

    {
        std::cout << "# PASS Final: Force and Energy\n";
        pw.start();

        particle_interaction_handle.compute_forces_energy();

        const auto &potential_nf = particle_interaction_handle.potential_nf();
        const auto &potential_ff = particle_interaction_handle.potential_ff();
        const auto &force_nf = particle_interaction_handle.force_nf();
        const auto &force_ff = particle_interaction_handle.force_ff();
        Real3 sumF;

        for (size_t i = 0; i < n; ++i)
        {
            potential[i] = potential_nf[i] + potential_ff[i];
            force[i] = force_nf[i] + force_ff[i];
            sumF += force[i];
        }
        std::cout << "∑F=" << sumF << std::endl;

        pw.stop();
        pw.report(n);
    }

    int prec = std::cout.precision();
    std::cout << "Energy=" << std::setprecision(16) << particle_interaction_handle.Ec()
              << "  Ec_nf=" << particle_interaction_handle.Ec_nf()
              << " Ec_ff=" << particle_interaction_handle.Ec_ff() << std::setprecision(prec)
              << std::endl;
    particle_interaction_handle.dump_front_back_particles(
            std::cout, get_env_int("DUMP_FRONT_BACK_PARTICLES", 5));

    t += timestamp();
    std::cout << "Elapsed FMM total time: " << t << std::endl;

    auto operm_potential = fmsolvr::make_vector_permutation_adaptor(potential, particle_handle.particle_permutation());
    auto operm_force = fmsolvr::make_vector_permutation_adaptor(force, particle_handle.particle_permutation());

    if (argc > 2) {
        fmsolvr::write_fphi<Real>(argv[2], operm_force, operm_potential);
    }

    if (argc > 3) {
#if defined(HAVE_FLOAT128)
        typedef __float128 RefReal;
#else
        typedef long double RefReal;
#endif
        strength_vector_type q(n);
        auto operm_particles = fmsolvr::make_vector_permutation_adaptor(particle_handle.ordered_particles(), particle_handle.particle_permutation());
        for (size_t i = 0; i < n; ++i) {
            q[i] = operm_particles[i].s();
        }
        Real Ec = particle_interaction_handle.Ec();
        fmsolvr::compare_with_reference<RefReal>(q, operm_force, operm_potential, Ec, argv[3]);
    }

    delete vq;
    delete vxyz;

    return 0;
}

int main(int argc, char ** argv)
{
    std::string simd = get_env_string("SIMD", "");
    if (simd == "NONE")
        return fmmtest<simd_engine_none>(argc, argv);
    return fmmtest<simd_engine_default>(argc, argv);
}

}  // namespace fmsolvr

int main(int argc, char **argv)
{
    return fmsolvr::main(argc, argv);
}

// vim: et:ts=4:sw=4
