/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */
#ifndef _AnBe_FMM_compare_with_reference_hpp_
#define _AnBe_FMM_compare_with_reference_hpp_

#include "fmsolvr/xyz.hpp"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <cstdlib>
#include <cmath>

#include "read_fphi.hpp"


namespace fmsolvr {

// compute relative error, fall back to absolute if reference is zero
template <typename Tp>
Tp rel(const Tp &delta, const Tp &ref)
{
    return ref != 0 ? delta / ref: delta;
}

template <typename ReferenceReal, typename VectorQ, typename VectorFxyz, typename VectorPhi, typename Real>
ReferenceReal compare_with_reference(
        const VectorQ & vq,
        const VectorFxyz & vF,
        const VectorPhi & vPhi,
        Real Ec,
        const char* f
        )
{
    typedef XYZ<ReferenceReal> ReferenceReal3;

    std::vector<ReferenceReal3> refF;
    std::vector<ReferenceReal> refPhi;

    fmsolvr::read_fphi<ReferenceReal>(f, refF, refPhi);

    ReferenceReal Ec_ref = 0.;
    ReferenceReal3 F(-0., -0., -0.);
    ReferenceReal maxFx(-0.), maxFy(-0.), maxFz(-0.);
    ReferenceReal maxdFx(-0.), maxdFy(-0.), maxdFz(-0.);
    ReferenceReal maxrdFx(-0.), maxrdFy(-0.), maxrdFz(-0.);
    ReferenceReal maxdFlen(-0.), maxrdFlen(-0.);
    ReferenceReal maxdPhi(-0.), maxrdPhi(-0.);

    if (refF.size() != vF.size()) {
        std::cout << "Particle count mismatch: N=" << vF.size() << "  Nref=" << refF.size() << std::endl;
        return nan("");
    }

    for (size_t i = 0; i < vF.size(); ++i) {
        ReferenceReal3 F_i(vF[i].x(), vF[i].y(), vF[i].z());
        ReferenceReal3 F_i_ref(refF[i]);
        ReferenceReal F_i_ref_len(length(F_i_ref));
        ReferenceReal Phi_i(vPhi[i]);
        ReferenceReal Phi_i_ref(refPhi[i]);
        ReferenceReal q_i(vq[i]);
        Ec_ref += Phi_i_ref * q_i;
        ReferenceReal3 dF_i = F_i_ref - F_i;
        ReferenceReal dF_i_len(length(dF_i));
        ReferenceReal dPhi_i = Phi_i_ref - Phi_i;
        F += F_i;
        maxFx = std::max(std::abs(F_i.x()), maxFx);
        maxFy = std::max(std::abs(F_i.y()), maxFy);
        maxFz = std::max(std::abs(F_i.z()), maxFz);
        maxdFx = std::max(std::abs(dF_i.x()), maxdFx);
        maxdFy = std::max(std::abs(dF_i.y()), maxdFy);
        maxdFz = std::max(std::abs(dF_i.z()), maxdFz);
        maxrdFx = std::max(std::abs(rel(dF_i.x(), F_i_ref.x())), maxrdFx);
        maxrdFy = std::max(std::abs(rel(dF_i.y(), F_i_ref.y())), maxrdFy);
        maxrdFz = std::max(std::abs(rel(dF_i.z(), F_i_ref.z())), maxrdFz);
        maxdFlen = std::max(dF_i_len, maxdFlen);
        maxrdFlen = std::max(rel(dF_i_len, F_i_ref_len), maxrdFlen);
        maxdPhi = std::max(std::abs(dPhi_i), maxdPhi);
        maxrdPhi = std::max(std::abs(rel(dPhi_i, Phi_i_ref)), maxrdPhi);
        //std::cerr << i << std::scientific << std::setprecision(40) << " " << Phi_i_ref << " " << Phi_i << " " << dPhi_i << " " << dPhi_i / Phi_i_ref << " " << maxdPhi << " " << maxrdPhi << std::endl;
        //deltaF += length(dF_i);
        //deltaPhi += std::abs(dPhi_i);
#if 0
        std::cout << i << ": (" << vF[i].x() << ", " << vF[i].y() << ", " << vF[i].z() << ") " << vPhi[i]
            << " <-> (" << x << ", " << y << ", " << z << ") " << p
            << "\tdF=(" << dF_i.x << ", " << dF_i.y << ", " << dF_i.z << ")"
            << "\trdF=(" << dF_i.x() / F_i_ref.x() << ", " << dF_i.y() / F_i_ref.y() << ", " << dF_i.z() / F_i_ref.z() << ")"
            << "\t|dF|=" << length(dF_i)
            << "\t|dF|/|F|=" << length(dF_i) / length(F_i_ref)
            << "\tdPhi=" << dPhi_i
            << "\trdPhi=" << dPhi_i / Phi_i_ref
            << "\tEc_r=" << Ec_ref
            << "\tmaxrdFx=" << maxrdFx
            << "\tmaxrdFy=" << maxrdFy
            << "\tmaxrdFz=" << maxrdFz
            << "\tmaxrdPhi=" << maxrdPhi
            << std::endl;
#endif
    }
    ReferenceReal deltaEc = std::abs(Ec_ref / 2. - ReferenceReal(Ec));
    std::cout << "ERRORS"
        << std::scientific << std::setprecision(40) << std::endl
        << " sumF=(" << F.x() << ", " << F.y() << ", " << F.z() << ")" << std::endl
        << " maxFx=" << maxFx << " maxFy=" << maxFy << " maxFz=" << maxFz << std::endl
        << " maxdFx=" << maxdFx << " maxdFy=" << maxdFy << " maxdFz=" << maxdFz << std::endl
        << " maxrdFx=" << maxrdFx << " maxrdFy=" << maxrdFy << " maxrdFz=" << maxrdFz << std::endl
        << " max|dF|=" << maxdFlen << " maxr|dF|=" << maxrdFlen << std::endl
        << " maxdPhi=" << maxdPhi << " maxrdPhi=" << maxrdPhi << std::endl
        << " adEc=" << deltaEc << " rdEc=" << (deltaEc / Ec_ref)
        << std::endl;
    std::cout.unsetf(std::ios::floatfield);

    // relative energy error
    return deltaEc / Ec_ref;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
