/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */
#ifndef _AnBe_FMM_homopole_hpp_
#define _AnBe_FMM_homopole_hpp_

#include "fmsolvr/math/parity_sign.hpp"

#include <cmath>
#include <cstddef>


namespace fmsolvr {

template <typename CoefficientMatrix>
void make_unitbox_homogeneous_charge_distribution_multipole_expansion(
        CoefficientMatrix &omega, size_t p)
{
    typedef ptrdiff_t index_type;
    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type Real;

    auto fac = [](size_t n)
    {
        Real res = 1.;
        for (size_t i = 2; i <= n; ++i)
        {
            res *= Real(i);
        }
        return res;
    };

    auto h = [&](index_type m, index_type k, index_type j) {
        Real res = 0.;
        for (index_type n = 0; n <= m; n += 2)
        {
            Real denominator =
                    fac(m - n) * fac(n) * Real(2 * j + n + 1) * Real(2 * k + m + 1 - 2 * j - n);
            res += reciprocal(toggle_sign_if_odd(n / 2, denominator));
        }
        return res;
    };

    auto g = [&](index_type m, index_type k) {
        Real res = 0.;
        for (index_type j = 0; j <= k; ++j)
        {
            res += reciprocal(fac(k - j) * fac(j)) * h(m, k, j);
        }
        return res;
    };

    auto f = [&](index_type l, index_type m) {
        Real res = 0.;
        for (index_type k = 0; k <= (l - m) / 2; ++k)
        {
            Real denominator = std::pow(Real(4), Real(k)) * fac(k + m) * fac(l - m - 2 * k + 1);
            res += reciprocal(toggle_sign_if_odd(k, denominator)) * g(m, k);
        }
        return res;
    };

    omega.zero();
    for (index_type l = 0; l <= index_type(p); l += 2)
    {
        for (index_type m = 0; m <= l; m += 4)
        {
            // would compute *incorrect* (non-zero) VALUES for l%2 != 0 or m%4 != 0
            omega.upper(l, m) = fac(m) * (toggle_sign_if_odd(l, 1.0) + toggle_sign_if_odd(m, 1.0)) *
                                std::pow(Real(2), Real(-(l + m + 1))) * f(l, m);
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
