/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_abc_hpp
#define FMSOLVR_abc_hpp

#include <iostream>

namespace fmsolvr
{

template <typename Tp>
class ABC
{
public:
    typedef Tp value_type;

    ABC(const value_type &a, const value_type &b, const value_type &c) : a_(a), b_(b), c_(c)
    { }

    template <typename Tother>
    ABC(const Tother &a, const Tother &b, const Tother &c)
            : a_(value_type(a)), b_(value_type(b)), c_(value_type(c))
    { }

    template <typename ABCother>
    ABC(const ABCother &abc) : ABC(abc.a(), abc.b(), abc.c())
    { }

    value_type center() const
    {
        return (a_ + b_ + c_) * 0.5;
    }

    const value_type &a() const
    { return a_; }

    const value_type &b() const
    { return b_; }

    const value_type &c() const
    { return c_; }

    value_type operator*(const value_type &xyz) const
    {
        return value_type(
                a_.x() * xyz.x() + b_.x() * xyz.y() + c_.x() * xyz.z(),
                a_.y() * xyz.x() + b_.y() * xyz.y() + c_.y() * xyz.z(),
                a_.z() * xyz.x() + b_.z() * xyz.y() + c_.z() * xyz.z());
    }

    bool operator ==(const ABC &abc) const
    {
        return a_ == abc.a_ && b_ == abc.b_ && c_ == abc.c_;
    }

    bool operator !=(const ABC &abc) const
    {
        return !(*this == abc);
    }

private:
    value_type a_, b_, c_;
};

template <typename Tp>
std::ostream & operator << (std::ostream & os, const ABC<Tp> & abc)
{
    os << "[" << abc.a() << ", " << abc.b() << ", " << abc.c() << "]";
    return os;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
