/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_aligned_alloc_hpp
#define FMSOLVR_aligned_alloc_hpp

#include <new>
#include <stdexcept>
#include <cstdlib>
#include <cerrno>

namespace fmsolvr
{

template <class T, unsigned alignment = alignof(T)>
class aligned_allocator;

//specialize for void:
template <unsigned alignment> class aligned_allocator<void, alignment>
{
	public:
	typedef void* pointer;
	typedef const void* const_pointer;
	// reference-to-void members are impossible.
	typedef void value_type;
	template <class U> struct rebind { typedef aligned_allocator<U, alignment> other; };
};

template <class T, unsigned alignment>
class aligned_allocator : public std::allocator<T>
{
public:
	typedef size_t size_type;
	typedef std::ptrdiff_t difference_type;

	typedef T* pointer;
	typedef const T* const_pointer;

	typedef T& reference;
	typedef const T& const_reference;

	typedef T value_type;

	template <class U>
	struct rebind { typedef aligned_allocator<U, alignment> other; };

	aligned_allocator() throw() { }
	aligned_allocator(const aligned_allocator& a) throw()
		: std::allocator<T>(a) { }
	template <class U>
	aligned_allocator(const aligned_allocator<U, alignment>&) throw() { }
	~aligned_allocator() throw() { }

//%	pointer address(reference r) const noexcept { return &r; }
//%	const_pointer address(const_reference r) const noexcept { return &r; }

	pointer allocate(size_type n, aligned_allocator<void, alignof(void*)>::const_pointer /*hint*/ = 0) // space for n Ts
	{
		void * p;
		unsigned ali = alignment;
		if (ali < alignof(void *))
			ali = alignof(void *);
		int ret = posix_memalign(&p, ali, n * sizeof(T));
		if (ret == ENOMEM)
			throw std::bad_alloc();
		else if (ret == EINVAL)
			throw std::invalid_argument("invalid alignemnt");
		return pointer(p);
	}
	void deallocate(pointer p, size_type /*n*/)   // deallocate n Ts, don't destroy
	{
		free(p);
	}

#if __cplusplus >= 201103L
// C++11
	template<typename U, typename... Args>
	void construct(U* p, Args&&... args)
	{ ::new(p) U(std::forward<Args>(args)...); }

	template<typename U>
	void destroy(U* p) { p->~U(); }
#else
// 4th ed
	void construct(pointer p, const T& val) { new(p) T(val); }          // initialize *p by val
	void destroy(pointer p) { p->~T(); }     // destroy *p but don't deallocate
#endif

//%	size_type max_size() const noexcept;

};

template<class T1, class T2, unsigned alignment>
bool operator==(const aligned_allocator<T1, alignment>&, const aligned_allocator<T2, alignment>&) noexcept
{
	return true;
}

template<class T1, class T2, unsigned alignment>
bool operator!=(const aligned_allocator<T1, alignment>&, const aligned_allocator<T2, alignment>&) noexcept
{
	return false;
}

}  // namespace fmsolvr

#endif
