/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_complex_vec_hpp
#define FMSOLVR_complex_vec_hpp

#include "fmsolvr/math/Complex.hpp"

#include <complex>

#define AVOID__muldc3



namespace fmsolvr
{

// This is something like Complex<Array<Real>>, but *not* anything like Array<Complex<Real>>.
template <typename ComplexTpVec, typename ComplexTpScalar>
class complex_vec : public ComplexTpVec
{
    typedef ComplexTpVec Base;
    typedef complex_vec Self;
    typedef typename ComplexTpScalar::value_type scalar_value_type;

public:
    typedef typename Base::value_type value_type;

//    using Base::Base;  // inherit constructors

    complex_vec() : Base()
    { }

    /*explicit*/ complex_vec(const ComplexTpVec &z)
            : Base(z)
    { }

    /*explicit*/ complex_vec(const value_type &re)
            : Base(re)
    { }

    complex_vec(const value_type &re, const value_type &im)
            : Base(re, im)
    { }

    explicit complex_vec(const ComplexTpScalar & z)
            : Base(broadcast(z.real()), broadcast(z.imag()))
    { }

    explicit complex_vec(const scalar_value_type &re)
            : Self(ComplexTpScalar(re))
    { }

    complex_vec(const scalar_value_type &re, const scalar_value_type &im)
            : Self(ComplexTpScalar(re, im))
    { }

    Self &operator*=(const ComplexTpScalar &z)
    {
        Self z_broadcast(z);  // broadcast z from Complex<Scalar> to Complex<Array<Scalar>>
        this->Base::operator*=(z_broadcast);
        return *this;
    }

    static value_type broadcast(scalar_value_type x)
    {
        return value_type(x);
    }
};

template <typename ComplexTpVec, typename ComplexTpScalar>
inline complex_vec<ComplexTpVec, ComplexTpScalar> operator*(const complex_vec<ComplexTpVec, ComplexTpScalar> &a, const ComplexTpScalar &b)
{
    complex_vec<ComplexTpVec, ComplexTpScalar> b_broadcast(b.real(), b.imag());
    return a * b_broadcast;
}

template <typename ComplexTpVec, typename ComplexTpScalar>
inline complex_vec<ComplexTpVec, ComplexTpScalar> operator*(const ComplexTpScalar &a, const complex_vec<ComplexTpVec, ComplexTpScalar> &b)
{
    complex_vec<ComplexTpVec, ComplexTpScalar> a_broadcast(a.real(), a.imag());
    return a_broadcast * b;
}

template <typename ComplexTpVec, typename ComplexTpScalar>
inline complex_vec<ComplexTpVec, ComplexTpScalar> operator*(const complex_vec<ComplexTpVec, ComplexTpScalar> &a, const typename ComplexTpScalar::value_type &b)
{
    typedef complex_vec<ComplexTpVec, ComplexTpScalar> Self;
    return a * Self::broadcast(b);
}

template <typename ComplexTpVec, typename ComplexTpScalar>
inline complex_vec<ComplexTpVec, ComplexTpScalar> operator*(const typename ComplexTpScalar::value_type &a, const complex_vec<ComplexTpVec, ComplexTpScalar> &b)
{
    typedef complex_vec<ComplexTpVec, ComplexTpScalar> Self;
    return Self::broadcast(a) * b;
}



template <typename Tp>
struct COMPLEX_GENERATOR1
{
    typedef std::complex<Tp> type;
};

#if defined(AVOID__muldc3)
template <>
struct COMPLEX_GENERATOR1<float>
{
    typedef Complex<float> type;
};

template <>
struct COMPLEX_GENERATOR1<double>
{
    typedef Complex<double> type;
};
#endif


template <typename TpVec, typename Tp>
struct COMPLEX_GENERATOR2
{
    typedef complex_vec<typename COMPLEX_GENERATOR1<TpVec>::type, typename COMPLEX_GENERATOR1<Tp>::type> type;
};


template <typename TpV, typename TpS = void>
struct COMPLEX_GENERATOR : public COMPLEX_GENERATOR2<TpV, TpS>
{ };

template <typename Tp>
struct COMPLEX_GENERATOR<Tp, void> : public COMPLEX_GENERATOR1<Tp>
{ };

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
