/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_CONTAINER_containermulticube_hpp
#define FMSOLVR_CONTAINER_containermulticube_hpp

#include "fmsolvr/container/containercube.hpp"

#include <cassert>
#include <vector>


namespace fmsolvr {

template <typename Generator>
class ContainerMultiCube
{
    typedef typename Generator::cube_type cube_type;
    typedef typename Generator::value_type value_type;

    typedef std::vector<cube_type *> cube_vector;
    typedef std::vector<cube_vector> cube_vector_vector;
    typedef std::vector<value_type *> value_vector;
public:
    typedef const cube_type & const_reference;
    typedef size_t size_type;

    ContainerMultiCube(Generator gen, size_type d = 0, size_type e = 0) : m_gen(gen)
    {
        if (e > 0)
            for (size_type dd = 1; dd <= d; ++dd)
                operator () (dd, e);
    }

    ~ContainerMultiCube()
    {
        for (auto it = m_cubes.begin(); it != m_cubes.end(); ++it) {
            for (auto it2 = it->begin(); it2 != it->end(); ++it2)
                delete *it2;
        }
        for (auto it = m_allocated.begin(); it != m_allocated.end(); ++it)
            delete *it;
    }

    const_reference operator () (size_type d, size_type e)
    {
        assert(d > 0 && e > 0);
        if (m_cubes.size() < d)
            m_cubes.resize(d);
        if (m_cubes[d - 1].size() < e)
            m_cubes[d - 1].resize(e);
        if (!m_cubes[d - 1][e - 1])
            m_cubes[d - 1][e - 1] = m_gen(d, e, std::back_inserter(m_allocated));
        return *m_cubes[d - 1][e - 1];
    }

private:
    Generator m_gen;
    cube_vector_vector m_cubes;
    value_vector m_allocated;
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
