/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_CONTAINER_pyramidal_array_hpp
#define FMSOLVR_CONTAINER_pyramidal_array_hpp

#include "fmsolvr/container/simple_vector.hpp"

#include <cstddef>
#include <cassert>
#include <iostream>

namespace fmsolvr
{

#if 1
#define PYRAMIDAL_ARRAY_RANGE_ASSERT(cond) assert(cond)
#else
#define PYRAMIDAL_ARRAY_RANGE_ASSERT(cond)
#endif


// pyramidal matrix with top (0,0,0) and bottom corners (p,0,0), (p,p,0), (p,0,p), (p,p,p)
// optimized for access in dimension order (first, second, third)
template <typename T, typename Allocator = std::allocator<T> >
class QuarterPyramid
{
    QuarterPyramid(const QuarterPyramid &) = delete;
    QuarterPyramid & operator=(const QuarterPyramid &) = delete;

public:
    typedef T value_type;
    typedef typename Allocator::template rebind<value_type>::other allocator_type;
    typedef size_t size_type;
    typedef ptrdiff_t index_type;
    typedef const value_type * const_k_iterator;

    QuarterPyramid(index_type p) : offsets_(p + 1), p_(p)
    {
        array_ = allocator_.allocate(size());
        for (size_type i = 0; i < size(); ++i)
            allocator_.construct(&array_[i], value_type());
        for (index_type l = 0; l <= p_; ++l)
            offsets_[l] = size_above(l);
    }

    ~QuarterPyramid()
    {
        for (size_type i = 0; i < size(); ++i)
            allocator_.destroy(&array_[i]);
        allocator_.deallocate(array_, size());
    }

    size_type p() const
    {
        return p_;
    }

    // 0 <= l <= p, 0 <= m <= l, 0 <= k <= l
    const value_type & operator () (index_type l, index_type m, index_type k) const
    {
        return array_[offset(l, m, k)];
    }

    // 0 <= l <= p, 0 <= m <= l, 0 <= k <= l
    value_type & operator () (index_type l, index_type m, index_type k)
    {
        return array_[offset(l, m, k)];
    }

    // 0 <= l <= p, 0 <= m <= l, 0 <= k <= l
    // the returned pointer allows direct access to the elements
    // (l, m, 0), ..., (l, m, l) which are stored consecutively in memory,
    // returns a pointer to the k'th element in this array
    const_k_iterator get_k_iterator(index_type l, index_type m, index_type k) const
    {
        return array_ + offset(l, m, k);
    }

private:
    value_type * array_;
    simple_vector<index_type, Allocator> offsets_;
    allocator_type allocator_;
    index_type p_;

    size_type size_above(index_type p) const
    {
        // sum(i=1, i<=p, i*i)
        return (p * (p + 1) * (2 * p + 1)) / 6;
    }

    size_type size() const
    {
        return size(p_);
    }

    size_type size(index_type p) const
    {
        return size_above(p + 1);
    }

    // 0 <= l <= p, 0 <= m <= l, 0 <= k <= l
    index_type offset(index_type l, index_type m, index_type k) const
    {
        PYRAMIDAL_ARRAY_RANGE_ASSERT(0 <= l && l <= p_ && 0 <= m && m <= l && 0 <= k && k <= l);
        // there are l+1 elements per dimension
        return offsets_[l] + (l + 1) * m + k;
    }
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
