/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_CONTAINER_simple_vector_hpp
#define FMSOLVR_CONTAINER_simple_vector_hpp

#include <cstddef>
#include <memory>

namespace fmsolvr
{

template <typename Tp, typename Allocator = std::allocator<Tp> >
class simple_vector
{
    simple_vector(const simple_vector &) = delete;
    simple_vector & operator = (const simple_vector &) = delete;

public:
    typedef Tp value_type;
    typedef typename Allocator::template rebind<value_type>::other allocator_type;
    typedef size_t size_type;
    typedef Tp * pointer;
    typedef Tp & reference;
    typedef const Tp * const_pointer;
    typedef const Tp & const_reference;
    typedef pointer iterator;
    typedef const_pointer const_iterator;

    simple_vector(size_type n) : size_(n)
    {
        array_ = allocator_.allocate(size());
        for (size_type i = 0; i < size(); ++i)
            allocator_.construct(&array_[i]);
    }

    template <typename ...Args>
    simple_vector(size_type n, Args... args) : size_(n)
    {
        array_ = allocator_.allocate(size());
        for (size_type i = 0; i < size(); ++i) {
            // DO NOT USE std::forward HERE!
            // It is undefined behavior to move an object twice
            allocator_.construct(&array_[i], args...);
        }

    }

    ~simple_vector()
    {
        for (size_type i = 0; i < size(); ++i)
            allocator_.destroy(&array_[i]);
        allocator_.deallocate(array_, size());
    }

    size_type size() const
    {
        return size_;
    }

    const_iterator cbegin() const
    {
        return array_;
    }

    const_iterator begin() const
    {
        return cbegin();
    }

    iterator begin()
    {
        return array_;
    }

    const_iterator cend() const
    {
        return array_ + size();
    }

    const_iterator end() const
    {
        return cend();
    }

    iterator end()
    {
        return array_ + size();
    }

    const_reference operator [] (size_type i) const
    {
        return array_[i];
    }

    reference operator [] (size_type i)
    {
        return array_[i];
    }

    const_pointer data() const
    {
        return array_;
    }

    pointer data()
    {
        return array_;
    }

private:
    pointer array_;
    allocator_type allocator_;
    size_type size_;
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
