/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_CONTAINER_triangular_array_hpp
#define FMSOLVR_CONTAINER_triangular_array_hpp

#include <cassert>
#include <iostream>

namespace fmsolvr
{

#if 1
#define TRIANGULAR_ARRAY_RANGE_ASSERT(cond) assert(cond)
#else
#define TRIANGULAR_ARRAY_RANGE_ASSERT(cond)
#endif

// accessors:
// T.upper(l, m)
// * the fastest and preferred way of element access
// * valid index range: 0 <= l <= p, 0 <= m <= l
// * no index checking
// * no index-dependent branching
// * no on-the-fly computation
// * returns a (modifiable) reference to actual storage

// triangular matrix with corners (0,0), (-p,-p), (p,p)
// optimized for column-major access
// only the elements in the upper triangle (0,0), (p,0), (p,p) are actually
// stored and efficiently accessible
template <typename T, typename ComputeLower, typename Allocator = std::allocator<T> >
class UpperTriangularArray
{
    UpperTriangularArray(const UpperTriangularArray &) = delete;
    UpperTriangularArray & operator=(const UpperTriangularArray &) = delete;
public:
    typedef T value_type;
    typedef ComputeLower compute_lower_functor;
    typedef typename Allocator::template rebind<value_type>::other allocator_type;
    typedef size_t size_type;
    typedef ssize_t index_type;
    typedef const value_type * const_column_iterator;

    UpperTriangularArray(index_type p) : p_(p)
    {
        array = allocator.allocate(size());
        for (size_type i = 0; i < size(); ++i)
            allocator.construct(&array[i], value_type());
    }

    ~UpperTriangularArray()
    {
        for (size_type i = 0; i < size(); ++i)
            allocator.destroy(&array[i]);
        allocator.deallocate(array, size());
    }

    size_type p() const
    {
        return p_;
    }

    void fill(const value_type & value = value_type())
    {
        std::fill(array, array + size(), value);
    }

    void zero()
    {
        fill();
    }

    void zero(index_type p)
    {
        std::fill(array, array + size(p), value_type());
    }

    // 0 <= l <= p, 0 <= m <= l
    const value_type & upper(index_type l, index_type m) const
    {
        return array[offset(l, m)];
    }

    // 0 <= l <= p, 0 <= m <= l
    value_type & upper(index_type l, index_type m)
    {
        return array[offset(l, m)];
    }

    // 0 <= l <= p, -l <= m <= -1
    value_type lower(index_type l, index_type m) const
    {
        return compute_lower_functor()(l, m, *this);
    }

    // 0 <= l <= p, -l <= m <= l
    value_type get(index_type l, index_type m) const
    {
        return (m < 0) ? lower(l, m) : upper(l, m);
    }

    void invalidate_lower()
    { }  // no-op

    void populate_lower()
    { }  // no-op

    bool require_lower() const
    { return true; }  // no-op

    const_column_iterator get_column_iterator_upper(index_type l, index_type m) const
    {
        return array + offset(l, m);
    }

    void operator+=(const UpperTriangularArray & right)
    {
        assert(p() == right.p());

        for (int i = 0; i != size(); ++i) {
            array[i] += right.array[i];
        }
    }

private:
    value_type * array;
    allocator_type allocator;
    index_type p_;

    size_type size() const
    {
        return size(p_);
    }

    size_type size(index_type p) const
    {
        // sum(1..p+1): there are (p+1) elements in the longest row/col
        return ((p + 1) * (p + 2)) / 2;
    }

    // 0 <= l <= p, 0 <= m <= l
    index_type offset(index_type l, index_type m) const
    {
        TRIANGULAR_ARRAY_RANGE_ASSERT((0 <= l) && (l <= p_) && (0 <= m) && (m <= l));
        return (l * (l + 1)) / 2 + m;
    }
};


// triangular matrix with corners (0,0), (-p,-p), (p,p)
// optimized for row-major access
// only the elements in the upper triangle (0,0), (p,0), (p,p) are actually
// stored and efficiently accessible
template <typename T, typename ComputeLower, typename Allocator = std::allocator<T> >
class UpperTriangularArrayRowMajor
{
    UpperTriangularArrayRowMajor(const UpperTriangularArrayRowMajor &) = delete;
    UpperTriangularArrayRowMajor & operator=(const UpperTriangularArrayRowMajor &) = delete;
public:
    typedef T value_type;
    typedef ComputeLower compute_lower_functor;
    typedef typename Allocator::template rebind<value_type>::other allocator_type;
    typedef size_t size_type;
    typedef ssize_t index_type;

    typedef T* row_iterator;

    UpperTriangularArrayRowMajor(index_type p) : p_(p)
    {
        array = allocator.allocate(size());
        for (size_type i = 0; i < size(); ++i)
            allocator.construct(&array[i], value_type());
    }

    ~UpperTriangularArrayRowMajor()
    {
        for (size_type i = 0; i < size(); ++i)
            allocator.destroy(&array[i]);
        allocator.deallocate(array, size());
    }

    size_type p() const
    {
        return p_;
    }

    void fill(const value_type & value = value_type())
    {
        std::fill(array, array + size(), value);
    }

    void zero()
    {
        fill();
    }

    void zero(index_type p)
    {
        std::fill(array, array + size(p), value_type());
    }

    // 0 <= l <= p, 0 <= m <= l
    const value_type & upper(index_type l, index_type m) const
    {
        return array[offset(l, m)];
    }

    // 0 <= l <= p, 0 <= m <= l
    value_type & upper(index_type l, index_type m)
    {
        return array[offset(l, m)];
    }

    // 0 <= l <= p, -l <= m <= -1
    value_type lower(index_type l, index_type m) const
    {
        return compute_lower_functor()(l, m, *this);
    }

    // 0 <= l <= p, -l <= m <= l
    value_type get(index_type l, index_type m) const
    {
        return (m < 0) ? lower(l, m) : upper(l, m);
    }

    row_iterator get_row_iterator(index_type l, index_type m)
    {
        return array + offset(l, m);
    }

    void invalidate_lower()
    { }  // no-op

    void populate_lower()
    { }  // no-op

    bool require_lower() const
    { return true; }  // no-op

private:
    value_type * array;
    allocator_type allocator;
    index_type p_;

    size_type size() const
    {
        return size(p_);
    }

    size_type size(index_type p) const
    {
        // sum(1..p+1): there are (p+1) elements in the longest row/col
        return ((p + 1) * (p + 2)) / 2;
    }

    // 0 <= l <= p, 0 <= m <= l
    index_type offset(index_type l, index_type m) const
    {
        TRIANGULAR_ARRAY_RANGE_ASSERT((0 <= l) && (l <= p_) && (0 <= m) && (m <= l));
        return m * (2 * p_ - m + 1) / 2 + l;
    }
};

// triangular matrix with corners (0,0), (-p,-p), (p,p)
// optimized for row-major access
template <typename T, typename ComputeLower, typename Allocator = std::allocator<T> >
class TriangularArrayRowMajor
{
    TriangularArrayRowMajor(const TriangularArrayRowMajor &) = delete;
    TriangularArrayRowMajor & operator=(const TriangularArrayRowMajor &) = delete;
public:
    typedef T value_type;
    typedef ComputeLower compute_lower_functor;
    typedef typename Allocator::template rebind<value_type>::other allocator_type;
    typedef size_t size_type;
    typedef ssize_t index_type;

    typedef T* row_iterator;

    TriangularArrayRowMajor(index_type p) : p_(p)
    {
        array = allocator.allocate(size());
        for (size_type i = 0; i < size(); ++i)
            allocator.construct(&array[i], value_type());
    }

    ~TriangularArrayRowMajor()
    {
        for (size_type i = 0; i < size(); ++i)
            allocator.destroy(&array[i]);
        allocator.deallocate(array, size());
    }

    size_type p() const
    {
        return p_;
    }

    void fill(const value_type & value = value_type())
    {
        std::fill(array, array + size(), value);
    }

    void zero()
    {
        fill();
    }

    void zero(index_type p)
    {
        std::fill(array, array + size(p), value_type());
    }

    // 0 <= l <= p, -l <= m <= l
    value_type & get(index_type l, index_type m) const
    {
        return array[offset(l,m)];
    }

    // 0 <= l <= p, 0 <= m <= l
    const value_type & operator () (index_type l, index_type m) const
    {
        return get(l, m);
    }

    // 0 <= l <= p, 0 <= m <= l
    value_type & operator () (index_type l, index_type m)
    {
        return get(l, m);
    }

    row_iterator get_row_iterator(index_type l, index_type m)
    {
        return array + offset(l, m);
    }

    void populate_lower();
    // UNIMPLEMENTED!

private:
    value_type * array;
    allocator_type allocator;
    index_type p_;

    size_type size() const
    {
        return size(p_);
    }

    size_type size(index_type p) const
    {
        // sum(1..p+1): there are (p+1) elements in the longest row/col
        return (p + 1) * (p + 1);
    }

    // 0 <= l <= p, 0 <= m <= l
    index_type offset(index_type l, index_type m) const
    {
        TRIANGULAR_ARRAY_RANGE_ASSERT((0 <= l) && (l <= p_) && (-l <= m) && (m <= l));
        if (0 <= m)
            return m * (2 * p_ - m + 1)/2+l;
        else
            return (p_+1)*(p_+1)-((p_+m+1)*(p_+m+2))/2+l+m;

    }
};


// triangular matrix with corners (0,0), (-p,-p), (p,p)
// for elements at positions (l,m) with 0 <= l <= p, -l <= m <= l
// optimized for column-major access
template <typename T, typename ComputeLower, typename Allocator = std::allocator<T> >
class TriangularArray
{
    TriangularArray(const TriangularArray &) = delete;
    TriangularArray & operator=(const TriangularArray &) = delete;
public:
    typedef T value_type;
    typedef ComputeLower compute_lower_functor;
    typedef typename Allocator::template rebind<value_type>::other allocator_type;
    typedef size_t size_type;
    typedef ssize_t index_type;
    typedef const T * const_column_iterator;

    TriangularArray(index_type p) : p_(p), lower_is_valid_(true)
    {
        array = allocator.allocate(size());
        for (size_type i = 0; i < size(); ++i)
            allocator.construct(&array[i], value_type());
    }

    ~TriangularArray()
    {
        for (size_type i = 0; i < size(); ++i)
            allocator.destroy(&array[i]);
        allocator.deallocate(array, size());
    }

    size_type p() const
    {
        return p_;
    }

    void fill(const value_type & value = value_type())
    {
        std::fill(array, array + size(), value);
    }

    void zero()
    {
        fill();
    }

    void zero(index_type p)
    {
        std::fill(array, array + size(p), value_type());
    }

    // 0 <= l <= p, 0 <= m <= l
    const value_type & upper(index_type l, index_type m) const
    {
        return get(l, m);
    }

    // 0 <= l <= p, 0 <= m <= l
    value_type & upper(index_type l, index_type m)
    {
        return get(l, m);
    }

    // 0 <= l <= p, -l <= m <= -1
    const value_type &lower(index_type l, index_type m) const
    {
        return get(l, m);
    }

    // 0 <= l <= p, -l <= m <= l
    const value_type & get(index_type l, index_type m) const
    {
        return array[offset(l, m)];
    }

    // 0 <= l <= p, -l <= m <= l
    value_type & get(index_type l, index_type m)
    {
        return array[offset(l, m)];
    }

    void invalidate_lower()
    {
        lower_is_valid_ = false;
    }

    void populate_lower()
    {
        compute_lower_functor compute_lower;
        for (index_type l = 1; l <= p_; ++l)
            for (index_type m = -1; m >= -l; --m)
                array[offset(l, m)] = compute_lower(l, m, *this);
        lower_is_valid_ = true;
    }

    bool require_lower() const
    {
        assert(lower_is_valid_);
        return lower_is_valid_;
    }

    const_column_iterator get_column_iterator_upper(index_type l, index_type m) const
    {
        return array + offset(l, m);
    }

    const_column_iterator get_column_iterator(index_type l, index_type m) const
    {
        return array + offset(l, m);
    }

private:
    value_type * array;
    allocator_type allocator;
    index_type p_;
    bool lower_is_valid_;

    size_type size() const
    {
        return size(p_);
    }

    size_type size(index_type p) const
    {
        // upper (0,0), (p,0), (p,p): sum(1..p+1)
        // lower (-1,-1), (-p,-p), (p,-1): sum(1..p)
        return (p + 1) * (p + 1);
    }

    // 0 <= l <= p, 0 <= m <= l
    index_type offset(index_type l, index_type m) const
    {
        TRIANGULAR_ARRAY_RANGE_ASSERT((0 <= l) && (l <= p_) && (-l <= m) && (m <= l));
        return l * l + l + m;
    }
};


template <typename CoefficientMatrixIn,
    typename CoefficientMatrixOut>
void copy_upper_triangle(
        const CoefficientMatrixIn & in,
        CoefficientMatrixOut & out,
        const size_t p)
{
    out.invalidate_lower();

    for (ssize_t l = 0; l <= ssize_t(p); ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            out.upper(l, m) = in.upper(l, m);
        }
    }
}


template <typename CoefficientMatrixIn, typename CoefficientMatrixOut, typename Scale>
void copy_scale_upper_triangle(
        const CoefficientMatrixIn &in,
        CoefficientMatrixOut &out,
        const size_t p,
        const Scale scale_factor)
{
    out.invalidate_lower();

    for (ssize_t l = 0; l <= ssize_t(p); ++l)
    {
        for (ssize_t m = 0; m <= l; ++m)
        {
            out.upper(l, m) = in.upper(l, m) * scale_factor;
        }
    }
}


template <typename CoefficientMatrix>
void dump_upper_triangle(const CoefficientMatrix & a, size_t p)
{
    std::cout << std::scientific;
    for (size_t m = p; m <= p; --m) {
        for (size_t l = 0; l <= p; ++l) {
            if (m <= l)
                std::cout << a.upper(l, m) << " ";
            else
                std::cout << "() ";
        }
        std::cout << std::endl;
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
