/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_dipole_compensation_hpp
#define FMSOLVR_dipole_compensation_hpp

#include <algorithm>

#include "fmsolvr/math/linear_algebra.hpp"
#include "fmsolvr/operator/particle2multipole.hpp"
#include "fmsolvr/operator/particle2local.hpp"


namespace fmsolvr {

template <typename SimulationBox, typename Callable>
class fake_particle_generator
{
    typedef typename SimulationBox::Real Real;
    typedef typename SimulationBox::Real3 Real3;

    const SimulationBox & simulationbox_;
    const Callable &callback_;

public:
    fake_particle_generator(const SimulationBox &simbox, const Callable &callback)
            : simulationbox_(simbox), callback_(callback)
    { }

    void generate(size_t ws, Real q0, Real qa, Real qb, Real qc)
    {
        gen(q0, -ws, -ws, -ws, -ws, -ws, -ws);
        gen(qa, ws+1, ws+1, -ws, +ws, -ws, +ws);
        gen(qb, -ws, +ws, ws+1, ws+1, -ws, +ws);
        gen(qc, -ws, +ws, -ws, +ws, ws+1, ws+1);
        gen(q0 + qa, -ws+1, +ws, -ws, -ws, -ws, -ws);
        gen(q0 + qb, -ws, -ws, -ws+1, +ws, -ws, -ws);
        gen(q0 + qc, -ws, -ws, -ws, -ws, -ws+1, +ws);
        gen(q0 + qa + qb, -ws+1, +ws, -ws+1, +ws, -ws, -ws);
        gen(q0 + qa + qc, -ws+1, +ws, -ws, -ws, -ws+1, +ws);
        gen(q0 + qb + qc, -ws, -ws, -ws+1, +ws, -ws+1, +ws);
    }

private:
    void gen(Real q, int alo, int ahi, int blo, int bhi, int clo, int chi)
    {
        Real3 shift = simulationbox_.reference_corner();
        if (!simulationbox_.periodic_a()) {
            alo = std::max(0, alo);
            ahi = std::min(0, ahi);
        }
        if (!simulationbox_.periodic_b()) {
            blo = std::max(0, blo);
            bhi = std::min(0, bhi);
        }
        if (!simulationbox_.periodic_c()) {
            clo = std::max(0, clo);
            chi = std::min(0, chi);
        }
        for (int i = alo; i <= ahi; ++i)
            for (int j = blo; j <= bhi; ++j)
                for (int k = clo; k <= chi; ++k) {
                    Real3 ijk(i, j, k);
                    Real3 pos = simulationbox_.abc() * ijk + shift;
                    callback_(pos, q);
                }
    }
};

template <typename SimulationBox, typename Real, typename Callable>
void generate_dipole_compensating_particles(
        const SimulationBox &simbox,
        size_t ws,
        Real q0,
        Real qa,
        Real qb,
        Real qc,
        const Callable &callback)
{
    fake_particle_generator<SimulationBox, Callable> gfp(simbox, callback);
    gfp.generate(ws, q0, qa, qb, qc);
}

template <typename Real, typename Real3, typename Real33>
void compute_dipole_compensating_charges(const Real3 &pdp, const Real33 &abc, Real &q0, Real &qa, Real & qb, Real & qc)
{
    linear_algebra::SqMatrix<Real> A(4);
	Real3 origin(0., 0., 0.);
    A[0][0] = A[0][1] = A[0][2] = A[0][3] = 1;
    A[1][0] = origin.x();
    A[1][1] = abc.a().x();
    A[1][2] = abc.b().x();
    A[1][3] = abc.c().x();
    A[2][0] = origin.y();
    A[2][1] = abc.a().y();
    A[2][2] = abc.b().y();
    A[2][3] = abc.c().y();
    A[3][0] = origin.z();
    A[3][1] = abc.a().z();
    A[3][2] = abc.b().z();
    A[3][3] = abc.c().z();
    Real dipole[4] = { 0, -pdp.x(), -pdp.y(), -pdp.z() };
    Real fiqtive[4];
    if (!linear_algebra::solve(4, A, fiqtive, dipole))
            throw "eqn system does not have a unique solution";
    q0 = fiqtive[0];
    qa = fiqtive[1];
    qb = fiqtive[2];
    qc = fiqtive[3];
    //linear_algebra::dump(4, A, fiqtive, dipole);
}

template <typename FMMHandle>
class dipole_compensation
{
    typedef typename FMMHandle::real3_type Real3;
    typedef typename FMMHandle::real_type Real;
    typedef typename FMMHandle::size_type size_type;
    typedef typename FMMHandle::simulationbox_type simulationbox_type;

public:
    dipole_compensation()
    { }

    dipole_compensation(const FMMHandle &fmm_handle, Real3 &dipole)
    {
        compute_dipole_compensating_charges(
                dipole, fmm_handle.simulationbox().abc(), q0_, qa_, qb_, qc_);
    }

    template <typename Omega>
    void add_compensating_omega(const FMMHandle &fmm_handle, Omega &omega)
    {
        const auto &abc = fmm_handle.simulationbox().abc();
        const auto &div_lut_p = fmm_handle.div_lut_p(fmm_handle.p());
        Real3 origin(0., 0., 0.);
        Real3 m = abc.center();
        P2M(q0_ * fmm_handle.omega_scaling_factor(), origin, omega, m, div_lut_p);
        P2M(qa_ * fmm_handle.omega_scaling_factor(), abc.a(), omega, m, div_lut_p);
        P2M(qb_ * fmm_handle.omega_scaling_factor(), abc.b(), omega, m, div_lut_p);
        P2M(qc_ * fmm_handle.omega_scaling_factor(), abc.c(), omega, m, div_lut_p);
    }

    template <typename Mu>
    void add_compensating_mu(const FMMHandle &fmm_handle, Mu &mu)
    {
        Real3 expn_root = fmm_handle.simulationbox().center();
        size_type p = fmm_handle.p();
        auto lambda = [&](Real3 xyz, Real q) { P2L(q, xyz, mu, expn_root, p); };
        generate_dipole_compensating_particles(
                fmm_handle.simulationbox(), fmm_handle.ws(), q0_, qa_, qb_, qc_, lambda);
    }

    Real q0() const
    { return q0_; }

    Real qa() const
    { return qa_; }

    Real qb() const
    { return qb_; }

    Real qc() const
    { return qc_; }

private:
    Real q0_ = 0., qa_ = 0., qb_ = 0., qc_ = 0.;
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
