/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_distribute_particles_hpp
#define FMSOLVR_distribute_particles_hpp

#include "fmsolvr/math/change_of_basis.hpp"
#include "fmsolvr/util/envtools.hpp"

#include <cmath>
#include <cassert>
#include <vector>
#include <algorithm>
#include <functional>
#include <iostream>



namespace fmsolvr {

namespace {

template <typename Pair>
struct compare_first : public std::less<Pair>
{
    bool operator () (const Pair & a, const Pair & b)
    {
        return a.first < b.first;
    }
};

template <typename SfcIndex, typename SiteFormIndex>
class boxSiteForm
{
    typedef boxSiteForm<SfcIndex, SiteFormIndex> Self;

public:
    boxSiteForm() : sfc_(0) , site_(0), form_(0)
    { }

    boxSiteForm(SfcIndex sfc, SiteFormIndex site, SiteFormIndex form)
            : sfc_(sfc), site_(site), form_(form)
    { }

    SfcIndex sfcindex() const
    { return sfc_; }

    SiteFormIndex site() const
    { return site_; }

    SiteFormIndex form() const
    { return form_; }

    bool operator<(const Self &b) const
    {
        return (sfcindex() < b.sfcindex()) ||
               ((sfcindex() == b.sfcindex()) &&
                ((site() < b.site()) || ((site() == b.site()) && (form() < b.form()))));
    }

private:
    SfcIndex sfc_;
    SiteFormIndex site_, form_;
};

}  // anonymous namespace


// no charge? no interactions
template <typename Particle>
bool is_chargeless(Particle ptcl)
{
    return (ptcl.s() == 0);
}

template <
    typename ParticleHandle,
    typename InputParticleVector>
void distribute_particles(
        ParticleHandle & particle_handle,
        const InputParticleVector & input)
{
    typedef typename ParticleHandle::real_type Real;
    typedef typename ParticleHandle::real3_type Real3;
    typedef typename ParticleHandle::real33_type Real33;
    typedef typename InputParticleVector::value_type particle_type;
    typedef typename ParticleHandle::boxid_type BoxID;
    typedef typename BoxID::sfcindex_type sfcindex_type;
    typedef typename BoxID::dimension_type dimension_type;
    typedef boxSiteForm<sfcindex_type, int> sortkey_type;
    typedef std::pair<sfcindex_type, particle_type> payload;
    typedef std::pair<sortkey_type, payload> sort_pair;
    typedef std::vector<sort_pair> sort_pair_vector;

    const auto & simbox = particle_handle.simulationbox();
    auto & permutation = particle_handle.particle_permutation();
    auto & boxsfcindex = particle_handle.boxsfcindex();
    auto & output = particle_handle.ordered_particles();
    auto & box_particle_offset = particle_handle.box_particle_offset();
    const int fmm_depth = particle_handle.depth();
    const size_t num_boxes_lowest = particle_handle.num_boxes_lowest();
    const size_t extra_boxes_lowest = particle_handle.extra_boxes_lowest();

    Real33 convert_to_abc = change_of_basis_from_standard_basis(simbox.abc());
    Real3 ref_corner = simbox.reference_corner();
    dimension_type box_dim_lowest = dimension_type(1) << fmm_depth;
    Real scale = box_dim_lowest;
    const bool wrap_particles = !simbox.open_boundary_conditions() && get_env_int("WRAP_PARTICLES", 1);
    const bool skip_chargeless_particles = particle_handle.extra_boxes_lowest() > 0 && get_env_int("SKIP_CHARGELESS_PARTICLES", 1);
    size_t n = input.size();
    assert(particle_handle.n() == n);
    sort_pair_vector tmp(n);
    Real zero(0);
    Real one(1);
    Real nearly_one = std::nextafter(one, zero);  // a little bit smaller than one
    size_t renormalized = 0;
    assert(zero < nearly_one && nearly_one < one);
    for (size_t i = 0; i < n; ++i) {
        particle_type ptcl = input[i];
        Real3 normalized = convert_to_abc * (Real3(ptcl) - ref_corner);
        if (wrap_particles) {
            // FIXME: missing support for 1D/2D periodicity
            Real pdx = std::floor(normalized.x());
            Real pdy = std::floor(normalized.y());
            Real pdz = std::floor(normalized.z());
            if (pdx != zero || pdy != zero || pdz != zero) {
                ++renormalized;
                Real3 wrap_offset(pdx, pdy, pdz);
                ptcl.setCoordinates(Real3(ptcl) - simbox.abc() * wrap_offset);
                normalized -= wrap_offset;
                // 0 <= r - floor(r) <= 1
                // Really! r = -tiny => floor(r) = -1 => -tiny - -1 = 1
                normalized =
                        Real3(normalized.x() == one ? nearly_one : normalized.x(),
                              normalized.y() == one ? nearly_one : normalized.y(),
                              normalized.z() == one ? nearly_one : normalized.z());
            }
        }
        else
        {
            // move particles on the "right" border a little bit "inside" the box
            normalized =
                    Real3(normalized.x() == one ? nearly_one : normalized.x(),
                          normalized.y() == one ? nearly_one : normalized.y(),
                          normalized.z() == one ? nearly_one : normalized.z());
            if (!(zero <= normalized.x() && normalized.x() < one && zero <= normalized.y() &&
                  normalized.y() < one && zero <= normalized.z() && normalized.z() < one))
            {
                std::cout << "particle #" << i << " is not normalized: " << normalized << std::endl;
            }
        }

        assert(zero <= normalized.x() && normalized.x() < one);
        assert(zero <= normalized.y() && normalized.y() < one);
        assert(zero <= normalized.z() && normalized.z() < one);
        dimension_type box_x(normalized.x() * scale);
        dimension_type box_y(normalized.y() * scale);
        dimension_type box_z(normalized.z() * scale);
        assert(0 <= box_x && box_x < box_dim_lowest);
        assert(0 <= box_y && box_y < box_dim_lowest);
        assert(0 <= box_z && box_z < box_dim_lowest);
        // this boxid is specific to the original input coordinates
        // recomputing it from the wrapped particle coordinates might
        // return a different result due to numerical precision
        BoxID boxid(box_x, box_y, box_z, fmm_depth);
        auto sfcindex = boxid.sfcindex();
        assert(sfcindex < num_boxes_lowest);
        if (skip_chargeless_particles && is_chargeless(ptcl)) {
            sfcindex = num_boxes_lowest;
        }
        sortkey_type sortkey(sfcindex, ptcl.site(), ptcl.form());
        tmp[i] = sort_pair(sortkey, payload(i, ptcl));
    }
    std::cout << "renormalized " << renormalized << " particles" << std::endl;
    std::sort(tmp.begin(), tmp.end(), compare_first<sort_pair>());
    for (size_t i = 0; i < n; ++i) {
        boxsfcindex[i] = tmp[i].first.sfcindex();
        permutation[tmp[i].second.first] = i;
        output[i] = tmp[i].second.second;
    }

    // populate box_particle_offset
    box_particle_offset[0] = 0;
    auto total_boxes_lowest = num_boxes_lowest + extra_boxes_lowest;
    for (size_t i = 0, id = 0; id < total_boxes_lowest; ++id)
    {
        while (i < boxsfcindex.size() && boxsfcindex[i] == id)
        {
            ++i;
        }
        box_particle_offset[id + 1] = i;
    }
    assert(box_particle_offset[total_boxes_lowest] == n);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
