/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_fake_stacked_complex_triangle_hpp
#define FMSOLVR_fake_stacked_complex_triangle_hpp

#include "fmsolvr/complex_vec.hpp"
#include "fmsolvr/simd/Vec_Array.hpp"

#include <algorithm>

namespace fmsolvr
{

template <unsigned N, typename Triangle>
class FakeStackedComplexTriangle
{
    typedef Triangle triangular_array_type;
    typedef typename triangular_array_type::value_type basic_complex_type;
    typedef typename basic_complex_type::value_type basic_scalar_type;
    typedef TrivialArray<basic_scalar_type, N> stacked_scalar_type;
    typedef typename COMPLEX_GENERATOR<stacked_scalar_type, basic_scalar_type>::type stacked_complex_type;
    typedef stacked_complex_type stacked_value_type;

    class fakeref
    {
        typedef typename triangular_array_type::index_type index_type;
        typedef FakeStackedComplexTriangle Container;
        typedef fakeref Self;

        typedef stacked_value_type value_type;

        Container * that_;
        index_type l_, m_;

    public:
        fakeref(Container * that, index_type l, index_type m) : that_(that), l_(l), m_(m)
        { }

        operator value_type () const
        {
            return that_->get_value(l_, m_);
        }

        Self operator = (const value_type & val)
        {
            that_->put_value(l_, m_, val);
            return *this;
        }

        Self operator += (const value_type & val)
        {
            value_type tmp = that_->get_value(l_, m_);
            tmp += val;
            that_->put_value(l_, m_, tmp);
            return *this;
        }
    };

    typedef fakeref wrapped_value_type;

public:
    typedef stacked_value_type value_type;
    typedef typename triangular_array_type::size_type size_type;
    typedef typename triangular_array_type::index_type index_type;

    FakeStackedComplexTriangle(triangular_array_type * ptr[N])
    {
        std::copy(ptr, ptr + N, ptr_array_);
        p_ = ptr_array_[0]->p();
        for (unsigned i = 1; i < N; ++i) {
            p_ = std::min(p_, ptr_array_[i]->p());
        }
    }

    size_type p() const
    {
        return p_;
    }

    const stacked_value_type upper(index_type l, index_type m) const
    {
        return get_value(l, m);
    }

    wrapped_value_type upper(index_type l, index_type m)
    {
        return wrapped_value_type(this, l, m);
    }

    void invalidate_lower()
    {
        for (unsigned i = 0; i < N; ++i) {
            ptr_array_[i]->invalidate_lower();
        }
    }

private:
    stacked_complex_type get_value(index_type l, index_type m) const
    {
        stacked_scalar_type re, im;
        for (unsigned i = 0; i < N; ++i) {
            basic_complex_type a = ptr_array_[i]->upper(l, m);
            re[i] = a.real();
            im[i] = a.imag();
        }
        return stacked_complex_type(re, im);
    }

    void put_value(index_type l, index_type m, const stacked_complex_type & val)
    {
        stacked_scalar_type re = val.real();
        stacked_scalar_type im = val.imag();
        for (unsigned i = 0; i < N; ++i) {
            ptr_array_[i]->upper(l, m) = basic_complex_type(re[i], im[i]);
        }
    }

    triangular_array_type * ptr_array_[N];
    size_type p_;
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
