/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_fmm_handle_hpp
#define FMSOLVR_fmm_handle_hpp

#include <ostream>
#include <type_traits>

#include "fmsolvr/handle_particle.hpp"
#include "fmsolvr/handle_precomputed.hpp"
#include "fmsolvr/handle_precomputed_p3.hpp"
#include "fmsolvr/handle_tree.hpp"
#include "fmsolvr/math/ipowlog.hpp"


#include "TreePartition.hpp"

#include "fmsolvr/p2p/base.hpp"
#include "fmsolvr/p2p/coulomb.hpp"
#include "fmsolvr/p2p/tile.hpp"
//#include "fmsolvr/p2p/tile_simd.hpp"
#include "fmsolvr/p2p/sequential.hpp"
//#include "fmsolvr/p2p/pthreads.hpp"


namespace fmsolvr {

template <typename FMMTypes>
class FMMHandle
{
public:
    typedef FMMParticleInteractionHandle<FMMTypes> ParticleInteractionHandle;
    typedef typename ParticleInteractionHandle::ParticleHandle ParticleHandle;
    typedef FMMPrecomputedDataHandle<FMMTypes> PrecomputedDataHandle;
    typedef FMMPrecomputedRotationDataHandle<FMMTypes> PrecomputedRotationDataHandle;
    typedef FMMTreeHandle<FMMTypes> TreeHandle;

    typedef typename FMMTypes::size_type size_type;
    typedef typename FMMTypes::index_type index_type;
    typedef typename FMMTypes::boxid_type boxid_type;
    typedef typename FMMTypes::operatorid_m2l_type operatorid_m2l_type;
    typedef typename FMMTypes::real_type real_type;
    typedef typename FMMTypes::real3_type real3_type;
    typedef typename FMMTypes::real33_type real33_type;
    typedef typename FMMTypes::simulationbox_type simulationbox_type;
    typedef typename FMMTypes::allocator_type allocator_type;
    typedef typename FMMTypes::multipole_coefficients_type multipole_coefficients_type;
    typedef typename FMMTypes::local_coefficients_type local_coefficients_type;
    typedef typename multipole_coefficients_type::value_type coefficient_type;

    typedef typename PrecomputedRotationDataHandle::RotationDataTypes RotationDataTypes;

    template <typename Tp>
    using SIMDengine = typename FMMTypes::template SIMDengine<Tp>;


    typedef TreePartition<boxid_type> tree_partition_type;

    typedef pairwise::Base<real_type,
                                coulomb_kernel,
                                typename ParticleHandle::outputadapter_type,
                                typename ParticleHandle::inputadapter_type,
                                typename ParticleHandle::inputadapter_type>
            P2PBase;
#if 1
    template <typename TileBase>
    using P2PTileProcessor = pairwise::tile<TileBase>;
#else
    template <typename TileBase>
    using P2PTileProcessor = pairwise::tile_simd<TileBase, simd_engine_default>;
#endif
    template <typename TileProcessor>
    using P2PTileSplitter = pairwise::sequential<TileProcessor>;
    typedef P2PTileSplitter<P2PTileProcessor<P2PBase>> P2P_type;

private:
    ParticleInteractionHandle & ptcl_;
    const PrecomputedDataHandle & pc_;
    PrecomputedRotationDataHandle & roth_;
    TreeHandle & th_;

    size_type d_;
    size_type ws_;
    size_type p_;

    bool verbose_;
    std::ostream &verbout_;

public:
    FMMHandle(
            ParticleInteractionHandle &ptcl,
            const PrecomputedDataHandle &pc,
            PrecomputedRotationDataHandle &roth,
            TreeHandle &th,
            size_type depth_compute,
            size_type ws_compute,
            size_type p_compute,
            bool verbose,
            std::ostream &verbout);

    ~FMMHandle();

    bool verbose() const
    { return verbose_; }

    std::ostream &verbout()
    { return verbout_; }

    size_type depth() const
    { return d_; }

    size_type ws() const
    { return ws_; }

    size_type p() const
    { return p_; }

    ParticleInteractionHandle & particle_handle()
    { return ptcl_; }

    const simulationbox_type & simulationbox() const
    { return ptcl_.simulationbox(); }

    real_type omega_scaling_factor() const
    { return pc_.omega_scaling_factor(); }

    real_type m2l_operator_scaling_factor() const
    { return pc_.m2l_operator_scaling_factor(); }

    auto m2m_operator_cube(size_type d, size_type dd) const
        -> typename std::add_const<decltype(pc_.m2m_operator_cube(d, dd))>::type
    { return pc_.m2m_operator_cube(d, dd); }

    auto m2l_operator_cube(size_type d, size_type ws) const
        -> typename std::add_const<decltype(pc_.m2l_operator_cube(d, ws))>::type
    { return pc_.m2l_operator_cube(d, ws); }

    auto l2l_operator_cube(size_type d, size_type dd) const
        -> typename std::add_const<decltype(pc_.l2l_operator_cube(d, dd))>::type
    { return pc_.l2l_operator_cube(d, dd); }

    auto lattice_operator() const
        -> typename std::add_const<decltype(pc_.lattice_operator())>::type
    { return pc_.lattice_operator(); }

    auto tree_omega() const
        -> typename std::add_const<decltype(th_.tree_omega())>::type
    { return th_.tree_omega(); }

    auto tree_omega()
        -> decltype(th_.tree_omega())
    { return th_.tree_omega(); }

    auto tree_mu() const
        -> typename std::add_const<decltype(th_.tree_mu())>::type
    { return th_.tree_mu(); }

    auto tree_mu()
        -> decltype(th_.tree_mu())
    { return th_.tree_mu(); }

    auto omega_dp()
        -> decltype(th_.omega_dp())
    { return th_.omega_dp(); }

    auto div_lut_p(size_type p) const
        -> typename std::add_const<decltype(pc_.div_lut_p(p))>::type
    { return pc_.div_lut_p(p); }

    auto RotationGeneratorM2M()
        -> decltype(roth_.RotationGeneratorM2M())
    { return roth_.RotationGeneratorM2M(); }

    auto RotationGeneratorM2L()
        -> decltype(roth_.RotationGeneratorM2L())
    { return roth_.RotationGeneratorM2L(); }

    auto RotationGeneratorL2L()
        -> decltype(roth_.RotationGeneratorL2L())
    { return roth_.RotationGeneratorL2L(); }

    size_type m2l_interaction_count() const
    {
        return icubed(2 * (2 * ws() + 1)) - icubed(2 * ws() + 1);
    }
};

template <typename FMMTypes>
FMMHandle<FMMTypes>::FMMHandle(
        ParticleInteractionHandle &ptcl,
        const PrecomputedDataHandle &pc,
        PrecomputedRotationDataHandle &roth,
        TreeHandle &th,
        size_type depth_compute,
        size_type ws_compute,
        size_type p_compute,
        bool verbose,
        std::ostream &verbout)
        : ptcl_(ptcl),
          pc_(pc),
          roth_(roth),
          th_(th),
          d_(depth_compute),
          ws_(ws_compute),
          p_(p_compute),
          verbose_(verbose),
          verbout_(verbout)
{
    assert(ptcl_.simulationbox() == pc_.simulationbox());
    assert(ptcl_.depth() == depth());
    assert(pc_.simulationbox() == roth_.simulationbox());
    assert(pc_.depth() == roth_.depth());
    assert(pc_.ws() == roth_.ws());
    assert(pc_.p() == roth_.p());

    assert(depth() <= pc_.depth());
    assert(depth() <= th_.depth());
    assert(ws() == pc_.ws());
    assert(p() <= pc_.p());
    assert(p() <= th_.p());
    assert(p() == div_lut_p(p()).p());
}

template <typename FMMTypes>
FMMHandle<FMMTypes>::~FMMHandle()
{
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
