/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */
#ifndef FMSOLVR_gen_operator_cube_hpp
#define FMSOLVR_gen_operator_cube_hpp

#include "fmsolvr/container/containercube.hpp"
#include "fmsolvr/operatorid_m2l.hpp"
#include "fmsolvr/tree/boxexpansionpoint.hpp"

namespace fmsolvr {

template <typename M2M_Operator, typename Real33>
class GenerateM2MOperatorCube
{
    const size_t p;
    const Real33 abc;

public:
    typedef M2M_Operator value_type;
    typedef ContainerCube<value_type> cube_type;

    GenerateM2MOperatorCube(size_t p, const Real33 & abc) : p(p), abc(abc)
    { }

  template <typename OutputIterator>
  cube_type * operator () (size_t d_c, size_t d_delta, OutputIterator allocated)
  {
    typedef typename Real33::value_type Real3;

    size_t opcubedim = 1 << d_delta;
    cube_type * m2m_cube = new cube_type(opcubedim);
    for (size_t i = 0; i < opcubedim; ++i) {
        for (size_t j = 0; j < opcubedim; ++j) {
            for (size_t k = 0; k < opcubedim; ++k) {
                Real3 source = box_expansion_point(abc, d_c, opcubedim, i, j, k);
                Real3 target(0., 0., 0.);
                value_type * m2m_op = new value_type(source, target, p);
                m2m_cube->set(i, j, k, *m2m_op);
                *(allocated++) = m2m_op;
            }
        }
    }
    return m2m_cube;
  }
};

template <typename M2L_Operator>
class ContainerCubeM2L : public ContainerCube<M2L_Operator>
{
    typedef ContainerCube<M2L_Operator> Base;

    // the storage (in Base) is used target-centric, but nothing
    // outside this class needs to know that

public:
    typedef typename Base::value_type value_type;
    typedef typename Base::const_reference const_reference;
    typedef typename Base::size_type size_type;

    ContainerCubeM2L(size_type ws) : Base(2 * (2 * ws + 1) + 1), offset_(2 * ws + 1)
    { }

    template <typename OpID>
    void set(const OpID &opID, const_reference container)
    {
        auto ijk = opID.relative_to_source();
        Base::set(offset_ + ijk.i(), offset_ + ijk.j(), offset_ + ijk.k(), container);
    }

    template <typename OpID>
    const_reference operator ()(const OpID &opID) const
    {
        auto ijk = opID.relative_to_source();
        return Base::operator()(offset_ + ijk.i(), offset_ + ijk.j(), offset_ + ijk.k());
    }

private:
    size_type offset_;  // for remapping possibly negative indices
                        // addressing in Base is nonnegative
};

template <typename M2L_Operator, typename Real33>
class GenerateM2LOperatorCube
{
    typedef typename M2L_Operator::value_type::value_type real_type;

    const size_t p;
    const Real33 abc;
    const real_type m2l_operator_scaling_factor_;


public:
    typedef M2L_Operator value_type;
    typedef ContainerCubeM2L<value_type> cube_type;

    GenerateM2LOperatorCube(
            size_t p, const Real33 &abc, const real_type m2l_operator_scaling_factor)
            : p(p), abc(abc), m2l_operator_scaling_factor_(m2l_operator_scaling_factor)
    { }

  template <typename OutputIterator>
  cube_type * operator () (size_t depth, size_t ws, OutputIterator allocated)
  {
    typedef typename Real33::value_type Real3;
    typedef M2L_OperatorID<DummyBoxID> OperatorID_M2L;
    typedef typename OperatorID_M2L::ijk_type IJK;

    size_t trg = 2 * ws + 1;
    IJK target_ijk(trg, trg, trg);
    size_t opcubedim = 2 * (2 * ws + 1) + 1;
    size_t inner_begin = ws + 1;
    size_t inner_end = inner_begin + 2 * ws + 1;
    cube_type * m2l_cube = new cube_type(opcubedim);
    for (size_t i = 0; i < opcubedim; ++i) {
        for (size_t j = 0; j < opcubedim; ++j) {
            for (size_t k = 0; k < opcubedim; ++k) {
                if (i < inner_begin || i >= inner_end
                        || j < inner_begin || j >= inner_end
                        || k < inner_begin || k >= inner_end) {
                    Real3 source = box_expansion_point(abc, depth, opcubedim, i, j, k);
                    IJK source_ijk(i, j, k);
                    Real3 target(0., 0., 0.);
                    value_type * m2l_op = new value_type(source, target, p, m2l_operator_scaling_factor_);
                    m2l_cube->set(OperatorID_M2L(source_ijk, target_ijk, depth), *m2l_op);
                    *(allocated++) = m2l_op;
                }
            }
        }
    }
    return m2l_cube;
  }
};

template <typename L2L_Operator, typename Real33>
class GenerateL2LOperatorCube
{
    const size_t p;
    const Real33 abc;

public:
    typedef L2L_Operator value_type;
    typedef ContainerCube<value_type> cube_type;

    GenerateL2LOperatorCube(size_t p, const Real33 & abc) : p(p), abc(abc)
    { }

  template <typename OutputIterator>
  cube_type * operator () (size_t d_c, size_t d_delta, OutputIterator allocated)
  {
    typedef typename Real33::value_type Real3;

    size_t opcubedim = 1 << d_delta;
    cube_type * l2l_cube = new cube_type(opcubedim);
    for (size_t i = 0; i < opcubedim; ++i) {
        for (size_t j = 0; j < opcubedim; ++j) {
            for (size_t k = 0; k < opcubedim; ++k) {
                Real3 source(0., 0., 0.);
                Real3 target = box_expansion_point(abc, d_c, opcubedim, i, j, k);
                value_type * l2l_op = new value_type(source, target, p);
                l2l_cube->set(i, j, k, *l2l_op);
                *(allocated++) = l2l_op;
            }
        }
    }
    return l2l_cube;
  }
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
