/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_handle_particle_hpp
#define FMSOLVR_handle_particle_hpp

#include <ostream>
#include <vector>
#include <cstdint>
#include <atomic>


namespace fmsolvr {

namespace {  // anonymous namespace

template <typename Real3, typename T4Vector>
struct inputadapter
{
    typedef T4Vector vector4_type;
    typedef typename vector4_type::value_type Real4;
    typedef typename Real4::value_type Real;
    typedef typename vector4_type::size_type size_type;

    const vector4_type & v4;

    inputadapter(const vector4_type & v4) : v4(v4) { }

    Real3 load_xyz(size_type i) const
    {
        return Real3(v4[i]);
    }

    Real4 load_xyzs(size_type i) const
    {
        return v4[i];
    }
};


template <typename T1Vector, typename T3Vector>
struct outputadapter
{
    typedef T1Vector vector1_type;
    typedef T3Vector vector3_type;
    typedef typename vector1_type::value_type Real;
    typedef typename vector3_type::value_type Real3;
    typedef typename vector3_type::size_type size_type;

    vector1_type & vPhi;
    vector3_type & vF;

    outputadapter(vector1_type & vPhi, vector3_type & vF) : vPhi(vPhi), vF(vF) { }

    void reduce_pf(size_type i, Real p, const Real3 & f)
    {
        vPhi[i] += p;
        vF[i] += f;
    }
};

}  // anonymous namespace


template <typename FMMTypes>
class FMMParticleHandle
{
public:
    typedef typename FMMTypes::size_type size_type;
    typedef typename FMMTypes::index_type index_type;
    typedef typename FMMTypes::real_type real_type;
    typedef typename FMMTypes::real3_type real3_type;
    typedef typename FMMTypes::real4_type real4_type;
    typedef typename FMMTypes::real33_type real33_type;
    typedef typename FMMTypes::simulationbox_type simulationbox_type;
    typedef typename FMMTypes::boxid_type boxid_type;

    // n elements
    typedef std::vector<uint32_t> particle_permutation_type;
    typedef std::vector<uint32_t> boxsfcindex_type;
    typedef std::vector<real4_type> particle_vector_type;
    typedef std::vector<real_type> potential_vector_type;
    typedef std::vector<real3_type> field_vector_type;
    typedef field_vector_type force_vector_type;

    // 8^d+1 elements
    typedef std::vector<uint32_t> box_particle_offset_type;

    typedef inputadapter<real3_type, particle_vector_type> inputadapter_type;
    typedef outputadapter<potential_vector_type, field_vector_type> outputadapter_type;

private:
    const simulationbox_type simulationbox_;
    const size_type n_;
    const size_type depth_;

    // the (forward) permutation as generated by distribute_particles()
    // the data for input particle i can be found at index perm[i] in ordered_particles (and related arrays)
    particle_permutation_type *particle_permutation_;

    // the following arrays contain "reordered" particle information
    boxsfcindex_type *boxsfcindex_;
    particle_vector_type *ordered_particles_;

    box_particle_offset_type *box_particle_offset_;

public:
    FMMParticleHandle(
            const simulationbox_type & simbox,
            size_type n,
            size_type depth);

    ~FMMParticleHandle();

    const simulationbox_type & simulationbox() const
    { return simulationbox_; }

    size_type depth() const
    { return depth_; }

    size_type n() const
    { return n_; }

    size_type n_active() const
    { return (*box_particle_offset_)[num_boxes_lowest()] - (*box_particle_offset_)[0]; }

    size_type num_boxes_lowest() const
    { return boxes_on_depth(depth()); }

    size_type extra_boxes_lowest() const
    { return 1; }

    const particle_permutation_type & particle_permutation() const
    { return *particle_permutation_; }

    particle_permutation_type & particle_permutation()
    { return *particle_permutation_; }

    const boxsfcindex_type & boxsfcindex() const
    { return *boxsfcindex_; }

    boxsfcindex_type & boxsfcindex()
    { return *boxsfcindex_; }

    const particle_vector_type & ordered_particles() const
    { return *ordered_particles_; }

    particle_vector_type & ordered_particles()
    { return *ordered_particles_; }

    const box_particle_offset_type & box_particle_offset() const
    { return *box_particle_offset_; }

    box_particle_offset_type & box_particle_offset()
    { return *box_particle_offset_; }

    size_type count_empty_boxes() const
    {
        size_type empty_boxes = 0;
        for (size_type id = 0; id < num_boxes_lowest(); ++id)
        {
            if ((*box_particle_offset_)[id] == (*box_particle_offset_)[id + 1])
            {
                ++empty_boxes;
            }
        }
        return empty_boxes;
    }

    size_type count_inactive_particles() const
    {
        return (*box_particle_offset_)[num_boxes_lowest() + extra_boxes_lowest()] -
               (*box_particle_offset_)[num_boxes_lowest()];
    }

    // compute monopole and (cartesian) dipole for reference point (0,0,0)
    // the dipole is independent of the reference point as long as the monopole is 0
    std::pair<real_type, real3_type> monopole_dipole() const
    {
        real_type mp = 0.;
        real3_type dp;
        const auto & ptcl = ordered_particles();
        for (size_type i = 0; i < n(); ++i) {
            real3_type xyz = real3_type(ptcl[i]);
            real_type q = ptcl[i].s();
            mp += q;
            dp += xyz * q;
        }
        return std::make_pair(mp, dp);
    }

    inputadapter_type particle_adapter() const
    {
        return inputadapter_type(ordered_particles());
    }

private:
    inline size_type boxes_on_depth(size_type d) const
    {
        // 8^d
        return size_type(1) << (3 * d);
    }
};

template <typename FMMTypes>
FMMParticleHandle<FMMTypes>::FMMParticleHandle(
        const simulationbox_type & simbox,
        size_type n,
        size_type depth) :
    simulationbox_(simbox),
    n_(n),
    depth_(depth)
{
    particle_permutation_ = new particle_permutation_type(n_);
    boxsfcindex_ = new boxsfcindex_type(n_);
    ordered_particles_ = new particle_vector_type(n_);
    box_particle_offset_ = new box_particle_offset_type(num_boxes_lowest() + extra_boxes_lowest() + 1);
}

template <typename FMMTypes>
FMMParticleHandle<FMMTypes>::~FMMParticleHandle()
{
    delete particle_permutation_;
    delete boxsfcindex_;
    delete ordered_particles_;
    delete box_particle_offset_;
}


// using the output_adapters will populate potential and efield only
// to populate forces and energies, compute_forces_energy() needs to be called
template <typename FMMTypes>
class FMMParticleInteractionHandle
{
public:
    typedef FMMParticleHandle<FMMTypes> ParticleHandle;

    typedef typename ParticleHandle::size_type size_type;
    typedef typename ParticleHandle::index_type index_type;
    typedef typename ParticleHandle::real_type real_type;
    typedef typename ParticleHandle::real3_type real3_type;
    typedef typename ParticleHandle::real4_type real4_type;
    typedef typename ParticleHandle::simulationbox_type simulationbox_type;

    // n elements
    typedef typename ParticleHandle::particle_permutation_type particle_permutation_type;
    typedef typename ParticleHandle::boxsfcindex_type boxsfcindex_type;
    typedef typename ParticleHandle::particle_vector_type particle_vector_type;
    typedef typename ParticleHandle::potential_vector_type potential_vector_type;
    typedef typename ParticleHandle::field_vector_type field_vector_type;
    typedef typename ParticleHandle::force_vector_type force_vector_type;

    // 8^d+1 elements
    typedef typename ParticleHandle::box_particle_offset_type box_particle_offset_type;

    typedef typename ParticleHandle::inputadapter_type inputadapter_type;
    typedef typename ParticleHandle::outputadapter_type outputadapter_type;

private:
    const ParticleHandle &particle_handle_;

    // the following arrays contain "reordered" particle information
    potential_vector_type *potential_nf_;
    potential_vector_type *potential_ff_;
    field_vector_type *efield_nf_;
    field_vector_type *efield_ff_;
    force_vector_type *force_nf_;
    force_vector_type *force_ff_;
    real_type Ec_nf_, Ec_ff_;
    std::atomic<size_type> interactions_nf_ = {0}, interactions_ff_ = {0};

public:
    FMMParticleInteractionHandle(
            const ParticleHandle &particle_handle);

    ~FMMParticleInteractionHandle();

    const simulationbox_type & simulationbox() const
    { return particle_handle_.simulationbox(); }

    size_type depth() const
    { return particle_handle_.depth(); }

    size_type n() const
    { return particle_handle_.n(); }

    size_type n_active() const
    { return particle_handle_.n_active(); }

    size_type num_boxes_lowest() const
    { return particle_handle_.num_boxes_lowest(); }

    const particle_permutation_type & particle_permutation() const
    { return particle_handle_.particle_permutation(); }

    const boxsfcindex_type & boxsfcindex() const
    { return particle_handle_.boxsfcindex(); }

    const particle_vector_type & ordered_particles() const
    { return particle_handle_.ordered_particles(); }

    const box_particle_offset_type & box_particle_offset() const
    { return particle_handle_.box_particle_offset(); }

    size_type count_empty_boxes() const
    { return particle_handle_.count_empty_boxes(); }

    std::pair<real_type, real3_type> monopole_dipole() const
    { return particle_handle_.monopole_dipole(); }

    const potential_vector_type &potential_nf() const
    { return *potential_nf_; }

    potential_vector_type &potential_nf()
    { return *potential_nf_; }

    const potential_vector_type &potential_ff() const
    { return *potential_ff_; }

    potential_vector_type &potential_ff()
    { return *potential_ff_; }

    const field_vector_type &efield_nf() const
    { return *efield_nf_; }

    field_vector_type &efield_nf()
    { return *efield_nf_; }

    const field_vector_type &efield_ff() const
    { return *efield_ff_; }

    field_vector_type &efield_ff()
    { return *efield_ff_; }

    const force_vector_type &force_nf() const
    { return *force_nf_; }

    force_vector_type &force_nf()
    { return *force_nf_; }

    const force_vector_type &force_ff() const
    { return *force_ff_; }

    force_vector_type &force_ff()
    { return *force_ff_; }

    real_type Ec() const
    { return Ec_nf_ + Ec_ff_; }

    real_type Ec_nf() const
    { return Ec_nf_; }

    real_type Ec_ff() const
    { return Ec_ff_; }

    size_type interactions_nf() const
    { return interactions_nf_; }

    size_type interactions_ff() const
    { return interactions_ff_; }

    size_type interactions() const
    { return interactions_nf_ + interactions_ff_; }

    void add_interactions_nf(size_type count)
    { interactions_nf_ += count; }

    void add_interactions_ff(size_type count)
    { interactions_ff_ += count; }

    void zero()
    {
        std::fill(potential_nf_->begin(), potential_nf_->end(), real_type(0.));
        std::fill(potential_ff_->begin(), potential_ff_->end(), real_type(0.));
        std::fill(efield_nf_->begin(), efield_nf_->end(), real3_type(0., 0., 0.));
        std::fill(efield_ff_->begin(), efield_ff_->end(), real3_type(0., 0., 0.));
        std::fill(force_nf_->begin(), force_nf_->end(), real3_type(0., 0., 0.));
        std::fill(force_ff_->begin(), force_ff_->end(), real3_type(0., 0., 0.));
        Ec_nf_ = Ec_ff_ = 0.;
        interactions_nf_ = interactions_ff_ = 0;
    }

    inputadapter_type particle_adapter() const
    {
        return particle_handle_.particle_adapter();
    }

    outputadapter_type potential_efield_nf_adapter()
    {
        return outputadapter_type(potential_nf(), efield_nf());
    }

    outputadapter_type potential_efield_ff_adapter()
    {
        return outputadapter_type(potential_ff(), efield_ff());
    }

    void compute_forces_energy();

    real3_type sum_forces() const;

    void dump_front_back_particles(std::ostream &os, size_type count) const;
};

template <typename FMMTypes>
FMMParticleInteractionHandle<FMMTypes>::FMMParticleInteractionHandle(
        const ParticleHandle &particle_handle) :
    particle_handle_(particle_handle)
{
    potential_nf_ = new potential_vector_type(n(), real_type(0.));
    potential_ff_ = new potential_vector_type(n(), real_type(0.));
    efield_nf_ = new field_vector_type(n(), real3_type(0., 0., 0.));
    efield_ff_ = new field_vector_type(n(), real3_type(0., 0., 0.));
    force_nf_ = new force_vector_type(n(), real3_type(0., 0., 0.));
    force_ff_ = new force_vector_type(n(), real3_type(0., 0., 0.));
    Ec_nf_ = Ec_ff_ = 0.;
    interactions_nf_ = interactions_ff_ = 0;
}

template <typename FMMTypes>
FMMParticleInteractionHandle<FMMTypes>::~FMMParticleInteractionHandle()
{
    delete potential_nf_;
    delete potential_ff_;
    delete efield_nf_;
    delete efield_ff_;
    delete force_nf_;
    delete force_ff_;
}

// populate force_nf_, force_ff_ from efield and q
// compute Ec_nf_, Ec_ff_ from potential and q
template <typename FMMTypes>
void FMMParticleInteractionHandle<FMMTypes>::compute_forces_energy()
{
    const auto &particles = ordered_particles();

    real_type Ec2_nf = 0., Ec2_ff = 0.;
    for (size_type i = 0; i < n(); ++i)
    {
        real_type q = -(-particles[i].s());  // negative gradient
        force_nf()[i] = efield_nf()[i] * q;
        force_ff()[i] = efield_ff()[i] * q;
        Ec2_nf += q * potential_nf()[i];
        Ec2_ff += q * potential_ff()[i];
    }
    Ec_nf_ = Ec2_nf * 0.5;
    Ec_ff_ = Ec2_ff * 0.5;
}

// compute sum of forces
// requires compute_forces_energy() to be called first
template <typename FMMTypes>
typename FMMParticleInteractionHandle<FMMTypes>::real3_type
FMMParticleInteractionHandle<FMMTypes>::sum_forces() const
{
    real3_type sumF(0., 0., 0.);
    for (size_type i = 0; i < n(); ++i)
    {
        sumF += force_nf()[i] + force_ff()[i];
    }
    return sumF;
}

template <typename FMMTypes>
void FMMParticleInteractionHandle<FMMTypes>::dump_front_back_particles(
        std::ostream &os, size_type count) const
{
    if (count == 0)
    {
        return;
    }
    size_type range = std::min<size_type>(n(), count);
    os << "# i\txyzs\tboxid\tefield\tforce\tpotential\n";
    for (size_type i = 0; i < n(); ++i)
    {
        if (i < range || i >= (n() - range))
        {
            size_type iperm = particle_permutation()[i];
            os << i << "\t" << ordered_particles()[iperm] << "\t" << boxsfcindex()[iperm] << "\t"
               << (efield_nf()[iperm] + efield_ff()[iperm]) << "\t"
               << (force_nf()[iperm] + force_ff()[iperm]) << "\t"
               << (potential_nf()[iperm] + potential_ff()[iperm]) << std::endl;
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
