/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_handle_precomputed_hpp
#define FMSOLVR_handle_precomputed_hpp

#include <type_traits>

#include "fmsolvr/container/containermulticube.hpp"
#include "fmsolvr/gen_operator_cube.hpp"
#include "fmsolvr/math/reciprocal.hpp"
#include "fmsolvr/operator/lattice/latticeoperator.hpp"
#include "fmsolvr/operator/p4/local2local.hpp"
#include "fmsolvr/operator/p4/multipole2local.hpp"
#include "fmsolvr/operator/p4/multipole2multipole.hpp"
#include "fmsolvr/util/checking_helpers.hpp"


namespace fmsolvr {

template <typename FMMTypes>
class FMMPrecomputedDataHandle
{
    typedef typename FMMTypes::size_type size_type;
    typedef typename FMMTypes::index_type index_type;
    typedef typename FMMTypes::real_type real_type;
    typedef typename FMMTypes::real3_type real3_type;
    typedef typename FMMTypes::real33_type real33_type;
    typedef typename FMMTypes::simulationbox_type simulationbox_type;
    typedef typename FMMTypes::triangular_array_type triangular_array_type;

    typedef GenerateM2MOperatorCube<M2M_Operator<triangular_array_type>, real33_type> M2Mp4OperatorCubeGenerator;
    typedef ContainerMultiCube<M2Mp4OperatorCubeGenerator> m2m_p4_multi_cube_type;

    typedef GenerateM2LOperatorCube<M2L_Operator<triangular_array_type>, real33_type> M2Lp4OperatorCubeGenerator;
    typedef ContainerMultiCube<M2Lp4OperatorCubeGenerator> m2l_p4_multi_cube_type;

    typedef GenerateL2LOperatorCube<L2L_Operator<triangular_array_type>, real33_type> L2Lp4OperatorCubeGenerator;
    typedef ContainerMultiCube<L2Lp4OperatorCubeGenerator> l2l_p4_multi_cube_type;

    typedef P2M_divtable<real_type> P2M_divtable_type;

private:
    //allocator_type allocator;

    // the base vectors of a primitive cell
    // (combining a primitive cell with a reference point and the particles
    // yields the simulation box)
    const simulationbox_type simulationbox_;

    // the FMM parameters depth, ws, p used for allocation and precomputation
    // actual computation could use lower p (and depth) values, but ws
    // probably has to stay unchanged
    const size_type depth_;
    const size_type ws_;
    const size_type p_;

    const real_type omega_scaling_factor_;
    const real_type m2l_operator_scaling_factor_;

    m2m_p4_multi_cube_type * m2m_p4_operator_multi_cube_;
    m2l_p4_multi_cube_type * m2l_p4_operator_multi_cube_;
    l2l_p4_multi_cube_type * l2l_p4_operator_multi_cube_;
    triangular_array_type * L_;
    P2M_divtable_type * div_lut_p_;

public:
    FMMPrecomputedDataHandle(
            const simulationbox_type & simulationbox,
            size_type depth,
            size_type ws,
            size_type p);

    ~FMMPrecomputedDataHandle();

    bool validate(size_type depth_compute, size_type ws_compute, size_type p_compute);

    const simulationbox_type & simulationbox() const
    { return simulationbox_; }

    size_type depth() const
    { return depth_; }

    size_type ws() const
    { return ws_; }

    size_type p() const
    { return p_; }

    real_type omega_scaling_factor() const
    { return omega_scaling_factor_; }

    real_type m2l_operator_scaling_factor() const
    { return m2l_operator_scaling_factor_; }

    auto m2m_operator_cube(size_type d, size_type dd) const
        -> typename std::add_const<decltype((*m2m_p4_operator_multi_cube_)(d, dd))>::type
    { return (*m2m_p4_operator_multi_cube_)(d, dd); }

    auto m2l_operator_cube(size_type d, size_type ws) const
        -> typename std::add_const<decltype((*m2l_p4_operator_multi_cube_)(d, ws))>::type
    { return (*m2l_p4_operator_multi_cube_)(d, ws); }

    auto l2l_operator_cube(size_type d, size_type dd) const
        -> typename std::add_const<decltype((*l2l_p4_operator_multi_cube_)(d, dd))>::type
    { return (*l2l_p4_operator_multi_cube_)(d, dd); }

    const triangular_array_type & lattice_operator() const
    { return *L_; }

    const P2M_divtable_type & div_lut_p(size_type p) const
    {
        p = p;
        assert(p == div_lut_p_->p());
        return *div_lut_p_;
    }

    // returns 2^k < (2p)!
    static real_type compute_scaling_factor(size_type p)
    {
        real_type fac(1.);
        real_type fac_max(std::numeric_limits<real_type>::max() / real_type(p ? 2 * p : 1));
        for (size_type i = 2; i <= 2 * p && fac < fac_max; ++i)
        {
            fac *= real_type(i);
        }
        // round down to the next power of two
        int exp;
        std::frexp(fac, &exp);
        real_type res = std::ldexp(real_type(.5), exp);
        return res;
    }
};

template <typename FMMTypes>
FMMPrecomputedDataHandle<FMMTypes>::FMMPrecomputedDataHandle(
        const simulationbox_type & simulationbox,
        size_type depth,
        size_type ws,
        size_type p) :
    simulationbox_(simulationbox),
    depth_(depth),
    ws_(ws),
    p_(p),
    omega_scaling_factor_(compute_scaling_factor(p_)),
    m2l_operator_scaling_factor_(reciprocal(omega_scaling_factor_)),
    L_(NULL)
{
    assert(omega_scaling_factor() * m2l_operator_scaling_factor() == real_type(1.));

    m2m_p4_operator_multi_cube_ = new m2m_p4_multi_cube_type(M2Mp4OperatorCubeGenerator(this->p(), this->simulationbox().abc()), this->depth(), 1);
    m2l_p4_operator_multi_cube_ = new m2l_p4_multi_cube_type(M2Lp4OperatorCubeGenerator(this->p(), this->simulationbox().abc(), this->m2l_operator_scaling_factor()), this->depth(), this->ws());
    l2l_p4_operator_multi_cube_ = new l2l_p4_multi_cube_type(L2Lp4OperatorCubeGenerator(this->p(), this->simulationbox().abc()), this->depth(), 1);

    if (!this->simulationbox().open_boundary_conditions())
    {
        // increase multipole order of the lattice operator if it does not contain the converging element (4, 0)
        size_type pL = std::max<size_type>(this->p(), 2);
        L_ = new triangular_array_type(2 * pL);
        size_type iterations = LatticeOperator(
                this->simulationbox().periodic_a(),
                this->simulationbox().periodic_b(),
                this->simulationbox().periodic_c(),
                this->simulationbox().abc(),
                this->ws(),
                *L_,
                pL,
                this->m2l_operator_scaling_factor(),
                42);
        L_->populate_lower();
        std::cout << "LOp#Diterations " << iterations << std::endl;
        if (!checking::is_finite(*L_, L_->p()))
        {
            throw std::domain_error("Lattice Operator is not finite");
        }
    }

    div_lut_p_ = new P2M_divtable_type(this->p());
}

template <typename FMMTypes>
FMMPrecomputedDataHandle<FMMTypes>::~FMMPrecomputedDataHandle()
{
    delete m2m_p4_operator_multi_cube_;
    delete m2l_p4_operator_multi_cube_;
    delete l2l_p4_operator_multi_cube_;
    delete L_;
    delete div_lut_p_;
}

template <typename FMMTypes>
bool FMMPrecomputedDataHandle<FMMTypes>::validate(size_type depth_compute, size_type ws_compute, size_type p_compute)
{
    if (p_compute != div_lut_p_->p())
    {
        delete div_lut_p_;
        div_lut_p_ = new P2M_divtable_type(p_compute);
    }
    return depth_compute <= depth() && ws_compute == ws() && p_compute <= p() && p_compute == div_lut_p_->p();
}


}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
