/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_handle_precomputed_p3_hpp
#define FMSOLVR_handle_precomputed_p3_hpp

#include "fmsolvr/operator/rotation/rotation_types.hpp"
#include "fmsolvr/operator/rotation/rotation_generator.hpp"


namespace fmsolvr {

template <typename FMMTypes>
class FMMPrecomputedRotationDataHandle
{
    typedef typename FMMTypes::size_type size_type;
    typedef typename FMMTypes::index_type index_type;
    typedef typename FMMTypes::real_type real_type;
    typedef typename FMMTypes::real3_type real3_type;
    typedef typename FMMTypes::real33_type real33_type;
    typedef typename FMMTypes::simulationbox_type simulationbox_type;
    typedef typename FMMTypes::triangular_array_type triangular_array_type;

public:
    typedef fmsolvr::RotationDataTypes<real_type, typename FMMTypes::allocator_type> RotationDataTypes;

private:
    typedef RG::RotationGeneratorX2X<typename RotationDataTypes::RotMatVec, real3_type, real33_type, typename FMMTypes::allocator_type> RotationGeneratorX2X_type;
    typedef RG::RotationGeneratorM2L<typename RotationDataTypes::RotMatVec, real3_type, real33_type, typename FMMTypes::allocator_type> RotationGeneratorM2L_type;

private:
    //allocator_type allocator;

    // the base vectors of a primitive cell
    // (combining a primitive cell with a reference point and the particles
    // yields the simulation box)
    const simulationbox_type simulationbox_;

    // the FMM parameters depth, ws, p used for allocation and precomputation
    // actual computation could use lower p (and depth) values, but ws
    // probably has to stay unchanged
    const size_type depth_;
    const size_type ws_;
    const size_type p_;

    RotationGeneratorX2X_type * RotGenX2X_;
    RotationGeneratorM2L_type * RotGenM2L_;

public:
    FMMPrecomputedRotationDataHandle(
            const simulationbox_type & simulationbox,
            size_type depth,
            size_type ws,
            size_type p);

    ~FMMPrecomputedRotationDataHandle();

    bool validate(size_type depth_compute, size_type ws_compute, size_type p_compute);

    const simulationbox_type & simulationbox() const
    { return simulationbox_; }

    size_type depth() const
    { return depth_; }

    size_type ws() const
    { return ws_; }

    size_type p() const
    { return p_; }

    RotationGeneratorX2X_type & RotationGeneratorM2M()
    { return *RotGenX2X_; }

    RotationGeneratorM2L_type & RotationGeneratorM2L()
    { return *RotGenM2L_; }

    RotationGeneratorX2X_type & RotationGeneratorL2L()
    { return *RotGenX2X_; }
};

template <typename FMMTypes>
FMMPrecomputedRotationDataHandle<FMMTypes>::FMMPrecomputedRotationDataHandle(
        const simulationbox_type & simulationbox,
        size_type depth,
        size_type ws,
        size_type p) :
    simulationbox_(simulationbox),
    depth_(depth),
    ws_(ws),
    p_(p)
{
    typedef typename FMMTypes::operatorid_m2l_type OperatorID_M2L;
    typedef typename OperatorID_M2L::ijk_type IJK;

    size_t d_delta = 1;
    RotGenX2X_ = new RotationGeneratorX2X_type(this->p(), d_delta, this->simulationbox().abc());

    size_t dim_op = size_t(1) << d_delta;
    for (size_t ii = 0; ii < dim_op; ++ii)
    {
        for (size_t jj = 0; jj < dim_op; ++jj)
        {
            for (size_t kk = 0; kk < dim_op; ++kk)
            {
                RotGenX2X_->generate_rotation_operator(IJK(ii, jj, kk), d_delta);
            }
        }
    }

    RotGenM2L_ = new RotationGeneratorM2L_type(this->p(), this->ws(), this->simulationbox().abc());

    index_type dim = index_type(1) << 1;
    for (index_type i = 0; i < dim; ++i)
    {
        for (index_type j = 0; j < dim; ++j)
        {
            for (index_type k = 0; k < dim; ++k)
            {
                // parent box coordinates
                index_type ip = i / 2;
                index_type jp = j / 2;
                index_type kp = k / 2;
                // outer ws shells
                index_type io_beg = (ip - ws) * 2;
                index_type io_end = (ip + ws + 1) * 2;
                index_type jo_beg = (jp - ws) * 2;
                index_type jo_end = (jp + ws + 1) * 2;
                index_type ko_beg = (kp - ws) * 2;
                index_type ko_end = (kp + ws + 1) * 2;
                // inner - no far field interactions
                index_type ii_beg = i - ws;
                index_type ii_end = i + ws + 1;
                index_type ji_beg = j - ws;
                index_type ji_end = j + ws + 1;
                index_type ki_beg = k - ws;
                index_type ki_end = k + ws + 1;
                for (index_type ii = io_beg; ii < io_end; ++ii)
                {
                    for (index_type jj = jo_beg; jj < jo_end; ++jj)
                    {
                        for (index_type kk = ko_beg; kk < ko_end; ++kk)
                        {
                            if (ii < ii_beg || ii >= ii_end || jj < ji_beg || jj >= ji_end ||
                                kk < ki_beg || kk >= ki_end)
                            {
                                IJK op_omega_ijk(ii, jj, kk);
                                IJK op_mu_ijk(i, j, k);
                                RotGenM2L_->generate_rotation_operator(OperatorID_M2L(op_omega_ijk, op_mu_ijk, -1));
                            }
                        }
                    }
                }
            }
        }
    }
}

template <typename FMMTypes>
FMMPrecomputedRotationDataHandle<FMMTypes>::~FMMPrecomputedRotationDataHandle()
{
    delete RotGenX2X_;
    delete RotGenM2L_;
}

template <typename FMMTypes>
bool FMMPrecomputedRotationDataHandle<FMMTypes>::validate(size_type depth_compute, size_type ws_compute, size_type p_compute)
{
    return depth_compute <= depth() && ws_compute == ws() && p_compute <= p();
}


}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
