/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_MATH_Complex_hpp
#define FMSOLVR_MATH_Complex_hpp

#include <complex>


namespace fmsolvr {

// This is a partial implementation of the interface defined by std::complex
// providing only the operators needed for the FMM.
// This implementation assumes only valid finite values are used for the real
// and imaginary parts and therefore does neither check for NaN/infinity
// arguments nor does it propagate them correctly.
template <typename Tp>
class Complex
{
public:
    typedef Tp value_type;

    Complex() : cmplx_() {}
    Complex(const Tp &r) : cmplx_(r) {}
    Complex(const Tp &r, const Tp &i) : cmplx_(r, i) {}
    explicit Complex(const std::complex<Tp> &c) : cmplx_(c) {}

    const std::complex<Tp> &cmplx() const
    { return cmplx_; }

    const Tp real() const
    { return cmplx_.real(); }

    const Tp imag() const
    { return cmplx_.imag(); }

    void real(const Tp &val)
    { cmplx_.real(val); }

    void imag(const Tp &val)
    { cmplx_.imag(val); }

    Complex<Tp> &operator+=(const Complex<Tp> &z)
    {
        cmplx_ += z.cmplx_;
        return *this;
    }

    Complex<Tp> &operator-=(const Complex<Tp> &z)
    {
        cmplx_ -= z.cmplx_;
        return *this;
    }

    // Does not use std::complex::operator*=(std::complex) which could
    // generate calls to __mulsc3()/__muldc3()/...
    Complex<Tp> &operator*=(const Complex<Tp> &z)
    {
        const Tp re = this->real() * z.real() - this->imag() * z.imag();
        const Tp im = this->real() * z.imag() + this->imag() * z.real();
        this->real(re);
        this->imag(im);
        return *this;
    }

    Tp abs() const
    {
        return std::abs(cmplx_);
    }

    Complex<Tp> conj() const
    {
        return Complex<Tp>(std::conj(cmplx_));
    }

private:
    std::complex<Tp> cmplx_;
};


template <typename Tp>
inline Complex<Tp> operator-(const Complex<Tp> &a)
{
    return Complex<Tp>(-a.real(), -a.imag());
}

template <typename Tp>
inline Complex<Tp> operator+(const Complex<Tp> &a, const Complex<Tp> &b)
{
    Complex<Tp> tmp = a;
    tmp += b;
    return tmp;
}

template <typename Tp>
inline Complex<Tp> operator-(const Complex<Tp> &a, const Complex<Tp> &b)
{
    Complex<Tp> tmp = a;
    tmp -= b;
    return tmp;
}

template <typename Tp>
inline Complex<Tp> operator*(const Complex<Tp> &a, const Complex<Tp> &b)
{
    Complex<Tp> tmp = a;
    tmp *= b;
    return tmp;
}

template <typename Tp>
inline Complex<Tp> operator*(const Complex<Tp> &a, const Tp &b)
{
    return Complex<Tp>(a.cmplx() * b);
}

template <typename Tp>
inline Complex<Tp> operator*(const Tp &a, const Complex<Tp> &b)
{
    return Complex<Tp>(a * b.cmplx());
}

template <typename Tp>
inline bool operator==(const Complex<Tp> &a, const Complex<Tp> &b)
{
    return a.cmplx() == b.cmplx();
}

template <typename Tp>
inline bool operator!=(const Complex<Tp> &a, const Complex<Tp> &b)
{
    return a.cmplx() != b.cmplx();
}

template <typename Tp>
inline std::ostream & operator<<(std::ostream & os, const Complex<Tp> &a)
{
    return os << a.cmplx();
}

}  // namespace fmsolvr


namespace fmm
{

template <typename Complex>
inline typename Complex::value_type abs(const Complex & c)
{
    return std::abs(c);
}

template <typename T>
inline T abs(const fmsolvr::Complex<T> & c)
{
    return c.abs();
}


template <typename Complex>
inline Complex conj(const Complex & c)
{
    return std::conj(c);
}

template <typename T>
inline fmsolvr::Complex<T> conj(const fmsolvr::Complex<T> & c)
{
    return c.conj();
}

}  // namespace fmm


#endif
// vim: et:ts=4:sw=4
