/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_MATH_linear_algebra_hpp
#define FMSOLVR_MATH_linear_algebra_hpp

#include "fmsolvr/container/simple_vector.hpp"
#include <cstddef>

namespace fmsolvr
{

namespace linear_algebra {

typedef simple_vector<size_t> IndexVector;

template <typename Real>
struct SqMatrix
{
    typedef Real value_type;
    typedef size_t size_type;

    value_type * values;
    size_t dim;

    SqMatrix(size_type dim) : dim(dim)
    {
        values = new value_type[dim * dim];
    }

    ~SqMatrix()
    {
        delete[] values;
    }

    value_type * operator [] (size_type row)
    {
        return values + row * dim;
    }

    const value_type * operator [] (size_type row) const
    {
        return values + row * dim;
    }

    SqMatrix(const SqMatrix &) = delete;
    SqMatrix & operator = (const SqMatrix &) = delete;
};

template <typename Matrix>
typename Matrix::value_type det(size_t dim, const size_t * rows, const size_t * cols, const Matrix & A)
{
    typedef typename Matrix::value_type Real;

    if (dim == 2)
        return A[rows[0]][cols[0]] * A[rows[1]][cols[1]] - A[rows[0]][cols[1]] * A[rows[1]][cols[0]];
    if (dim == 1)
        return A[rows[0]][cols[0]];

    IndexVector newrows(dim - 1);
    for (size_t i = 0; i < dim - 1; ++i)
        newrows[i] = rows[i + 1];
    Real result = A[rows[0]][cols[0]] * det(dim - 1, newrows.data(), cols + 1, A);
    for (size_t i = 1; i < dim; ++i) {
        newrows[i - 1] = rows[i - 1];
        Real a_minor = A[rows[i]][cols[0]] * det(dim - 1, newrows.data(), cols + 1, A);
        if (i & 1)
            result -= a_minor;
        else
            result += a_minor;
    }
    return result;
}

template <typename Matrix>
typename Matrix::value_type det(size_t dim, const Matrix & A)
{
    IndexVector rc(dim);
    for (size_t i = 0; i < dim; ++i)
        rc[i] = i;
    return det(dim, rc.data(), rc.data(), A);
}

template <typename Matrix>
typename Matrix::value_type matminor(size_t dim, size_t row, size_t col, const size_t * rows, const size_t * cols, const Matrix & A)
{
    IndexVector newrows(dim - 1);
    IndexVector newcols(dim - 1);
    {
        size_t r;
        for (r = 0; r < row; ++r)
            newrows[r] = rows[r];
        for ( ; r < dim - 1; ++r)
            newrows[r] = rows[r + 1];
    }
    {
        size_t c;
        for (c = 0; c < col; ++c)
            newcols[c] = cols[c];
        for ( ; c < dim - 1; ++c)
            newcols[c] = cols[c + 1];
    }
    return det(dim - 1, newrows.data(), newcols.data(), A);
}


template <typename Matrix>
void adjugate(size_t dim, const Matrix & A, Matrix & Aadj)
{
    typedef typename Matrix::value_type Real;

    IndexVector rc(dim);
    for (size_t i = 0; i < dim; ++i)
        rc[i] = i;
    for (size_t j = 0; j < dim; ++j)
        for (size_t i = 0; i < dim; ++i) {
            // adjugate matrix = transpose of the cofactor matrix
            Real minor_ij = matminor(dim, i, j, rc.data(), rc.data(), A);
            if ((i + j) & 1)
                Aadj[j][i] = - minor_ij;
            else
                Aadj[j][i] = minor_ij;
        }
}

template <typename Matrix>
bool inverse(size_t dim, const Matrix & A, Matrix & Ainv)
{
    typedef typename Matrix::value_type Real;
    // A^-1 = 1 / det(A) * adj(A)
    Real detA = det(dim, A);
    if (detA == 0.)
        return false;
    Real scale = Real(1.) / det(dim, A);
    adjugate(dim, A, Ainv);
    for (size_t i = 0; i < dim; ++i)
        for (size_t j = 0; j < dim; ++j)
            Ainv[i][j] *= scale;
    return true;
}

// compute  b = Ax
template <typename Real, typename Matrix>
void multiply(size_t dim, const Matrix & A, const Real * x, Real * b, Real scale)
{
    for (size_t i = 0; i < dim; ++i) {
        Real r = 0;
        for (size_t j = 0; j < dim; ++j)
            r += A[i][j] * x[j];
        b[i] = scale * r;
    }
}

// solve  Ax = b  for x
template <typename Real, typename Matrix>
bool solve(size_t dim, const Matrix & A, Real * x, const Real * b)
{
    // Ax = b
    // A^-1 * A * x = A^-1 * b
    // x = 1 / det(A) * adj(A) * b
    Real detA = det(dim, A);
    if (detA == 0.)
        return false;
    Real scale = Real(1.) / detA;
    Matrix Aadj(dim);
    adjugate(dim, A, Aadj);
    multiply(dim, Aadj, b, x, scale);
    return true;
}

}  // namespace linear_algebra

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
