/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_MISC_MultidimensionalCoordinates_hpp
#define FMSOLVR_MISC_MultidimensionalCoordinates_hpp

#include <cassert>
#include <iostream>
#include <initializer_list>
#include <algorithm>

namespace fmsolvr
{

// little endian: coord[0] = least significant dim
template <unsigned NDims, typename Integral = ssize_t>
struct MultidimensionalCoordinates
{
    typedef Integral integral_type;

    MultidimensionalCoordinates()
    {
        std::fill(coord, coord + ndims(), 0);
    }

    MultidimensionalCoordinates(std::initializer_list<integral_type> il)
    {
        assert(il.size() == ndims());
        std::copy(il.begin(), il.end(), coord);
    }

    static size_t ndims()
    {
        return NDims;
    }

    const integral_type & operator [] (size_t index) const
    {
        return coord[index];
    }

    integral_type & operator [] (size_t index)
    {
        return coord[index];
    }

    // linearize coordinates inside a ndims()-dimensional cube with side length 'length'
    integral_type linearize(integral_type length) const
    {
        integral_type res = 0;
        for (ssize_t i = ndims() - 1; i >= 0; --i)
            res = res * length + coord[i];
        return res;
    }

    MultidimensionalCoordinates & operator += (const MultidimensionalCoordinates & a)
    {
        for (size_t i = 0; i < ndims(); ++i)
            coord[i] += a[i];
        return *this;
    }

    MultidimensionalCoordinates & operator -= (const MultidimensionalCoordinates & a)
    {
        for (size_t i = 0; i < ndims(); ++i)
            coord[i] -= a[i];
        return *this;
    }

    MultidimensionalCoordinates operator + (const MultidimensionalCoordinates & a) const
    {
        MultidimensionalCoordinates b(*this);
        b += a;
        return b;
    }

    MultidimensionalCoordinates operator - (const MultidimensionalCoordinates & a) const
    {
        MultidimensionalCoordinates b(*this);
        b -= a;
        return b;
    }

    bool operator == (const MultidimensionalCoordinates & a) const
    {
        return !(*this != a);
    }

    bool operator != (const MultidimensionalCoordinates & a) const
    {
        for (size_t i = 0; i < ndims(); ++i)
            if (coord[i] != a[i])
                return true;
        return false;
    }

private:
    integral_type coord[NDims];
};

template <unsigned NDims, typename Integral>
std::ostream & operator << (std::ostream & os, const MultidimensionalCoordinates<NDims, Integral> & a)
{
    os << "(";
    for (size_t i = 0; i < a.ndims(); ++i) {
        if (i > 0)
            os << ",";
        os << a[i];
    }
    os << ")";
    return os;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
