/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_MISC_energy_stress_triangular_array_group_hpp
#define FMSOLVR_MISC_energy_stress_triangular_array_group_hpp

#include <cstddef>
#include <functional>

namespace fmsolvr
{

template <typename Tp>
class EnergyStressTriangularArrayGroup
{
public:
    typedef Tp value_type;

private:
    struct _ABC
    {
        _ABC(int p)
            : a(p), b(p), c(p)
        { }

        value_type a;
        value_type b;
        value_type c;
    };

    typedef EnergyStressTriangularArrayGroup _Self;
    EnergyStressTriangularArrayGroup(const _Self &) = delete;
    _Self & operator = (const _Self &) = delete;

public:
    EnergyStressTriangularArrayGroup(int p)
        : energy(p), stress(p)
    { }

    void zero()
    {
        energy.zero();
        stress.a.zero();
        stress.b.zero();
        stress.c.zero();
    }

    void zero(size_t p)
    {
        energy.zero(p);
        stress.a.zero(p);
        stress.b.zero(p);
        stress.c.zero(p);
    }

    void invalidate_lower()
    {
        energy.invalidate_lower();
        stress.a.invalidate_lower();
        stress.b.invalidate_lower();
        stress.c.invalidate_lower();
    }

    void populate_lower()
    {
        energy.populate_lower();
        stress.a.populate_lower();
        stress.b.populate_lower();
        stress.c.populate_lower();
    }

    value_type energy;
    _ABC stress;
};


template <typename ESGroupStorage, typename Value, template <typename, typename> class Operation>
class EnergyStressStorageOperator : public std::binary_function<ESGroupStorage, Value, void>
{
    typedef ESGroupStorage es_group_storage_type;
    typedef typename es_group_storage_type::value_type storage_type;
    typedef typename storage_type::value_type storage_value_type;
    typedef Value compute_value_type;
    typedef Operation<storage_value_type, compute_value_type> Op;
    // FIXME: use traits to get complex::value_type || type
    typedef typename compute_value_type::value_type scale_type;
public:
    typedef typename storage_type::index_type index_type;

    EnergyStressStorageOperator(scale_type a, scale_type b, scale_type c)
        : scale{a, b, c}
    { }

    void operator () (es_group_storage_type & st, index_type l, index_type m, const compute_value_type & value) const
    {
        const Op op;
        op(st.energy.upper(l, m), value);
        op(st.stress.a.upper(l, m), value * scale.a);
        op(st.stress.b.upper(l, m), value * scale.b);
        op(st.stress.c.upper(l, m), value * scale.c);
    }

private:
    struct { scale_type a, b, c; } scale;
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
