/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_MISC_split_complex_mul_acc_simd_asm_m256d_hpp
#define FMSOLVR_MISC_split_complex_mul_acc_simd_asm_m256d_hpp

#include "fmsolvr/simd/Vec.hpp"

namespace fmsolvr
{

#ifdef __AVX__
// n is the number of SIMD-words to process
inline
//__attribute__((noinline))
void split_complex_mul_acc_simd_asm(
        __m256d & are,
        __m256d & aim,
        const double * pra,
        const double * pia,
        const double * prb,
        const double * pib,
        size_t n)
{
    typedef __m256d simd_type;
    typedef typename Vec_traits<simd_type>::scalar_type scalar_type;
    Vec_traits<simd_type> VT;
    const size_t s = VT.VecSize;

    simd_type arr = VT.zero();
    simd_type aii = VT.zero();
    simd_type ari = VT.zero();
    simd_type air = VT.zero();

    // process tail that is not divisible by 3
    size_t tail = n % 3;
    for (size_t i = n - tail; i < n; ++i)
    {
        simd_type ra = VT.unaligned_load(pra + i * s);
        simd_type rb = VT.unaligned_load(prb + i * s);
        simd_type ib = VT.unaligned_load(pib + i * s);
        simd_type ia = VT.unaligned_load(pia + i * s);
        arr += ra * rb;
        ari += ra * ib;
        air += ia * rb;
        aii += ia * ib;
    }
    n -= tail;

    if (n == 0) {
        are += arr - aii;
        aim += ari + air;
        return;
    }

    simd_type y0;
    simd_type y1;
    simd_type y2;
    simd_type y3;
    simd_type y4;
    simd_type y5;
    simd_type y6;
    simd_type y7;
    simd_type y8;
    simd_type y9;
    simd_type yA;
    simd_type yB;

#define LOADOP(load, from, displacement) \
    "       vmovupd     " displacement "(%[" # from "],%[ofs],1), %[" # load "]\n"
#define ACCUOP(accu, val) \
    "       vaddpd      %[" # val "],%[" # accu "], %[" # accu "]\n"
#define MULOP(res, a, b) \
    "       vmulpd      %[" # a "],%[" # b "], %[" # res "]\n"
    __asm__(
LOADOP(y0, pra, "     ")
LOADOP(y1, prb, "     ")
LOADOP(y2, pib, "     ")
LOADOP(y3, pia, "     ")
LOADOP(y4, pra, " 0x20")
LOADOP(y5, prb, " 0x20")                   MULOP(yB, y0, y1)
LOADOP(y6, pib, " 0x20")                   MULOP(y0, y0, y2)
LOADOP(y7, pia, " 0x20")                   MULOP(y1, y1, y3)
"       add         %[ofsincr], %[ofs]\n"
LOADOP(y8, pra, "-0x20")                   MULOP(y2, y2, y3)
LOADOP(y9, prb, "-0x20")                   MULOP(y3, y4, y5)
LOADOP(yA, pib, "-0x20")  ACCUOP(arr, yB)  MULOP(y4, y4, y6)
LOADOP(yB, pia, "-0x20")  ACCUOP(ari, y0)  MULOP(y5, y5, y7)
"       cmp         %[ofs],%[ofsend]\n"  // ofsend - ofs
"       jna         2f\n"
".align 32\n"
"1:\n"
LOADOP(y0, pra, "     ")  ACCUOP(air, y1)  MULOP(y6, y6, y7)
LOADOP(y1, prb, "     ")  ACCUOP(aii, y2)  MULOP(y7, y8, y9)
LOADOP(y2, pib, "     ")  ACCUOP(arr, y3)  MULOP(y8, y8, yA)
LOADOP(y3, pia, "     ")  ACCUOP(ari, y4)  MULOP(y9, y9, yB)
LOADOP(y4, pra, " 0x20")  ACCUOP(air, y5)  MULOP(yA, yA, yB)
LOADOP(y5, prb, " 0x20")  ACCUOP(aii, y6)  MULOP(yB, y0, y1)
LOADOP(y6, pib, " 0x20")  ACCUOP(arr, y7)  MULOP(y0, y0, y2)
LOADOP(y7, pia, " 0x20")  ACCUOP(ari, y8)  MULOP(y1, y1, y3)
"       add         %[ofsincr], %[ofs]\n"
LOADOP(y8, pra, "-0x20")  ACCUOP(air, y9)  MULOP(y2, y2, y3)
LOADOP(y9, prb, "-0x20")  ACCUOP(aii, yA)  MULOP(y3, y4, y5)
LOADOP(yA, pib, "-0x20")  ACCUOP(arr, yB)  MULOP(y4, y4, y6)
LOADOP(yB, pia, "-0x20")  ACCUOP(ari, y0)  MULOP(y5, y5, y7)
"       cmp         %[ofs],%[ofsend]\n"  // ofsend - ofs
"       ja          1b\n"
"2:\n"
                          ACCUOP(air, y1)  MULOP(y6, y6, y7)
                          ACCUOP(aii, y2)  MULOP(y7, y8, y9)
                          ACCUOP(arr, y3)  MULOP(y8, y8, yA)
                          ACCUOP(ari, y4)  MULOP(y9, y9, yB)
                          ACCUOP(air, y5)  MULOP(yA, yA, yB)
                          ACCUOP(aii, y6)
                          ACCUOP(arr, y7)
                          ACCUOP(ari, y8)
                          ACCUOP(air, y9)
                          ACCUOP(aii, yA)
            :
            [y0] "=x"(y0),
            [y1] "=x"(y1),
            [y2] "=x"(y2),
            [y3] "=x"(y3),
            [y4] "=x"(y4),
            [y5] "=x"(y5),
            [y6] "=x"(y6),
            [y7] "=x"(y7),
            [y8] "=x"(y8),
            [y9] "=x"(y9),
            [yA] "=x"(yA),
            [yB] "=x"(yB),
            [arr] "+x"(arr),
            [aii] "+x"(aii),
            [ari] "+x"(ari),
            [air] "+x"(air)
            :
            [pra] "R"(pra),
            [pia] "R"(pia),
            [prb] "R"(prb),
            [pib] "R"(pib),
            [ofs] "a"(size_t(0)),
            [ofsincr] "i"(3 * s * sizeof(scalar_type)),
            [ofsend] "r"(n * s * sizeof(scalar_type))
            :
            );

    are += arr - aii;
    aim += ari + air;
}
#endif  // __AVX__

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
