/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_multipole_hpp
#define FMSOLVR_multipole_hpp

#include "fmsolvr/complex_vec.hpp"
#include "fmsolvr/container/triangular_array.hpp"
#include "fmsolvr/math/parity_sign.hpp"
#include "fmsolvr/sourcecounters.hpp"

namespace fmsolvr
{

struct MultipoleComputeLower
{
    // 0 <= l <= p, -l <= m <= -1
    template <typename UTA>
    typename UTA::value_type operator () (typename UTA::index_type l, typename UTA::index_type m, const UTA & uta)
    {
        return toggle_sign_if_odd(-m, fmm::conj(uta.upper(l, -m)));
    }
};


template <typename Real, typename Allocator = std::allocator<Real> >
class MultipoleCoefficientsUpperLower : public TriangularArray<typename COMPLEX_GENERATOR<Real>::type, MultipoleComputeLower, Allocator>
{
    typedef TriangularArray<typename COMPLEX_GENERATOR<Real>::type, MultipoleComputeLower, Allocator> Base;
    typedef fmsolvr::SourceCounters SourceCounters;
    SourceCounters sourceCounters_;

public:
    typedef Real real_type;
    typedef typename Base::value_type complex_type;
    typedef typename Base::index_type index_type;

    MultipoleCoefficientsUpperLower(index_type p) : Base(p)
    { }

    template <typename Table>
    MultipoleCoefficientsUpperLower(index_type p, const Table & table) : Base(p)
    {
        for (size_t i = 0; i < sizeof(table) / sizeof(table[0]); ++i) {
            index_type l = table[i].l, m = table[i].m;
            if (l >= 0 && l <= p && m >= -l && m <= l)
                this->get(l, m) = table[i].val;
        }
        this->populate_lower();
    }

    // add a single particle
    void incSourceCounters()
    { sourceCounters_.inc(); }

    void addSourceCounters(const SourceCounters & sc)
    { sourceCounters_.add(sc); }

    const SourceCounters & sourceCounters() const
    { return sourceCounters_; }

    bool empty() const
    { return sourceCounters_.empty(); }

    void clearSourceCounters()
    { sourceCounters_.clear(); }
};

template <typename Real, typename Allocator = std::allocator<Real> >
class MultipoleCoefficientsUpper : public UpperTriangularArray<typename COMPLEX_GENERATOR<Real>::type, MultipoleComputeLower, Allocator>
{
    typedef UpperTriangularArray<typename COMPLEX_GENERATOR<Real>::type, MultipoleComputeLower, Allocator> Base;
    typedef fmsolvr::SourceCounters SourceCounters;
    SourceCounters sourceCounters_;

public:
    typedef Real real_type;
    typedef typename Base::value_type complex_type;
    typedef typename Base::index_type index_type;

    MultipoleCoefficientsUpper(index_type p) : Base(p)
    { }

    template <typename Table>
    MultipoleCoefficientsUpper(index_type p, const Table & table) : Base(p)
    {
        for (size_t i = 0; i < sizeof(table) / sizeof(table[0]); ++i) {
            index_type l = table[i].l, m = table[i].m;
            if (l >= 0 && l <= p && m >= 0 && m <= l)
                this->upper(l, m) = table[i].val;
        }
        this->populate_lower();
    }

    // add a single particle
    void incSourceCounters()
    { sourceCounters_.inc(); }

    void addSourceCounters(const SourceCounters & sc)
    { sourceCounters_.add(sc); }

    const SourceCounters & sourceCounters() const
    { return sourceCounters_; }

    bool empty() const
    { return sourceCounters_.empty(); }

    void clearSourceCounters()
    { sourceCounters_.clear(); }

    void operator+=(const MultipoleCoefficientsUpper &right)
    {
        if (!right.sourceCounters_.empty())
        {
            this->Base::operator+=(right);
            addSourceCounters(right.sourceCounters_);
        }
    }
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
