/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P3_LOCAL2LOCAL_P3_HPP
#define FMSOLVR_OPERATOR_P3_LOCAL2LOCAL_P3_HPP

#include "fmsolvr/container/linear_array.hpp"
#include "fmsolvr/operator/p3/local2local_p3_impl.hpp"
#include "fmsolvr/operator/rotation/rotation_generator.hpp"
#include "fmsolvr/operator/rotation/rotation_matrix.hpp"
#include "fmsolvr/operator/rotation/rotation_matrix_vector.hpp"
#include "fmsolvr/simd/simd_engine.hpp"

namespace fmsolvr
{

// Note: In contrast to P2M, P2L, M2M, and M2L that all use (source - target),
// for L2L it is required to use (target - source) instead. Really.

// The forward rotation of an M2M source (child) and the forward rotation of
// an L2L source (parent) are identical. The L2L rotation thereby seems off
// by 180°, but is intrinsically compensated by the shift vector pointing
// into opposite directions: M2M (source-target), L2L (target-source).
// The p3 shift operator consists only of the distance (non-negative,
// undirected), not the (target-source) vector itself. Hence the p4 vector
// property (different directions for M2M and L2L) is not visible and is
// handled via the seemingly wrong rotation.
// The backward rotation is handled accordingly. It is always defined
// (M2M, M2L, L2L) to "undo" the corresponding forward rotation.

// Implements the rotation-based L2L operator. The operation consists of
// three steps:
// - Rotation of the reference frame of the local expansion
//     so that the z-axis is aligned with the shift vector
// - Application of the simplified L2L operator
// - Backward rotation of the local expansion into the original reference frame
// Those three steps are represented by 2 operators, a RotationOperator
// representing the two rotations and an L2Lp3_Operator representing the
// simplified L2L operator.

// L2L-operator for a shift along the quantization axis
// of the local expansion.
// Holds precomputed factors (R^i/i!) for i = 0,...,p where R is
// the length of the shift vector.
template <typename Real, typename Allocator = std::allocator<Real> >
class L2Lp3_Operator : public LinearArray<Real, Allocator>
{
    typedef LinearArray<Real, Allocator> Base;

    L2Lp3_Operator(const Real R, const size_t p) : Base(p)
    {
        (*this)[0] = Real(1.);
        for (size_t i = 1; i <= p; ++i)
        {
            (*this)[i] = R * (*this)[i - 1] / Real(i);
        }
    }

public:
    template <typename Real3>
    L2Lp3_Operator(
            const Real3 &expansion_point_source,
            const Real3 &expansion_point_target,
            const size_t p)
            : L2Lp3_Operator(length(expansion_point_target - expansion_point_source), p)
    { }
};


// Rotation-based L2L operator for a given source and target point. All operators
// are computed on the fly, which makes the opration costly.
template <typename Real3,
          typename CoefficientMatrix,
          typename Scratch>
void L2Lp3(
        const CoefficientMatrix &mu_source,
        const Real3 &expansion_point_source,
        CoefficientMatrix &mu_target,
        const Real3 &expansion_point_target,
        const size_t p,
        Scratch &scratch)
{

    typedef typename CoefficientMatrix::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef RotationMatrixUpper<real_type> rotation_matrix;
    typedef RotationMatrixVector<rotation_matrix> rotation_matrix_vector;
    typedef RG::RotationGeneratorNoPrecomp<rotation_matrix_vector, Real3> rotation_generator;
    typedef typename rotation_generator::rotation_operator rotation_operator;

    rotation_generator G(p);
    rotation_operator R = G.generate_rotation(expansion_point_target - expansion_point_source);
    L2Lp3_Operator<real_type> C(expansion_point_source, expansion_point_target, p);

    L2Lp3<simd_engine_none>(mu_source, R, C, mu_target, p, scratch);
}


// Rotation-based L2L operator for a given rotation operator and simplified
// L2L operator. The operator matrices can be precomputed and reused.
template <template <typename> class SIMDengine,
          typename LocalCoefficientsIn,
          typename RotationOperator,
          typename L2LOperator,
          typename LocalCoefficientsOut,
          typename Scratch>
void L2Lp3(
        const LocalCoefficientsIn &mu_source,
        const RotationOperator &R,
        const L2LOperator &C,
        LocalCoefficientsOut &mu_target,
        const size_t p,
        Scratch &scratch)
{
    L2Lp3_impl_default<SIMDengine>(mu_source, R, C, mu_target, p, scratch);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
