/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P3_LOCAL2LOCAL_P3_IMPL_HPP
#define FMSOLVR_OPERATOR_P3_LOCAL2LOCAL_P3_IMPL_HPP

#include "fmsolvr/operator/p3/local2local_p3_shift.hpp"
#include "fmsolvr/operator/rotation/rotation.hpp"

namespace fmsolvr
{

// Applies the rotation-based L2L operator given by the RotationOperator R
// and the simplified L2L operator to the local expansion given in mu_in
// and writes the results to mu_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = target - source.
// This is an accumulating operation.
template <typename LocalCoefficientsIn,
          typename RotationOperator,
          typename L2LOperator,
          typename LocalCoefficientsOut>
void L2Lp3_impl_reference(
        const LocalCoefficientsIn &mu_source,
        const RotationOperator &R,
        const L2LOperator &C,
        LocalCoefficientsOut &mu_target,
        const size_t p)
{
    LocalCoefficientsIn mu_source_rot(p);
    LocalCoefficientsOut mu_target_rot(p);

    ApplyRotationLocalForward_reference(mu_source, R, mu_source_rot, p);
    L2Lp3_apply_C(mu_source_rot, C, mu_target_rot, p);
    mu_target_rot.populate_lower();
    ApplyRotationLocalBackward_reference(mu_target_rot, R, mu_target, p);
}


// Applies the rotation-based L2L operator given by the RotationOperator R
// and the simplified L2L operator to the local expansion given in mu_in
// and writes the results to mu_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = target - source.
// This is an accumulating operation.
template <template <typename> class SIMDengine,
          typename LocalCoefficientsIn,
          typename RotationOperator,
          typename L2LOperator,
          typename LocalCoefficientsOut,
          typename Scratch>
void L2Lp3_impl_precomp(
        const LocalCoefficientsIn &mu_source,
        const RotationOperator &R,
        const L2LOperator &C,
        LocalCoefficientsOut &mu_target,
        const size_t p,
        Scratch &scratch)
{
    auto &mu_source_rot = scratch.omuga_source_rot();
    // scratch does not guarantee access to the lower part
    LocalCoefficientsOut mu_target_rot(p);

    ApplyRotationLocalForward_phi_g(mu_source, R, mu_source_rot, p);
    L2Lp3_apply_C(mu_source_rot, C, mu_target_rot, p);
    mu_target_rot.populate_lower();
    ApplyRotationLocalBackward_phi_g(mu_target_rot, R, mu_target, p);
}


// Applies the rotation-based L2L operator given by the RotationOperator R
// and the simplified L2L operator to the local expansion given in mu_in
// and writes the results to mu_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = target - source.
// This is an accumulating operation.
template <template <typename> class SIMDengine,
          typename LocalCoefficientsIn,
          typename RotationOperator,
          typename L2LOperator,
          typename LocalCoefficientsOut,
          typename Scratch>
void L2Lp3_impl_folded(
        const LocalCoefficientsIn &mu_source,
        const RotationOperator &R,
        const L2LOperator &C,
        LocalCoefficientsOut &mu_target,
        const size_t p,
        Scratch &scratch)
{
    auto &mu_source_rot = scratch.omuga_source_rot();
    auto &mu_target_rot = scratch.omuga_target_rot();
    auto &scratcharray = scratch.array();

    ApplyRotationLocalForward_folded<SIMDengine>(mu_source, R, mu_source_rot, p, scratcharray);
    L2Lp3_apply_C_ml<SIMDengine>(mu_source_rot, C, mu_target_rot, p);
    ApplyRotationLocalBackward_folded<SIMDengine>(mu_target_rot, R, mu_target, p);
}


// Selects the default L2Lp3 implementation.
template <template <typename> class SIMDengine,
          typename LocalCoefficientsIn,
          typename RotationOperator,
          typename L2LOperator,
          typename LocalCoefficientsOut,
          typename Scratch>
void L2Lp3_impl_default(
        const LocalCoefficientsIn &mu_source,
        const RotationOperator &R,
        const L2LOperator &C,
        LocalCoefficientsOut &mu_target,
        const size_t p,
        Scratch &scratch)
{
#ifdef FORCE_VALIDATE_REQUIRE_LOWER
    mu_source.require_lower();
#endif
    L2Lp3_impl_folded<SIMDengine>(mu_source, R, C, mu_target, p, scratch);
    //L2Lp3_impl_precomp<SIMDengine>(mu_source, R, C, mu_target, p, scratch);
    //L2Lp3_impl_reference(mu_source, R, C, mu_target, p);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
