/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P3_LOCAL2LOCAL_P3_SHIFT_HPP
#define FMSOLVR_OPERATOR_P3_LOCAL2LOCAL_P3_SHIFT_HPP

#include <cassert>
#include <cstddef>

namespace fmsolvr
{

// Apply L2L shift along the quantization axis.
// This is an assigning operation.
template <typename LocalCoefficientsIn,
          typename OperatorMatrix,
          typename LocalCoefficientsOut>
__attribute__((flatten))
void L2Lp3_apply_C(
        const LocalCoefficientsIn &mu_in,
        const OperatorMatrix &C,
        LocalCoefficientsOut &mu_out,
        const size_t p_compute)
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);

    assert(p_compute <= mu_in.p());
    assert(p_compute <= C.p());
    assert(p_compute <= mu_out.p());
    mu_out.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type mu_l_m(0.);
            for (index_type j = l; j <= p; ++j)
            {
                mu_l_m += mu_in.upper(j, m) * C[j - l];
            }
            mu_out.upper(l, m) = mu_l_m;
        }
    }
}


// Apply L2L shift along the quantization axis.
// This is an assigning operation.
template <template <typename> class SIMDengine,
          typename LocalCoefficientsIn,
          typename OperatorMatrix,
          typename LocalCoefficientsOut>
__attribute__((flatten))
void L2Lp3_apply_C_ml(
        const LocalCoefficientsIn &mu_in,
        const OperatorMatrix &C,
        LocalCoefficientsOut &mu_out,
        const size_t p_compute)
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);

    assert(p_compute <= mu_in.p());
    assert(p_compute <= C.p());
    assert(p_compute <= mu_out.p());
    mu_out.invalidate_lower();

    for (index_type m = 0; m <= p; ++m)
    {
        for (index_type l = m; l <= p; ++l)
        {
            complex_type mu_l_m(0.);
            for (index_type j = l; j <= p; ++j)
            {
                mu_l_m += mu_in.upper(j, m) * C[j - l];
            }
            mu_out.upper(l, m) = mu_l_m;
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
