/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P3_MULTIPOLE2LOCAL_P3_HPP
#define FMSOLVR_OPERATOR_P3_MULTIPOLE2LOCAL_P3_HPP

#include "fmsolvr/container/linear_array.hpp"
#include "fmsolvr/math/reciprocal.hpp"
#include "fmsolvr/operator/p3/multipole2local_p3_impl.hpp"
#include "fmsolvr/operator/rotation/rotation_generator.hpp"
#include "fmsolvr/operator/rotation/rotation_matrix.hpp"
#include "fmsolvr/operator/rotation/rotation_matrix_vector.hpp"
#include "fmsolvr/simd/simd_engine.hpp"

namespace fmsolvr
{

// Implements the rotation-based M2L operator. The operation consists of
// three steps:
// - Rotation of the reference frame of the multipole expansion
//     so that the z-axis is aligned with the shift vector
// - Application of the simplified M2L operator
// - Backward rotation of the local expansion into the original reference frame
// Those three steps are represented by 2 operators, a RotationOperator
// representing the two rotations and an M2Lp3_Operator representing the
// simplified M2L operator.

// Rotation-based M2L operator. Shifts a multipole expansion
// along the quantization axis and converts it to a local expansion.
// Holds precomputed factors (i!/R^{i+1}) for i = 0,...,2*p where R is
// the length of the shift vector.
template <typename Real, typename Allocator = std::allocator<Real> >
class M2Lp3_Operator : public LinearArray<Real, Allocator>
{
    typedef LinearArray<Real, Allocator> Base;
    using typename Base::size_type;

public:
    M2Lp3_Operator(const Real R, const size_t p, const Real m2l_operator_scaling_factor = Real(1.))
            : Base(2 * p)
    {
        Real R_recip = reciprocal(R);
        (*this)[0] = R_recip * m2l_operator_scaling_factor;
        for (size_t i = 1; i <= 2 * p; ++i)
        {
            (*this)[i] = (*this)[i - 1] * Real(i) * R_recip;
        }
    }

    template <typename Real3>
    M2Lp3_Operator(
            const Real3 &expansion_point_source,
            const Real3 &expansion_point_target,
            const size_t p)
            : M2Lp3_Operator(length(expansion_point_target - expansion_point_source), p)
    { }

    size_type p() const
    { return Base::p() / 2; }
};


// Rotation-based M2L operator for a given source and target point. All operators
// are computed on the fly, which makes the opration costly.
template <typename Real3,
          typename CoefficientMatrix,
          typename Scratch>
void M2Lp3(
        const CoefficientMatrix &omega_source,
        const Real3 &expansion_point_source,
        CoefficientMatrix &mu_target,
        const Real3 &expansion_point_target,
        const size_t p,
        Scratch &scratch)
{
    typedef typename CoefficientMatrix::value_type complex_type;
    typedef typename complex_type::value_type scalar_type;
    typedef RotationMatrixUpper<scalar_type> rotation_matrix;
    typedef RotationMatrixVector<rotation_matrix> rotation_matrix_vector;
    typedef RG::RotationGeneratorNoPrecomp<rotation_matrix_vector, Real3> rotation_generator;
    typedef typename rotation_generator::rotation_operator rotation_operator;

    rotation_generator G(p);
    rotation_operator R = G.generate_rotation(expansion_point_source - expansion_point_target);
    M2Lp3_Operator<scalar_type> B(expansion_point_source, expansion_point_target, p);

    M2Lp3<simd_engine_none>(omega_source, R, B, mu_target, p, scratch);
}


// Rotation-based M2L operator for a given rotation operator and simplified
// M2L operator. The operator matrices can be precomputed and reused.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2LOperator,
          typename LocalCoefficientsOut,
          typename Scratch>
void M2Lp3(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2LOperator &B,
        LocalCoefficientsOut &mu_target,
        const size_t p,
        Scratch &scratch)
{
    M2Lp3_impl_default<SIMDengine>(omega_source, R, B, mu_target, p, scratch);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
