/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P3_MULTIPOLE2LOCAL_P3_IMPL_HPP
#define FMSOLVR_OPERATOR_P3_MULTIPOLE2LOCAL_P3_IMPL_HPP

#include "fmsolvr/operator/p3/multipole2local_p3_shift.hpp"
#include "fmsolvr/operator/rotation/rotation.hpp"

#if defined(M2LP3_IMPL_DUMP)
#include "fmsolvr/util/dump_helpers.hpp"

#include <iomanip>
#include <iostream>
#include <limits>
#endif

namespace fmsolvr {

// Applies the rotation-based M2L operator given by the RotationOperator R
// and the simplified M2L operator to the multipole expansion given in omega_in
// and writes the results to mu_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = source - target.
// This is an accumulating operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2LOperator,
          typename LocalCoefficientsOut>
void M2Lp3_impl_reference(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2LOperator &B,
        LocalCoefficientsOut &mu_target,
        const size_t p)
{
    MultipoleCoefficientsIn omega_source_rot(p);
    LocalCoefficientsOut mu_target_rot(p);

    ApplyRotationMultipoleForward_reference(omega_source, R, omega_source_rot, p);
    omega_source_rot.populate_lower();  // M2L requires lower triangle.
    M2Lp3_apply_B(omega_source_rot, B, mu_target_rot, p);
    mu_target_rot.populate_lower();
    ApplyRotationLocalBackward_reference(mu_target_rot, R, mu_target, p);
}


// Applies the rotation-based M2L operator given by the RotationOperator R
// and the simplified M2L operator to the multipole expansion given in omega_in
// and writes the results to mu_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = source - target.
// This is an accumulating operation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2LOperator,
          typename LocalCoefficientsOut,
          typename Scratch>
void M2Lp3_impl_folded(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2LOperator &B,
        LocalCoefficientsOut &mu_target,
        const size_t p,
        Scratch &scratch)
{
    auto &omega_source_rot = scratch.omuga_source_rot();
    auto &mu_target_rot = scratch.omuga_target_rot();
    auto &scratcharray = scratch.array();

    ApplyRotationMultipoleForward_folded<SIMDengine>(
            omega_source, R, omega_source_rot, p, scratcharray);
    M2Lp3_apply_B_mlcached<SIMDengine>(omega_source_rot, B, mu_target_rot, p, scratcharray);
    ApplyRotationLocalBackward_folded<SIMDengine>(mu_target_rot, R, mu_target, p);
}


#if defined(M2LP3_IMPL_DUMP)
// Applies the rotation-based M2L operator given by the RotationOperator R
// and the simplified M2L operator to the multipole expansion given in omega_in
// and writes the results to mu_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = source - target.
// This is an accumulating operation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2LOperator,
          typename LocalCoefficientsOut,
          typename Scratch>
void M2Lp3_impl_folded_dump(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2LOperator &B,
        LocalCoefficientsOut &mu_target,
        const size_t p,
        Scratch &scratch)
{
    auto &omega_source_rot = scratch.omuga_source_rot();
    auto &mu_target_rot = scratch.omuga_target_rot();
    auto &scratcharray = scratch.array();

    std::cout << "# omega_source" << std::endl;
    dump_list_upper_triangle_complex(omega_source, p);
    std::cout << "# R.FoldedCoefficientsRotateMultipoleForward()" << std::endl;
    dump_list_quarter_pyramid_real2(R.FoldedCoefficientsRotateMultipoleForward(), p);
    std::cout << "# R.e_ikphi()" << std::endl;
    dump_list_array_complex(R.e_ikphi(), p);

    ApplyRotationMultipoleForward_folded<SIMDengine>(
            omega_source, R, omega_source_rot, p, scratcharray);

    std::cout << "# omega_source_rot" << std::endl;
    dump_list_upper_triangle_complex(omega_source_rot, p);
    std::cout << "# B" << std::endl;
    dump_list_array(B, 2 * p - 1);

    M2Lp3_apply_B_mlcached<SIMDengine>(omega_source_rot, B, mu_target_rot, p, scratcharray);

    std::cout << "# mu_target_rot" << std::endl;
    dump_list_upper_triangle_complex(mu_target_rot, p);
    std::cout << "# R.FoldedCoefficientsRotateLocalBackward()" << std::endl;
    dump_list_quarter_pyramid_real2(R.FoldedCoefficientsRotateLocalBackward(), p);
    std::cout << "# R.e_ikphi()" << std::endl;
    dump_list_array_complex(R.e_ikphi(), p);

    ApplyRotationLocalBackward_folded<SIMDengine>(mu_target_rot, R, mu_target, p);

    std::cout << "# mu_target" << std::endl;
    dump_list_upper_triangle_complex(mu_target, p);
}
#endif


// Selects the default M2Lp3 implementation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2LOperator,
          typename LocalCoefficientsOut,
          typename Scratch>
void M2Lp3_impl_default(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2LOperator &B,
        LocalCoefficientsOut &mu_target,
        const size_t p,
        Scratch &scratch)
{
#ifdef FORCE_VALIDATE_REQUIRE_LOWER
    omega_source.require_lower();
#endif
    M2Lp3_impl_folded<SIMDengine>(omega_source, R, B, mu_target, p, scratch);
    //M2Lp3_impl_reference(omega_source, R, B, mu_target, p);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
