/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P3_MULTIPOLE2LOCAL_P3_SHIFT_HPP
#define FMSOLVR_OPERATOR_P3_MULTIPOLE2LOCAL_P3_SHIFT_HPP

#include <cassert>

#include "fmsolvr/math/Complex.hpp"
#include "fmsolvr/math/parity_sign.hpp"

namespace fmsolvr
{

// Apply M2L shift along the quantization axis.
// This is an assigning operation.
template <typename MultipoleCoefficientsIn,
          typename OperatorMatrix,
          typename LocalCoefficientsOut>
__attribute__((flatten))
void M2Lp3_apply_B(
        const MultipoleCoefficientsIn &omega_in,
        const OperatorMatrix &B,
        LocalCoefficientsOut &mu_out,
        const size_t p_compute)
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);

    assert(p_compute <= omega_in.p());
    assert(p_compute <= B.p());
    assert(p_compute <= mu_out.p());
    omega_in.require_lower();
    mu_out.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type mu_l_m(0.);
            for (index_type j = m; j <= p; ++j)
            {
                mu_l_m += toggle_sign_if_odd(j, omega_in.lower(j, -m) * B[l + j]);
            }
            mu_out.upper(l, m) = mu_l_m;
        }
    }
}


// Apply M2L shift along the quantization axis.
// This is an assigning operation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename OperatorMatrix,
          typename LocalCoefficientsOut,
          typename ScratchArray>
__attribute__((flatten))
void M2Lp3_apply_B_mlcached(
        const MultipoleCoefficientsIn &omega_in,
        const OperatorMatrix &B,
        LocalCoefficientsOut &mu_out,
        const size_t p_compute,
        ScratchArray &scratch)
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    auto &omega_m = scratch;

    static_assert(
            std::is_same<complex_type, typename ScratchArray::value_type>::value,
            "mismatching scratch value_type");
    assert(p_compute <= omega_in.p());
    assert(p_compute <= B.p());
    assert(p_compute <= mu_out.p());
    assert(p_compute <= omega_m.p());
    mu_out.invalidate_lower();

    for (index_type m = 0; m <= p; ++m)
    {
        for (index_type j = m; j <= p; ++j)
        {
            omega_m[j] = toggle_sign_if_odd(j + m, fmm::conj(omega_in.upper(j, m)));
        }
        for (index_type l = m; l <= p; ++l)
        {
            complex_type mu_l_m(0.);
            for (index_type j = m; j <= p; ++j)
            {
                mu_l_m += omega_m[j] * B[l + j];
            }
            mu_out.upper(l, m) = mu_l_m;
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
