/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P3_MULTIPOLE2MULTIPOLE_P3_HPP
#define FMSOLVR_OPERATOR_P3_MULTIPOLE2MULTIPOLE_P3_HPP

#include "fmsolvr/container/linear_array.hpp"
#include "fmsolvr/operator/p3/multipole2multipole_p3_impl.hpp"
#include "fmsolvr/operator/rotation/rotation_generator.hpp"
#include "fmsolvr/operator/rotation/rotation_matrix.hpp"
#include "fmsolvr/operator/rotation/rotation_matrix_vector.hpp"
#include "fmsolvr/simd/simd_engine.hpp"

namespace fmsolvr {

// Implements the rotation-based M2M operator. The operation consists of
// three steps:
// - Rotation of the reference frame so that the z-axis is aligned with the shift vector
// - Application of the simplified M2M operator
// - Backward rotation into the original reference frame
// Those three steps are represented by 2 operators, a RotationOperator
// representing the two rotations and an M2Mp3_Operator representing the
// simplified M2M operator.

// M2M-operator for a shift along the quantization axis
// of the multipole expansion.
// Holds precomputed factors (R^i/i!) for i = 0,...,p where R is
// the length of the shift vector.
template <typename Real, typename Allocator = std::allocator<Real> >
class M2Mp3_Operator : public LinearArray<Real, Allocator>
{
    typedef LinearArray<Real, Allocator> Base;

    M2Mp3_Operator(const Real R, const size_t p) : Base(p)
    {
        (*this)[0] = Real(1.);
        for (size_t i = 1; i <= p; ++i)
        {
            (*this)[i] = R * (*this)[i - 1] / Real(i);
        }
    }

public:
    template <typename Real3>
    M2Mp3_Operator(
            const Real3 &expansion_point_source,
            const Real3 &expansion_point_target,
            const size_t p)
            : M2Mp3_Operator(length(expansion_point_target - expansion_point_source), p)
    { }
};


// Rotation-based M2M operator for a given source and target point. All operators
// are computed on the fly, which makes the opration costly.
template <typename Real3,
          typename CoefficientMatrix,
          typename Scratch>
void M2Mp3(
        const CoefficientMatrix &omega_source,
        const Real3 &expansion_point_source,
        CoefficientMatrix &omega_target,
        const Real3 &expansion_point_target,
        const size_t p,
        Scratch &scratch)
{
    typedef typename CoefficientMatrix::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef RotationMatrixUpper<real_type> rotation_matrix;
    typedef RotationMatrixVector<rotation_matrix> rotation_matrix_vector;
    typedef RG::RotationGeneratorNoPrecomp<rotation_matrix_vector, Real3> rotation_generator;
    typedef typename rotation_generator::rotation_operator rotation_operator;

    rotation_generator G(p);
    rotation_operator R = G.generate_rotation(expansion_point_source - expansion_point_target);
    M2Mp3_Operator<real_type> A(expansion_point_source, expansion_point_target, p);

    M2Mp3<simd_engine_none>(omega_source, R, A, omega_target, p, scratch);
}


// Rotation-based M2M operator for a given rotation operator and simplified
// M2M operator. The operator matrices can be precomputed and reused.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2MOperator,
          typename MultipoleCoefficientsOut,
          typename Scratch>
void M2Mp3(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2MOperator &A,
        MultipoleCoefficientsOut &omega_target,
        const size_t p,
        Scratch &scratch)
{
    M2Mp3_impl_default<SIMDengine>(omega_source, R, A, omega_target, p, scratch);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
