/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P3_MULTIPOLE2MULTIPOLE_P3_IMPL_HPP
#define FMSOLVR_OPERATOR_P3_MULTIPOLE2MULTIPOLE_P3_IMPL_HPP

#include <memory>

#include "fmsolvr/operator/p3/multipole2multipole_p3_shift.hpp"
#include "fmsolvr/operator/rotation/rotation.hpp"

namespace fmsolvr
{

// Applies the rotation-based M2M operator given by the RotationOperator R
// and the simplified M2M operator to the multipole expansion given in omega_in
// and writes the results to omega_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = source - target.
// This is an accumulating operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2MOperator,
          typename MultipoleCoefficientsOut>
void M2Mp3_impl_reference(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2MOperator &A,
        MultipoleCoefficientsOut &omega_target,
        const size_t p)
{
    MultipoleCoefficientsIn omega_source_rot(p);
    MultipoleCoefficientsOut omega_target_rot(p);

    ApplyRotationMultipoleForward_reference(omega_source, R, omega_source_rot, p);
    M2Mp3_apply_A(omega_source_rot, A, omega_target_rot, p);
    omega_target_rot.populate_lower();
    ApplyRotationMultipoleBackward_reference(omega_target_rot, R, omega_target, p);
}


template <typename CoefficientMatrix>
class M2Mp3_functor_reference
{
public:
    typedef CoefficientMatrix scratch1_type;
    typedef CoefficientMatrix scratch2_type;

    template <typename RotationOperator, typename M2MOperator>
    void operator()(
            const CoefficientMatrix &omega_source,
            const RotationOperator &R,
            const M2MOperator &A,
            CoefficientMatrix &omega_target,
            const size_t p,
            scratch1_type &omega_source_rot,
            scratch2_type &omega_target_rot)
    {
        ApplyRotationMultipoleForward(omega_source, R, omega_source_rot, p);
        M2Mp3_apply_A(omega_source_rot, A, omega_target_rot, p);
        omega_target_rot.populate_lower();
        ApplyRotationMultipoleBackward(omega_target_rot, R, omega_target, p);
    }
};

// Applies the rotation-based M2M operator given by the RotationOperator R
// and the simplified M2M operator to the multipole expansion given in omega_in
// and writes the results to omega_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = source - target.
// This is an accumulating operation.
template <typename CoefficientMatrix,
          typename RotationOperator,
          typename M2MOperator>
void M2Mp3_impl_reference_functor(
        const CoefficientMatrix &omega_source,
        const RotationOperator &R,
        const M2MOperator &A,
        CoefficientMatrix &omega_target,
        const size_t p)
{
    typedef M2Mp3_functor_reference<CoefficientMatrix> M2Mp3functor;
    typedef typename M2Mp3functor::scratch1_type scratch1_type;
    typedef typename M2Mp3functor::scratch2_type scratch2_type;

    std::allocator<int> Alloc;
    auto scratch1 = std::allocate_shared<scratch1_type>(Alloc, p);
    auto scratch2 = std::allocate_shared<scratch2_type>(Alloc, p);

    M2Mp3functor()(omega_source, R, A, omega_target, p, *scratch1, *scratch2);
}


// Applies the rotation-based M2M operator given by the RotationOperator R
// and the simplified M2M operator to the multipole expansion given in omega_in
// and writes the results to omega_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = source - target.
// This is an accumulating operation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2MOperator,
          typename MultipoleCoefficientsOut,
          typename Scratch>
void M2Mp3_impl_precomp(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2MOperator &A,
        MultipoleCoefficientsOut &omega_target,
        const size_t p,
        Scratch &scratch)
{
    auto &omega_source_rot = scratch.omuga_source_rot();
    // scratch does not guarantee access to the lower part
    MultipoleCoefficientsOut omega_target_rot(p);

    ApplyRotationMultipoleForward_phi_f(omega_source, R, omega_source_rot, p);
    M2Mp3_apply_A(omega_source_rot, A, omega_target_rot, p);
    omega_target_rot.populate_lower();
    ApplyRotationMultipoleBackward_phi_f(omega_target_rot, R, omega_target, p);
}


// Applies the rotation-based M2M operator given by the RotationOperator R
// and the simplified M2M operator to the multipole expansion given in omega_in
// and writes the results to omega_out. Will only yield a correct result if the angles
// theta and phi of the RotationOperator correspond to the azimuthal and polar angles of the
// shift vector d = source - target.
// This is an accumulating operation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2MOperator,
          typename MultipoleCoefficientsOut,
          typename Scratch>
void M2Mp3_impl_folded(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2MOperator &A,
        MultipoleCoefficientsOut &omega_target,
        const size_t p,
        Scratch &scratch)
{
    auto &omega_source_rot = scratch.omuga_source_rot();
    auto &omega_target_rot = scratch.omuga_target_rot();
    auto &scratcharray = scratch.array();

    ApplyRotationMultipoleForward_folded<SIMDengine>(
            omega_source, R, omega_source_rot, p, scratcharray);
    M2Mp3_apply_A_ml<SIMDengine>(omega_source_rot, A, omega_target_rot, p);
    ApplyRotationMultipoleBackward_folded<SIMDengine>(omega_target_rot, R, omega_target, p);
}


// Selects the default M2Mp3 implementation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename M2MOperator,
          typename MultipoleCoefficientsOut,
          typename Scratch>
void M2Mp3_impl_default(
        const MultipoleCoefficientsIn &omega_source,
        const RotationOperator &R,
        const M2MOperator &A,
        MultipoleCoefficientsOut &omega_target,
        const size_t p,
        Scratch &scratch)
{
#ifdef FORCE_VALIDATE_REQUIRE_LOWER
    omega_source.require_lower();
#endif
    M2Mp3_impl_folded<SIMDengine>(omega_source, R, A, omega_target, p, scratch);
    //M2Mp3_impl_precomp<SIMDengine>(omega_source, R, A, omega_target, p, scratch);
    //M2Mp3_impl_reference(omega_source, R, A, omega_target, p);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
