/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P3_MULTIPOLE2MULTIPOLE_P3_SHIFT_HPP
#define FMSOLVR_OPERATOR_P3_MULTIPOLE2MULTIPOLE_P3_SHIFT_HPP

#include <cstddef>

namespace fmsolvr
{

// Apply simplified M2M operator.
// This is an assigning operation.
template <typename MultipoleCoefficientsIn,
          typename OperatorMatrix,
          typename MultipoleCoefficientsOut>
__attribute__((flatten))
void M2Mp3_apply_A(
        const MultipoleCoefficientsIn &omega_in,
        const OperatorMatrix &A,
        MultipoleCoefficientsOut &omega_out,
        const size_t p_compute)
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);

    assert(p_compute <= omega_in.p());
    assert(p_compute <= A.p());
    assert(p_compute <= omega_out.p());
    omega_out.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type omega_l_m(0.);
            for (index_type j = m; j <= l; ++j)
            {
                omega_l_m += omega_in.upper(j, m) * A[l - j];
            }
            omega_out.upper(l, m) = omega_l_m;
        }
    }
}


// Apply simplified M2M operator.
// This is an assigning operation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename OperatorMatrix,
          typename MultipoleCoefficientsOut>
__attribute__((flatten))
void M2Mp3_apply_A_ml(
        const MultipoleCoefficientsIn &omega_in,
        const OperatorMatrix &A,
        MultipoleCoefficientsOut &omega_out,
        const size_t p_compute)
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);

    assert(p_compute <= omega_in.p());
    assert(p_compute <= A.p());
    assert(p_compute <= omega_out.p());
    omega_out.invalidate_lower();

    for (index_type m = 0; m <= p; ++m)
    {
        for (index_type l = m; l <= p; ++l)
        {
            complex_type omega_l_m(0.);
            for (index_type j = m; j <= l; ++j)
            {
                omega_l_m += omega_in.upper(j, m) * A[l - j];
            }
            omega_out.upper(l, m) = omega_l_m;
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
