/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P4_local2local_hpp
#define FMSOLVR_OPERATOR_P4_local2local_hpp

#include <cassert>
#include <algorithm>

#include "fmsolvr/operator/particle2multipole.hpp"

namespace fmsolvr
{

// L2L(mu, (0, 0, 0)) is an idempotent operation

// Note: In contrast to P2M, P2L, M2M, and M2L that all use (source - target),
// for L2L it is required to use (target - source) instead. Really.

// expansion point shift vector orientation:
// for a point xyz reasonably far away from expansion_point_source (and expansion_point_target)
// P2L (@expansion_point_source) followed by L2L (@expansion_point_target)
//        P2L(make_xyzq(xyz - expansion_point_source, q), m_s, p);
//        L2L_reference(m_s, expansion_point_target - expansion_point_source, m_st, p);
// is equivalent to direct P2L (@expansion_point_target) (within numerics)
//        P2L(make_xyzq(xyz - expansion_point_target, q), m_t, p);

// all values O(l,m) with (l < 0) or (|m| > l) are defined as (0,0i)
// but they are not being accessed anyway
template <typename CoefficientMatrix>
class L2L_Operator : public CoefficientMatrix
{
public:
    template <typename Real3>
    L2L_Operator(
        const Real3 & expansion_point_source,
        const Real3 & expansion_point_target,
        const size_t p)
    : CoefficientMatrix(p)
    {
        rawP2M_nodivtable(expansion_point_target - expansion_point_source, *static_cast<CoefficientMatrix *>(this), p);
        this->populate_lower();
    }
};


// ensure the correct orientation of the shifting vector
template <typename Real3, typename CoefficientMatrix>
void L2L(
        const CoefficientMatrix & mu_source,
        const Real3 & expansion_point_source,
        CoefficientMatrix & mu_target,
        const Real3 & expansion_point_target,
        const ssize_t p,
        const ssize_t p_out)
{
    assert(p >= p_out);
    L2L_Operator<CoefficientMatrix> C(expansion_point_source, expansion_point_target, p);
    L2L_impl_reference(mu_source, C, mu_target, p, p_out);
}


// ensure the correct orientation of the shifting vector
template <typename Real3, typename CoefficientMatrix>
void L2L(
        const CoefficientMatrix & mu_source,
        const Real3 & expansion_point_source,
        CoefficientMatrix & mu_target,
        const Real3 & expansion_point_target,
        const ssize_t p)
{
    L2L_Operator<CoefficientMatrix> C(expansion_point_source, expansion_point_target, p);
    L2L_impl_reference(mu_source, C, mu_target, p, p);
}


// p^4
template <typename CoefficientMatrix>
void L2L(
        const CoefficientMatrix & mu_source,
        const L2L_Operator<CoefficientMatrix> & C,
        CoefficientMatrix & mu_target,
        const ssize_t p)
{
    L2L_impl_reference(mu_source, C, mu_target, p, p);
}


// p^4 (actually p^2 * p_out^2), k=-j..j
template <typename CoefficientMatrix>
void L2L_impl_reference(
        const CoefficientMatrix & mu_in,
        const L2L_Operator<CoefficientMatrix> & C,
        CoefficientMatrix & mu_out,
        const ssize_t p,
        const ssize_t p_out_)
{
    mu_in.require_lower();
    C.require_lower();
    mu_out.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    // fail if mu_out(0,0) cannot be computed correctly due to insufficient input order
    assert(p <= std::min<ssize_t>(mu_in.p(), C.p()));
    // limit output order to values that can be stored and computed correctly
    const ssize_t p_out = std::min<ssize_t>(std::min(p_out_, p), mu_out.p());

    for (ssize_t l = 0; l <= p_out; ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            complex_type mu_l_m(0.);
            for (ssize_t j = l; j <= p; ++j) {
                const ssize_t k_min = m - (j - l);
                const ssize_t k_max = m + (j - l);
                for (ssize_t k = k_min; k <= k_max; ++k)
                    mu_l_m += C.get(j - l, k - m) * mu_in.get(j, k);
            }
            mu_out.upper(l, m) += mu_l_m;
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
