/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P4_local2local_derivative_hpp
#define FMSOLVR_OPERATOR_P4_local2local_derivative_hpp

#include <sys/types.h>

namespace fmsolvr
{

template <typename CoefficientMatrix>
void DxL2L_reference(
        const CoefficientMatrix & mu,
        CoefficientMatrix & dmu,
        const ssize_t p)
{
    mu.require_lower();
    dmu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;

    for (ssize_t l = 0; l < p; ++l) {
        for (ssize_t m = 0; m <= l; ++m)
            dmu.upper(l, m) = (mu.get(l + 1, m + 1) - mu.get(l + 1, m - 1)) * scalar_type(-0.5);
    }
    for (ssize_t m = 0; m <= p; ++m)
        dmu.upper(p, m) = complex_type(0.);
}

template <typename CoefficientMatrix>
void DyL2L_reference(
        const CoefficientMatrix & mu,
        CoefficientMatrix & dmu,
        const ssize_t p)
{
    mu.require_lower();
    dmu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;

    for (ssize_t l = 0; l < p; ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            complex_type tmp = (mu.get(l + 1, m + 1) + mu.get(l + 1, m - 1)) * scalar_type(0.5);
            dmu.upper(l, m) = complex_type(-tmp.imag(), tmp.real());
        }
    }
    for (ssize_t m = 0; m <= p; ++m)
        dmu.upper(p, m) = complex_type(0.);
}

template <typename CoefficientMatrix>
void DzL2L_reference(
        const CoefficientMatrix & mu,
        CoefficientMatrix & dmu,
        const ssize_t p)
{
    mu.require_lower();
    dmu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;

    for (ssize_t l = 0; l < p; ++l) {
        for (ssize_t m = 0; m <= l; ++m)
            dmu.upper(l, m) = -mu.get(l + 1, m);
    }
    for (ssize_t m = 0; m <= p; ++m)
        dmu.upper(p, m) = complex_type(0.);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
