/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P4_multipole2local_hpp
#define FMSOLVR_OPERATOR_P4_multipole2local_hpp

#include <algorithm>

#include "fmsolvr/math/parity_sign.hpp"
#include "fmsolvr/operator/particle2local.hpp"

namespace fmsolvr
{

template <typename CoefficientMatrix>
class M2L_Operator : public CoefficientMatrix
{
public:
    template <typename Real3>
    M2L_Operator(
        const Real3 & xyz,
            const size_t p,
            const typename Real3::value_type m2l_operator_scaling_factor =
                    typename Real3::value_type(1.))
            : CoefficientMatrix(2 * p)
    {
        rawP2L(xyz, *this, 2 * p, m2l_operator_scaling_factor);
        this->populate_lower();
    }

    template <typename Real3>
    M2L_Operator(
        const Real3 & expansion_point_source,
        const Real3 & expansion_point_target,
            const size_t p,
            const typename Real3::value_type m2l_operator_scaling_factor =
                    typename Real3::value_type(1.))
            : M2L_Operator(
                      expansion_point_source - expansion_point_target,
                      p,
                      m2l_operator_scaling_factor)
    { }
};


// problematic to get the shifting vector xyz in the right orientation
template <typename Real3, typename CoefficientMatrix>
void M2L_reference_DONOTUSE(
        const CoefficientMatrix & omega,
        const Real3 & xyz,
        CoefficientMatrix & mu,
        const ssize_t p)
{
    M2L_Operator<CoefficientMatrix> B(xyz, p);
    M2L_reference(omega, B, mu, p);
}


// ensure the correct orientation of the shifting vector
template <typename Real3, typename CoefficientMatrix>
void M2L_reference(
        const CoefficientMatrix & omega_source,
        const Real3 & expansion_point_source,
        CoefficientMatrix & mu_target,
        const Real3 & expansion_point_target,
        const ssize_t p)
{
    M2L_Operator<CoefficientMatrix> B(expansion_point_source, expansion_point_target, p);
    M2L_reference(omega_source, B, mu_target, p);
}


// p^4, k=-j..j, with custom B operator
// classical M2L operation on actual *charges*, does include the (-1)^j factor
template <typename CoefficientMatrix, typename OperatorMatrix>
__attribute__ ((noinline))
void M2L_reference(
        const CoefficientMatrix & omega,
        const OperatorMatrix & B,
        CoefficientMatrix & mu,
        const ssize_t p)
{
    omega.require_lower();
    B.require_lower();
    mu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    const ssize_t p_omega = std::min<ssize_t>(p, omega.p());
    const ssize_t p_B = B.p();
    const ssize_t p_mu = std::min<ssize_t>(p, mu.p());

    for (ssize_t l = 0; l <= p_mu; ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            complex_type mu_l_m(0.);
            const ssize_t j_max = std::min(p_omega, p_B - l);
            for (ssize_t j = 0; j <= j_max; ++j) {
                complex_type mu_l_m_j(0.);
                for (ssize_t k = -j; k <= j; ++k)
                    mu_l_m_j += B.get(j + l, k + m) * omega.get(j, k);
                mu_l_m += toggle_sign_if_odd(j, mu_l_m_j);
            }
            mu.upper(l, m) += toggle_sign_if_odd(0, mu_l_m);
        }
    }
}


// p^4, k=-j..j, with custom B operator
// raw == for operator construction (chargeless), does *not* include (-1)^{j,l} factor
template <typename CoefficientMatrix, typename OperatorMatrix>
void rawM2L_reference(
        const CoefficientMatrix & omega,
        const OperatorMatrix & B,
        CoefficientMatrix & mu,
        const ssize_t p)
{
    omega.require_lower();
    B.require_lower();
    mu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    const ssize_t p_omega = std::min<ssize_t>(p, omega.p());
    const ssize_t p_B = B.p();
    const ssize_t p_mu = std::min<ssize_t>(p, mu.p());

    for (ssize_t l = 0; l <= p_mu; ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            complex_type mu_l_m(0.);
            const ssize_t j_max = std::min(p_omega, p_B - l);
            for (ssize_t j = 0; j <= j_max; ++j) {
                for (ssize_t k = -j; k <= j; ++k)
                    mu_l_m += B.get(j + l, k + m) * omega.get(j, k);
            }
            mu.upper(l, m) += mu_l_m;
        }
    }
}


// classical M2L operation on actual *charges*, does include the (-1)^l factor
template <typename CoefficientMatrix>
void
__attribute__ ((noinline))
__attribute ((flatten))
    single_M2L(
        const CoefficientMatrix & omega,
        const CoefficientMatrix & B,
        CoefficientMatrix & mu,
        const ssize_t p)
{
    M2L_reference(omega, B, mu, p);
}


// p^4, k=-j..j, with custom B operator
// raw == for operator construction (chargeless), does *not* include (-1)^l factor
template <typename CoefficientMatrix>
void
__attribute__ ((noinline))
__attribute ((flatten))
    single_rawM2L(
        const CoefficientMatrix & omega,
        const CoefficientMatrix & B,
        CoefficientMatrix & mu,
        const ssize_t p)
{
    rawM2L_reference(omega, B, mu, p);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
