/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P4_multipole2local_optimized_hpp
#define FMSOLVR_OPERATOR_P4_multipole2local_optimized_hpp

#include "fmsolvr/container/simple_vector.hpp"
#include "fmsolvr/container/triangular_array.hpp"
#include "fmsolvr/math/Complex.hpp"
#include "fmsolvr/misc/split_complex_mul_acc_simd_asm_m256d.hpp"
#include "fmsolvr/operator/p4/multipole2local.hpp"
#include "fmsolvr/simd/Vec.hpp"

namespace fmsolvr
{

// round n up to the next multiple of padd
inline ssize_t pad(ssize_t n, ssize_t padd)
{
    n += padd - 1;
    return n - n % padd;
}


// p^4, k=-j..j, with custom B operator
// classical M2L operation on actual *charges*, does include the (-1)^l factor
template <typename CoefficientMatrix, typename OperatorMatrix>
void M2L_toggle_l_THIS_IS_WRONG(
        const CoefficientMatrix & omega,
        const OperatorMatrix & B,
        CoefficientMatrix & mu,
        const ssize_t p,
        const ssize_t pB)
{
    omega.require_lower();
    B.require_lower();
    mu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;

    for (ssize_t l = 0; l <= p; ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            complex_type mu_l_m(0.);
            const ssize_t j_max = std::min(p, pB - l);
            for (ssize_t j = 0; j <= j_max; ++j) {
                for (ssize_t k = -j; k <= j; ++k)
                    mu_l_m += B.get(j + l, k + m) * omega.get(j, k);
            }
            mu.upper(l, m) += toggle_sign_if_odd(l, mu_l_m);
        }
    }
}


// FIXME: all optimized variants are still toggle_l based !!! XXX

// p^4, k=-j..j, with custom B operator
template <typename CoefficientMatrix, typename OperatorMatrix>
void single_M2L_SIMD_lmjk(
        const CoefficientMatrix & omega,
        const OperatorMatrix & B,
        CoefficientMatrix & mu,
        const ssize_t p,
        const ssize_t pB)
{
    omega.require_lower();
    B.require_lower();
    mu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;
    typedef typename simd_engine_default<scalar_type>::type simd_base_type;
    typedef simd_base_type simd_type;
#if 1
    typedef SuperSIMD<simd_base_type, 1> super_simd_type;
#else
    typedef SuperSIMD<simd_base_type, 2> super_simd_type;
#endif
    typedef TriangularArray<scalar_type, void> scalar_triangle_type;
    typedef typename scalar_triangle_type::const_column_iterator const_column_iterator;
    Vec_traits<simd_type> VT;
    Vec_traits<super_simd_type> SuVT;

    scalar_triangle_type omega_re(p);
    scalar_triangle_type omega_im(p);
    scalar_triangle_type B_re(pB);
    scalar_triangle_type B_im(pB);

    for (ssize_t l = 0; l <= p; ++l) {
        for (ssize_t m = -l; m <= l; ++m) {
            complex_type olm = omega.get(l, m);
            omega_re.get(l, m) = olm.real();
            omega_im.get(l, m) = olm.imag();
        }
    }

    for (ssize_t l = 0; l <= pB; ++l) {
        for (ssize_t m = -l; m <= l; ++m) {
            complex_type Blm = B.get(l, m);
            B_re.get(l, m) = Blm.real();
            B_im.get(l, m) = Blm.imag();
        }
    }

    for (ssize_t l = 0; l <= p; ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            scalar_type mu_l_m_rr = 0.;
            scalar_type mu_l_m_ri = 0.;
            scalar_type mu_l_m_ir = 0.;
            scalar_type mu_l_m_ii = 0.;
            simd_type mu_l_m_rr_ = VT.zero();
            simd_type mu_l_m_ri_ = VT.zero();
            simd_type mu_l_m_ir_ = VT.zero();
            simd_type mu_l_m_ii_ = VT.zero();
            const ssize_t j_max = std::min(p, pB - l);
            for (ssize_t j = 0; j <= j_max; ++j) {
                //for (ssize_t k = -j; k <= j; ++k)
                //    mu_l_m += B.get(j + l, k + m) * omega.get(j, k);
                const ssize_t k_begin = -j;
                const ssize_t k_end = j + 1;
                const_column_iterator omega_re_col_j = omega_re.get_column_iterator(j, k_begin);
                const_column_iterator omega_im_col_j = omega_im.get_column_iterator(j, k_begin);
                const_column_iterator B_re_col_j = B_re.get_column_iterator(j + l, k_begin + m);
                const_column_iterator B_im_col_j = B_im.get_column_iterator(j + l, k_begin + m);
                const ssize_t k0_begin = 0;
                const ssize_t k0_end = k_end - k_begin;
                const ssize_t k0_end_simd = k0_end - k0_end % SuVT.VecSize;
                ssize_t k0 = k0_begin;
                for ( ; k0 < k0_end_simd; k0 += SuVT.VecSize) {
                    super_simd_type re0_ = SuVT.unaligned_load(omega_re_col_j + k0);
                    super_simd_type im0_ = SuVT.unaligned_load(omega_im_col_j + k0);
                    super_simd_type re1_ = SuVT.unaligned_load(B_re_col_j + k0);
                    super_simd_type im1_ = SuVT.unaligned_load(B_im_col_j + k0);

                    mu_l_m_rr_ += SuVT.supersum(re0_ * re1_);
                    mu_l_m_ii_ += SuVT.supersum(im0_ * im1_);
                    mu_l_m_ri_ += SuVT.supersum(re0_ * im1_);
                    mu_l_m_ir_ += SuVT.supersum(re1_ * im0_);
                }
                for ( ; k0 < k0_end; ++k0) {
                    scalar_type re0 = omega_re_col_j[k0];
                    scalar_type im0 = omega_im_col_j[k0];
                    scalar_type re1 = B_re_col_j[k0];
                    scalar_type im1 = B_im_col_j[k0];

                    mu_l_m_rr += re0 * re1;
                    mu_l_m_ii += im0 * im1;
                    mu_l_m_ri += re0 * im1;
                    mu_l_m_ir += re1 * im0;
                }
            }
            mu_l_m_rr += VT.sum(mu_l_m_rr_);
            mu_l_m_ii += VT.sum(mu_l_m_ii_);
            mu_l_m_ri += VT.sum(mu_l_m_ri_);
            mu_l_m_ir += VT.sum(mu_l_m_ir_);
            mu.get(l, m) += toggle_sign_if_odd(l, complex_type(mu_l_m_rr - mu_l_m_ii, mu_l_m_ri + mu_l_m_ir));
        }
    }
}

// p^4, k=-j..j, with custom B operator
template <typename CoefficientMatrix, typename OperatorMatrix>
void single_M2L_SIMD_ljmk(
        const CoefficientMatrix & omega,
        const OperatorMatrix & B,
        CoefficientMatrix & mu,
        const ssize_t p,
        const ssize_t pB)
{
    omega.require_lower();
    B.require_lower();
    mu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;
    typedef typename simd_engine_default<scalar_type>::type simd_base_type;
    typedef simd_base_type simd_type;
#if 1
    typedef SuperSIMD<simd_base_type, 1> super_simd_type;
#else
    typedef SuperSIMD<simd_base_type, 2> super_simd_type;
#endif
    typedef TriangularArray<scalar_type, void> scalar_triangle_type;
    typedef typename scalar_triangle_type::const_column_iterator const_column_iterator;
    Vec_traits<simd_type> VT;
    Vec_traits<super_simd_type> SuVT;

    scalar_triangle_type omega_re(p);
    scalar_triangle_type omega_im(p);
    scalar_triangle_type B_re(pB);
    scalar_triangle_type B_im(pB);

    for (ssize_t l = 0; l <= p; ++l) {
        for (ssize_t m = -l; m <= l; ++m) {
            complex_type olm = omega.get(l, m);
            omega_re.get(l, m) = olm.real();
            omega_im.get(l, m) = olm.imag();
        }
    }

    for (ssize_t l = 0; l <= pB; ++l) {
        for (ssize_t m = -l; m <= l; ++m) {
            complex_type Blm = B.get(l, m);
            B_re.get(l, m) = Blm.real();
            B_im.get(l, m) = Blm.imag();
        }
    }

    for (ssize_t l = 0; l <= p; ++l) {
      const ssize_t j_max = std::min(p, pB - l);
      for (ssize_t j = 0; j <= j_max; ++j)
      {
        const ssize_t k_begin = -j;
        const ssize_t k_end = j + 1;
        const_column_iterator omega_re_col_j = omega_re.get_column_iterator(j, k_begin);
        const_column_iterator omega_im_col_j = omega_im.get_column_iterator(j, k_begin);
        const ssize_t m_begin = 0;
        const_column_iterator B_re_col_j_m0 = B_re.get_column_iterator(j + l, k_begin + m_begin);
        const_column_iterator B_im_col_j_m0 = B_im.get_column_iterator(j + l, k_begin + m_begin);

        for (ssize_t m = 0; m <= l; ++m) {
            scalar_type mu_l_m_rr = 0.;
            scalar_type mu_l_m_ri = 0.;
            scalar_type mu_l_m_ir = 0.;
            scalar_type mu_l_m_ii = 0.;
            simd_type mu_l_m_rr_ = VT.zero();
            simd_type mu_l_m_ri_ = VT.zero();
            simd_type mu_l_m_ir_ = VT.zero();
            simd_type mu_l_m_ii_ = VT.zero();
            {
                //for (ssize_t k = -j; k <= j; ++k)
                //    mu_l_m += B.get(j + l, k + m) * omega.get(j, k);
                const_column_iterator B_re_col_j = B_re_col_j_m0 + (m - m_begin);
                const_column_iterator B_im_col_j = B_im_col_j_m0 + (m - m_begin);
                const ssize_t k0_begin = 0;
                const ssize_t k0_end = k_end - k_begin;
                const ssize_t k0_end_simd = k0_end - k0_end % SuVT.VecSize;
                ssize_t k0 = k0_begin;
                for ( ; k0 < k0_end_simd; k0 += SuVT.VecSize) {
                    super_simd_type re0_ = SuVT.unaligned_load(omega_re_col_j + k0);
                    super_simd_type im0_ = SuVT.unaligned_load(omega_im_col_j + k0);
                    super_simd_type re1_ = SuVT.unaligned_load(B_re_col_j + k0);
                    super_simd_type im1_ = SuVT.unaligned_load(B_im_col_j + k0);

                    mu_l_m_rr_ += SuVT.supersum(re0_ * re1_);
                    mu_l_m_ii_ += SuVT.supersum(im0_ * im1_);
                    mu_l_m_ri_ += SuVT.supersum(re0_ * im1_);
                    mu_l_m_ir_ += SuVT.supersum(re1_ * im0_);
                }
                for ( ; k0 < k0_end; ++k0) {
                    scalar_type re0 = omega_re_col_j[k0];
                    scalar_type im0 = omega_im_col_j[k0];
                    scalar_type re1 = B_re_col_j[k0];
                    scalar_type im1 = B_im_col_j[k0];

                    mu_l_m_rr += re0 * re1;
                    mu_l_m_ii += im0 * im1;
                    mu_l_m_ri += re0 * im1;
                    mu_l_m_ir += re1 * im0;
                }
            }
            mu_l_m_rr += VT.sum(mu_l_m_rr_);
            mu_l_m_ii += VT.sum(mu_l_m_ii_);
            mu_l_m_ri += VT.sum(mu_l_m_ri_);
            mu_l_m_ir += VT.sum(mu_l_m_ir_);
            mu.get(l, m) += toggle_sign_if_odd(l, complex_type(mu_l_m_rr - mu_l_m_ii, mu_l_m_ri + mu_l_m_ir));
        }
      }
    }
}

// p^4, k=-j..j, with custom B operator
template <typename CoefficientMatrix, typename OperatorMatrix>
void single_M2L_SIMD_lmjk_ljmk(
        const CoefficientMatrix & omega,
        const OperatorMatrix & B,
        CoefficientMatrix & mu,
        const ssize_t p,
        const ssize_t pB)
{
    omega.require_lower();
    B.require_lower();
    mu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;
    typedef typename simd_engine_default<scalar_type>::type simd_base_type;
    typedef simd_base_type simd_type;
#if 1
    typedef SuperSIMD<simd_base_type, 1> super_simd_type;
#else
    typedef SuperSIMD<simd_base_type, 2> super_simd_type;
#endif
    typedef TriangularArray<scalar_type, void> scalar_triangle_type;
    typedef typename scalar_triangle_type::const_column_iterator const_column_iterator;
    Vec_traits<simd_type> VT;
    Vec_traits<super_simd_type> SuVT;

    const ssize_t pad_omega = SuVT.VecSize * 3;
    const size_t padded_linearized_omega_size = (p + 1) * (p + 1) + (p + 1) * (pad_omega - 1);

    simple_vector<size_t> omega_offsets(p + 1);
    simple_vector<scalar_type> omega_re(padded_linearized_omega_size);
    simple_vector<scalar_type> omega_im(padded_linearized_omega_size);
    typedef typename simple_vector<scalar_type>::const_iterator const_linearized_omega_iterator;

    for (ssize_t l = 0, i = 0; l <= p; ++l) {
        omega_offsets[l] = i;
        for (ssize_t m = -l; m <= l; ++m, ++i) {
            complex_type olm = omega.get(l, m);
            omega_re[i] = olm.real();
            omega_im[i] = olm.imag();
        }
        for ( ; i % pad_omega != 0; ++i) {
            omega_re[i] = omega_im[i] = 0.;
        }
    }

    ssize_t pad_B = 1;
    while ((2 * pB + 1) * pad_B + pad_B * (pad_B + 1) < pad_omega)
        ++pad_B;

    scalar_triangle_type B_re(pB + pad_B);
    scalar_triangle_type B_im(pB + pad_B);

    for (ssize_t l = 0; l <= pB; ++l) {
        for (ssize_t m = -l; m <= l; ++m) {
            complex_type Blm = B.get(l, m);
            B_re.get(l, m) = Blm.real();
            B_im.get(l, m) = Blm.imag();
        }
    }

    // lmjk loop with "few" updates for "small" p
    const ssize_t p_switch = 16;

    for (ssize_t l = 0; l <= p; ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            scalar_type mu_l_m_rr = 0.;
            scalar_type mu_l_m_ri = 0.;
            scalar_type mu_l_m_ir = 0.;
            scalar_type mu_l_m_ii = 0.;
            simd_type mu_l_m_rr_ = VT.zero();
            simd_type mu_l_m_ri_ = VT.zero();
            simd_type mu_l_m_ir_ = VT.zero();
            simd_type mu_l_m_ii_ = VT.zero();
            const ssize_t j_max = std::min(std::min(p, pB - l), p_switch - 1);
            for (ssize_t j = 0; j <= j_max; ++j) {
                //for (ssize_t k = -j; k <= j; ++k)
                //    mu_l_m += B.get(j + l, k + m) * omega.get(j, k);
                const ssize_t k_begin = -j;
                const ssize_t k_end = j + 1;
                const_linearized_omega_iterator omega_re_col_j = omega_re.cbegin() + omega_offsets[j];
                const_linearized_omega_iterator omega_im_col_j = omega_im.cbegin() + omega_offsets[j];
                const_column_iterator B_re_col_j = B_re.get_column_iterator(j + l, k_begin + m);
                const_column_iterator B_im_col_j = B_im.get_column_iterator(j + l, k_begin + m);
                const ssize_t k0_begin = 0;
                const ssize_t k0_end = k_end - k_begin;
                const ssize_t k0_end_simd = pad(k0_end, SuVT.VecSize);
                ssize_t k0 = k0_begin;
                for ( ; k0 < k0_end_simd; k0 += SuVT.VecSize) {
                    super_simd_type re0_ = SuVT.unaligned_load(omega_re_col_j + k0);
                    super_simd_type im0_ = SuVT.unaligned_load(omega_im_col_j + k0);
                    super_simd_type re1_ = SuVT.unaligned_load(B_re_col_j + k0);
                    super_simd_type im1_ = SuVT.unaligned_load(B_im_col_j + k0);

                    mu_l_m_rr_ += SuVT.supersum(re0_ * re1_);
                    mu_l_m_ii_ += SuVT.supersum(im0_ * im1_);
                    mu_l_m_ri_ += SuVT.supersum(re0_ * im1_);
                    mu_l_m_ir_ += SuVT.supersum(re1_ * im0_);
                }
#if 0
                for ( ; k0 < k0_end; ++k0) {
                    scalar_type re0 = omega_re_col_j[k0];
                    scalar_type im0 = omega_im_col_j[k0];
                    scalar_type re1 = B_re_col_j[k0];
                    scalar_type im1 = B_im_col_j[k0];

                    mu_l_m_rr += re0 * re1;
                    mu_l_m_ii += im0 * im1;
                    mu_l_m_ri += re0 * im1;
                    mu_l_m_ir += re1 * im0;
                }
#endif
            }
            mu_l_m_rr += VT.sum(mu_l_m_rr_);
            mu_l_m_ii += VT.sum(mu_l_m_ii_);
            mu_l_m_ri += VT.sum(mu_l_m_ri_);
            mu_l_m_ir += VT.sum(mu_l_m_ir_);
            mu.get(l, m) += toggle_sign_if_odd(l, complex_type(mu_l_m_rr - mu_l_m_ii, mu_l_m_ri + mu_l_m_ir));
        }
    }

    // ljmk loop with more updates but better cache efficiency for larger "p"
    for (ssize_t l = 0; l <= p; ++l) {
      const ssize_t j_begin = p_switch;
      const ssize_t j_max = std::min(p, pB - l);
      for (ssize_t j = j_begin; j <= j_max; ++j)
      {
        const ssize_t k_begin = -j;
        const ssize_t k_end = j + 1;
        const_linearized_omega_iterator omega_re_col_j = omega_re.cbegin() + omega_offsets[j];
        const_linearized_omega_iterator omega_im_col_j = omega_im.cbegin() + omega_offsets[j];
        const ssize_t m_begin = 0;
        const_column_iterator B_re_col_j_m0 = B_re.get_column_iterator(j + l, k_begin + m_begin);
        const_column_iterator B_im_col_j_m0 = B_im.get_column_iterator(j + l, k_begin + m_begin);

        for (ssize_t m = 0; m <= l; ++m) {
            scalar_type mu_l_m_rr = 0.;
            scalar_type mu_l_m_ri = 0.;
            scalar_type mu_l_m_ir = 0.;
            scalar_type mu_l_m_ii = 0.;
            simd_type mu_l_m_rr_ = VT.zero();
            simd_type mu_l_m_ri_ = VT.zero();
            simd_type mu_l_m_ir_ = VT.zero();
            simd_type mu_l_m_ii_ = VT.zero();
            {
                //for (ssize_t k = -j; k <= j; ++k)
                //    mu_l_m += B.get(j + l, k + m) * omega.get(j, k);
                const_column_iterator B_re_col_j = B_re_col_j_m0 + (m - m_begin);
                const_column_iterator B_im_col_j = B_im_col_j_m0 + (m - m_begin);
                const ssize_t k0_begin = 0;
                const ssize_t k0_end = k_end - k_begin;
                const ssize_t k0_end_simd = pad(k0_end, pad_omega);
                const ssize_t simd_words = (k0_end_simd - k0_begin) / VT.VecSize;
                if (true || simd_words > 17) {
                split_complex_mul_acc_simd_asm(
                        mu_l_m_rr_,
                        mu_l_m_ri_,
                        omega_re_col_j + k0_begin,
                        omega_im_col_j + k0_begin,
                        B_re_col_j + k0_begin,
                        B_im_col_j + k0_begin,
                        simd_words);
                } else {
                ssize_t k0 = k0_begin;
                for ( ; k0 < k0_end_simd; k0 += SuVT.VecSize) {
                    super_simd_type re0_ = SuVT.unaligned_load(omega_re_col_j + k0);
                    super_simd_type im0_ = SuVT.unaligned_load(omega_im_col_j + k0);
                    super_simd_type re1_ = SuVT.unaligned_load(B_re_col_j + k0);
                    super_simd_type im1_ = SuVT.unaligned_load(B_im_col_j + k0);

                    mu_l_m_rr_ += SuVT.supersum(re0_ * re1_);
                    mu_l_m_ii_ += SuVT.supersum(im0_ * im1_);
                    mu_l_m_ri_ += SuVT.supersum(re0_ * im1_);
                    mu_l_m_ir_ += SuVT.supersum(re1_ * im0_);
                }
                }
#if 0
                for ( ; k0 < k0_end; ++k0) {
                    scalar_type re0 = omega_re_col_j[k0];
                    scalar_type im0 = omega_im_col_j[k0];
                    scalar_type re1 = B_re_col_j[k0];
                    scalar_type im1 = B_im_col_j[k0];

                    mu_l_m_rr += re0 * re1;
                    mu_l_m_ii += im0 * im1;
                    mu_l_m_ri += re0 * im1;
                    mu_l_m_ir += re1 * im0;
                }
#endif
            }
            mu_l_m_rr += VT.sum(mu_l_m_rr_);
            mu_l_m_ii += VT.sum(mu_l_m_ii_);
            mu_l_m_ri += VT.sum(mu_l_m_ri_);
            mu_l_m_ir += VT.sum(mu_l_m_ir_);
            mu.get(l, m) += toggle_sign_if_odd(l, complex_type(mu_l_m_rr - mu_l_m_ii, mu_l_m_ri + mu_l_m_ir));
        }
      }
    }
}


// p^4, k=-j..j, with custom B operator
// classical M2L operation on actual *charges*, does include the (-1)^l factor
template <typename CoefficientMatrix>
void
__attribute__ ((noinline))
__attribute ((flatten))
    single_M2L_optimized(
        const CoefficientMatrix & omega,
        const CoefficientMatrix & B,
        CoefficientMatrix & mu,
        const ssize_t p,
        const ssize_t pB)
{
#if 1
    M2L_reference(omega, B, mu, p, pB);
#elif 1
    // hybrid: lmjk for j < p_switch, ljmk for j >= p_switch
    single_M2L_SIMD_lmjk_ljmk(omega, B, mu, p, pB);
#elif 1
    // column-wise processing of omega
    // better cache utilization, p^3 updates in mu
    single_M2L_SIMD_ljmk(omega, B, mu, p, pB);
#else
    // whole-omega processing
    // bad cache utilization if omega does not fit, p^2 updates in mu
    single_M2L_SIMD_lmjk(omega, B, mu, p, pB);
#endif
}


// p^4, k=-j..j, with custom B operator
// classical M2L operation on actual *charges*, does include the (-1)^j factor
template <typename CoefficientMatrix, typename OperatorMatrix>
void M2L_opt_m0(
        const CoefficientMatrix & omega,
        const OperatorMatrix & B,
        CoefficientMatrix & mu,
        const size_t p)
{
    omega.require_lower();
    B.require_lower();
    mu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type Real;
    assert(p <= omega.p());
    assert(2 * p <= B.p());
    assert(p <= mu.p());
    const ssize_t p_omega = p;
    const ssize_t p_B = 2 * p;
    const ssize_t p_mu = p;

    for (ssize_t l = 0; l <= p_mu; ++l) {
        {
            Real mu_l_0(0.);
            const ssize_t j_max = std::min(p_omega, p_B - l);
            for (ssize_t j = 0; j <= j_max; ++j) {
                Real mu_l_0_j(0.);
                for (ssize_t k = -j; k <= j; ++k)
                    mu_l_0_j += (B.get(j + l, k + 0) * omega.get(j, k)).real();
                mu_l_0 += toggle_sign_if_odd(j, mu_l_0_j);
            }
            mu.upper(l, 0) += toggle_sign_if_odd(0, mu_l_0);
        }
        for (ssize_t m = 1; m <= l; ++m) {
            complex_type mu_l_m(0.);
            const ssize_t j_max = std::min(p_omega, p_B - l);
            for (ssize_t j = 0; j <= j_max; ++j) {
                complex_type mu_l_m_j(0.);
                for (ssize_t k = -j; k <= j; ++k)
                    mu_l_m_j += B.get(j + l, k + m) * omega.get(j, k);
                mu_l_m += toggle_sign_if_odd(j, mu_l_m_j);
            }
            mu.upper(l, m) += toggle_sign_if_odd(0, mu_l_m);
        }
    }
}


// p^4, k=-j..j, with custom B operator
// classical M2L operation on actual *charges*, does include the (-1)^j factor
template <ssize_t p, typename CoefficientMatrix, typename OperatorMatrix>
void M2L_fix_p(
        const CoefficientMatrix & omega,
        const OperatorMatrix & B,
        CoefficientMatrix & mu)
{
    omega.require_lower();
    B.require_lower();
    mu.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    const ssize_t p_omega = p;
    const ssize_t p_B = 2 * p;
    const ssize_t p_mu = p;

    for (ssize_t l = 0; l <= p_mu; ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            complex_type mu_l_m(0.);
            const ssize_t j_max = std::min(p_omega, p_B - l);
            for (ssize_t j = 0; j <= j_max; ++j) {
                complex_type mu_l_m_j(0.);
                for (ssize_t k = -j; k <= j; ++k)
                    mu_l_m_j += B.get(j + l, k + m) * omega.get(j, k);
                mu_l_m += toggle_sign_if_odd(j, mu_l_m_j);
            }
            mu.upper(l, m) += toggle_sign_if_odd(0, mu_l_m);
        }
    }
}


// p^4, k=-j..j, with custom B operator
// classical M2L operation on actual *charges*, does include the (-1)^j factor
template <typename CoefficientMatrix, typename OperatorMatrix>
void M2L_switch(
        const CoefficientMatrix & omega,
        const OperatorMatrix & B,
        CoefficientMatrix & mu,
        const size_t p)
{
    assert(p <= omega.p());
    assert(2 * p <= B.p());
    assert(p <= mu.p());

    switch (p) {
    case 0:
        return M2L_fix_p<0>(omega, B, mu);
    case 1:
        return M2L_fix_p<1>(omega, B, mu);
    case 2:
        return M2L_fix_p<2>(omega, B, mu);
    case 3:
        return M2L_fix_p<3>(omega, B, mu);
    case 4:
        return M2L_fix_p<4>(omega, B, mu);
    case 5:
        return M2L_fix_p<5>(omega, B, mu);
    case 6:
        return M2L_fix_p<6>(omega, B, mu);
    case 7:
        return M2L_fix_p<7>(omega, B, mu);
    case 8:
        return M2L_fix_p<8>(omega, B, mu);
    case 9:
        return M2L_fix_p<9>(omega, B, mu);
    case 10:
        return M2L_fix_p<10>(omega, B, mu);
    case 11:
        return M2L_fix_p<11>(omega, B, mu);
    case 12:
        return M2L_fix_p<12>(omega, B, mu);
    case 13:
        return M2L_fix_p<13>(omega, B, mu);
    case 14:
        return M2L_fix_p<14>(omega, B, mu);
    case 15:
        return M2L_fix_p<15>(omega, B, mu);
    case 16:
        return M2L_fix_p<16>(omega, B, mu);
    case 17:
        return M2L_fix_p<17>(omega, B, mu);
    case 18:
        return M2L_fix_p<18>(omega, B, mu);
    case 19:
        return M2L_fix_p<19>(omega, B, mu);
    case 20:
        return M2L_fix_p<20>(omega, B, mu);
    default:
        return M2L_reference(omega, B, mu, p);
    }
}


// p^4
template <typename Real3, typename CoefficientMatrix>
void XM2L(
        const CoefficientMatrix & omega,
        const Real3 & xyz,
        CoefficientMatrix & mu,
        const size_t p)
{
    typedef typename CoefficientMatrix::complex_type complex_type;

    M2L_Operator<CoefficientMatrix> B(xyz, p);
    for (size_t l = 0; l <= p; ++l) {
        for (size_t m = 0; m <= l; ++m) {
            complex_type mu_l_m(0.);
            for (size_t j = 0; j <= p; ++j) {
                // k=0 .. j
                for (size_t k = 0; k <= j; ++k)
                    mu_l_m += B.upper(j + l, k + m) * omega.upper(j, k);

                // k=-j .. -1  => remapped to positive indices
                // k <= m
                size_t min_m_j = std::min(m, j);
                for (size_t k = 1; k <= min_m_j; ++k) {
                    complex_type tmp = B.upper(j + l, m - k) * fmm::conj(omega.upper(j, k));
                    mu_l_m += toggle_sign_if_odd(k, tmp);
                }
                // k > m
                complex_type tmp(0.);
                for (size_t k = min_m_j + 1; k <= j; ++k) {
                    tmp += fmm::conj(B.upper(j + l, k - m)) * fmm::conj(omega.upper(j, k));
                }
                mu_l_m += toggle_sign_if_odd(m, tmp);
            }
            mu.upper(l, m) += toggle_sign_if_odd(l, mu_l_m);
        }
    }
}


// p^4
// omega1 --(x,y,z)--> mu2
// mu1 <--(-x,-y,-z)-- omega2
template <typename Real3, typename CoefficientMatrix>
void M2L_pair_reference(
        const CoefficientMatrix & omega1,
        const CoefficientMatrix & omega2,
        const Real3 & xyz,
        CoefficientMatrix & mu2,
        CoefficientMatrix & mu1,
        const ssize_t p)
{
    typedef typename CoefficientMatrix::complex_type complex_type;
    M2L_Operator<CoefficientMatrix> B(xyz, p);

    for (ssize_t l = 0; l <= p; ++l) {
        for (ssize_t m = 0; m <= l; ++m) {
            complex_type mu2_l_m(0.);
            complex_type mu1_l_m(0.);
            for (ssize_t j = 0; j <= p; ++j) {
                complex_type mu1_l_m_j(0.);
                for (ssize_t k = -j; k <= j; ++k) {
                    complex_type Bjlkm = B.get(j + l, k + m);
                    mu2_l_m += Bjlkm * omega1.get(j, k);
                    mu1_l_m_j += Bjlkm * omega2.get(j, k);
                }
                mu1_l_m += toggle_sign_if_odd(l + j, mu1_l_m_j);
            }
            mu2.upper(l, m) += toggle_sign_if_odd(l, mu2_l_m);
            mu1.upper(l, m) += toggle_sign_if_odd(l, mu1_l_m);
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
