/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P4_multipole2multipole_hpp
#define FMSOLVR_OPERATOR_P4_multipole2multipole_hpp

#include <algorithm>

#include "fmsolvr/operator/particle2multipole.hpp"

namespace fmsolvr
{

// M2M(omega, (0, 0, 0)) is an idempotent operation

// expansion point shift vector orientation:
// for a point xyz reasonably near to expansion_point_source (and expansion_point_target)
// P2M (particle ==> expansion_point_source) followed by M2M (==> expansion_point_target)
//        P2M(q, xyz, o_s, expansion_point_source, p);
//        M2M(o_s, expansion_point_source, o_st, expansion_point_target, p);
// is equivalent to direct P2M (particle ==> expansion_point_target) (within numerics)
//        P2M(q, xyz, o_t, expansion_point_target, p);


// all values O(l,m) with (l < 0) or (|m| > l) are defined as (0,0i)
// but they are not being accessed anyway
template <typename CoefficientMatrix>
class M2M_Operator : public CoefficientMatrix
{
public:
    template <typename Real3>
    M2M_Operator(
        const Real3 & expansion_point_source,
        const Real3 & expansion_point_target,
        const size_t p)
    : CoefficientMatrix(p)
    {
        rawP2M_nodivtable(expansion_point_source - expansion_point_target, *static_cast<CoefficientMatrix *>(this), p);
        this->populate_lower();
    }
};


// p^4
template <typename Real3, typename CoefficientMatrix>
void M2M(
        const CoefficientMatrix & omega_source,
        const Real3 & expansion_point_source,
        CoefficientMatrix & omega_target,
        const Real3 & expansion_point_target,
        const ssize_t p)
{
    M2M_Operator<CoefficientMatrix> A(expansion_point_source, expansion_point_target, p);
    M2M_impl_reference(omega_source, A, omega_target, p);
}


// p^4
template <typename CoefficientMatrix>
void M2M(
        const CoefficientMatrix & omega_in,
        const M2M_Operator<CoefficientMatrix> & A,
        CoefficientMatrix & omega_out,
        const ssize_t p)
{
    M2M_impl_reference(omega_in, A, omega_out, p);
}


// p^4, k=-j..j
// this function may also be used to perform P2M of (mono-/)di-/quadrupoles ...
// by feeding them in via omega_in with omega_in.p() \in {1,2,3,...}
template <typename CoefficientMatrix>
void M2M_impl_reference(
        const CoefficientMatrix & omega_in,
        const M2M_Operator<CoefficientMatrix> & A,
        CoefficientMatrix & omega_out,
        const ssize_t p)
{
    omega_in.require_lower();
    A.require_lower();
    omega_out.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    const ssize_t p_in = omega_in.p();
    // limit output order to values that can be stored and computed correctly
    const ssize_t p_out = std::min<ssize_t>(p, std::min(omega_out.p(), A.p()));

    for (ssize_t l = 0; l <= p_out; ++l) {
        const ssize_t j_max = std::min(l, p_in);
        for (ssize_t m = 0; m <= l; ++m) {
            complex_type omega_l_m(0.);
            for (ssize_t j = 0; j <= j_max; ++j) {
                const ssize_t k_min = std::max(-j, m - (l - j));
                const ssize_t k_max = std::min(+j, m + (l - j));
                for (ssize_t k = k_min; k <= k_max; ++k)
                    omega_l_m += A.get(l - j, m - k) * omega_in.get(j, k);
            }
            omega_out.upper(l, m) += omega_l_m;
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
