/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_P4_multipole2multipole_derivative_hpp
#define FMSOLVR_OPERATOR_P4_multipole2multipole_derivative_hpp

#include <sys/types.h>
#include <algorithm>

namespace fmsolvr
{

template <typename CoefficientMatrix>
void DxM2M_reference(
        const CoefficientMatrix & omega,
        CoefficientMatrix & domega,
        const ssize_t p)
{
    omega.require_lower();
    domega.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;

    if (p >= 0)
        domega.upper(0, 0) = complex_type(0.);
    if (p >= 1)
        domega.upper(1, 0) = complex_type(0.);
    for (ssize_t l = 1; l <= p; ++l) {
        for (ssize_t m = 0; m <= l - 2; ++m)
            domega.upper(l, m) = (omega.get(l - 1, m - 1) - omega.get(l - 1, m + 1)) * scalar_type(0.5);
        for (ssize_t m = std::max<ssize_t>(l - 1, 1); m <= l; ++m)
            domega.upper(l, m) = omega.get(l - 1, m - 1) * scalar_type(0.5);
    }
}

template <typename CoefficientMatrix>
void DyM2M_reference(
        const CoefficientMatrix & omega,
        CoefficientMatrix & domega,
        const ssize_t p)
{
    omega.require_lower();
    domega.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;

    if (p >= 0)
        domega.upper(0, 0) = complex_type(0.);
    if (p >= 1)
        domega.upper(1, 0) = complex_type(0.);
    for (ssize_t l = 1; l <= p; ++l) {
        for (ssize_t m = 0; m <= l - 2; ++m) {
            complex_type tmp = (omega.get(l - 1, m - 1) + omega.get(l - 1, m + 1)) * scalar_type(0.5);
            domega.upper(l, m) = complex_type(tmp.imag(), -tmp.real());
        }
        for (ssize_t m = std::max<ssize_t>(l - 1, 1); m <= l; ++m) {
            complex_type tmp = omega.get(l - 1, m - 1) * scalar_type(0.5);
            domega.upper(l, m) = complex_type(tmp.imag(), -tmp.real());
        }
    }
}

template <typename CoefficientMatrix>
void DzM2M_reference(
        const CoefficientMatrix & omega,
        CoefficientMatrix & domega,
        const ssize_t p)
{
    omega.require_lower();
    domega.invalidate_lower();

    typedef typename CoefficientMatrix::complex_type complex_type;

    if (p >= 0)
        domega.upper(0, 0) = complex_type(0.);
    for (ssize_t l = 1; l <= p; ++l) {
        for (ssize_t m = 0; m < l; ++m)
            domega.upper(l, m) = omega.get(l - 1, m);
        domega.upper(l, l) = complex_type(0.);
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
