/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_ROTATION_eikphitable_hpp
#define FMSOLVR_OPERATOR_ROTATION_eikphitable_hpp

#include "fmsolvr/complex_vec.hpp"
#include "fmsolvr/container/linear_array.hpp"
#include "fmsolvr/math/angle.hpp"

#include <memory>


namespace fmsolvr {

// stores e^(i*k*phi) = (cos(phi) + i*sin(phi))^k
template <typename Real, typename Allocator = std::allocator<Real> >
class EikPhiTable : public LinearArray<typename COMPLEX_GENERATOR<Real>::type, Allocator>
{
    typedef LinearArray<typename COMPLEX_GENERATOR<Real>::type, Allocator> Base;
    using typename Base::value_type;
    typedef value_type complex_type;
    typedef Angle<Real> angle_type;

public:
    template <typename Angle>
    EikPhiTable(
        const Angle & phi,
        const size_t p_max)
        : Base(p_max), phi_(phi)
    {
        complex_type eiphi(phi.cos(), phi.sin());  // e^(i*phi)
        (*this)[0] = complex_type(1);
        for (size_t k = 1; k <= p_max; ++k)
        {
            (*this)[k] = (*this)[k - 1] * eiphi;  // e^(i*k*phi)
        }
    }

    const angle_type & phi() const
    { return phi_; }

private:
    angle_type phi_;
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
