/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_ROTATION_rotation_hpp
#define FMSOLVR_OPERATOR_ROTATION_rotation_hpp

#include "fmsolvr/math/Complex.hpp"
#include "fmsolvr/math/parity_sign.hpp"
#include "fmsolvr/reduce_complex_scale.hpp"

#include <vector>
#include <cmath>
#include <cassert>

namespace fmsolvr
{

////////////////////////////////////////////////////////////////////////////

// Multipole expansion forward rotation around phi, theta
//
// omega_out_rot(l,m) = sum_{k=-l}^{l}( f_{l}^{k,m} *
//                                      d(theta,l,m,k) *
//                                      exp(i*k*phi) *
//                                      omega_in(l,k) )
// f_{l}^{k,m} = sqrt(( (l-k)!(l+k)! ) / ( (l-m)!(l+m)! ))
// d(theta,l,m,k) = d_l^{m,k}(theta) (Wigner small d rotation coefficients)
//
// For the forward rotation the expansion is first rotated about the
// z-axis (phi) and then about the y-axis (theta).
// Z-rotation (exp(i*k*phi)) is applied to the input elements
// omega_in(l,k).

// Multipole expansion backward rotation around -theta, -phi
//
// omega_out(l,m) = sum_{k=-l}^{l}( f_{l}^{k,m} *
//                                  (-1)^{k+m} * d(theta,l,m,k) *
//                                  exp(-i*m*phi) *
//                                  omega_in_rot(l,k) )
// f_{l}^{k,m} = sqrt(( (l-k)!(l+k)! ) / ( (l-m)!(l+m)! ))
// d(theta,l,m,k) = d_l^{m,k}(theta) (Wigner small d rotation coefficients)
//
// For the backward rotation the expansion is first rotated about the
// y-axis (theta) and then about the z-axis (phi).
// Z-rotation (exp(-i*m*phi)) is applied to the output elements
// omega_out(l,m). This is not obvious!

// Local expansion forward rotation around phi, theta
//
// mu_out_rot(l,m) = sum_{k=-l}^{l}( g_{l}^{k,m} *
//                                   d(theta,l,m,k) *
//                                   exp(-i*k*phi) *
//                                   mu_in(l,k) )
// g_{l}^{k,m} = sqrt(( (l-m)!(l+m)! ) / ( (l-k)!(l+k)! ))
// d(theta,l,m,k) = d_l^{m,k}(theta) (Wigner small d rotation coefficients)
//
// For the forward rotation the expansion is first rotated about the
// z-axis (phi) and then about the y-axis (theta).
// Z-rotation (exp(-i*k*phi)) is applied to the input elements
// mu_in(l,k).

// Local expansion backward rotation around -theta, -phi
//
// mu_out(l,m) = sum_{k=-l}^{l}( g_{l}^{k,m} *
//                               (-1)^{k+m} * d(theta,l,m,k) *
//                               exp(i*m*phi) *
//                               mu_in_rot(l,k) )
// g_{l}^{k,m} = sqrt(( (l-m)!(l+m)! ) / ( (l-k)!(l+k)! ))
// d(theta,l,m,k) = d_l^{m,k}(theta) (Wigner small d rotation coefficients)
//
// For the backward rotation the expansion is first rotated about the
// y-axis (theta) and then about the z-axis (phi).
// Z-rotation (exp(i*m*phi)) is applied to the output elements
// mu_out(l,m). This is not obvious!

// Note:
// f_{l}^{k,m} == f_{l}^{-k,m}
// g_{l}^{k,m} == g_{l}^{-k,m}

// Properties of Wigner small d rotation coefficients:
// d(-theta,l,m,k) == (-1)^{k+m} * d(theta,l,m,k)

// Complex trigonometry:
// exp(i*phi) == cos(phi) + i*sin(phi)
// exp(-i*k*phi) == conj(exp(i*k*phi))

// Properties of multipole expansions and local expansions:
// omuga(l,-k) == (-1)^k * conj(omuga(l, k))

////////////////////////////////////////////////////////////////////////////


struct NoScratchDummy
{ };


////////////////////////////////////////////////////////////////////////////


// Apply the rotation given by the RotationOperator object R to the
// multipole expansion of order p given by omega_in and write the results to
// omega_out.
// omega_out_rot(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * d(theta,l,m,k) * exp(i*k*phi) * omega_in(l,k) )
// This is an assigning operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationMultipoleForward_reference(
        const MultipoleCoefficientsIn &omega_in,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out_rot,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();

    assert(p_compute <= omega_in.p());
    assert(p_compute <= d.p());
    assert(p_compute <= omega_out_rot.p());
    omega_in.require_lower();
    omega_out_rot.invalidate_lower();

    const real_type cos_phi(R.phi().cos());
    const real_type sin_phi(R.phi().sin());
    complex_type e_ilphi(1);
    real_type    f_ll0(1);

    for (index_type l = 0; l <= p; ++l)
    {
        real_type f_llm(f_ll0);

        for (index_type m = 0; m <= l; ++m)
        {
            complex_type omega(0, 0);
            complex_type e_ikphi(e_ilphi);
            real_type    f_lkm(f_llm);

            for (index_type k = -l; k <= l; ++k)
            {
                omega += f_lkm * e_ikphi * d(l, m, k) * omega_in.get(l, k);

                // (for the next iteration)
                // exp(i*(k+1)*phi)
                e_ikphi *= complex_type(cos_phi, sin_phi);
                // f_l^{k+1,m}
                f_lkm *= std::sqrt(real_type(l+k+1)/real_type(l-k));
            }
            omega_out_rot.upper(l, m) = omega;

            // (for the next iteration)
            // f_l^{-l,m+1}
            f_llm *= std::sqrt(real_type(l-m)/real_type(l+m+1));
        }

        // (for the next iteration)
        // exp(-i*(l+1)*phi)
        e_ilphi *= complex_type(cos_phi, -sin_phi);
        // f_{l+1}^{-l-1,0}
        f_ll0 *= std::sqrt(real_type((2*l+1)*(2*l+2))/real_type((l+1)*(l+1)));
    }
}


// Apply the rotation given by the RotationOperator object R backwards. This operation
// is inverse to ApplyRotationMultipole with identical RotationOperator R.
// For the backwards rotation the expansion is first rotated about the y-axis and then
// about the z-axis.
// omega_out(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(-i*m*phi) * omega_in_rot(l,k) )
// This is an accumulating operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationMultipoleBackward_reference(
        const MultipoleCoefficientsIn &omega_in_rot,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();

    assert(p_compute <= omega_in_rot.p());
    assert(p_compute <= d.p());
    assert(p_compute <= omega_out.p());
    omega_in_rot.require_lower();
    omega_out.invalidate_lower();

    const real_type cos_phi(R.phi().cos());
    const real_type sin_phi(R.phi().sin());
    real_type f_ll0(1);

    for (index_type l = 0; l <= p; ++l)
    {
        // f_l^{-l,m}
        real_type    f_llm(f_ll0);
        // exp(-i*m*phi)
        complex_type e_imphi(1);

        for (index_type m = 0; m <= l; ++m)
        {
            complex_type omega(0, 0);
            real_type    f_lkm(f_llm);

            for (index_type k = -l; k <= l; ++k)
            {
                omega += toggle_sign_if_odd(k + m, f_lkm * d(l, m, k) * omega_in_rot.get(l, k));

                // (for the next iteration)
                // f_l^{k+1,m}
                f_lkm *= std::sqrt(real_type(l+k+1)/real_type(l-k));
            }
            omega           *= e_imphi;
            omega_out.upper(l, m) += omega;

            // (for the next iteration)
            // f_l^{-l,m+1}
            f_llm *= std::sqrt(real_type(l-m)/real_type(l+m+1));
            // exp(-i*(m+1)*phi)
            e_imphi *= complex_type(cos_phi, -sin_phi);
        }

        // (for the next iteration)
        // f_{l+1}^{-l-1,m+1}
        f_ll0 *= std::sqrt(real_type((2*l+1)*(2*l+2))/real_type((l+1)*(l+1)));
    }
}


// Apply the rotation given by the RotationOperator object R to the
// local expansion of degree p given by mu_in and write the results to
// mu_out.
// mu_out_rot(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * d(theta,l,m,k) * exp(-i*k*phi) * mu_in(l,k) )
// This is an assigning operation.
template <typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationLocalForward_reference(
        const LocalCoefficientsIn &mu_in,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out_rot,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();

    assert(p_compute <= mu_in.p());
    assert(p_compute <= d.p());
    assert(p_compute <= mu_out_rot.p());
    mu_in.require_lower();
    mu_out_rot.invalidate_lower();

    const real_type cos_phi(R.phi().cos());
    const real_type sin_phi(R.phi().sin());
    // exp(i*0*phi)
    complex_type e_ilphi(1);
    // g_0^{0,0} = 0
    real_type    g_ll0(1);

    for (index_type l = 0; l <= p; ++l)
    {
        real_type g_llm(g_ll0);

        for (index_type m = 0; m <= l; ++m)
        {
            complex_type mu(0, 0);
            real_type    g_lkm(g_llm);
            complex_type e_ikphi = e_ilphi;

            for (index_type k = -l; k <= l; ++k)
            {
                mu += g_lkm * e_ikphi * d(l, m, k) * mu_in.get(l, k);

                // (for the next iteration)
                // exp(-i*(k+1)*phi)
                e_ikphi *= complex_type(cos_phi, -sin_phi);
                // g_l^{k+1,m}
                g_lkm *= std::sqrt(real_type(l-k)/real_type(l+k+1));
            }
            mu_out_rot.upper(l, m) = mu;

            // (for the next iteration)
            // g_l^{-l,m+1}
            g_llm       *= std::sqrt(real_type(l+m+1)/(l-m));
        }

        // (for the next iteration)
        // exp(i*(l+1)*phi)
        e_ilphi = e_ilphi*complex_type(cos_phi, sin_phi);
        // g_{l+1}^{-l-1,0}
        g_ll0 *= std::sqrt(real_type((l+1)*(l+1))/real_type((2*l+1)*(2*l+2)));
    }
}


// Apply the rotation given by the RotationOperator object R backwards. This operation
// is inverse to ApplyRotationLocal with identical RotationOperator R.
// For the backwards rotation the expansion is first rotated about the y-axis and then
// about the z-axis.
// mu_out(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(i*m*phi) * mu_in_rot(l,k) )
// This is an accumulating operation.
template <typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationLocalBackward_reference(
        const LocalCoefficientsIn &mu_in_rot,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();

    assert(p_compute <= mu_in_rot.p());
    assert(p_compute <= d.p());
    assert(p_compute <= mu_out.p());
    mu_in_rot.require_lower();
    mu_out.invalidate_lower();

    const real_type cos_phi(R.phi().cos());
    const real_type sin_phi(R.phi().sin());
    // g_0^{0,0} = 1
    real_type g_ll0(1);

    for (index_type l = 0; l <= p; ++l)
    {
        real_type    g_llm(g_ll0);
        complex_type e_imphi(1);

        for (index_type m = 0; m <= l; ++m)
        {
            complex_type mu(0, 0);
            real_type    g_lkm(g_llm);

            for (index_type k = -l; k <= l; ++k)
            {
                mu += toggle_sign_if_odd(k + m, g_lkm * d(l, m, k) * mu_in_rot.get(l, k));

                // (for the next iteration)
                // g_l^{k+1,m}
                g_lkm *= std::sqrt(real_type(l-k)/real_type(l+k+1));
            }
            mu           *= e_imphi;
            mu_out.upper(l, m) += mu;

            // (for the next iteration)
            // g_l^{l,m+1}
            g_llm *= std::sqrt(real_type(l+m+1)/real_type(l-m));
            // exp(i*(m+1)*phi)
            e_imphi *= complex_type(cos_phi, sin_phi);
        }

        // (for the next iteration)
        // g_l+1^{-l-1,0}
        g_ll0 *= std::sqrt(real_type((l+1)*(l+1))/real_type((2*l+1)*(2*l+2)));
    }
}


////////////////////////////////////////////////////////////////////////////


// Optimized version of ApplyRotationMultipoleForward_reference
// omega_out_rot(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * d(theta,l,m,k) * exp(i*k*phi) * omega_in(l,k) )
// * summing only over k = 0,...,l
// This is an assigning operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationMultipoleForward(
        const MultipoleCoefficientsIn &omega_in,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out_rot,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();

    assert(p_compute <= omega_in.p());
    assert(p_compute <= d.p());
    assert(p_compute <= omega_out_rot.p());
    omega_in.require_lower();
    omega_out_rot.invalidate_lower();

    const real_type cos_phi(R.phi().cos());
    const real_type sin_phi(R.phi().sin());

    for (index_type l = 0; l <= p; ++l)
    {
        real_type f_l0m(1);

        for (index_type m = 0; m <= l; ++m)
        {
            complex_type e_ikphi(1);
            real_type    f_lkm(f_l0m);
            complex_type omega = f_lkm * d(l, m, 0) * omega_in.upper(l, 0);

            for (index_type k = 1; k <= l; ++k)
            {
                // exp(i*k*phi)
                e_ikphi *= complex_type(cos_phi, sin_phi);
                // f_l^{k,m}
                f_lkm *= std::sqrt(real_type(l+k)/real_type(l-k+1));

                omega += f_lkm * (d(l, m, k)  * e_ikphi            * omega_in.upper(l, k) +
                                  d(l, m, -k) * fmm::conj(e_ikphi) * omega_in.lower(l, -k));
            }
            omega_out_rot.upper(l, m) = omega;

            // (for the next iteration)
            // f_l^{0,m+1}
            f_l0m *= std::sqrt(real_type(l-m)/real_type(l+m+1));
        }
    }
}


// Optimized version of ApplyRotationMultipoleBackward_reference
// omega_out(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(-i*m*phi) * omega_in_rot(l,k) )
// * summing only over k = 0,...,l
// This is an accumulating operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationMultipoleBackward(
        const MultipoleCoefficientsIn &omega_in_rot,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();

    assert(p_compute <= omega_in_rot.p());
    assert(p_compute <= d.p());
    assert(p_compute <= omega_out.p());
    omega_in_rot.require_lower();
    omega_out.invalidate_lower();

    const real_type cos_phi(R.phi().cos());
    const real_type sin_phi(R.phi().sin());

    for (index_type l = 0; l <= p; ++l)
    {
        real_type    f_l0m(1);
        complex_type e_imphi(1);

        for (index_type m = 0; m <= l; ++m)
        {
            real_type    f_lkm(f_l0m);
            complex_type omega = toggle_sign_if_odd(m, f_lkm * d(l, m, 0) * omega_in_rot.upper(l, 0));

            for (index_type k = 1; k <= l; ++k)
            {
                // f_l^{k,m}
                f_lkm *= std::sqrt(real_type(l+k)/real_type(l-k+1));

                omega += toggle_sign_if_odd(k + m, f_lkm * (d(l, m, k)  * omega_in_rot.upper(l, k) +
                                                            d(l, m, -k) * omega_in_rot.lower(l, -k)));
            }
            omega           *= e_imphi;
            omega_out.upper(l, m) += omega;

            // (for the next iteration)
            // f_l^{0,m+1}
            f_l0m *= std::sqrt(real_type(l-m)/real_type(l+m+1));
            // exp(-i*(m+1)*phi)
            e_imphi *= complex_type(cos_phi, -sin_phi);
        }
    }
}


// Optimized version of ApplyRotationLocalForward_reference
// mu_out_rot(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * d(theta,l,m,k) * exp(-i*k*phi) * mu_in(l,k) )
// * summing only over k = 0,...,l
// This is an assigning operation.
template <typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationLocalForward(
        const LocalCoefficientsIn &mu_in,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out_rot,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();

    assert(p_compute <= mu_in.p());
    assert(p_compute <= d.p());
    assert(p_compute <= mu_out_rot.p());
    mu_in.require_lower();
    mu_out_rot.invalidate_lower();

    const real_type cos_phi(R.phi().cos());
    const real_type sin_phi(R.phi().sin());

    for (index_type l = 0; l <= p; ++l)
    {
        real_type g_l0m(1);

        for (index_type m = 0; m <= l; ++m)
        {
            complex_type e_ikphi(1);
            real_type    g_lkm(g_l0m);
            complex_type mu = g_lkm * d(l, m, 0) * mu_in.upper(l, 0);

            for (index_type k = 1; k <= l; ++k)
            {
                // exp(-i*k*phi)
                e_ikphi *= complex_type(cos_phi, -sin_phi);
                // g_l^{k,m}
                g_lkm *= std::sqrt(real_type(l-k+1)/real_type(l+k));

                mu += g_lkm * (d(l, m, k)  * e_ikphi            * mu_in.upper(l, k) +
                               d(l, m, -k) * fmm::conj(e_ikphi) * mu_in.lower(l, -k));
            }
            mu_out_rot.upper(l, m) = mu;

            // (for the next iteration)
            // g_l^{0,m+1}
            g_l0m *= std::sqrt(real_type(l+m+1)/real_type(l-m));
        }
    }
}


// Optimized version of ApplyRotationLocalBackward_reference
// mu_out(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(i*m*phi) * mu_in_rot(l,k) )
// * summing only over k = 0,...,l
// This is an accumulating operation.
template <typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationLocalBackward(
        const LocalCoefficientsIn &mu_in_rot,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();

    assert(p_compute <= mu_in_rot.p());
    assert(p_compute <= d.p());
    assert(p_compute <= mu_out.p());
    mu_in_rot.require_lower();
    mu_out.invalidate_lower();

    const real_type cos_phi(R.phi().cos());
    const real_type sin_phi(R.phi().sin());

    for (index_type l = 0; l <= p; ++l)
    {
        real_type    g_l0m(1);
        complex_type e_imphi(1);

        for (index_type m = 0; m <= l; ++m)
        {
            real_type    g_lkm(g_l0m);
            complex_type mu = toggle_sign_if_odd(m, g_lkm * d(l, m, 0) * mu_in_rot.upper(l, 0));

            for (index_type k = 1; k <= l; ++k)
            {
                // g_l^{k,m}
                g_lkm *= std::sqrt(real_type(l-k+1)/real_type(l+k));

                mu += toggle_sign_if_odd(k + m, g_lkm * (d(l, m, k)  * mu_in_rot.upper(l, k) +
                                                         d(l, m, -k) * mu_in_rot.lower(l, -k)));
            }
            mu           *= e_imphi;
            mu_out.upper(l, m) += mu;

            // (for the next iteration)
            // g_l^{0,m+1}
            g_l0m *= std::sqrt(real_type(l+m+1)/real_type(l-m));
            // exp(i*(m+1)*phi)
            e_imphi *= complex_type(cos_phi, sin_phi);
        }
    }
}


////////////////////////////////////////////////////////////////////////////


// Optimized version of ApplyRotationMultipoleForward_reference
// omega_out_rot(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * d(theta,l,m,k) * exp(i*k*phi) * omega_in(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi)
// This is an assigning operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationMultipoleForward_phi(
        const MultipoleCoefficientsIn &omega_in,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out_rot,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();
    const auto & e_ikphi = R.e_ikphi();

    assert(p_compute <= omega_in.p());
    assert(p_compute <= d.p());
    assert(p_compute <= e_ikphi.p());
    assert(p_compute <= omega_out_rot.p());
    omega_in.require_lower();
    omega_out_rot.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        real_type f_l0m(1);

        for (index_type m = 0; m <= l; ++m)
        {
            real_type    f_lkm(f_l0m);
            complex_type omega = f_lkm * d(l, m, 0) * omega_in.upper(l, 0);

            for (index_type k = 1; k <= l; ++k)
            {
                // e_ikphi[k] = exp(i*k*phi), conj(e_ikphi[k]) = exp(-i*k*phi)
                // f_l^{k,m}
                f_lkm *= std::sqrt(real_type(l+k)/real_type(l-k+1));

                omega += f_lkm * (d(l, m, k)  * e_ikphi[k]            * omega_in.upper(l, k) +
                                  d(l, m, -k) * fmm::conj(e_ikphi[k]) * omega_in.lower(l, -k));
            }
            omega_out_rot.upper(l, m) = omega;

            // (for the next iteration)
            // f_l^{0,m+1}
            f_l0m *= std::sqrt(real_type(l-m)/real_type(l+m+1));
        }
    }
}


// Optimized version of ApplyRotationMultipoleBackward_reference
// omega_out(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(-i*m*phi) * omega_in_rot(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi)
// This is an accumulating operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationMultipoleBackward_phi(
        const MultipoleCoefficientsIn &omega_in_rot,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();
    const auto & e_imphi = R.e_ikphi();

    assert(p_compute <= omega_in_rot.p());
    assert(p_compute <= d.p());
    assert(p_compute <= e_imphi.p());
    assert(p_compute <= omega_out.p());
    omega_in_rot.require_lower();
    omega_out.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        real_type    f_l0m(1);

        for (index_type m = 0; m <= l; ++m)
        {
            real_type    f_lkm(f_l0m);
            complex_type omega = toggle_sign_if_odd(m, f_lkm * d(l, m, 0) * omega_in_rot.upper(l, 0));

            for (index_type k = 1; k <= l; ++k)
            {
                // f_l^{k,m}
                f_lkm *= std::sqrt(real_type(l+k)/real_type(l-k+1));

                omega += toggle_sign_if_odd(k + m, f_lkm * (d(l, m, k)  * omega_in_rot.upper(l, k) +
                                                            d(l, m, -k) * omega_in_rot.lower(l, -k)));
            }
            // conj(e_imphi[m]) = exp(-i*m*phi)
            omega           *= fmm::conj(e_imphi[m]);
            omega_out.upper(l, m) += omega;

            // (for the next iteration)
            // f_l^{0,m+1}
            f_l0m *= std::sqrt(real_type(l-m)/real_type(l+m+1));
        }
    }
}


// Optimized version of ApplyRotationLocalForward_reference
// mu_out_rot(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * d(theta,l,m,k) * exp(-i*k*phi) * mu_in(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi)
// This is an assigning operation.
template <typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationLocalForward_phi(
        const LocalCoefficientsIn &mu_in,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out_rot,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();
    const auto & e_ikphi = R.e_ikphi();

    assert(p_compute <= mu_in.p());
    assert(p_compute <= d.p());
    assert(p_compute <= e_ikphi.p());
    assert(p_compute <= mu_out_rot.p());
    mu_in.require_lower();
    mu_out_rot.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        real_type g_l0m(1);

        for (index_type m = 0; m <= l; ++m)
        {
            real_type g_lkm(g_l0m);
            complex_type mu = g_lkm * d(l, m, 0) * mu_in.upper(l, 0);

            for (index_type k = 1; k <= l; ++k)
            {
                // e_ikphi[k] = exp(i*k*phi), conj(e_ikphi[k]) = exp(-i*k*phi)
                // g_l^{k,m}
                g_lkm *= std::sqrt(real_type(l - k + 1) / real_type(l + k));

                mu += g_lkm * (d(l, m, k)  * fmm::conj(e_ikphi[k]) * mu_in.upper(l, k) +
                               d(l, m, -k) * e_ikphi[k]            * mu_in.lower(l, -k));
            }
            mu_out_rot.upper(l, m) = mu;

            // (for the next iteration)
            // g_l^{0,m+1}
            g_l0m *= std::sqrt(real_type(l + m + 1) / real_type(l - m));
        }
    }
}


// Optimized version of ApplyRotationLocalBackward_reference
// mu_out(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(i*m*phi) * mu_in_rot(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi)
// This is an accumulating operation.
template <typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationLocalBackward_phi(
        const LocalCoefficientsIn &mu_in_rot,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef typename complex_type::value_type real_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d = R.d();
    const auto & e_imphi = R.e_ikphi();

    assert(p_compute <= mu_in_rot.p());
    assert(p_compute <= d.p());
    assert(p_compute <= e_imphi.p());
    assert(p_compute <= mu_out.p());
    mu_in_rot.require_lower();
    mu_out.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        real_type g_l0m(1);

        for (index_type m = 0; m <= l; ++m)
        {
            real_type g_lkm(g_l0m);
            complex_type mu = toggle_sign_if_odd(m, g_lkm * d(l, m, 0) * mu_in_rot.upper(l, 0));

            for (index_type k = 1; k <= l; ++k)
            {
                // g_l^{k,m}
                g_lkm *= std::sqrt(real_type(l - k + 1) / real_type(l + k));

                mu += toggle_sign_if_odd(k + m, g_lkm * (d(l, m, k)  * mu_in_rot.upper(l, k) +
                                                         d(l, m, -k) * mu_in_rot.lower(l, -k)));
            }
            // e_imphi[m] = exp(i*m*phi)
            mu *= e_imphi[m];
            mu_out.upper(l, m) += mu;

            // (for the next iteration)
            // g_l^{0,m+1}
            g_l0m *= std::sqrt(real_type(l + m + 1) / real_type(l - m));
        }
    }
}


////////////////////////////////////////////////////////////////////////////


// Optimized version of ApplyRotationMultipoleForward_reference
// omega_out_rot(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * d(theta,l,m,k) * exp(i*k*phi) * omega_in(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi) and f_l^{k,m}
// This is an assigning operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationMultipoleForward_phi_f(
        const MultipoleCoefficientsIn &omega_in,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out_rot,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & f = R.f_forward();
    const auto & d = R.d();
    const auto & e_ikphi = R.e_ikphi();

    assert(p_compute <= omega_in.p());
    assert(p_compute <= f.p());
    assert(p_compute <= d.p());
    assert(p_compute <= e_ikphi.p());
    assert(p_compute <= omega_out_rot.p());
    omega_in.require_lower();
    omega_out_rot.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type omega = f(l, m, 0) * d(l, m, 0) * omega_in.upper(l, 0);

            for (index_type k = 1; k <= l; ++k)
            {
                // e_ikphi[k] = exp(i*k*phi), conj(e_ikphi[k]) = exp(-i*k*phi)
                omega += f(l, m, k) * (d(l, m, k)  * e_ikphi[k]            * omega_in.upper(l, k) +
                                       d(l, m, -k) * fmm::conj(e_ikphi[k]) * omega_in.lower(l, -k));
            }
            omega_out_rot.upper(l, m) = omega;
        }
    }
}


// Optimized version of ApplyRotationMultipoleBackward_reference
// omega_out(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(-i*m*phi) * omega_in_rot(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi) and f_l^{k,m}
// This is an accumulating operation.
template <typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationMultipoleBackward_phi_f(
        const MultipoleCoefficientsIn &omega_in_rot,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & f = R.f_backward();
    const auto & d = R.d();
    const auto & e_imphi = R.e_ikphi();

    assert(p_compute <= omega_in_rot.p());
    assert(p_compute <= f.p());
    assert(p_compute <= d.p());
    assert(p_compute <= e_imphi.p());
    assert(p_compute <= omega_out.p());
    omega_in_rot.require_lower();
    omega_out.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type omega = f(l, m, 0) * d(l, m, 0) * omega_in_rot.upper(l, 0);

            for (index_type k = 1; k <= l; ++k)
            {
                omega += f(l, m, k) * (d(l, m, k)  * omega_in_rot.upper(l, k) +
                                       d(l, m, -k) * omega_in_rot.lower(l, -k));
            }
            // conj(e_imphi[m]) = exp(-i*m*phi)
            omega           *= fmm::conj(e_imphi[m]);
            omega_out.upper(l, m) += omega;
        }
    }
}


// Optimized version of ApplyRotationLocalForward_reference
// mu_out_rot(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * d(theta,l,m,k) * exp(-i*k*phi) * mu_in(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi) and g_l^{k,m}
// This is an assigning operation.
template <typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationLocalForward_phi_g(
        const LocalCoefficientsIn &mu_in,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out_rot,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & g = R.g_forward();
    const auto & d = R.d();
    const auto & e_ikphi = R.e_ikphi();

    assert(p_compute <= mu_in.p());
    assert(p_compute <= g.p());
    assert(p_compute <= d.p());
    assert(p_compute <= e_ikphi.p());
    assert(p_compute <= mu_out_rot.p());
    mu_in.require_lower();
    mu_out_rot.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type mu = g(l, m, 0) * d(l, m, 0) * mu_in.upper(l, 0);

            for (index_type k = 1; k <= l; ++k)
            {
                // e_ikphi[k] = exp(i*k*phi), conj(e_ikphi[k]) = exp(-i*k*phi)
                mu += g(l, m, k) * (d(l, m, k)  * fmm::conj(e_ikphi[k]) * mu_in.upper(l, k) +
                                    d(l, m, -k) * e_ikphi[k]            * mu_in.lower(l, -k));
            }
            mu_out_rot.upper(l, m) = mu;
        }
    }
}


// Optimized version of ApplyRotationLocalBackward_reference
// mu_out(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(i*m*phi) * mu_in_rot(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi) and g_l^{k,m}
// This is an accumulating operation.
template <typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
void ApplyRotationLocalBackward_phi_g(
        const LocalCoefficientsIn &mu_in_rot,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & g = R.g_backward();
    const auto & d = R.d();
    const auto & e_imphi = R.e_ikphi();

    assert(p_compute <= mu_in_rot.p());
    assert(p_compute <= g.p());
    assert(p_compute <= d.p());
    assert(p_compute <= e_imphi.p());
    assert(p_compute <= mu_out.p());
    mu_in_rot.require_lower();
    mu_out.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type mu = g(l, m, 0) * d(l, m, 0) * mu_in_rot.upper(l, 0);

            for (index_type k = 1; k <= l; ++k)
            {
                mu += g(l, m, k) * (d(l, m, k)  * mu_in_rot.upper(l, k) +
                                    d(l, m, -k) * mu_in_rot.lower(l, -k));
            }
            // e_imphi[m] = exp(i*m*phi)
            mu *= e_imphi[m];
            mu_out.upper(l, m) += mu;
        }
    }
}


////////////////////////////////////////////////////////////////////////////


// Optimized version of ApplyRotationMultipoleForward_reference
// omega_out_rot(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * d(theta,l,m,k) * exp(i*k*phi) * omega_in(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi) and folding f_l^{k,m} with d(l,m,k)
// * using externally provided scratch array
// This is an assigning operation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray>
__attribute__((flatten))
void ApplyRotationMultipoleForward_folded(
        const MultipoleCoefficientsIn &omega_in,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out_rot,
        const size_t p_compute,
        ScratchArray & scratch)
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d_f = R.FoldedCoefficientsRotateMultipoleForward();
    const auto & e_ikphi = R.e_ikphi();
    auto & e_ikphi_omega_l_k = scratch;

    static_assert(std::is_same<complex_type, typename ScratchArray::value_type>::value,
                  "mismatching scratch value_type");
    assert(p_compute <= omega_in.p());
    assert(p_compute <= d_f.p());
    assert(p_compute <= e_ikphi.p());
    assert(p_compute <= omega_out_rot.p());
    assert(p_compute <= e_ikphi_omega_l_k.p());
    omega_out_rot.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type k = 0; k <= l; ++k)
        {
            // e_ikphi[k] = exp(i*k*phi)
            e_ikphi_omega_l_k[k] = e_ikphi[k] * omega_in.upper(l, k);
        }
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type omega = scale_real(d_f(l, m, 0), e_ikphi_omega_l_k[0]);

            for (index_type k = 1; k <= l; ++k)
            {
                omega += scale_complex(d_f(l, m, k), e_ikphi_omega_l_k[k]);
            }
            omega_out_rot.upper(l, m) = omega;
        }
    }
}


// Optimized version of ApplyRotationMultipoleBackward_reference
// omega_out(l,m) = sum_{k=-l}^{l}(
//     f_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(-i*m*phi) * omega_in_rot(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi) and folding f_l^{k,m} with d(l,m,k)
// This is an accumulating operation.
template <template <typename> class SIMDengine,
          typename MultipoleCoefficientsIn,
          typename RotationOperator,
          typename MultipoleCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
__attribute__((flatten))
void ApplyRotationMultipoleBackward_folded(
        const MultipoleCoefficientsIn &omega_in_rot,
        const RotationOperator  &R,
        MultipoleCoefficientsOut &omega_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename MultipoleCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d_f = R.FoldedCoefficientsRotateMultipoleBackward();
    const auto & e_imphi = R.e_ikphi();

    assert(p_compute <= omega_in_rot.p());
    assert(p_compute <= d_f.p());
    assert(p_compute <= e_imphi.p());
    assert(p_compute <= omega_out.p());
    omega_out.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type omega = scale_real(d_f(l, m, 0), omega_in_rot.upper(l, 0));

            for (index_type k = 1; k <= l; ++k)
            {
                omega += scale_complex(d_f(l, m, k), omega_in_rot.upper(l, k));
            }
            // conj(e_imphi[m]) = exp(-i*m*phi)
            omega           *= fmm::conj(e_imphi[m]);
            omega_out.upper(l, m) += omega;
        }
    }
}


// Optimized version of ApplyRotationLocalForward_reference
// mu_out_rot(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * d(theta,l,m,k) * exp(-i*k*phi) * mu_in(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi) and folding g_l^{k,m} with d(l,m,k)
// * using externally provided scratch array
// This is an assigning operation.
template <template <typename> class SIMDengine,
          typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray>
__attribute__((flatten))
void ApplyRotationLocalForward_folded(
        const LocalCoefficientsIn &mu_in,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out_rot,
        const size_t p_compute,
        ScratchArray & scratch)
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d_g = R.FoldedCoefficientsRotateLocalForward();
    const auto & e_ikphi = R.e_ikphi();
    auto & e_ikphi_mu_l_k = scratch;

    static_assert(std::is_same<complex_type, typename ScratchArray::value_type>::value,
                  "mismatching scratch value_type");
    assert(p_compute <= mu_in.p());
    assert(p_compute <= d_g.p());
    assert(p_compute <= e_ikphi.p());
    assert(p_compute <= mu_out_rot.p());
    assert(p_compute <= e_ikphi_mu_l_k.p());
    mu_out_rot.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type k = 0; k <= l; ++k)
        {
            // e_ikphi[k] = exp(i*k*phi), conj(e_ikphi[k]) = exp(-i*k*phi)
            e_ikphi_mu_l_k[k] = fmm::conj(e_ikphi[k]) * mu_in.upper(l, k);
        }
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type mu = scale_real(d_g(l, m, 0), e_ikphi_mu_l_k[0]);

            for (index_type k = 1; k <= l; ++k)
            {
                mu += scale_complex(d_g(l, m, k), e_ikphi_mu_l_k[k]);
            }
            mu_out_rot.upper(l, m) = mu;
        }
    }
}


// Optimized version of ApplyRotationLocalBackward_reference
// mu_out(l,m) = sum_{k=-l}^{l}(
//     g_{l}^{k,m} * (-1)^{k+m} * d(theta,l,m,k) * exp(i*m*phi) * mu_in_rot(l,k) )
// * summing only over k = 0,...,l
// * using precomputed e^(i*k*phi) and folding g_l^{k,m} with d(l,m,k)
// This is an accumulating operation.
template <template <typename> class SIMDengine,
          typename LocalCoefficientsIn,
          typename RotationOperator,
          typename LocalCoefficientsOut,
          typename ScratchArray = NoScratchDummy>
__attribute__((flatten))
void ApplyRotationLocalBackward_folded(
        const LocalCoefficientsIn &mu_in_rot,
        const RotationOperator  &R,
        LocalCoefficientsOut &mu_out,
        const size_t p_compute,
        const ScratchArray & /* unused */ = NoScratchDummy())
{
    typedef typename LocalCoefficientsOut::value_type complex_type;
    typedef ptrdiff_t index_type;

    const index_type p(p_compute);
    const auto & d_g = R.FoldedCoefficientsRotateLocalBackward();
    const auto & e_imphi = R.e_ikphi();

    assert(p_compute <= mu_in_rot.p());
    assert(p_compute <= d_g.p());
    assert(p_compute <= e_imphi.p());
    assert(p_compute <= mu_out.p());
    mu_out.invalidate_lower();

    for (index_type l = 0; l <= p; ++l)
    {
        for (index_type m = 0; m <= l; ++m)
        {
            complex_type mu = scale_real(d_g(l, m, 0), mu_in_rot.upper(l, 0));

            for (index_type k = 1; k <= l; ++k)
            {
                mu += scale_complex(d_g(l, m, k), mu_in_rot.upper(l, k));
            }
            // e_imphi[m] = exp(i*m*phi)
            mu *= e_imphi[m];
            mu_out.upper(l, m) += mu;
        }
    }
}


////////////////////////////////////////////////////////////////////////////

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
