/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_ROTATION_rotation_coefficients_hpp
#define FMSOLVR_OPERATOR_ROTATION_rotation_coefficients_hpp

#include "fmsolvr/container/triangular_array.hpp"
#include "fmsolvr/math/angle.hpp"
#include "fmsolvr/math/parity_sign.hpp"

#include <sys/types.h>
#include <cmath>
#include <vector>


namespace fmsolvr
{

namespace rotation_coefficients_utility {

// Compute the value  of the scaled associated Legendre polynomial
// L_l^m(cos(theta)) of degree l and order m at cos(theta).
template <typename Real, typename UInt>
Real ScaledLegendrePlm(
                       const UInt l_,
                       const UInt m_,
                       const Angle<Real> theta)
{

    Real result = 1;
    Real l(l_);
    Real m(m_);

    for (UInt i_ = 0; i_ < m; i_++) {
        Real i(i_);
        result = -std::sqrt((i * 2 + 1) / (i * 2 + 2)) * theta.sin() * result;
    }

    if (l == m) return result;

    Real p_l_minus_1, p_l_minus_2;
    p_l_minus_1 = result;
    result = result * std::sqrt(m * 2 + 1) * theta.cos();

    if (l == m + 1) return result;

    for (UInt i_ = m + 1; i_ < l; i_++) {
        Real i(i_);
        p_l_minus_2 = p_l_minus_1;
        p_l_minus_1 = result;
        result = ((i * 2 + 1) * theta.cos() * p_l_minus_1
                  - std::sqrt((i - m) * (i + m)) * p_l_minus_2)
                 / std::sqrt((i + m + 1) * (i - m + 1));
    }

    return result;
}

// Heaviside step function
template <typename Real, typename SInt>
Real heaviside_step(SInt m)
{
    if (m >= 0) {
        return 1;
    } else {
        return -1;
    }
}

// Function a as defined in the documentation
template <typename Real, typename SInt, typename UInt>
Real a(
    const UInt l_,
    const SInt m_)
{
    Real l(l_);
    Real m(m_);
    if (l < UInt(std::abs(m))) {
        return 0;
    } else {
        return std::sqrt((l + m + 1) * (l + 1 - m)
                         / ((l * 2 + 1) * (l * 2 + 3)));
    }
}

// Function b as defined in the documentation
template <typename Real, typename SInt, typename UInt>
Real b(
    UInt l_,
    SInt m_)
{
    Real l(l_);
    Real m(m_);
    return heaviside_step<Real>(m_) * std::sqrt((l - m - 1) * (l - m)
            / ((l * 2 - 1) * (l * 2 + 1)));
}

// Function d as defined in the documentation
template <typename Real, typename SInt, typename UInt>
Real d(
    UInt l_,
    SInt m_)
{
    Real m(m_);
    Real l(l_);
    return heaviside_step<Real>(m_) / 2 * std::sqrt((l - m) * (l + m + 1));
}

// Compute H_l^{0,m}(theta) for angle theta given by cos(theta), sin(theta)
template <typename Real, typename SInt, typename UInt>
Real H_0(
    UInt l,
    SInt m,
    const Angle<Real> theta)
{
    UInt m_abs(std::abs(m));
    return toggle_sign_if_odd(m, ScaledLegendrePlm(l, m_abs, theta));
}

// Compute H_l^{1,m}(theta) for angle theta given by cos(theta),sin(theta) from
// the values H_{l+1}^{0,m-1}, H_{l+1}^{0,m}, H_{l+1}^{0,m+1} given by H_m_minus_1,H_m,H_m_plus_1, respectively.
template <typename Real, typename SInt, typename UInt>
Real H_1(
    UInt l,
    SInt m,
    Real H_m_minus_1,
    Real H_m,
    Real H_m_plus_1,
    const Angle<Real> theta)
{
    Real result = b<Real>(l + 1, -m - 1) * (1 - theta.cos()) / 2 * H_m_plus_1;
    result -= b<Real>(l + 1, m - 1) * (1 + theta.cos()) / 2 * H_m_minus_1;
    result -= a<Real>(l, m) * theta.sin() * H_m;
    result /= b<Real>(l + 1, 0);
    return result;
}

// Compute H_l^{k,m}  from H_l^{k-1,m-1},
// H_l^{k-2,m} and  H_l^{k-1,m+1} given by H_m_minus_1,H_m,H_m_plus_1,
// respectively.
template <typename Real, typename SInt, typename UInt>
Real H_kPos(
    UInt l,
    SInt k,
    SInt m,
    Real H_m_minus_1,
    Real H_m,
    Real H_m_plus_1)
{
    Real result = d<Real>(l, k - 2) * H_m;
    result -= d<Real>(l, m - 1) * H_m_minus_1;
    result += d<Real>(l, m) * H_m_plus_1;
    result /= d<Real>(l, k - 1);
    return result;
}

// Compute H_l^{k,m}(theta) from H_l^{k+1,m-1},
// H_l^{k+2,m}, H_l^{k+1,m+1} given by H_m_minus_1,H_m,H_m_plus_1,
// respectively. Implements eq. (?) in the documentation.
template <typename Real, typename SInt, typename UInt>
Real H_kNeg(
    UInt l,
    SInt k,
    SInt m,
    Real H_m_minus_1,
    Real H_m,
    Real H_m_plus_1)
{
    Real result = d<Real>(l, k + 1) * H_m;
    result += d<Real>(l, m - 1) * H_m_minus_1;
    result -= d<Real>(l, m) * H_m_plus_1;
    result /= d<Real>(l, k);
    return result;
}
}

// Compute the rotation coefficients H_l^{k,m}(theta) for the computational triangle
// k = p,...,-p and m = abs(k),...,p and store them in the given triangular array.
template <
    typename Real,
    typename UInt,
    typename TriangularArray,
    typename Allocator = std::allocator<Real >>
void RotationCoefficients(
    UInt l,
    const Angle<Real> theta,
    TriangularArray & d)
{
    using namespace rotation_coefficients_utility;
    typedef ssize_t index_type;

    typedef typename Allocator::template rebind<Real>::other allocator_type;
    allocator_type allocator;

    Real * H_l_plus_1 = allocator.allocate(l + 2);

    // H_l^{0,m},H_{l+1}^{0,m} for m = 0,...,l
    for (index_type m = 0; m <= l; ++m) {
        d(m, 0) = H_0<Real>(l, m, theta);
        H_l_plus_1[m] = H_0<Real>(l + 1, m, theta);
    }

    // H_{l+1}^{0,l+1}
    H_l_plus_1[l + 1] = H_0<Real>(l + 1, ssize_t(l + 1), theta);

    // H_l^{1,m} for m = 1,...,l
    for (index_type m = 1; m <= l; ++m) {
        d(m, 1) = H_1<Real>(l, m,
                            H_l_plus_1[m - 1],
                            H_l_plus_1[m],
                            H_l_plus_1[m + 1],
                            theta);
    }

    // H_l^{k,m} for k=2,...,m and m = k,...,l
    for (index_type k = 2; k <= l; ++k) {
        for (index_type m = k; m < l; ++m) {
            d(m, k) = H_kPos<Real>(l, k, m,
                                   d.get(m - 1, k - 1),
                                   d.get(m, k - 2),
                                   d.get(m + 1, k - 1));
        }
        d(l, k) = H_kPos<Real>(l, k, ssize_t(l),
                               d.get(l - 1, k - 1),
                               d.get(l, k - 2),
                               0);
    }

    // H_l^{k,m} for k=-1,...,-m and m = k,...,l
    for (index_type k = -1; k >= -l; --k) {
        for (index_type m = -k; m < l; ++m) {
            d(m, k) = H_kNeg<Real>(l, k, m,
                                   d.get(m - 1, k + 1),
                                   d.get(m, k + 2),
                                   d.get(m + 1, k + 1));
        }
        d(l, k) = H_kNeg<Real>(l, k, ssize_t(l),
                               d.get(l - 1, k + 1),
                               d.get(l, k + 2),
                               0);
    }

    allocator.deallocate(H_l_plus_1, l + 2);
}


// Compute Wigner small d-rotation coefficients in the triangle
// m = 0,...,l and k = -m,...,m for degree l and rotational angle
// theta and copy values into the given RotationMatrix using its
// copy_from_triangle function.
template <
    typename Real,
    typename RotationMatrix >
void WignerSmallD_reference(
    const ssize_t l,
    const Angle<Real> theta,
    RotationMatrix & d)
{
    TriangularArrayRowMajor<Real, void> H(l);
    RotationCoefficients(l, theta, H);
    d.copy_from_triangle(H);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
