/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_ROTATION_rotation_constants_hpp
#define FMSOLVR_OPERATOR_ROTATION_rotation_constants_hpp

#include <cmath>

#include "fmsolvr/container/pyramidal_array.hpp"
#include "fmsolvr/math/parity_sign.hpp"

namespace fmsolvr
{

// These constants are independent of angles or distances and
// are used for the different rotation operations.
//
// The accessors return 3D-addressable datastructures which can be indexed with
// (l, m, k) where 0 <= l <= p, 0 <= m <= l, 0 <= k <= l.
//
// With
//   f_{l}^{k,m} = sqrt(( (l-k)!(l+k)! ) / ( (l-m)!(l+m)! ))
// and
//   g_{l}^{k,m} = sqrt(( (l-m)!(l+m)! ) / ( (l-k)!(l+k)! ))
// the different constants returned are:
//  * f_forward:  f_{l}^{k,m}
//  * f_backward: f_{l}^{k,m} * (-1)^{k+m}
//  * g_forward:  g_{l}^{k,m}
//  * g_backward: g_{l}^{k,m} * (-1)^{k+m}
//
template <typename Real, typename Allocator = std::allocator<Real>>
class RotationConstants
{
    typedef size_t size_type;
    typedef Real real_type;
    typedef QuarterPyramid<real_type, Allocator> Pyramid;
    typedef std::shared_ptr<Pyramid> shared_pyramid;
    typedef std::shared_ptr<const Pyramid> const_shared_pyramid;

public:
    typedef Pyramid real_pyramid_type;

    RotationConstants(size_type p_max, bool only_m2l = false)
    {
        f_forward_ = createMultipoleForwardRotationConstantsPyramid(p_max);
        if (!only_m2l) {
            f_backward_ = createMultipoleBackwardRotationConstantsPyramid(p_max);
            g_forward_ = createLocalForwardRotationConstantsPyramid(p_max);
        }
        g_backward_ = createLocalBackwardRotationConstantsPyramid(p_max);
    }

    const Pyramid & f_forward() const
    { return *f_forward_; }

    const Pyramid & f_backward() const
    { return *f_backward_; }

    const Pyramid & g_forward() const
    { return *g_forward_; }

    const Pyramid & g_backward() const
    { return *g_backward_; }

private:
    // f(l,m,k) = f_{l}^{k,m}
    // f_{l}^{k,m} = sqrt(( (l-k)!(l+k)! ) / ( (l-m)!(l+m)! ))
    const_shared_pyramid createMultipoleForwardRotationConstantsPyramid(const size_type p)
    {
        shared_pyramid pyramid = std::allocate_shared<Pyramid>(Allocator(), p);

        for (size_type l = 0; l <= p; ++l)
        {
            real_type f_l0m(1);
            for (size_type m = 0; m <= l; ++m)
            {
                real_type    f_lkm(f_l0m);
                (*pyramid)(l, m, 0) = f_lkm;
                for (size_type k = 1; k <= l; ++k)
                {
                    // f_l^{k,m}
                    f_lkm *= std::sqrt(real_type(l+k)/real_type(l-k+1));
                    (*pyramid)(l, m, k) = f_lkm;
                }

                // (for the next iteration)
                // f_l^{0,m+1}
                f_l0m *= std::sqrt(real_type(l-m)/real_type(l+m+1));
            }
        }
        return pyramid;
    }

    // f(l,m,k) = f_{l}^{k,m} * (-1)^{k+m}
    // f_{l}^{k,m} = sqrt(( (l-k)!(l+k)! ) / ( (l-m)!(l+m)! ))
    const_shared_pyramid createMultipoleBackwardRotationConstantsPyramid(const size_type p)
    {
        shared_pyramid pyramid = std::allocate_shared<Pyramid>(Allocator(), p);

        for (size_type l = 0; l <= p; ++l)
        {
            real_type    f_l0m(1);
            for (size_type m = 0; m <= l; ++m)
            {
                real_type    f_lkm(f_l0m);
                (*pyramid)(l, m, 0) = toggle_sign_if_odd(m, f_lkm);
                for (size_type k = 1; k <= l; ++k)
                {
                    // f_l^{k,m}
                    f_lkm *= std::sqrt(real_type(l+k)/real_type(l-k+1));
                    (*pyramid)(l, m, k) = toggle_sign_if_odd(k+m, f_lkm);
                }

                // (for the next iteration)
                // f_l^{0,m+1}
                f_l0m *= std::sqrt(real_type(l-m)/real_type(l+m+1));
            }
        }
        return pyramid;
    }

    // g(l,m,k) = g_{l}^{k,m}
    // g_{l}^{k,m} = sqrt(( (l-m)!(l+m)! ) / ( (l-k)!(l+k)! ))
    const_shared_pyramid createLocalForwardRotationConstantsPyramid(const size_type p)
    {
        shared_pyramid pyramid = std::allocate_shared<Pyramid>(Allocator(), p);

        for (size_type l = 0; l <= p; ++l) {
            real_type g_l0m(1);
            for (size_type m = 0; m <= l; ++m) {
                real_type g_lkm(g_l0m);
                (*pyramid)(l, m, 0) = g_lkm;
                for (size_type k = 1; k <= l; ++k) {
                    // g_l^{k,m}
                    g_lkm *= std::sqrt(real_type(l - k + 1) / real_type(l + k));
                    (*pyramid)(l, m, k) = g_lkm;
                }

                // (for the next iteration)
                // g_l^{0,m+1}
                if (m < l)
                {
                    g_l0m *= std::sqrt(real_type(l + m + 1) / real_type(l - m));
                }
            }
        }
        return pyramid;
    }

    // g(l,m,k) = g_{l}^{k,m} * (-1)^{k+m}
    // g_{l}^{k,m} = sqrt(( (l-m)!(l+m)! ) / ( (l-k)!(l+k)! ))
    const_shared_pyramid createLocalBackwardRotationConstantsPyramid(const size_type p)
    {
        shared_pyramid pyramid = std::allocate_shared<Pyramid>(Allocator(), p);

        for (size_type l = 0; l <= p; ++l)
        {
            real_type    g_l0m(1);
            for (size_type m = 0; m <= l; ++m)
            {
                real_type    g_lkm(g_l0m);
                (*pyramid)(l, m, 0) = toggle_sign_if_odd(m, g_lkm);
                for (size_type k = 1; k <= l; ++k)
                {
                    // g_l^{k,m}
                    g_lkm *= std::sqrt(real_type(l-k+1)/real_type(l+k));
                    (*pyramid)(l, m, k) = toggle_sign_if_odd(k+m, g_lkm);
                }

                // (for the next iteration)
                // g_l^{0,m+1}
                if (m < l)
                {
                    g_l0m *= std::sqrt(real_type(l + m + 1) / real_type(l - m));
                }
            }
        }
        return pyramid;
    }

    const_shared_pyramid f_forward_;
    const_shared_pyramid f_backward_;
    const_shared_pyramid g_forward_;
    const_shared_pyramid g_backward_;
};

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
