/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_ROTATION_rotation_matrix_hpp
#define FMSOLVR_OPERATOR_ROTATION_rotation_matrix_hpp

#include "fmsolvr/math/parity_sign.hpp"

#include <cassert>
#include <iomanip>
#include <iostream>
#include <sys/types.h>

namespace fmsolvr
{

// Contains classes to represent a Wigner small d-rotation matrix d_l^{k,m} for
// a fixed degree l. The full matrix has size (2l + 1) x (2l + 1) and
// the indecese k and m range from -l to l.
// RotationMatrices have to provide a function copy_from_triangle() that copies
// the rotation coefficients H_l^{k,m} to the matrix and converts them to
// Wigner small d-coefficients using the relation given in the documentation.
// Furthermore, a RotationMatrix has to provide a function populate_dependent()
// to compute the rotation coefficients that lie outside of the computational triangle.

// Full rotation matrix  of size (2*l+1)*(2l+1)  indexed
// by (k,m) with k = -l,...,l and m = -m,...,m. Uses built-in
// type array to store coefficients, linearization is performed
// row-wise.
// COEFFICIENTS ARE UNDEFINED UNTIL COMPUTED!

template <typename Real, typename Allocator = std::allocator<Real> >
class RotationMatrix
{
    RotationMatrix(const RotationMatrix &)            = delete;
    RotationMatrix &operator=(const RotationMatrix &) = delete;
    public:
        typedef Real value_type;
        typedef typename Allocator::template rebind<value_type>::other allocator_type;
        typedef size_t size_type;
        typedef ssize_t index_type;

        RotationMatrix(size_type l) : l_(l)
        {
            array = allocator.allocate(size());
            for (size_type i = 0; i < size(); ++i)
                allocator.construct(&array[i], value_type());
        }

        ~RotationMatrix()
        {
            for (size_type i = 0; i < size(); ++i)
                allocator.destroy(&array[i]);
            allocator.deallocate(array, size());
        }

        size_type l() const
        {
            return l_;
        }

        value_type get(index_type k, index_type m) const
        {
            return array[offset(k, m)];
        }

        // d_l^{k,m}
        const value_type &operator() (index_type k, index_type m) const
        {
            return array[offset(k, m)];
        }

        // d_l^{k,m}
        value_type &operator() (index_type k, index_type m)
        {
            return array[offset(k, m)];
        }

        // Copy the H_l^{k,m} rotation coefficients in the computational triangle
        // to the rotation matrix and convert to Wigner small d-coefficients d_l^{k,m}.
        template <typename TriangularArray>
        void copy_from_triangle(const TriangularArray &H)
        {
            for (index_type k = -index_type(l_); k <= index_type(l_); ++k)
            {
                for (index_type m = std::abs(k); m <= index_type(l_); ++m)
                {
                    // d_l^{k,m} = epsilon_k epsilon_{-m} H_l^{k,m}
                    array[offset(k, m)] = toggle_sign_if_odd(std::max(ssize_t(0), k), H(m, k));
                }
            }
        }

        // Compute values lying outside the computational triangle of the
        // recursion.
        void populate_dependent()
        {
            index_type l = index_type(l_);
            for (index_type k = -l; k <= l; ++k)
            {
                for (index_type m = -l; m <= l; ++m)
                {
                    array[offset(k, m)] = compute_dependent(k, m);
                }
            }
        }

    private:
        value_type * array;
        allocator_type allocator;
        size_type    l_;


        size_type size() const
        {
            return (2 * l_ + 1) * (2 * l_ + 1);
        }

        // Store elements row-wise.
        index_type offset(index_type k, index_type m) const
        {
            index_type l = l_;
            // Assert in-range access
            assert((-l <= k) && (k <= l));
            assert((-l <= m) && (m <= l));

            return (k + l) * (2 * l + 1) + (m + l);
        }

        // Compute value of Wigner small d-rotation d_l^{k,m} using the values in
        // the computational triangle k = -l,...,l and m = abs(k),...,l.
        value_type compute_dependent(
            index_type k,
            index_type m)
        {
            if (std::abs(k) <= std::abs(m))
            {
                if (m < 0)
                {
                    // d_^{k,m} = (-1)^{|k|+|m||} d_^{-k,-m}
                    return toggle_sign_if_odd(k + m, (*this).get(-k, -m));
                }
                else
                {
                    // d_^{k,m} = d_^{k,m}
                    return (*this).get(k, m);
                }
            }
            else if (k < 0)
            {
                // d_^{k,m} = d_^{-m,-k}
                return (*this).get(-m, -k);
            }
            else
            {
                // d_^{k,m} = (-1)^{|k|+|m||} d_^{m,k}
                return toggle_sign_if_odd(k + m, (*this).get(m, k));
            }
        }

};

// Print RotationMatrix
template <typename Real>
std::ostream &operator<< (
        std::ostream               &os,
        const RotationMatrix<Real> &ra)
{
    typedef ssize_t index_type;

    os.setf(std::ios::fixed, std::ios::floatfield);
    os << std::setprecision(4);

    index_type l = ra.l();

    for (index_type k = l; k >= -l; k--)
    {
        for (index_type m = -l; m <= l; m++)
        {
            os << std::setw(8) << ra(k, m) << " ";
        }
        os << std::endl;
    }
    return os;
}

// Upper half of the rotation matrix of size (l+1)*(2l+1)  indexed
// by (k,m) with k = 0,...,l and m = -m,...,m. Uses built-in type
// array to store the coefficients, linearization is performed row-wise.
// COEFFICIENTS ARE UNDEFINED UNTIL COMPUTED!
template <typename Real, typename Allocator = std::allocator<Real> >
class RotationMatrixUpper
{
    RotationMatrixUpper(const RotationMatrixUpper &)            = delete;
    RotationMatrixUpper &operator=(const RotationMatrixUpper &) = delete;
    public:
        typedef Real value_type;
        typedef typename Allocator::template rebind<value_type>::other allocator_type;
        typedef size_t size_type;
        typedef ssize_t index_type;

        RotationMatrixUpper(size_type l) : l_(l)
        {
            array = allocator.allocate(size());
            for (size_type i = 0; i < size(); ++i)
                allocator.construct(&array[i], value_type());
        }

        ~RotationMatrixUpper()
        {
            for (size_type i = 0; i < size(); ++i)
                allocator.destroy(&array[i]);
            allocator.deallocate(array, size());
        }

        size_type l() const
        {
            return l_;
        }

        value_type get(index_type k, index_type m) const
        {
            return array[offset(k, m)];
        }

        const value_type &operator() (index_type k, index_type m) const
        {
            return array[offset(k, m)];
        }

        value_type &operator() (index_type k, index_type m)
        {
            return array[offset(k, m)];
        }

        // The original computational triangle does not fit into the matrix.
        // Since the H_l^{k,m} matrices are symmetric, simply use transpose of
        // of the computational triangle to obtain the independent the coefficients
        // d_l^{k,m} for k = k,...,0 and m = -k,...,k.
        template <typename TriangularArray>
        void copy_from_triangle(const TriangularArray &H)
        {
            for (index_type k = 0; k <= index_type(l_); ++k)
            {
                for (index_type m = -k; m <= k; ++m)
                {
                    // d_l^{k,m} = epsilon_k epsilon_{-m} H_l^{k,m}
                    array[offset(k, m)] = toggle_sign_if_odd(std::max(ssize_t(0), -m) + k, H(k, m));
                }
            }
        }

        // Compute values lying outside the computational triangle of the
        // recursion.
        void populate_dependent()
        {
            for (index_type k = 0; k <= index_type(l_); ++k)
            {
                for (index_type m = -index_type(l_); m <= index_type(l_); ++m)
                {
                    array[offset(k, m)] = compute_dependent(k, m);
                }
            }
        }

    private:
        value_type * array;
        allocator_type allocator;
        size_type    l_;

        size_type size() const
        {
            return (l_ + 1) * (2 * l_ + 1);
        }

        // Matrix is linearized row-wise.
        index_type offset(index_type k, index_type m) const
        {
            // Assert in-range access
            assert((0 <= k) && (k <= index_type(l_)));
            assert((-index_type(l_) <= m) && (m <= index_type(l_)));

            return k * (2 * l_ + 1) + m + l_;
        }

        value_type compute_dependent(
            index_type k,
            index_type m)
        {
            if (std::abs(m) <= k)
            {
                return (*this).get(k, m);
            }
            else if (m > 0)
            {
                // d_^{k,m} = (-1)^{|k|+|m||} d_^{-k,-m}
                return toggle_sign_if_odd(k + m, (*this).get(m, k));
            }
            else
            {
                // d_^{k,m} = d_^{k,m}
                return (*this).get(-m, -k);
            }
        }
};

// Print RotationMatrix
template <typename Real>
std::ostream &operator<< (
        std::ostream                    &os,
        const RotationMatrixUpper<Real> &ra)
{
    typedef ssize_t index_type;

    os.setf(std::ios::fixed, std::ios::floatfield);
    os << std::setprecision(4);

    index_type l = ra.l();

    for (index_type k = l; k >= 0; k--)
    {
        for (index_type m = -l; m <= l; m++)
        {
            os << std::setw(8) << ra(k, m) << " ";
        }
        os << std::endl;
    }
    return os;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
