/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_ROTATION_rotation_matrix_vector_hpp
#define FMSOLVR_OPERATOR_ROTATION_rotation_matrix_vector_hpp

#include "fmsolvr/container/pyramidal_array.hpp"
#include "fmsolvr/container/simple_vector.hpp"
#include "fmsolvr/math/angle.hpp"
#include "fmsolvr/operator/rotation/rotation.hpp"
#include "fmsolvr/operator/rotation/rotation_coefficients.hpp"
#include "fmsolvr/realpair.hpp"

namespace fmsolvr
{

// A RotationMatrixVector represent a sequence of Wigner
// small d-rotation matrices d_l^{k,m}(theta)  of degrees l= 0,...,p.
// A RotationMatrixVector represents a rotation about the Y-axis
// for a given angle theta. To simplify computation, the angles are
// given by their cosine and sine, theta_cos and theta_sin, respectively.

// Uses simple_vector to store pointers to RotationMatrix objects.
template <typename RotationMatrix,
          typename Allocator = std::allocator<RotationMatrix> >
class RotationMatrixVector
{
    RotationMatrixVector &operator= (const RotationMatrixVector &RV) = delete;
    RotationMatrixVector(const RotationMatrixVector &RV)             = delete;
    public:
        typedef typename Allocator::template rebind<RotationMatrix>::other RotMat_allocator_type;
        typedef typename RotationMatrix::size_type size_type;
        typedef typename RotationMatrix::index_type index_type;
        typedef typename RotationMatrix::value_type value_type;
	typedef Angle<value_type> angle_type;
    typedef RealPair<value_type> packed_value_type;
    typedef QuarterPyramid<packed_value_type, Allocator> packed_pyramid_type;

        //
        RotationMatrixVector(size_type p) : p_(p), d_(size())
        { }

        // RotationMatrixVector of degree p with precomputed
        // RotationMatrix objects for angle theta
        template <typename RotationConstants>
        RotationMatrixVector(size_type p, angle_type theta, const RotationConstants & rc)
            : p_(p), d_(size())
        {
            compute(theta, rc);
        }

        ~RotationMatrixVector()
        {
            for (size_type i = 0; i < size(); ++i)
            {
		if (d_[i]) {
			RotMat_allocator.destroy(d_[i]);
			RotMat_allocator.deallocate(d_[i], 1);
		}
            }
        }

        // Compute rotation matrices for the rotational
        // angle theta
        template <typename RotationConstants>
        void compute(angle_type theta, const RotationConstants & rc)
        {
            for (size_t l = 0; l <= p_; ++l)
            {
		if (d_[l])
			RotMat_allocator.destroy(d_[l]);
		else
                	d_[l] = RotMat_allocator.allocate(1);
		RotMat_allocator.construct(d_[l], l);
                WignerSmallD_reference(l, theta, *d_[l]);
                d_[l]->populate_dependent();
            }
        BuildPackedPyramids(rc);
            is_computed_ = true;
        }

        // Returns true if the RotationMatrix objects in the vector are
        // computed or not.
        bool is_computed() const
        {
            return is_computed_;
        }

        value_type get (
            index_type l,
            index_type k,
            index_type m) const
        {
            // Assert that only computed matrices are read.
            assert(is_computed_);
            // Assert that access is in range
            assert((0 <= l) && (l <= index_type(p_)));

            return d_[l]->get(k, m);
        }

        const RotationMatrix &get (index_type l) const
        {
            return *d_[l];
        }


        value_type operator()  (
            index_type l,
            index_type k,
            index_type m) const
        {
            return get(l, k, m);
        }

        size_type p() const
        {
            return p_;
        }

    const packed_pyramid_type & FoldedCoefficientsRotateMultipoleForward() const
    { return *rot_multipole_pyramid_; }

    const packed_pyramid_type & FoldedCoefficientsRotateMultipoleBackward() const
    { return *rot_multipole_backwards_pyramid_; }

    const packed_pyramid_type & FoldedCoefficientsRotateLocalForward() const
    { return *rot_local_pyramid_; }

    const packed_pyramid_type & FoldedCoefficientsRotateLocalBackward() const
    { return *rot_local_backwards_pyramid_; }

    private:
	RotMat_allocator_type RotMat_allocator;
        size_type                    p_;
        bool                         is_computed_ = false;
        simple_vector<RotationMatrix*, RotMat_allocator_type> d_;

    std::shared_ptr<packed_pyramid_type> rot_multipole_pyramid_;
    std::shared_ptr<packed_pyramid_type> rot_multipole_backwards_pyramid_;
    std::shared_ptr<packed_pyramid_type> rot_local_pyramid_;
    std::shared_ptr<packed_pyramid_type> rot_local_backwards_pyramid_;

        size_type size() const
        {
            return p_ + 1;
        }

private:
    template <typename RotationConstants>
    void BuildPackedPyramids(const RotationConstants & rc)
    {
        rot_multipole_pyramid_ = std::allocate_shared<packed_pyramid_type>(Allocator(), p());
        rot_multipole_backwards_pyramid_ = std::allocate_shared<packed_pyramid_type>(Allocator(), p());
        rot_local_pyramid_ = std::allocate_shared<packed_pyramid_type>(Allocator(), p());
        rot_local_backwards_pyramid_ = std::allocate_shared<packed_pyramid_type>(Allocator(), p());

        FillPyramid(*rot_multipole_pyramid_, rc.f_forward());
        FillPyramid(*rot_multipole_backwards_pyramid_, rc.f_backward());
        FillPyramid(*rot_local_pyramid_, rc.g_forward());
        FillPyramid(*rot_local_backwards_pyramid_, rc.g_backward());
    }

    template <typename Pyramid, typename Flkm>
    void FillPyramid(Pyramid & pyramid, const Flkm & flkm)
    {
        typedef typename Pyramid::value_type pair_type;
        typedef typename pair_type::value_type real_type;

        assert(p() <= pyramid.p());
        assert(p() <= flkm.p());

        for (index_type l = 0; l <= index_type(p()); ++l) {
            const auto & d_l = get(l);
            for (index_type j = 0; j <= l; ++j) {
                pyramid(l, j, 0) = pair_type(flkm(l, j, 0) * d_l(j, 0), 0.);
                for (index_type k = 1; k <= l; ++k) {
                    real_type f = flkm(l, j, k);
                    real_type x = d_l(j, k);
                    real_type y = toggle_sign_if_odd(k, d_l(j, -k));
                    pyramid(l, j, k) = pair_type(f * (x + y), f * (x - y));
                }
            }
        }
    }
};

// Print rotation matrices of rotation operator
template <typename RotationMatrix,
          typename Allocator>
std::ostream &operator<< (std::ostream                               &os,
                          const RotationMatrixVector<RotationMatrix, Allocator> &rv)
{
    typedef ssize_t index_type;

    for (index_type l = 0; l <= rv.p(); l++)
    {
        os << "l = " << l << ":" << std::endl;
        os << rv(l) << std::endl;
    }
    return os;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
