/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_ROTATION_rotation_operator_hpp
#define FMSOLVR_OPERATOR_ROTATION_rotation_operator_hpp

#include "fmsolvr/math/angle.hpp"

#include <memory>
#include <iostream>

namespace fmsolvr
{

struct Nothing { };


// RotationOperator

// Rotation operator for a given degree p. Represents a rotation of angle
// phi about the z-axis followed by a rotation of angle theta about the
// y-axis. Holds a reference
// to a RotationMatrixVector containing the Wigner small d-rotation
// matrices for the rotation about the Y-axis as well as the cosine
// and sine of phi for the rotation about the z-axis.
// The referenced RotationMatrixVector objects should normally be managed
// by the RotationGenerator, nevertheless for rotating about arbitrary angles
// the RotationOperator has to manage the RotationMatrixVector itself.
template <
    typename RotationMatrixVector,
    typename EikPhiTable,
    typename RotationConstants,
    typename OwnedStuff = Nothing>
class RotationOperator
{
public:
    typedef EikPhiTable e_ikphi_table_type;
        typedef typename RotationMatrixVector::size_type size_type;
        typedef typename RotationMatrixVector::index_type index_type;
        typedef typename RotationMatrixVector::value_type value_type;
        typedef Angle<value_type> angle_type;
        typedef typename RotationConstants::real_pyramid_type real_pyramid_type;
        typedef typename RotationMatrixVector::packed_pyramid_type packed_pyramid_type;
    typedef std::ptrdiff_t id_type;

        RotationOperator(
            const RotationMatrixVector & d,
            const EikPhiTable & eikphitable,
            const RotationConstants & rc,
            const id_type id,
            OwnedStuff owned_stuff = Nothing()) :
            d_(d),
            eikphitable_(eikphitable),
            rot_constants_(rc),
            id_(id),
            owned_stuff_(owned_stuff)
        {}

    const real_pyramid_type & f_forward() const
    { return rot_constants_.f_forward(); }

    const real_pyramid_type & f_backward() const
    { return rot_constants_.f_backward(); }

    const real_pyramid_type & g_forward() const
    { return rot_constants_.g_forward(); }

    const real_pyramid_type & g_backward() const
    { return rot_constants_.g_backward(); }

        const RotationMatrixVector & d() const
        {
            return d_;
        }

    const packed_pyramid_type & FoldedCoefficientsRotateMultipoleForward() const
    { return d().FoldedCoefficientsRotateMultipoleForward(); }

    const packed_pyramid_type & FoldedCoefficientsRotateMultipoleBackward() const
    { return d().FoldedCoefficientsRotateMultipoleBackward(); }

    const packed_pyramid_type & FoldedCoefficientsRotateLocalForward() const
    { return d().FoldedCoefficientsRotateLocalForward(); }

    const packed_pyramid_type & FoldedCoefficientsRotateLocalBackward() const
    { return d().FoldedCoefficientsRotateLocalBackward(); }

        const angle_type & phi() const
        { return eikphitable_.phi(); }

    const EikPhiTable & e_ikphi() const
    { return eikphitable_; }

    id_type id() const
    { return id_; }

    private:
        const RotationMatrixVector & d_;
        const EikPhiTable & eikphitable_;
        const RotationConstants & rot_constants_;
        const id_type id_;
        OwnedStuff owned_stuff_;
};

// Print rotation matrices of rotation operator
template <typename... Args>
std::ostream &operator<< (std::ostream                                 &os,
                          const RotationOperator<Args...> &R)
{
    os << "RotationOperator:" << std::endl;
    os << R.d() << std::endl;
    return os;
}

}  // namespae fmsolvr

#endif
// vim: et:ts=4:sw=4
