/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_P2P_coulomb_hpp
#define FMSOLVR_P2P_coulomb_hpp

#include "fmsolvr/stackedmultireal.hpp"
#include "fmsolvr/math/rcplength.hpp"


namespace fmsolvr {

    // generic Coulomb (or '1/r') kernel for the directed interaction between
    // two distinct particles (target, source)
template <typename StackedReal>
class coulomb_kernel
{
    typedef StackedMultiReal<StackedReal> RealGen;
    typedef typename RealGen::stacked_real_type Real;
    typedef typename RealGen::stacked_real3_type Real3;
    typedef std::pair<Real, Real3> potential_efield_pair;

public:
        template <typename Target, typename Source>
        potential_efield_pair operator () (const Target & target, const Source & source)
        {
            Real3 x_i(target);
            Real3 x_j(source);
            Real q_j(source.s());

            Real3 diff = x_i - x_j;                 // (x_i - x_j)
            Real rlen = rcplength(diff);            // 1 / |x_i - x_j|
            Real q_j_rlen = q_j * rlen;             // q_j / |x_i - x_j|
            Real rlen2 = rlen * rlen;               // 1 / |x_i - x_j|^2
            Real q_j_rlen3 = q_j_rlen * rlen2;      // q_j / |x_i - x_j|^3
            Real3 f_i = diff * q_j_rlen3;           // (x_i - x_j) * q_j / |x_i - x_j|^3
            Real phi_i = q_j_rlen;
            return potential_efield_pair(phi_i, f_i);
        }
};

}  // namespace fmm

#endif
// vim: et:ts=4:sw=4
