/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_P2P_pthreads_hpp
#define FMSOLVR_P2P_pthreads_hpp

#include <pthread.h>
#include <cstdlib>
#include <cassert>
#include <vector>


namespace fmsolvr {
namespace pairwise {

        template <typename object_type>
        struct tile_worker
        {
            typedef size_t size_type;
            typedef void (object_type::*member_function)(size_type, size_type, size_type, size_type, size_type);

            object_type * object;
            member_function function;
            size_type i_begin, i_end, j_begin, j_end;

            tile_worker(object_type & object, member_function function,
                    size_type i_begin, size_type i_end, size_type j_begin, size_type j_end)
                : object(&object), function(function),
                i_begin(i_begin), i_end(i_end), j_begin(j_begin), j_end(j_end)
            { }

            void operator () ()
            {
                (object->*function)(i_begin, 1, i_end, j_begin, j_end);
            }
        };

        template <typename object_type, typename data_type>
        struct tile_worker2
        {
            typedef size_t size_type;
            typedef void (object_type::*member_function)(size_type, size_type, size_type, size_type, size_type, data_type);

            object_type * object;
            member_function function;
            size_type i_begin, i_end, j_begin, j_end;
            data_type data;

            tile_worker2(object_type & object, member_function function,
                    size_type i_begin, size_type i_end, size_type j_begin, size_type j_end, data_type data)
                : object(&object), function(function),
                i_begin(i_begin), i_end(i_end), j_begin(j_begin), j_end(j_end), data(data)
            { }

            void operator () ()
            {
                (object->*function)(i_begin, 1, i_end, j_begin, j_end, data);
            }
        };

        template <typename Worker>
        void* pthreads_worker(void * data)
        {
            Worker * worker = reinterpret_cast<Worker *>(data);
            (*worker)();
            return 0;
        }

// this is a "tile generator", i.e. it splits the input tile into
// smaller bits that get handed off (in parallel) to a "tile processor"
template <typename TileProcessor>
class pthreads
{
    typedef typename TileProcessor::size_type size_type;
    typedef typename TileProcessor::Real Real;
    typedef typename TileProcessor::Real3 Real3;
    typedef TileProcessor tile_processor_type;

            typedef void (tile_processor_type::*tile_processor_member_function)(size_type, size_type, size_type, size_type, size_type);
            typedef void (tile_processor_type::*tile_processor_member_function5)(size_type, size_type, size_type, size_type, size_type, Real3);

            tile_processor_type tile_processor;

            int nthreads;

public:
    template <typename... Args>
    pthreads(Args&&... args) : tile_processor(std::forward<Args>(args)...), nthreads(0)
            {
                if (getenv("OMP_NUM_THREADS"))
                    nthreads = atoi(getenv("OMP_NUM_THREADS"));
                if (nthreads < 1)
                    nthreads = 1;
            }

            void overlapping_interactions(size_type i_begin, size_type i_end, size_type j_begin, size_type j_end)
            {
                run_threaded(&tile_processor_type::overlapping_interactions, i_begin, i_end, j_begin, j_end);
            }

            void disjoint_interactions(size_type i_begin, size_type i_end, size_type j_begin, size_type j_end)
            {
                run_threaded(&tile_processor_type::disjoint_interactions, i_begin, i_end, j_begin, j_end);
            }

            void offsetted_interactions(size_type i_begin, size_type i_end, size_type j_begin, size_type j_end, Real3 offset)
            {
                run_threaded2(&tile_processor_type::offsetted_interactions, i_begin, i_end, j_begin, j_end, offset);
            }

            void single_scaled_interaction(size_type i, size_type j, Real scale, Real3 offset)
            {
                tile_processor.single_scaled_interaction(i, j, scale, offset);
            }

private:
            void run_threaded(tile_processor_member_function processor_function,
                    size_type i_begin, size_type i_end, size_type j_begin, size_type j_end)
            {
                typedef tile_worker<tile_processor_type> worker_type;
                if (i_end <= i_begin)
                    return;
                size_type i_len = i_end - i_begin;
                int num_threads = std::min(nthreads, int(i_len));
                std::vector<worker_type> thread_workers;
                std::vector<pthread_t> threads;

                thread_workers.reserve(num_threads);
                threads.reserve(num_threads);
                size_type i_done = 0;
                for (int t = 0; t < num_threads; ++t)
                {
                    size_type i_slice = (i_len - i_done) / (num_threads - t);
                    thread_workers.push_back(worker_type(tile_processor, processor_function, i_begin + i_done, i_begin + i_done + i_slice, j_begin, j_end));
                    pthread_t thr;
                    if (pthread_create(&thr, 0, pthreads_worker<worker_type>, &thread_workers[t]) == 0)
                        threads.push_back(thr);
                    else
                        throw "pthread_create";
                    i_done += i_slice;
                }
                assert(i_begin + i_done == i_end);
                for (size_t t = 0; t < threads.size(); ++t) {
                    pthread_join(threads[t], 0);
                }
            }

            template <typename data_type>
            void run_threaded2(tile_processor_member_function5 processor_function,
                    size_type i_begin, size_type i_end, size_type j_begin, size_type j_end, data_type data)
            {
                typedef tile_worker2<tile_processor_type, data_type> worker_type;
                if (i_end <= i_begin)
                    return;
                size_type i_len = i_end - i_begin;
                int num_threads = std::min(nthreads, int(i_len));
                std::vector<worker_type> thread_workers;
                std::vector<pthread_t> threads;

                thread_workers.reserve(num_threads);
                threads.reserve(num_threads);
                size_type i_done = 0;
                for (int t = 0; t < num_threads; ++t)
                {
                    size_type i_slice = (i_len - i_done) / (num_threads - t);
                    thread_workers.push_back(worker_type(tile_processor, processor_function, i_begin + i_done, i_begin + i_done + i_slice, j_begin, j_end, data));
                    pthread_t thr;
                    if (pthread_create(&thr, 0, pthreads_worker<worker_type>, &thread_workers[t]) == 0)
                        threads.push_back(thr);
                    else
                        throw "pthread_create";
                    i_done += i_slice;
                }
                assert(i_begin + i_done == i_end);
                for (size_t t = 0; t < threads.size(); ++t) {
                    pthread_join(threads[t], 0);
                }
            }
};

}  // namespace pairwise
}  // namespace fmm

#endif
// vim: et:ts=4:sw=4
