/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_P2P_tile_hpp
#define FMSOLVR_P2P_tile_hpp

#include "fmsolvr/p2p/reducer.hpp"

namespace fmsolvr {
namespace pairwise {

// this is a "tile processor", i.e. it performs (single-threaded) the
// actual field+potential computation
// (these small tiles are created by a "tile generator")
template <typename TileBase>
class tile
{
public:
    typedef typename TileBase::size_type size_type;
    typedef typename TileBase::real_type Real;
    typedef typename TileBase::real3_type Real3;

private:
    typedef typename TileBase::real4_type Real4;
    typedef typename TileBase::result_container result_container;
    typedef reducer<Real, result_container> reducer_type;

    TileBase tbase_;

public:
    template <typename... Args>
    tile(Args&&... args) : tbase_(std::forward<Args>(args)...)
    { }

            // source == target and i and j ranges may overlap
            void overlapping_interactions(size_type i_begin, size_type i_step, size_type i_end, size_type j_begin, size_type j_end)
            {
                auto &result = tbase_.result();
                const auto &target = tbase_.target();
                for (size_type i = i_begin; i < i_end; i += i_step) {
                    Real3 xyz_i = target.load_xyz(i);
                    reducer_type reducer_i(result, i);
                    interaction_sequence(xyz_i, reducer_i, j_begin, std::min(j_end, i));
                    // skip j == i
                    interaction_sequence(xyz_i, reducer_i, std::max(j_begin, i + 1), j_end);
                }
            }

            // source != target or i and j ranges are disjoint
            void disjoint_interactions(size_type i_begin, size_type i_step, size_type i_end, size_type j_begin, size_type j_end)
            {
                auto &result = tbase_.result();
                const auto &target = tbase_.target();
                for (size_type i = i_begin; i < i_end; i += i_step) {
                    Real3 xyz_i = target.load_xyz(i);
                    reducer_type reducer_i(result, i);
                    interaction_sequence(xyz_i, reducer_i, j_begin, j_end);
                }
            }

            // like disjoint_interactions, but add offset vector to the distance vector between particles i and j
            void offsetted_interactions(size_type i_begin, size_type i_step, size_type i_end, size_type j_begin, size_type j_end, Real3 offset)
            {
                const auto &target = tbase_.target();
                auto &result = tbase_.result();
                for (size_type i = i_begin; i < i_end; i += i_step) {
                    Real3 xyz_i = target.load_xyz(i) + offset;
                    reducer_type reducer_i(result, i);
                    interaction_sequence(xyz_i, reducer_i, j_begin, j_end);
                }
            }

            // a single interaction between target i and source j, can be scaled
            void single_scaled_interaction(size_type i, size_type j, Real scale = 1.0, Real3 offset = Real3(0., 0., 0.))
            {
                auto &result = tbase_.result();
                const auto &target = tbase_.target();
                const auto &source = tbase_.source();
                Real3 xyz_i = target.load_xyz(i) + offset;
                reducer_type reducer_i(result, i);
                Real4 xyzs_j = source.load_xyzs(j);
                auto res = single_interaction(xyz_i, xyzs_j);  // returns a pair (potential, field)
                reducer_i(res.first * scale, res.second * scale);
            }

protected:
            void interaction_sequence(const Real3 & xyz_i, reducer_type & reducer_i, size_type j_begin, size_type j_end)
            {
                const auto &source = tbase_.source();
                for (size_type j = j_begin; j < j_end; ++j) {
                    Real4 xyzs_j = source.load_xyzs(j);
                    auto res = single_interaction(xyz_i, xyzs_j);  // returns a pair (potential, field)
                    reducer_i(res.first, res.second);
                }
            }

            // returns a pair (potential, field)
            auto single_interaction(const Real3 & xyz_i, const Real4 & xyzs_j)
                -> decltype(TileBase::template Kernel<Real>()(xyz_i, xyzs_j))
            {
                auto kernel = TileBase::template Kernel<Real>();
                return kernel(xyz_i, xyzs_j);
            }
};

}  // namespace pairwise
}  // namespace fmm

#endif
// vim: et:ts=4:sw=4
