/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SEQUENTIAL_pass12345_hpp
#define FMSOLVR_SEQUENTIAL_pass12345_hpp

#include "fmsolvr/fmm_enum.hpp"
#include "fmsolvr/sequential/pass1.hpp"
#include "fmsolvr/sequential/pass2.hpp"
#include "fmsolvr/sequential/pass2dl.hpp"
#include "fmsolvr/sequential/pass3.hpp"
#include "fmsolvr/sequential/pass4.hpp"
#include "fmsolvr/sequential/pass5.hpp"
#include "fmsolvr/util/TimeLine.hpp"
#include "fmsolvr/util/checking_helpers.hpp"
#include "fmsolvr/util/envtools.hpp"
#include "fmsolvr/util/wrpapi.hpp"


namespace fmsolvr {

namespace {

template <typename T>
void unused(const T &)
{ }

}  // anonymous namespace

template <typename FMMHandle>
void pass12345(papiwrapper<> &pw, FMMHandle &fmm_handle)
{
    double t = -timestamp();

    auto &verbout = fmm_handle.verbout();

    auto & particle_handle = fmm_handle.particle_handle();
    const auto & simulationbox = particle_handle.simulationbox();

    // regular passes
    const bool doCompensateDipole = !simulationbox.open_boundary_conditions() &&
                                    (get_env_int("DIPOLE_COMPENSATION", 1) != 0);
    const bool doPass1P2M = (get_env_int("PASS1P2M", 1) != 0);
    const bool doPass1M2M = (get_env_int("PASS1M2M", 1) != 0);
    const bool doPass2 = (get_env_int("PASS2", 1) != 0);
    const bool doPass2Lattice = (get_env_int("PASS2LATTICE", 1) != 0);
    const bool doPass3 = (get_env_int("PASS3", 1) != 0);
    const bool doPass4 = (get_env_int("PASS4", 1) != 0);
    const bool doPass5 = (get_env_int("PASS5", 1) != 0);
    const bool doCheckInteractionCount = (get_env_int("CHECKINTERACTIONCOUNT", 1) != 0);

    typedef typename FMMHandle::size_type size_type;
    typedef typename FMMHandle::boxid_type BoxID;
    typedef typename FMMHandle::P2P_type P2P_type;

    //typedef timeline::TimeLineNone<FMMActivity> TimeLine;
    typedef timeline::TimeLinePapiPrint<FMMActivity> TimeLine;
    //typedef timeline::TimeLineAccumulated<FMMActivity> TimeLine;
    //typedef timeline::TimeLineDetailed<FMMActivity> TimeLine;

    const size_type n = particle_handle.n_active();
    const size_type depth = fmm_handle.depth();
    const size_type ws = fmm_handle.ws();
    const size_type p = fmm_handle.p();

    TimeLine timeline(fmm_handle.verbose() ? &pw : NULL);

    verbout << "# Features:";
    verbout << " Pass1P2M:" << doPass1P2M;
    verbout << " Pass1M2M:" << doPass1M2M;
    verbout << " Pass2:" << doPass2;
    verbout << " Pass2Lattice:" << doPass2Lattice;
    verbout << " Pass3:" << doPass3;
    verbout << " Pass4:" << doPass4;
    verbout << " Pass5:" << doPass5;
    verbout << " DipoleCompensation:" << doCompensateDipole;
    verbout << std::endl;

    // *** near field ***

    // PASS 5: P2P
    if (doPass5)
    {
        verbout << "# PASS 5: P2P\n";
        auto activity = timeline.startActivity(FMMActivity::P2P);
        size_t p2pcount = 0;
        auto particles = particle_handle.particle_adapter();
        auto result = particle_handle.potential_efield_nf_adapter();

        P2P_type P2P(result, particles, particles);
        auto pass5result = pass5(fmm_handle, P2P);
        p2pcount += pass5result.first;
        particle_handle.add_interactions_nf(pass5result.second);
        timeline.stopActivity(activity, pass5result.second);
        pw.report(verbout, p2pcount);  // alternative normalization
    }

    // *** far field ***

    // PASS 1: P2M
    if (doPass1P2M)
    {
        verbout << "# PASS 1: P2M\n";
        auto activity = timeline.startActivity(FMMActivity::P2M);
        pass1_P2M(fmm_handle);
        timeline.stopActivity(activity, n);
    }

    // PASS 1: M2M
    if (doPass1M2M)
    {
        verbout << "# PASS 1: M2M\n";
        auto activity = timeline.startActivity(FMMActivity::M2M);
        size_t m2mcount = 0;
        m2mcount += pass1_M2M_tree(fmm_handle);
        timeline.stopActivity(activity, m2mcount);
    }

    assert(checking::is_finite<BoxID>(fmm_handle.tree_omega(), depth, p));

    // tree_omega has had populate_lower called for all elements, the omegas are ready to use

    // PASS 2: M2L
    if (doPass2)
    {
        double t_pass2 = -timestamp();
        //all M2L calls are independent: for each target box all contributors determined and combined into target box
        //with WS=1 each target box has 189 contributors
        verbout << "# PASS 2: M2L\n";
        for (size_t d = 1; d <= depth; ++d) {
            auto activity = timeline.startActivity(FMMActivity::M2L);
            size_t m2lcount = 0;
            m2lcount += pass2_M2L_one_level(fmm_handle, d);
            timeline.stopActivity(activity, m2lcount);
        }
        t_pass2 += timestamp();
        verbout << "Pass 2 M2L elapsed: " << t_pass2 << std::endl;
    }

    // PASS 2: Lattice (+ Dipole Compensation)
    if (doPass2Lattice)
    {
        if (!simulationbox.open_boundary_conditions())
        {
            verbout << "# PASS 2: Lattice\n";
            auto activity = timeline.startActivity(FMMActivity::Lattice);
            pass2_dipole_lattice(fmm_handle);
            timeline.stopActivity(activity, 1);
        }
    }

    // tree_mu needs populate_lower() to be called for all elements before being used
    // only tree_mu[root = BoxID(0, 0, 0, 0)] is complete at this point, L2L will accumulate more into the deeper levels

    assert(checking::is_finite<BoxID>(fmm_handle.tree_mu(), depth, p));

    // PASS 3: L2L
    if (doPass3)
    {
        verbout << "# PASS 3: L2L\n";
        auto activity = timeline.startActivity(FMMActivity::L2L);
        size_t l2lcount = 0;
        l2lcount += pass3_L2L_tree(fmm_handle);
        timeline.stopActivity(activity, l2lcount);
    }

    assert(checking::is_finite<BoxID>(fmm_handle.tree_mu(), depth, p));

    // tree_mu has had populate_lower called for all elements, the are ready to use

    // PASS 4: Potential and Field
    if (doPass4)
    {
        verbout << "# PASS 4: phi & f\n";
        auto activity = timeline.startActivity(FMMActivity::L2P);
        pass4(fmm_handle);
        timeline.stopActivity(activity, n);
    }

    size_t nf_interactions = particle_handle.interactions_nf();
    size_t ff_interactions = particle_handle.interactions_ff();
    verbout << "# interactions: n=" << n << " nf=" << nf_interactions << " ff=" << ff_interactions << std::endl;
    if (doCheckInteractionCount)
    {
        if (simulationbox.open_boundary_conditions())
            assert(nf_interactions + ff_interactions == n * (n - 1));
        else
            assert(nf_interactions + ff_interactions ==
                   n * (n - 1) + n * n * ((2 * ws + 1) * (2 * ws + 1) * (2 * ws + 1) - 1));
    }

    t += timestamp();
    verbout << "Elapsed FMM pass 1-5 time: " << t << std::endl;

#ifdef NDEBUG
    unused(BoxID(0, 0, 0, 0));
    unused(ws);
    unused(p);
#endif
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
