/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SEQUENTIAL_pass2_hpp
#define FMSOLVR_SEQUENTIAL_pass2_hpp

#include "fmsolvr/operator/p3/multipole2local_p3.hpp"
#include "fmsolvr/operator/p4/multipole2local.hpp"
#include "fmsolvr/util/envtools.hpp"

#include <cassert>

namespace fmsolvr {

namespace {

// remap periodic index i into [0,..,dim)
template <typename Integral>
inline Integral periodic_remap(Integral i, Integral dim)
{
    while (i < 0)
        i += dim;
    while (i >= dim)
        i -= dim;
    return i;
}

}  // anonymous namespace

// this routine processes a single tree level
template <typename FMMHandle>
size_t pass2_M2L_one_level(FMMHandle & fmm_handle, size_t d)
{
    const bool p4 = get_env_int("P4", 0);
    typedef typename FMMHandle::real_type Real;
    typedef typename FMMHandle::size_type size_type;
    typedef typename FMMHandle::index_type index_type;
    typedef typename FMMHandle::boxid_type BoxID;
    typedef typename FMMHandle::operatorid_m2l_type OperatorID_M2L;
    typedef typename OperatorID_M2L::ijk_type IJK;
    typedef typename FMMHandle::allocator_type allocator_type;
    typedef typename FMMHandle::coefficient_type coefficient_type;
    typedef typename FMMHandle::RotationDataTypes::Rot_Operator RotationOperator;
    typedef typename FMMHandle::RotationDataTypes::M2L_Operator M2Lp3_Operator;
    typedef typename FMMHandle::RotationDataTypes::template ScratchTypeGenerator<coefficient_type>::Scratch scratch_type;

    const auto &simulationbox = fmm_handle.simulationbox();
    const auto & tree_omega = fmm_handle.tree_omega();
    auto & tree_mu = fmm_handle.tree_mu();
    auto &GM2L = fmm_handle.RotationGeneratorM2L();
    const size_type ws = fmm_handle.ws();
    const size_type p = fmm_handle.p();
    const auto & m2l_operator = fmm_handle.m2l_operator_cube(d, ws);
    const Real m2l_operator_scaling_factor = fmm_handle.m2l_operator_scaling_factor();
    scratch_type scratch(p);

    size_t m2lcount = 0;
    index_type dim = index_type(1) << d;
    // on each level there are two different index coordinate systems for cubic data structures
    // with index (0, 0, 0) at the distinguished (lower left front) corner
    // * two cubes of dimension dim^3 for omegas and mus
    // * one cube of dimension (2(2ws+1)+1)^3 for the B operators
    // the distinguished corner (0, 0, 0) of the B cube is at (i0, j0, k0) in the omega cube
    // * (i, j, k) is the index of some mu (and the B cube is centered here, too)
    // * (ii, jj, kk) is the index of some omega
    // * (ii - i0, jj - j0, kk - k0) is the index of the corresponding B
    for (index_type i = 0; i < dim; ++i)
    {
        for (index_type j = 0; j < dim; ++j)
        {
            for (index_type k = 0; k < dim; ++k)
            {
                IJK op_mu_ijk(i, j, k);  // for addressing the operator
                BoxID mu_bid(i, j, k, d);
                auto & mu = tree_mu[mu_bid];
                mu.zero(p);
                mu.clearSourceCounters();
                // no particles in the target box?
                // (this HACK only works if source tree and target tree are identical
                // and Pass 1 is finished (at least for mu_bid))
                if (tree_omega[mu_bid].empty())
                    continue;
                // parent box coordinates
                index_type ip = i / 2;
                index_type jp = j / 2;
                index_type kp = k / 2;
                // outer ws shells
                index_type io_beg = (ip - ws) * 2;
                index_type io_end = (ip + ws + 1) * 2;
                index_type jo_beg = (jp - ws) * 2;
                index_type jo_end = (jp + ws + 1) * 2;
                index_type ko_beg = (kp - ws) * 2;
                index_type ko_end = (kp + ws + 1) * 2;
                // inner - no far field interactions
                index_type ii_beg = i - ws;
                index_type ii_end = i + ws + 1;
                index_type ji_beg = j - ws;
                index_type ji_end = j + ws + 1;
                index_type ki_beg = k - ws;
                index_type ki_end = k + ws + 1;
                for (index_type ii = io_beg; ii < io_end; ++ii)
                {
                    for (index_type jj = jo_beg; jj < jo_end; ++jj)
                    {
                        for (index_type kk = ko_beg; kk < ko_end; ++kk)
                        {
                            if (ii < ii_beg || ii >= ii_end
                                    || jj < ji_beg || jj >= ji_end
                                    || kk < ki_beg || kk >= ki_end) {
                                // FIXME: missing support for 1D/2D periodicity
                                if (simulationbox.open_boundary_conditions() && (ii < 0 || ii >= dim || jj < 0 || jj >= dim || kk < 0 || kk >= dim))
                                    continue;
                                IJK op_omega_ijk(ii, jj, kk);  // for addressing the operator, no wrapping!
                                OperatorID_M2L op_ijk(op_omega_ijk, op_mu_ijk, d);
                                // FIXME: skip if out of range and not periodic in that direction
                                index_type iio = periodic_remap(ii, dim);
                                index_type jjo = periodic_remap(jj, dim);
                                index_type kko = periodic_remap(kk, dim);
                                assert(0 <= iio && iio < dim && 0 <= jjo && jjo < dim && 0 <= kko && kko < dim);
                                BoxID omega_bid(iio, jjo, kko, d);
                                const auto & omega = tree_omega[omega_bid];
                                if (omega.empty())
                                    continue;

                                //M2L is similar to M2M -> see above contains parallelism which should be exploited at least on warp level to ensure good memory access pattern
                                if (p4) {
                                    M2L_reference(omega, m2l_operator(op_ijk), mu, p);
                                } else {
                                    RotationOperator R =
                                        GM2L.generate_rotation_operator(op_ijk);
                                    Real shift_distance = simulationbox.box_center_distance(op_ijk);
                                    std::shared_ptr<M2Lp3_Operator> B_p3 =
                                        std::allocate_shared<M2Lp3_Operator>(allocator_type(), shift_distance, p, m2l_operator_scaling_factor);
                                    M2Lp3<FMMHandle::template SIMDengine>(omega, R, *B_p3, mu, p, scratch);
                                }
                                ++m2lcount;
                                mu.addSourceCounters(omega.sourceCounters());
                            }
                        }
                    }
                }
            }
        }
    }
    return m2lcount;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
