/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SEQUENTIAL_pass2dl_hpp
#define FMSOLVR_SEQUENTIAL_pass2dl_hpp

#include <iostream>

#include "fmsolvr/dipole_compensation.hpp"
#include "fmsolvr/operator/p4/multipole2local.hpp"
#include "fmsolvr/util/envtools.hpp"


namespace fmsolvr {

template <typename FMMHandle>
void pass2_dipole_lattice(FMMHandle &fmm_handle)
{
    typedef typename FMMHandle::real33_type Real33;
    typedef typename FMMHandle::real3_type Real3;
    typedef typename FMMHandle::real_type Real;
    typedef typename FMMHandle::size_type size_type;
    typedef typename FMMHandle::boxid_type BoxID;
    typedef dipole_compensation<FMMHandle> DipoleCompensation;

    auto &verbout = fmm_handle.verbout();

    const auto &simulationbox = fmm_handle.simulationbox();
    const bool doCompensateDipole = !simulationbox.open_boundary_conditions() &&
                                    (get_env_int("DIPOLE_COMPENSATION", 1) != 0);
    DipoleCompensation dipole_compensation;

    // NOTE: this only works if the system is chargeless (sum of charges == 0)
    if (doCompensateDipole)
    {
        verbout << "# PASS 2: Dipole Compensation\n";
        const auto &abc = simulationbox.abc();
        Real33 convert_to_abc = change_of_basis_from_standard_basis(abc);
        Real33 convert_from_abc = change_of_basis_to_standard_basis(abc);
        // compute monopole and (cartesian) dipole for reference point (0,0,0)
        // the dipole is independent of the reference point as long as the monopole is 0
        auto mpdp = fmm_handle.particle_handle().monopole_dipole();
        Real totalq = mpdp.first;
        Real3 dp = mpdp.second;
        Real3 dp_abc = convert_to_abc * dp;
        Real3 pdp_abc(
                simulationbox.periodic_a() ? dp_abc.x() : 0.,
                simulationbox.periodic_b() ? dp_abc.y() : 0.,
                simulationbox.periodic_c() ? dp_abc.z() : 0.);
        Real3 pdp = convert_from_abc * pdp_abc;
        if (simulationbox.periodic_a() && simulationbox.periodic_b() && simulationbox.periodic_c())
        {
            pdp = dp;
        }
        verbout << "monopole: " << totalq << std::endl;
        verbout << "dipole: " << dp << "_xyz  " << dp_abc << "_abc" << std::endl;
        verbout << "dipole to be compensated: " << pdp << "_xyz  " << pdp_abc << "_abc"
                  << std::endl;

        Real3 origin(0., 0., 0.);
        dipole_compensation = DipoleCompensation(fmm_handle, pdp);
        const auto &dc = dipole_compensation;
        verbout << "fictive charges: " << dc.q0() << " " << dc.qa() << " " << dc.qb() << " "
                  << dc.qc() << std::endl;
        Real3 rdp =
                dp + origin * dc.q0() + abc.a() * dc.qa() + abc.b() * dc.qb() + abc.c() * dc.qc();
        verbout << "remaining dipole: " << rdp << "_xyz  " << convert_to_abc * rdp << "_abc"
                  << std::endl;
    }

    const size_type p = fmm_handle.p();
    BoxID bid_root(0, 0, 0, 0);
    const auto &omega = fmm_handle.tree_omega()[bid_root];
    auto &omega_dp = fmm_handle.omega_dp();

    // Dipole Compensation
    if (doCompensateDipole)
    {
        copy_upper_triangle(omega, omega_dp, p);
        dipole_compensation.add_compensating_omega(fmm_handle, omega_dp);
        omega_dp.populate_lower();
    }

    auto &mu = fmm_handle.tree_mu()[bid_root];
    mu.zero(p);
    mu.clearSourceCounters();
    single_M2L(doCompensateDipole ? omega_dp : omega, fmm_handle.lattice_operator(), mu, p);

    // Dipole Compensation
    if (doCompensateDipole)
    {
        dipole_compensation.add_compensating_mu(fmm_handle, mu);
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
