/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SEQUENTIAL_pass4_hpp
#define FMSOLVR_SEQUENTIAL_pass4_hpp

#include "fmsolvr/operator/p4/local2local.hpp"
#include "fmsolvr/operator/p4/local2local_derivative.hpp"


namespace fmsolvr {

// this routine processes all particles
template <typename FMMHandle>
void pass4(FMMHandle & fmm_handle)
{
    typedef typename FMMHandle::real3_type Real3;
    typedef typename FMMHandle::real_type Real;
    typedef typename FMMHandle::size_type size_type;
    typedef typename FMMHandle::boxid_type BoxID;
    typedef typename FMMHandle::multipole_coefficients_type multipole_coefficients_type;
    typedef typename FMMHandle::local_coefficients_type local_coefficients_type;
    typedef typename local_coefficients_type::complex_type complex_type;

    auto & particle_handle = fmm_handle.particle_handle();
    const auto & tree_mu = fmm_handle.tree_mu();
    const size_type depth = fmm_handle.depth();
    const size_type p = fmm_handle.p();
    const auto & div_lut_p = fmm_handle.div_lut_p(p);
    const size_type num_boxes_lowest = particle_handle.num_boxes_lowest();
    const auto & ordered_particles = particle_handle.ordered_particles();
    const auto & box_particle_offset = particle_handle.box_particle_offset();
    auto result = particle_handle.potential_efield_ff_adapter();

    size_t ff_interactions = 0;
    //This is parallel for all particles
    //-> one thread per paticle each particle reads the taylor moment generated in PASS 1-3 of its box
    //the taylor moment will likely fit into shared memory
    local_coefficients_type dmux(p);
    local_coefficients_type dmuy(p);
    local_coefficients_type dmuz(p);
    multipole_coefficients_type omega(p);
    for (size_t id = 0; id < num_boxes_lowest; ++id)
    {
        if (box_particle_offset[id] == box_particle_offset[id + 1])
        {
            // no target particles in this box
            continue;
        }

        BoxID bid(id, depth);

        const auto & mu = tree_mu[bid];
        const size_t source_particles = mu.sourceCounters().particles();
        Real3 expansion_point_mu_ = particle_handle.simulationbox().expansion_point(bid);

        DxL2L_reference(mu, dmux, p);
        DyL2L_reference(mu, dmuy, p);
        DzL2L_reference(mu, dmuz, p);
        dmux.populate_lower();
        dmuy.populate_lower();
        dmuz.populate_lower();

        for (size_t i = box_particle_offset[id]; i < box_particle_offset[id + 1]; ++i)
        {
            Real3 xyz = Real3(ordered_particles[i]);
            Real q = -1.;  // Field computation only

            // far field Field
            // FIXME: excluding the sign, omega and O should be the same
            omega.zero(p);
            P2M(q, xyz, omega, expansion_point_mu_, div_lut_p);
            omega.populate_lower();

            // far field Potential
            L2L_Operator<local_coefficients_type> O(expansion_point_mu_, xyz, p);

            // FIXME: optimize multiplication: (re, 0) = (re, im) * (re, im)
            complex_type Phi_i(0.);
            complex_type Fx(0.);
            complex_type Fy(0.);
            complex_type Fz(0.);

            for (ssize_t l = 0; l <= ssize_t(p); ++l)
            {
                for (ssize_t m = -l; m <= l; ++m)
                {
                    Phi_i += mu.get(l, m) * O.get(l, m);
                    complex_type omega_lm = omega.get(l, m);
                    Fx += dmux.get(l, m) * omega_lm;
                    Fy += dmuy.get(l, m) * omega_lm;
                    Fz += dmuz.get(l, m) * omega_lm;
                }
            }

            // FIXME: verify imag() part is (reasonably) zero
            result.reduce_pf(i, Phi_i.real(), -Real3(Fx.real(), Fy.real(), Fz.real()));
            ff_interactions += source_particles;
        }
    }
    particle_handle.add_interactions_ff(ff_interactions);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
