/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SIMD_Vec_hpp
#define FMSOLVR_SIMD_Vec_hpp

#include "fmsolvr/util/float128.hpp"

#include "fmsolvr/simd/Vec_base.hpp"
#include "fmsolvr/simd/Vec_config.hpp"
#include "fmsolvr/simd/Vec_super.hpp"

#include "fmsolvr/simd/Vec_Array.hpp"
#include "fmsolvr/simd/Vec_m128.hpp"        // float[4], SSE
#include "fmsolvr/simd/Vec_m128d.hpp"		// double[2], SSE2
#include "fmsolvr/simd/Vec_m256.hpp"		// float[8], AVX
#include "fmsolvr/simd/Vec_m256d.hpp"		// double[4], AVX
#include "fmsolvr/simd/Vec_m512.hpp"		// float[16], AVX512
#include "fmsolvr/simd/Vec_m512d.hpp"		// double[8], AVX512
#include "fmsolvr/simd/Vec_wrap_m512.hpp"	// float[16], AVX512
#include "fmsolvr/simd/Vec_wrap_m512d.hpp"	// double[8], AVX512

namespace fmsolvr
{

template <typename Real>
struct simd_engine_none
{
    template <typename T>
    using engine = simd_engine_none<T>;
    typedef Real type;
};

template <typename Real>
struct simd_engine_array1
{
    template <typename T>
    using engine = simd_engine_array1<T>;
    typedef TrivialArray<Real, 1> type;
};

template <typename Real>
struct simd_engine_array2
{
    template <typename T>
    using engine = simd_engine_array2<T>;
    typedef TrivialArray<Real, 2> type;
};

template <typename Real>
struct simd_engine_array4
{
    template <typename T>
    using engine = simd_engine_array4<T>;
    typedef TrivialArray<Real, 4> type;
};

template <typename Real>
struct simd_engine_array8
{
    template <typename T>
    using engine = simd_engine_array8<T>;
    typedef TrivialArray<Real, 8> type;
};

#ifdef FMSOLVR_HAVE_AVX512F
template <typename Real>
struct simd_engine_avx512 : public simd_engine_none<Real>
{
};

template <>
struct simd_engine_avx512<float>
{
    template <typename T>
    using engine = simd_engine_avx512<T>;
#if defined(__ICC) && (__ICC < 1800)
    typedef wrap_m512 type;
#else
    typedef __m512 type;
#endif
};

template <>
struct simd_engine_avx512<double>
{
    template <typename T>
    using engine = simd_engine_avx512<T>;
#if defined(__ICC) && (__ICC < 1800)
    typedef wrap_m512d type;
#else
    typedef __m512d type;
#endif
};
#endif  // FMSOLVR_HAVE_AVX512F

#ifdef FMSOLVR_HAVE_AVX
template <typename Real>
struct simd_engine_avx : public simd_engine_none<Real>
{
};

template <>
struct simd_engine_avx<float>
{
    template <typename T>
    using engine = simd_engine_avx<T>;
    typedef __m256 type;
};

template <>
struct simd_engine_avx<double>
{
    template <typename T>
    using engine = simd_engine_avx<T>;
    typedef __m256d type;
};
#endif  // FMSOLVR_HAVE_AVX

#ifdef FMSOLVR_HAVE_SSE
template <typename Real>
struct simd_engine_sse : public simd_engine_none<Real>
{
};

template <>
struct simd_engine_sse<float>
{
    template <typename T>
    using engine = simd_engine_sse<T>;
    typedef __m128 type;
};
#endif  // FMSOLVR_HAVE_SSE

#ifdef FMSOLVR_HAVE_SSE2
template <>
struct simd_engine_sse<double>
{
    template <typename T>
    using engine = simd_engine_sse<T>;
    typedef __m128d type;
};
#endif  // FMSOLVR_HAVE_SSE2

template <typename Real>
struct simd_engine_default
{
#if 0
#elif defined(FMSOLVR_HAVE_AVX512F)
    template <typename T>
    using engine = simd_engine_avx512<T>;
#elif defined(FMSOLVR_HAVE_AVX)
    template <typename T>
    using engine = simd_engine_avx<T>;
#elif defined(FMSOLVR_HAVE_SSE)
    template <typename T>
    using engine = simd_engine_sse<T>;
#else
    template <typename T>
    using engine = simd_engine_none<T>;
#endif
    typedef typename engine<Real>::type type;
};

template <>
struct simd_engine_default<long double> : public simd_engine_none<long double>
{
};

#if defined(HAVE_FLOAT128)
template <>
struct simd_engine_default<__float128> : public simd_engine_none<__float128>
{
};
#endif

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
