/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SIMD_Vec_Array_hpp
#define FMSOLVR_SIMD_Vec_Array_hpp

#include "fmsolvr/simd/Vec_base.hpp"

namespace fmsolvr
{

template <typename Scalar, unsigned k>
class TrivialArray
{
public:
    TrivialArray() = default;

    TrivialArray(const Scalar &b)
    {
        for (size_t i = 0; i < k; ++i)
            a[i] = b;
    }

	Scalar & operator[] (size_t i)
	{
		return a[i];
	}

	const Scalar & operator[] (size_t i) const
	{
		return a[i];
	}

    TrivialArray operator-() const
    {
        TrivialArray c;
        for (size_t i = 0; i < k; ++i)
            c[i] = -a[i];
        return c;
    }

	TrivialArray operator + (const TrivialArray & b) const
	{
		TrivialArray c;
		for (size_t i = 0; i < k; ++i)
			c[i] = a[i] + b[i];
		return c;
	}

	TrivialArray operator - (const TrivialArray & b) const
	{
		TrivialArray c;
		for (size_t i = 0; i < k; ++i)
			c[i] = a[i] - b[i];
		return c;
	}

	TrivialArray operator * (const TrivialArray & b) const
	{
		TrivialArray c;
		for (size_t i = 0; i < k; ++i)
			c[i] = a[i] * b[i];
		return c;
	}

	TrivialArray & operator += (const TrivialArray & b)
	{
		for (size_t i = 0; i < k; ++i)
			a[i] += b[i];
		return *this;
	}

	TrivialArray & operator -= (const TrivialArray & b)
	{
		for (size_t i = 0; i < k; ++i)
			a[i] -= b[i];
		return *this;
	}

	TrivialArray & operator *= (const TrivialArray & b)
	{
		for (size_t i = 0; i < k; ++i)
			a[i] *= b[i];
		return *this;
	}

private:
    Scalar a[k];
};


template <typename Scalar, unsigned k>
inline TrivialArray<Scalar, k> operator*(const TrivialArray<Scalar, k> &a, const Scalar &b)
{
    TrivialArray<Scalar, k> c;
    for (size_t i = 0; i < k; ++i)
        c[i] = a[i] * b;
    return c;
}

template <typename Scalar, unsigned k>
inline TrivialArray<Scalar, k> operator*(const Scalar &a, const TrivialArray<Scalar, k> &b)
{
    TrivialArray<Scalar, k> c;
    for (size_t i = 0; i < k; ++i)
        c[i] = a * b[i];
    return c;
}


template <typename Scalar, unsigned k>
struct Vec_traits<TrivialArray<Scalar, k> > : public Vec_base<TrivialArray<Scalar, k>, Scalar>
{
	typedef Vec_base<TrivialArray<Scalar, k>, Scalar> base;
	typedef typename base::vec_type vec_type;
	typedef typename base::scalar_type scalar_type;

	vec_type zero() const
	{
		vec_type z;
		for (size_t i = 0; i < k; ++i)
			z[i] = 0;
		return z;
	}

	vec_type same(scalar_type s) const
	{
		vec_type v;
		for (size_t i = 0; i < k; ++i)
			v[i] = s;
		return v;
	}

	scalar_type sum(const vec_type & v) const
	{
		scalar_type s = 0;
		for (size_t i = 0; i < k; ++i)
			s += v[i];
		return s;
	}
};

template <typename Scalar, unsigned k>
TrivialArray<Scalar, k> rsqrt(const TrivialArray<Scalar, k> & a)
{
	TrivialArray<Scalar, k> r;
	for (size_t i = 0; i < k; ++i) {
		r[i] = rsqrt(a[i]);
	}
	return r;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
