/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SIMD_Vec_base_hpp
#define FMSOLVR_SIMD_Vec_base_hpp

#include "fmsolvr/math/rsqrt.hpp"

#include <cstddef>
#include <type_traits>

namespace fmsolvr
{

template <typename Vec, typename Scalar>
struct Vec_base
{
    typedef typename std::remove_const<Vec>::type vec_type;
    typedef typename std::remove_const<Scalar>::type scalar_type;

	enum { VecSize = sizeof(vec_type) / sizeof(scalar_type) };

    static vec_type load(const scalar_type *p)
    {
        return at(p);
    }

    static void store(scalar_type *p, const vec_type v)
    {
        at(p) = v;
    }

    static vec_type unaligned_load(const scalar_type *p)
	{
		return *p;
	}

    static vec_type from(const scalar_type *p)
	{
		return *(reinterpret_cast<const vec_type *>(p));
	}

    static const vec_type &at(const scalar_type *p)
	{
		return *(reinterpret_cast<const vec_type *>(p));
	}

    static vec_type &at(scalar_type *p)
	{
		return *(reinterpret_cast<vec_type *>(p));
	}

    static vec_type partial_load(const scalar_type *p, size_t k)
    {
        vec_type v = vec_type();
        for (size_t i = 0; i < k; ++i)
        {
            element(v, i) = p[i];
        }
        return v;
    }

    static void partial_store(scalar_type *p, const vec_type &v, size_t k)
    {
        for (size_t i = 0; i < k; ++i)
        {
            p[i] = element(v, i);
        }
    }

    static const scalar_type &element(const vec_type &v, size_t i)
    {
        return reinterpret_cast<const scalar_type *>(&v)[i];
    }

    static scalar_type &element(vec_type &v, size_t i)
    {
        return reinterpret_cast<scalar_type *>(&v)[i];
    }

	size_t index_floor(size_t x) const
	{
		size_t skewed = x % VecSize;
		return x - skewed;
	}

	size_t index_ceil(size_t x) const
	{
		size_t skewed = x % VecSize;
		return x + ((skewed == 0) ? (0) : (VecSize - skewed));
	}

};

template <typename Vec>
struct Vec_traits : public Vec_base<Vec, Vec>
{
	typedef Vec_base<Vec, Vec> base;
	typedef typename base::vec_type vec_type;
	typedef typename base::scalar_type scalar_type;

	vec_type zero() const
	{
		return 0;
	}

	vec_type same(scalar_type s) const
	{
		return s;
	}

	scalar_type sum(const vec_type & v) const
	{
		return v;
	}
};

template <typename SIMD>
struct SIMDtraits : public Vec_traits<SIMD>
{
    enum
    {
        SIMDwidth = Vec_traits<SIMD>::VecSize
    };
};

template <typename Vec>
Vec rcplength(const Vec & dx, const Vec & dy, const Vec & dz)
{
	Vec ssq = dx * dx + dy * dy + dz * dz;
	return rsqrt(ssq);
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
