/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SIMD_Vec_m256_hpp
#define FMSOLVR_SIMD_Vec_m256_hpp

// __m256 = float[8] (AVX)

#include "fmsolvr/simd/Vec_config.hpp"

#ifdef FMSOLVR_HAVE_AVX

#include "fmsolvr/aligned_alloc.hpp"
#include "fmsolvr/simd/Vec_base.hpp"
#include <immintrin.h>

namespace fmsolvr
{

template <>
struct Vec_traits<__m256> : public Vec_base<__m256, float>
{
	typedef Vec_base<__m256, float> base;
	typedef typename base::vec_type vec_type;
	typedef typename base::scalar_type scalar_type;
	typedef aligned_allocator<vec_type> allocator_type;

	vec_type zero() const
	{
		vec_type z = {0., 0., 0., 0., 0., 0., 0., 0.};
		return z;
	}

	vec_type same(scalar_type s) const
	{
		vec_type v = {s, s, s, s, s, s, s, s};
		return v;
	}

	scalar_type sum(const vec_type & v) const
	{
		const scalar_type * p = reinterpret_cast<const scalar_type *>(&v);
		scalar_type s = 0.;
		for (size_t i = 0; i < VecSize; ++i)
			s += p[i];
		return s;
	}
};

template <>
inline
__m256 rsqrt_exact(__m256 a)
{
	__m256 sqr = _mm256_sqrt_ps(a);
	__m256 ones = {1., 1., 1., 1., 1., 1., 1., 1.};
	return ones / sqr;
}

inline __m256 rsqrt_estimate_11bits(__m256 a)
{
	return _mm256_rsqrt_ps(a);
}

inline __m256 rsqrt_estimate_22bits(__m256 a)
{
	__m256 ymm0;
	__m256 ymm2;
	__m256 ymm3 = {3.f, 3.f, 3.f, 3.f, 3.f, 3.f, 3.f, 3.f};
	__m256 ymm4 = {0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f};
	__m256 ymm5;
	ymm5 = a;
	ymm0 = _mm256_rsqrt_ps(ymm5);
	ymm2 = ymm0 * ymm0;
	ymm2 = ymm2 * ymm5;
	ymm2 = ymm3 - ymm2;
	ymm0 = ymm0 * ymm2;
	ymm0 = ymm0 * ymm4;
	return ymm0;
}

inline __m256 rsqrt_estimate_23bits(__m256 a)
{
	__m256 x = a;
	__m256 x0, x1, x2, r;
	__m256 three = {3.f, 3.f, 3.f, 3.f, 3.f, 3.f, 3.f, 3.f};
	__m256 half = {0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f};
	x0 = _mm256_rsqrt_ps(x);
	r = x0 * x0;
	r = r * x;
	r = three - r;
	r = r * x0;
	x1 = r * half;
	r = x1 * x1;
	r = r * x;
	r = three - r;
	r = r * x1;
	x2 = r * half;
	return x2;
}

template <>
inline
__m256 rsqrt(__m256 a)
{
#if 0
	return rsqrt_estimate_11bits(a);
#elif 0
	return rsqrt_estimate_22bits(a);
#elif 0
	return rsqrt_estimate_23bits(a);
#else
	return rsqrt_exact(a);
#endif
}

}  // namespace fmsolvr

#endif  // FMSOLVR_HAVE_AVX

#endif
