/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SIMD_Vec_m256d_hpp
#define FMSOLVR_SIMD_Vec_m256d_hpp

// __m256d = double[4] (AVX)

#include "fmsolvr/simd/Vec_config.hpp"

#ifdef FMSOLVR_HAVE_AVX

#include "fmsolvr/aligned_alloc.hpp"
#include "fmsolvr/simd/Vec_base.hpp"
#include <immintrin.h>

namespace fmsolvr
{

template <>
struct Vec_traits<__m256d> : public Vec_base<__m256d, double>
{
	typedef Vec_base<__m256d, double> base;
	typedef typename base::vec_type vec_type;
	typedef typename base::scalar_type scalar_type;
	typedef aligned_allocator<vec_type> allocator_type;

	vec_type unaligned_load(const scalar_type * p) const
	{
		return _mm256_loadu_pd(p);
	}

	vec_type zero() const
	{
		vec_type z = {0., 0., 0., 0.};
		return z;
	}

	vec_type same(scalar_type s) const
	{
		vec_type v = {s, s, s, s};
		return v;
	}

	scalar_type sum(const vec_type & v) const
	{
		const scalar_type * p = reinterpret_cast<const scalar_type *>(&v);
		scalar_type s = 0.;
		for (size_t i = 0; i < VecSize; ++i)
			s += p[i];
		return s;
	}


	// vvvv wwww => {0,0,(w+w+w+w),(v+v+v+v)}
	vec_type sumpair(const vec_type & v, const vec_type & w) const
	{
		typedef __m128d half_vec_type;
		vec_type vhadd = _mm256_hadd_pd(v, w);
		half_vec_type upper = _mm256_extractf128_pd(vhadd, 1);
		half_vec_type lower = _mm256_castpd256_pd128(vhadd);
		half_vec_type sum = upper + lower;
		return _mm256_castpd128_pd256(sum);
	}

	// vvvv wwww => {(w+w),(v+v),(w+w),(v+v)}
	vec_type interleavedsumpair(const vec_type & v, const vec_type & w) const
	{
		return _mm256_hadd_pd(v, w);
	}
};

template <>
inline
__m256d rsqrt_exact(__m256d a)
{
	__m256d sqr = _mm256_sqrt_pd(a);
	__m256d ones = {1., 1., 1., 1.};
	return ones / sqr;
}

}  // namespace fmsolvr

#endif  // FMSOLVR_HAVE_AVX

#endif
