/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SIMD_Vec_Super_hpp
#define FMSOLVR_SIMD_Vec_Super_hpp

#include "fmsolvr/simd/Vec_base.hpp"

namespace fmsolvr
{

template <typename SIMD, unsigned k>
class SuperSIMD
{
	typedef Vec_traits<SIMD> simd_traits;

public:
	typedef typename simd_traits::scalar_type Scalar;
	enum { SIMD_width = simd_traits::VecSize };

    SuperSIMD() = default;

    explicit SuperSIMD(const Scalar &b)
    {
        SIMD b_simd = simd_traits().same(b);
        for (size_t i = 0; i < k; ++i)
            a[i] = b_simd;
    }

	Scalar & operator[] (size_t i)
	{
		size_t simd_elem = i / SIMD_width;
		size_t scalar_elem = i % SIMD_width;
        simd_traits ST;
        return ST.element(a[simd_elem], scalar_elem);
	}

	const Scalar & operator[] (size_t i) const
	{
		size_t simd_elem = i / SIMD_width;
		size_t scalar_elem = i % SIMD_width;
        simd_traits ST;
        return ST.element(a[simd_elem], scalar_elem);
	}

	SuperSIMD operator - () const
	{
		SuperSIMD c;
		for (size_t i = 0; i < k; ++i)
			c.a[i] = -a[i];
		return c;
	}

	SuperSIMD operator + (const SuperSIMD & b) const
	{
		SuperSIMD c;
		for (size_t i = 0; i < k; ++i)
			c.a[i] = a[i] + b.a[i];
		return c;
	}

	SuperSIMD operator - (const SuperSIMD & b) const
	{
		SuperSIMD c;
		for (size_t i = 0; i < k; ++i)
			c.a[i] = a[i] - b.a[i];
		return c;
	}

	SuperSIMD operator * (const SuperSIMD & b) const
	{
		SuperSIMD c;
		for (size_t i = 0; i < k; ++i)
			c.a[i] = a[i] * b.a[i];
		return c;
	}

	SuperSIMD operator / (const SuperSIMD & b) const
	{
		SuperSIMD c;
		for (size_t i = 0; i < k; ++i)
			c.a[i] = a[i] / b.a[i];
		return c;
	}

	SuperSIMD & operator += (const SuperSIMD & b)
	{
		for (size_t i = 0; i < k; ++i)
			a[i] += b.a[i];
		return *this;
	}

	SuperSIMD & operator -= (const SuperSIMD & b)
	{
		for (size_t i = 0; i < k; ++i)
			a[i] -= b.a[i];
		return *this;
	}

	SuperSIMD & operator *= (const SuperSIMD & b)
	{
		for (size_t i = 0; i < k; ++i)
			a[i] *= b.a[i];
		return *this;
	}

    SIMD a[k];
};


template <typename SIMD, unsigned k>
inline SuperSIMD<SIMD, k> operator*(const SuperSIMD<SIMD, k> &a, const typename SuperSIMD<SIMD, k>::Scalar &b)
{
    return a * SuperSIMD<SIMD, k>(b);
}

template <typename SIMD, unsigned k>
inline SuperSIMD<SIMD, k> operator*(const typename SuperSIMD<SIMD, k>::Scalar &a, const SuperSIMD<SIMD, k> &b)
{
    return SuperSIMD<SIMD, k>(a) * b;
}


template <typename SIMD, unsigned k>
struct Vec_traits<SuperSIMD<SIMD, k> > : public Vec_base<SuperSIMD<SIMD, k>, typename Vec_traits<SIMD>::scalar_type>
{
	typedef Vec_base<SuperSIMD<SIMD, k>, typename Vec_traits<SIMD>::scalar_type> base;
	typedef typename base::vec_type vec_type;
	typedef typename base::scalar_type scalar_type;
	typedef SIMD nested_simd_type;
	typedef Vec_traits<nested_simd_type> nested_simd_traits;

	vec_type unaligned_load(const scalar_type * p) const
	{
		nested_simd_traits nested;
		vec_type z;
		for (size_t i = 0; i < k; ++i, p += nested.VecSize)
			z.a[i] = nested.unaligned_load(p);
		return z;
	}

	vec_type zero() const
	{
		nested_simd_traits nested;
		vec_type z;
		for (size_t i = 0; i < k; ++i)
			z.a[i] = nested.zero();
		return z;
	}

	vec_type same(scalar_type s) const
	{
		nested_simd_traits nested;
		vec_type v;
		for (size_t i = 0; i < k; ++i)
			v.a[i] = nested.same(s);
		return v;
	}

	scalar_type sum(const vec_type & v) const
	{
		nested_simd_traits nested;
		scalar_type s = 0;
		for (size_t i = 0; i < k; ++i)
			s += nested.sum(v.a[i]);
		return s;
	}

	nested_simd_type supersum(const vec_type & v) const
	{
		nested_simd_type s = v.a[0];
		for (size_t i = 1; i < k; ++i)
			s += v.a[i];
		return s;
	}
};

template <typename SIMD, unsigned k>
SuperSIMD<SIMD, k> rsqrt(const SuperSIMD<SIMD, k> & a)
{
	SuperSIMD<SIMD, k> r;
	for (size_t i = 0; i < k; ++i) {
		r.a[i] = rsqrt(a.a[i]);
	}
	return r;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
