#ifndef _AnBe_FMM_pass2_hpp_
#define _AnBe_FMM_pass2_hpp_

#include "fmsolvr/operator/p3/multipole2local_p3.hpp"
#include "fmsolvr/operator/p4/multipole2local.hpp"
#include "fmsolvr/util/envtools.hpp"

#include <cassert>
#include <tuple>
#include <queue>

namespace fmsolvr {

namespace {

// remap periodic index i into [0,..,dim)
template <typename Integral>
inline Integral periodic_remap(Integral i, Integral dim)
{
    while (i < 0)
        i += dim;
    while (i >= dim)
        i -= dim;
    return i;
}

}  // anonymous namespace

/**
 * Computes M2L for one source box.
 * We use mutex locks on target boxes and set the dependency counters of source and target taylor
 * moments.
 */
template <typename FMMHandle, typename TaskFactory, typename MutexContainer, typename DependencyManager>
__attribute__ ((flatten))
void  pass2_M2LSource(
        FMMHandle &fmm_handle,
        size_t id,
        size_t d,
        TaskFactory &task_factory,
        MutexContainer &mutex_container,
        DependencyManager &dm)
{
    const bool p4 = get_env_int("P4", 0);
    typedef typename FMMHandle::real_type Real;
    typedef typename FMMHandle::size_type size_type;
    typedef typename FMMHandle::index_type index_type;
    typedef typename FMMHandle::boxid_type BoxID;
    typedef typename FMMHandle::operatorid_m2l_type OperatorID_M2L;
    typedef typename OperatorID_M2L::ijk_type IJK;
    typedef typename FMMHandle::allocator_type allocator_type;
    typedef typename FMMHandle::coefficient_type coefficient_type;
    typedef typename FMMHandle::RotationDataTypes::Rot_Operator RotationOperator;
    typedef typename FMMHandle::RotationDataTypes::M2L_Operator M2Lp3_Operator;
    typedef typename FMMHandle::RotationDataTypes::template ScratchTypeGenerator<coefficient_type>::Scratch scratch_type;

    const auto &simulationbox = fmm_handle.simulationbox();
    const auto & tree_omega = fmm_handle.tree_omega();
    auto &GM2L = fmm_handle.RotationGeneratorM2L();
    const size_type ws = fmm_handle.ws();
    const size_type p = fmm_handle.p();
    const auto & m2l_operator = fmm_handle.m2l_operator_cube(d, ws);
    scratch_type scratch(p);

    auto & tree_ow_ = mutex_container.tree_ow_;
    index_type dim = index_type(1) << d;
    {
        {
            {
                BoxID source_bid(id, d);
                const auto i = source_bid.i();
                const auto j = source_bid.j();
                const auto k = source_bid.k();

                const auto &omega = tree_omega[source_bid];
                IJK op_omega_ijk(i, j, k);  // for addressing the operator

                // parent box coordinates
                index_type ip = i / 2;
                index_type jp = j / 2;
                index_type kp = k / 2;
                // outer ws shells
                index_type io_beg = (ip - ws) * 2;
                index_type io_end = (ip + ws + 1) * 2;
                index_type jo_beg = (jp - ws) * 2;
                index_type jo_end = (jp + ws + 1) * 2;
                index_type ko_beg = (kp - ws) * 2;
                index_type ko_end = (kp + ws + 1) * 2;
                // inner - no far field interactions
                index_type ii_beg = i - ws;
                index_type ii_end = i + ws + 1;
                index_type ji_beg = j - ws;
                index_type ji_end = j + ws + 1;
                index_type ki_beg = k - ws;
                index_type ki_end = k + ws + 1;


                for (index_type ii = io_beg; ii < io_end; ++ii)
                {
                    for (index_type jj = jo_beg; jj < jo_end; ++jj)
                    {
                        for (index_type kk = ko_beg; kk < ko_end; ++kk)
                        {
                            if (ii < ii_beg || ii >= ii_end
                                    || jj < ji_beg || jj >= ji_end
                                    || kk < ki_beg || kk >= ki_end) {
                                // FIXME: missing support for 1D/2D periodicity
                                if (simulationbox.open_boundary_conditions() && (ii < 0 || ii >= dim || jj < 0 || jj >= dim || kk < 0 || kk >= dim))
                                {

                                    /**
                                     * If the box is not inside the simulationbox and we have open
                                     * boundary conditions we have one less M2L for the source box
                                     * as a target.
                                     */
                                    dm.ResolvePartialMu(source_bid, task_factory, tree_ow_);
                                    continue;

                                }


                                IJK op_mu_ijk(ii, jj, kk);  // for addressing the operator, no wrapping!
                                OperatorID_M2L op_ijk(op_omega_ijk, op_mu_ijk, d);

                                // FIXME: skip if out of range and not periodic in that direction
                                index_type iio = periodic_remap(ii, dim);
                                index_type jjo = periodic_remap(jj, dim);
                                index_type kko = periodic_remap(kk, dim);
                                assert(0 <= iio && iio < dim && 0 <= jjo && jjo < dim && 0 <= kko && kko < dim);
                                BoxID target_bid(iio, jjo, kko, d);
                                auto &ow = tree_ow_[target_bid];


                                // TODO: When omega is empty it would be faster to resolve
                                // dependencies separately and avoid 3 nested for loops here. We
                                // need a fast method for iterating M2L interaction list
                                if (omega.empty())
                                {
                                    dm.ResolvePartialMu(target_bid, task_factory, tree_ow_);
                                    continue;
                                }
                                /**
                                 * If we are not able to aquire the lock here we will put this
                                 * computation aside and compute it later.
                                 */
                                {
                                auto reservation = ow.AcquireAccess();
                                auto & tmp_mu = reservation.Obj();

                                if (p4) {
                                    M2L_reference(omega, m2l_operator(op_ijk), tmp_mu, p);
                                } else {
                                    RotationOperator R =
                                        GM2L.generate_rotation_operator(op_ijk);
                                    Real shift_distance = simulationbox.box_center_distance(op_ijk);
                                    std::shared_ptr<M2Lp3_Operator> B_p3 =
                                        std::allocate_shared<M2Lp3_Operator>(allocator_type(), shift_distance, p);
                                    M2Lp3<FMMHandle::template SIMDengine>(omega, R, *B_p3, tmp_mu, p, scratch);
                                }


                                tmp_mu.addSourceCounters(omega.sourceCounters());
                                /**
                                 * Comment corresponding the populate_lower()
                                 *
                                 * This is only relevant for UpperLower triangular array.
                                 *
                                 * For the sequential version (WIP/tree) only one populate lower
                                 * after L2L is required. In parallel we do not know if L2L or M2L
                                 * is the last operation on the local moment. Therefor we need to
                                 * populate lower after each operation. Another argument for
                                 * populate lower after each operation is, that the local moment has
                                 * a valid lower after each operation. This means we do not have any
                                 * assumptions on the operators. Another possibility would be to
                                 * populate lower local moments before we use them in the next L2L
                                 * or L2P operation. The disadvantage is, that we need to lock this
                                 * step, because we do not know if we are the only thread using the
                                 * moment. Additional we would need to load the mu non const to
                                 * populate lower and one would not expect an operator to modify the
                                 * input data. It would also be possible to add another task for
                                 * populating lower, but this seems to be a lot of additional work
                                 * for UpperLower version.
                                 */
                                tmp_mu.populate_lower();
                                } //ow.Oap().WithdrawObjectAccess(reservation);
                                dm.ResolvePartialMu(
                                        target_bid, task_factory, tree_ow_);
                            }
                        }
                    }
                }
            }
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
