#ifndef _AnBe_FMM_pass5_hpp_
#define _AnBe_FMM_pass5_hpp_

#include <algorithm>
#include <cassert>

namespace fmsolvr {

template <typename FMMHandle, typename MutexContainer, typename DependencyManager>
void pass5_P2P(FMMHandle &fmm_handle, typename FMMHandle::tree_partition_type tp, MutexContainer &mutex_container, DependencyManager & dm)
{
    typedef typename FMMHandle::real3_type Real3;
    typedef typename FMMHandle::size_type size_type;
    typedef typename FMMHandle::index_type index_type;
    typedef typename FMMHandle::boxid_type BoxID;
    typedef typename FMMHandle::ParticleHandle::inputadapter_type inputadapter_type;

    auto & particle_handle = fmm_handle.particle_handle();
    inputadapter_type particles = particle_handle.particle_adapter();
    auto result = particle_handle.potential_efield_nf_adapter();


    typedef typename FMMHandle::P2P_type P2P_type;
    P2P_type P2P(result, particles, particles);

    const size_type depth = fmm_handle.depth();
    const index_type ws = fmm_handle.ws();
    const auto & box_particle_offset = particle_handle.box_particle_offset();

    size_t nf_interactions = 0;
    ssize_t dim = ssize_t(1) << depth;
    {
        {
            for (auto box : tp.box_range())
            {
                //map 3D box id to 1D box id
                size_t idt = box.sfcindex();
                auto i = box.i();
                auto j = box.j();
                auto k = box.k();
                nf_interactions = 0;
                auto &mutex = mutex_container.lowest_lvl_particle_mutexes_[box.sfcindex()];
                //particles of one box are stored in a contigous sub array [target_beg:target_end)
                size_t target_beg = box_particle_offset[idt];
                size_t target_end = box_particle_offset[idt + 1];
                //Call all to all Kernel interactions within each box - overlapping interactions
                //here we need to take care that a particle does not interacte with itselfe
                //only output values are changed so this call has (target_end-target_beg) parallelism if we avoid atomics
                //and (target_end-target_beg)**2 if we use atomics
                //intermediate solutions with multiple output arrays are also possible
                mutex.lock();
                P2P.overlapping_interactions(target_beg, target_end, target_beg, target_end);
                nf_interactions += (target_end - target_beg) * (target_end - target_beg - 1);
                //loop over neighbour hood (parameter ws defines size of neighbour hood)
                //First guess: the follwoing loop nest should be propably in a CUDA kernel (safe multiple writes of target particle)
                //This would result in a single kernel with dim*dim*dim*(target_end-target_beg) threads
                //size within each box can vary but we are targeting homegenous systems so there should no or only few empty boxes.
                //We should aim for k*32 particles per box with k >= 1
                for (ssize_t ii = i - ws; ii < i + ws + 1; ++ii) {
                    for (ssize_t jj = j - ws; jj < j + ws + 1; ++jj) {
                        for (ssize_t kk = k - ws; kk < k + ws + 1; ++kk) {
                            //box should not interact with itselfe -> already handled with P2P.overlapping_interactions
                            if (ii != i || jj != j || kk != k) {
                                //Begin handling boundaries: open_boundary_conditions -> boxes on the boundary do not have all neighbours
                                //periodic boundary condition: we need to wrap around
                                // FIXME: missing support for 1D/2D periodicity
                                if (particle_handle.simulationbox().open_boundary_conditions() &&
                                        (ii < 0 || ii >= dim || jj < 0 || jj >= dim || kk < 0 || kk >= dim))
                                {
                                        continue;
                                }
                                const auto & abc = particle_handle.simulationbox().abc();
                                Real3 periodic_fold_shift(0., 0., 0.);
                                ssize_t iif = ii;
                                if (iif < 0) {
                                    iif += dim;
                                    periodic_fold_shift -= abc.a();
                                } else if (iif >= dim) {
                                    iif -= dim;
                                    periodic_fold_shift += abc.a();
                                }
                                ssize_t jjf = jj;
                                if (jjf < 0) {
                                    jjf += dim;
                                    periodic_fold_shift -= abc.b();
                                } else if (jjf >= dim) {
                                    jjf -= dim;
                                    periodic_fold_shift += abc.b();
                                }
                                ssize_t kkf = kk;
                                if (kkf < 0) {
                                    kkf += dim;
                                    periodic_fold_shift -= abc.c();
                                } else if (kkf >= dim) {
                                    kkf -= dim;
                                    periodic_fold_shift += abc.c();
                                }
                                assert(0 <= iif && iif < dim && 0 <= jjf && jjf < dim && 0 <= kkf && kkf < dim);
                                //End handling boundaries
                                size_t ids = BoxID(iif, jjf, kkf, depth).sfcindex();
                                size_t source_beg = box_particle_offset[ids];
                                size_t source_end = box_particle_offset[ids + 1];
                                //target particles are the particles for which potential and field is calculated
                                //this process owns the target particles. source particles are the particles we are interacting with
                                //source particles might be owned by another process
                                //Same parallelism than P2P.overlapping_interactions if we allow atomics or have multiple output arrays
                                //which are later reduced P2P.overlapping_interactions and all or a subset P2P.offsetted_interactions can be handled
                                //in parallel.
                                //Possibly more efficient is to start one thread per target particle and handle P2P.overlapping_interactions and all P2P.offsetted_interactions
                                //serially in one kernel
                                // These P2P.offsetted_interactions() calls include all box2box interactions
                                // (with periodic_fold_shift == (0, 0, 0)) that could also be done with P2P.disjoint_interactions().
                                P2P.offsetted_interactions(target_beg, target_end, source_beg, source_end, -periodic_fold_shift);
                                nf_interactions += (target_end - target_beg) * (source_end - source_beg);
                            }
                        }
                    }
                }
                mutex.unlock();
                fmm_handle.particle_handle().add_interactions_nf(nf_interactions);
                dm.dc().force_computation_counter -= target_end - target_beg;
            }
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
