/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_TREE_boxid_4d_hpp
#define FMSOLVR_TREE_boxid_4d_hpp

#include "fmsolvr/tree/boxid.hpp"

#include <algorithm>
#include <limits>
#include <ostream>
#include <type_traits>


namespace fmsolvr {

template <typename Integral, typename SfcIntegral = size_t>
class BoxID4D {
    typedef BoxID<BoxID4D<Integral, SfcIntegral>> Self;

public:
    typedef Integral depth_type;
    typedef Integral dimension_type;
    typedef typename std::make_signed<Integral>::type signed_dimension_type;
    typedef SfcIntegral sfcindex_type;

    BoxID4D(dimension_type i, dimension_type j, dimension_type k, depth_type d) :
        i_(i), j_(j), k_(k), d_(d)
    { }

    BoxID4D(sfcindex_type sfcidx, depth_type d) :
        d_(d)
    {
        i_ = sfcidx & mask();
        j_ = (sfcidx >> d_) & mask();
        k_ = (sfcidx >> (d_ * 2)) & mask();
    }

    static constexpr size_t max_depth()
    {
        return std::min(
                std::numeric_limits<sfcindex_type>::digits / 3,
                std::numeric_limits<dimension_type>::digits
                );
    }

    dimension_type i() const
    { return i_; }

    dimension_type j() const
    { return j_; }

    dimension_type k() const
    { return k_; }

    depth_type d() const
    { return d_; }

    sfcindex_type sfcindex() const
    {
        return (k_ << (d_ * 2)) + (j_ << d_) + i_;
    }
    bool operator==(const Self &box_r) const
    {
        return d_ == box_r.d_ && i_ == box_r.i_ && j_ == box_r.j_ && k_ == box_r.k_;
    }
    bool operator!=(const Self &box_r) const
    {
        return !((*this) == box_r);
    }

    Self parent() const
    {
        return Self(i_ >> 1, j_ >> 1, k_ >> 1, d_ - 1);
    }

    Self child(dimension_type i, dimension_type j, dimension_type k) const
    {
        return Self((i_ << 1) + i, (j_ << 1) + j, (k_ << 1) + k, d_ + 1);
    }

    Self
    descendant(dimension_type i, dimension_type j, dimension_type k, depth_type delta_depth) const
    {
        return Self(
                (i_ << delta_depth) + i,
                (j_ << delta_depth) + j,
                (k_ << delta_depth) + k,
                d_ + delta_depth);
    }

private:
    dimension_type i_, j_, k_;
    depth_type d_;

    dimension_type mask() const
    {
        return (dimension_type(1) << d_) - 1;
    }
};

template <typename Integral, typename SfcIntegral = size_t>
std::ostream & operator << (std::ostream & os, const BoxID4D<Integral, SfcIntegral> & bid)
{
    os << "(" << bid.d() << ": " << bid.i() << ", " << bid.j() << ", " << bid.k() << ")";
    return os;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
