/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_TREE_denseoctree_hpp
#define FMSOLVR_TREE_denseoctree_hpp

#include <memory>

#include "fmsolvr/tree/boxid.hpp"


namespace fmsolvr {

template <typename Container, typename Allocator = std::allocator<Container *> >
class OctreeStorageDense {
public:
    typedef OctreeStorageDense Self;
    typedef Container value_type;
    typedef typename value_type::size_type size_type;

    class levelref
    {
        Self * ots_;
        size_type d_;
    public:
        levelref(Self * ots, size_type d) : ots_(ots), d_(d)
        { }

        const value_type & operator [] (size_type i) const
        {
            return ots_->element(d_, i);
        }

        value_type & operator [] (size_type i)
        {
            return ots_->element(d_, i);
        }
    };

    class levelrange
    {
        value_type ** ptr_;
        size_type level_size_;
    public:
        levelrange(value_type ** ptr, size_type level_size) : ptr_(ptr), level_size_(level_size)
        {
        }

        class levelrange_iterator
        {
            value_type **ptr_;

        public:
            levelrange_iterator(value_type **ptr) : ptr_(ptr)
            {
            }
            // TODO const increment
            levelrange_iterator &operator++()
            {
                ++ptr_;
                return *this;
            }
            //TODO implement operator+ for random access

            bool operator==(const levelrange_iterator &other) const
            {
                return other.ptr_ == ptr_;
            }
            bool operator!=(const levelrange_iterator &other) const
            {
                return !operator==(other);
            }
            value_type & operator*()
            {
                return **ptr_;
            }
            const value_type & operator*() const
            {
                return **ptr_;
            }
        };

        levelrange_iterator begin()
        {
            return levelrange_iterator(ptr_);
        }
        const levelrange_iterator begin() const
        {
            return levelrange_iterator(ptr_);
        }

        levelrange_iterator end()
        {
            return levelrange_iterator(ptr_ + level_size_);
        }
        const levelrange_iterator end() const
        {
            return levelrange_iterator(ptr_ + level_size_);
        }
    };

    levelrange LevelRange(size_type d)
    {
        return levelrange(ptr_ + boxes_above_depth(d), boxes_on_depth(d));
    }

    const levelrange LevelRange(size_type d) const
    {
        return levelrange(ptr_ + boxes_above_depth(d), boxes_on_depth(d));
    }

private:
    static size_type boxes_on_depth(size_type d)
    {
        // 8^d
        return size_type(1) << (3 * d);
    }

    static size_type boxes_above_depth(size_type d)
    {
        // sum(0, d-1, 8^i)
        return ((size_type(1) << (3 * d)) - 1 ) / 7;
    }

public:

    template <typename... Args>
    OctreeStorageDense(size_type d, Args... args);

    template <typename... CoupledTreeTypes, typename... Args>
    OctreeStorageDense(size_type d, OctreeStorageDense<CoupledTreeTypes...> &coupled_tree, Args... args);


    ~OctreeStorageDense();

    size_type d() const
    { return d_; }

private:
    size_type lin_index(size_type d, size_type i) const
    {
        return boxes_above_depth(d) + i;
    }

    template <typename BoxIDImpl>
    size_type lin_index(const BoxID<BoxIDImpl> &bid) const
    {
        return lin_index(bid.d(), bid.sfcindex());
    }
public:
    const value_type & lin_element(size_type i) const
    {
        return *ptr_[i];
    }

    value_type & lin_element(size_type i)
    {
        return *ptr_[i];
    }
private:
    const value_type & element(size_type d, size_type i) const
    {
        return lin_element(lin_index(d, i));
    }

    value_type & element(size_type d, size_type i)
    {
        return lin_element(lin_index(d, i));
    }

    const levelref & operator [] (size_type d) const
    {
        return levelref(this, d);
    }

    levelref operator [] (size_type d)
    {
        return levelref(this, d);
    }

public:
    template <typename BoxIDImpl>
    const value_type & operator [] (const BoxID<BoxIDImpl> & bid) const
    {
        return lin_element(lin_index(bid));
    }

    template <typename BoxIDImpl>
    value_type & operator [] (const BoxID<BoxIDImpl> & bid)
    {
        return lin_element(lin_index(bid));
    }

private:
    size_type d_;
    value_type **ptr_;
};

template <typename Container, typename Allocator>
template <typename... Args>
OctreeStorageDense<Container, Allocator>::OctreeStorageDense(size_type d, Args... args)
    : d_(d)
{
    size_type n = boxes_above_depth(d_ + 1);
    ptr_ = new value_type*[n];
    for (size_type i = 0; i < n; ++i)
    {
        ptr_[i] = nullptr;
        ptr_[i] = new value_type(args...);
    }
}

template <typename Container, typename Allocator>
template <typename... CoupledTreeTypes, typename... Args>
OctreeStorageDense<Container, Allocator>::OctreeStorageDense(
        size_type d, OctreeStorageDense<CoupledTreeTypes...> & coupled_tree, Args... args)
        : d_(d)
{
    size_type n = boxes_above_depth(d_ + 1);
    ptr_ = new value_type*[n];
    for (size_type i = 0; i < n; ++i)
    {
        ptr_[i] = nullptr;
        ptr_[i] = new value_type(coupled_tree.lin_element(i), args...);
    }
}



template <typename Container, typename Allocator>
OctreeStorageDense<Container, Allocator>::~OctreeStorageDense()
{
    size_type n = boxes_above_depth(d_ + 1);
    for (size_type i = 0; i < n; ++i)
        delete ptr_[i];
    delete[] ptr_;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
