/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_TREE_simulationbox_hpp
#define FMSOLVR_TREE_simulationbox_hpp

#include "fmsolvr/tree/boxexpansionpoint.hpp"

namespace fmsolvr {

// a SimulationBox lives in a global orthonormal coordinate system
// and may have an arbitrary triclinic shape with an arbitrary center
template <typename Real33_>
class SimulationBox
{
public:
    typedef Real33_ Real33;
    typedef typename Real33::value_type Real3;
    typedef typename Real3::value_type Real;

    SimulationBox(const Real33 & abc, const Real3 & center, bool periodic_a, bool periodic_b, bool periodic_c)
        : abc_(abc), center_(center), periodic_a_(periodic_a), periodic_b_(periodic_b), periodic_c_(periodic_c)
    { }

    const Real33 & abc() const
    { return abc_; }

    const Real3 & center() const
    { return center_; }

    const Real3 reference_corner() const
    {
        return center_ - abc_.center();
    }

    template <typename BoxID>
    Real3 expansion_point(const BoxID & bid) const
    {
        size_t d = bid.d();
        return center_ + box_expansion_point(abc_, d, size_t(1) << d, bid.i(), bid.j(), bid.k());
    }

    template <typename OpID>
    Real box_center_distance(const OpID &opID) const
    {
        auto ijk = opID.relative_to_source();
        Real3 ortho_distance = Real3(ijk.i(), ijk.j(), ijk.k()) * reciprocal<Real>(size_t(1) << opID.d());
        return length(abc_ * ortho_distance);
    }

    bool periodic_a() const
    {
        return periodic_a_;
    }

    bool periodic_b() const
    {
        return periodic_b_;
    }

    bool periodic_c() const
    {
        return periodic_c_;
    }

    bool open_boundary_conditions() const
    {
        return !(periodic_a_ || periodic_b_ || periodic_c_);
    }

    bool operator ==(const SimulationBox &sb) const
    {
        return abc_ == sb.abc_
               && center_ == sb.center_
               && periodic_a_ == sb.periodic_a_ && periodic_b_ == sb.periodic_b_ && periodic_c_ == sb.periodic_c_;
    }

    bool operator !=(const SimulationBox &sb) const
    {
        return !(*this == sb);
    }

private:
    Real33 abc_;
    Real3 center_;
    bool periodic_a_, periodic_b_, periodic_c_;
};

} // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
