/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_UTIL_TimeLine_hpp
#define FMSOLVR_UTIL_TimeLine_hpp

#include "fmsolvr/util/envtools.hpp"
#include "fmsolvr/util/wrpapi.hpp"

#include <chrono>
#include <iostream>
#include <vector>

namespace fmsolvr
{

namespace timeline
{

template <typename ActivityEnum>
class ActivityHRClock
{
    typedef std::chrono::high_resolution_clock clock;
    typedef std::chrono::time_point<clock> time_point;

public:
    typedef ActivityEnum activity_enum;
    typedef uint64_t amount_of_work_type;

private:
    struct start_event
    {
        activity_enum what_;
        time_point when_;

        start_event(activity_enum what, time_point when) : what_(what), when_(when)
        { }

        void changeActivity(activity_enum what)
        {
            what_ = what;
        }
    };

    struct stop_event
    {
        time_point when_;

        stop_event(time_point when) : when_(when)
        { }
    };

public:
    static start_event startActivity(activity_enum e)
    {
        time_point now = clock::now();
        return start_event(e, now);
    }

    static stop_event stopActivity()
    {
        time_point now = clock::now();
        return stop_event(now);
    }

    ActivityHRClock(const start_event &start, const stop_event &stop, amount_of_work_type amount)
            : activity_id_(start.what_),
              amount_of_work_(amount),
              t_start_(start.when_),
              t_stop_(stop.when_)
    { }

    ~ActivityHRClock()
    { }

    uint64_t getDuration_ns()
    {
        return std::chrono::nanoseconds(t_stop_ - t_start_).count();
    }

    amount_of_work_type getAmountOfWork()
    {
        return amount_of_work_;
    }

    activity_enum getActivityID()
    {
        return activity_id_;
    }

private:
    activity_enum activity_id_;
    amount_of_work_type amount_of_work_;
    time_point t_start_;
    time_point t_stop_;
};

template <typename ActivityEnum>
class ActivityScoreP;

template <typename ActivityEnum>
class ActivityNone
{
public:
    typedef ActivityEnum activity_enum;
    typedef uint64_t amount_of_work_type;

private:
    struct start_event
    {
        void changeActivity(activity_enum /* unused */)
        { }
    };

    struct stop_event
    { };

public:
    static start_event startActivity(activity_enum /* unused */)
    {
        return start_event();
    }

    static stop_event stopActivity()
    {
        return stop_event();
    }

    ActivityNone(
            const start_event & /* unused */,
            const stop_event & /* unused */,
            amount_of_work_type /* unused */)
    { }

    ~ActivityNone()
    { }

    uint64_t getDuration_ns()
    {
        return 0;
    }
};

template <typename ActivityEnum, template <typename> class Activity = ActivityNone>
class TimeLineNone
{
public:
    typedef Activity<ActivityEnum> activity_type;
    typedef typename activity_type::activity_enum activity_enum;
    typedef typename activity_type::amount_of_work_type amount_of_work_type;
    typedef decltype(activity_type::startActivity(activity_enum())) start_event_type;

    start_event_type startActivity(activity_enum e)
    {
        return activity_type::startActivity(e);
    }

    void stopActivity(const start_event_type &start_event, amount_of_work_type aow)
    {
        auto stop_event = activity_type::stopActivity();
        addActivity(activity_type(start_event, stop_event, aow));
    }

    TimeLineNone(papiwrapper<> * = nullptr)
    { }

    ~TimeLineNone()
    { }

    void addActivity(activity_type /* unused */)
    { }

    void clear()
    { }

    void reserve(size_t /* unused */)
    { }

    size_t getSize()
    {
        return 0;
    }
};

template <typename ActivityEnum, template <typename> class Activity = ActivityHRClock>
class TimeLineDetailed
{
public:
    typedef Activity<ActivityEnum> activity_type;
    typedef typename activity_type::activity_enum activity_enum;
    typedef typename activity_type::amount_of_work_type amount_of_work_type;
    typedef decltype(activity_type::startActivity(activity_enum())) start_event_type;

    TimeLineDetailed(papiwrapper<> * = nullptr)
    {
        int reserve;
        if ((reserve = get_env_int("TIMELINE_RESERVE", 0)) > 0)
        {
            activities_.reserve(reserve);
        }
    }

    ~TimeLineDetailed()
    { }

    start_event_type startActivity(activity_enum e)
    {
        return activity_type::startActivity(e);
    }

    void stopActivity(const start_event_type &start_event, amount_of_work_type aow)
    {
        auto stop_event = activity_type::stopActivity();
        addActivity(activity_type(start_event, stop_event, aow));
    }

    void addActivity(activity_type e)
    {
        activities_.emplace_back(e);
    }

    void clear()
    {
        activities_.clear();
    }

    void reserve(size_t num_activities)
    {
        activities_.reserve(num_activities);
    }

    size_t getSize()
    {
        return activities_.size();
    }

    const std::vector<activity_type> &getRecords() const
    {
        return activities_;
    }
private:
    std::vector<activity_type> activities_;
};

template <typename ActivityEnum, template <typename> class Activity = ActivityHRClock>
class TimeLineAccumulated
{
public:
    typedef Activity<ActivityEnum> activity_type;
    typedef typename activity_type::activity_enum activity_enum;
    typedef typename activity_type::amount_of_work_type amount_of_work_type;
    typedef decltype(activity_type::startActivity(activity_enum())) start_event_type;

    struct accu
    {
        uint64_t num_activities_;
        uint64_t accu_time;
        uint64_t accu_work;

        accu() : num_activities_(0), accu_time(0), accu_work(0)
        { }

        accu(uint64_t num_activities, uint64_t accu_time, uint64_t accu_work)
                : num_activities_(num_activities), accu_time(accu_time), accu_work(accu_work)
        { }
    };

    TimeLineAccumulated(papiwrapper<> * = nullptr)
            : accus_(static_cast<size_t>(activity_enum::_last_enum))
    { }

    ~TimeLineAccumulated()
    { }

    start_event_type startActivity(activity_enum e)
    {
        return activity_type::startActivity(e);
    }

    void stopActivity(const start_event_type &start_event, amount_of_work_type aow)
    {
        auto stop_event = activity_type::stopActivity();
        addActivity(activity_type(start_event, stop_event, aow));
    }

    void addActivity(activity_type e)
    {
        accu &a = accus_[static_cast<size_t>(e.getActivityID())];
        ++a.num_activities_;
        a.accu_time += e.getDuration_ns();
        a.accu_work += e.getAmountOfWork();
    }

    void clear()
    {
        for (auto &a : accus_)
        {
            a = accu();
        }
    }

    void reserve(size_t /* unused */)
    { }

    size_t getSize()
    {
        return accus_.size();
    }

    const std::vector<accu> &getRecords()
    {
        return accus_;
    }

private:
    std::vector<accu> accus_;
};

template <typename ActivityEnum, template <typename> class Activity = ActivityNone>
class TimeLinePapiPrint
{
public:
    typedef Activity<ActivityEnum> activity_type;
    typedef typename activity_type::activity_enum activity_enum;
    typedef typename activity_type::amount_of_work_type amount_of_work_type;
    typedef decltype(activity_type::startActivity(activity_enum())) start_event_type;

    start_event_type startActivity(activity_enum e)
    {
        start_event_type start_event = activity_type::startActivity(e);
        if (pw_)
        {
            pw_->start();
        }
        return start_event;
    }

    void stopActivity(const start_event_type &start_event, amount_of_work_type aow)
    {
        if (pw_)
        {
            pw_->stop();
        }
        auto stop_event = activity_type::stopActivity();
        addActivity(activity_type(start_event, stop_event, aow));
        if (pw_)
        {
            pw_->report(aow);
        }
    }

    TimeLinePapiPrint(papiwrapper<> *pw = nullptr) : pw_(pw)
    { }

    ~TimeLinePapiPrint()
    { }

    void addActivity(activity_type /* unused */)
    { }

    void clear()
    { }

    void reserve(size_t /* unused */)
    { }

    size_t getSize()
    {
        return 0;
    }

private:
    papiwrapper<> *pw_;
};

}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
