/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_UTIL_buildinfo_hpp
#define FMSOLVR_UTIL_buildinfo_hpp

#include <iostream>
#include <cxxabi.h>
#include "fmsolvr/simd/Vec_base.hpp"
#include "fmsolvr/util/float128.hpp"

namespace fmsolvr
{

template <typename T>
std::string get_name_of_type()
{
	int status;
	char *realname;
	realname = abi::__cxa_demangle(typeid(T).name(), 0, 0, &status);
	std::string name = (status == 0) ? realname : "***unknown***";
	free(realname);
	return name;
}

#if defined(HAVE_FLOAT128) && defined(__GNUG__) && __GNUG__ < 5
// PR43622
template <>
inline
std::string get_name_of_type<__float128>()
{
    return "__float128";
}
#endif

template <typename Real>
struct buildinfo { };

template <typename Real>
std::ostream & operator << (std::ostream & os, buildinfo<Real>)
{
    os <<
#ifdef GITBRANCH
        GITBRANCH ":" GITREVISION
#endif
        " Real=" << get_name_of_type<Real>() <<
        " sizeof(Real)=" << sizeof(Real) <<
#ifdef USE_UPPER
        " UPPER"
#else
        " UPPERLOWER"
#endif
#ifdef NDEBUG
        " NDEBUG"
#else
        " assert() is enabled"
#endif
#ifdef __FAST_MATH__
        " __FAST_MATH__"
#endif
       ;
    return os;
}


template <typename SIMD>
struct SIMDinfo { };

template <typename SIMD>
std::ostream & operator << (std::ostream & os, SIMDinfo<SIMD>)
{
    typedef SIMDtraits<SIMD> ST;
    os <<
        "SIMD width=" << ST::SIMDwidth <<
        " scalar_type=" << get_name_of_type<typename ST::scalar_type>() <<
        " vec_type=" << get_name_of_type<typename ST::vec_type>();
    return os;
}


struct compilerinfo { };

inline
std::ostream & operator << (std::ostream & os, compilerinfo)
{
#ifdef COMPILER
    os << "Compiled with " COMPILER " (" COMPILER_VERSION ") using " COMPILER_FLAGS;
#endif
    return os;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
