/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_UTIL_checking_helpers_hpp
#define FMSOLVR_UTIL_checking_helpers_hpp

#include <cmath>
#include <iostream>


namespace fmsolvr {

namespace checking {

template <typename Complex>
bool is_finite(const Complex & z)
{
    return std::isfinite(z.real()) && std::isfinite(z.imag());
}

template <typename OMuga>
bool is_finite(const OMuga & omuga, size_t p)
{
    bool finite = true;
    size_t mlast = p + 1;
    for (size_t l = 0; l <= p; ++l)
    {
        for (size_t m = 0; m < std::min(mlast, l + 1); ++m)
        {
            if (!is_finite(omuga.upper(l, m)))
            {
                finite = false;
                mlast = m;
                std::cout << "Coefficient (" << l << ", " << m << ") is not finite: " << omuga.upper(l, m) << std::endl;
            }
        }
    }
    return finite;
}

template <typename BoxID, typename TreeOMuga>
bool is_finite(const TreeOMuga & tree_omuga, size_t depth, size_t p)
{
    bool finite = true;
    for (size_t d = 0; d <= depth; ++d)
    {
        size_t dim = size_t(1) << d;
        for (size_t i = 0; i < dim; ++i)
        {
            for (size_t j = 0; j < dim; ++j)
            {
                for (size_t k = 0; k < dim; ++k)
                {
                    BoxID bid(i, j, k, d);
                    const auto & omuga = tree_omuga[bid];
                    if (!is_finite(omuga, p))
                    {
                        std::cout << "Expansion " << bid << " is not finite" << std::endl;
                        finite = false;
                    }
                }
            }
        }
    }
    return finite;
}

}  // namespace checking

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
