/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_UTIL_float128_hpp
#define FMSOLVR_UTIL_float128_hpp

#include <cstddef>

#if defined(_GLIBCXX_USE_FLOAT128) && !defined(__clang__) && !defined(__ICC)
#define HAVE_FLOAT128
#endif

#if defined(HAVE_FLOAT128)

extern "C" {
#include <quadmath.h>
}

#include <cmath>

namespace std
{
    inline __float128 abs(__float128 q)
    {
        return fabsq(q);
    }

    inline __float128 floor(__float128 q)
    {
        return floorq(q);
    }

    inline __float128 sqrt(__float128 q)
    {
        return sqrtq(q);
    }

    inline __float128 pow(__float128 x, __float128 y)
    {
        return powq(x, y);
    }

    inline __float128 sin(__float128 q)
    {
        return sinq(q);
    }

    inline __float128 cos(__float128 q)
    {
        return cosq(q);
    }

    inline __float128 nextafter(__float128 q, __float128 d)
    {
        return nextafterq(q, d);
    }

    inline bool isfinite(__float128 q)
    {
        return !isinfq(q) && !isnanq(q);
    }
}

#include <iostream>

namespace std
{
    inline std::ostream& operator << (std::ostream & os, const __float128 & q)
    {
        char buf[128];
        std::ios_base::fmtflags flags = os.flags();
        std::streamsize width = os.width();
        std::streamsize precision = os.precision();
        std::string fmt;
        fmt += '%';
        if (flags & std::ios::showpos)
            fmt += ' ';  // HACK for printing leading space-or-minus (correct showpos implementation would be '+')
        fmt += "*.*Q";
        if (flags & std::ios::scientific)
            fmt += 'e';
        else if (flags & std::ios::fixed)
            fmt += 'f';
        else
            fmt += 'g';
        quadmath_snprintf(buf, 128, fmt.c_str(), width, precision, q);
        os << buf;
        return os;
    }

    inline std::istream &operator>>(std::istream &is, __float128 &q)
    {
        std::string buf;
        is >> buf;
        q = strtoflt128(buf.c_str(), nullptr);
        return is;
    }
}

#include <limits>

namespace std
{

// this is only a stub!
template <>
struct numeric_limits<__float128>
{
#if __cplusplus >= 201103L
    static constexpr int max_digits10 = 36;
#endif
};

}

#endif  // HAVE_FLOAT128

#endif
// vim: et:ts=4:sw=4
