/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_UTIL_wrpapi_impl_hpp
#define FMSOLVR_UTIL_wrpapi_impl_hpp

#include "fmsolvr/util/wrpapi.hpp"

#include <cstring>
#include <iostream>

namespace fmsolvr
{

#ifdef WITH_PAPI

template <typename T>
papiwrapper<T>::papiwrapper() : EventSet(PAPI_NULL), start_result(-1), stop_result(-1)
{
	int retval = -1;

	if ( ( retval = PAPI_library_init( PAPI_VER_CURRENT ) ) != PAPI_VER_CURRENT )
		throw PAPI_strerror(retval);
	if ( ( retval = PAPI_create_eventset( &EventSet ) ) != PAPI_OK)
		throw PAPI_strerror(retval);

	if ( PAPI_query_event( PAPI_FP_OPS ) == PAPI_OK )
		event[0] = PAPI_FP_OPS;
	else if ( PAPI_query_event( PAPI_FP_INS ) == PAPI_OK )
		event[0] = PAPI_FP_INS;
	else
		event[0] = PAPI_TOT_INS;
	event[1] = PAPI_TOT_CYC;
	event[2] = PAPI_L1_DCM;
	event[3] = PAPI_L2_TCM;
	event[3] = PAPI_FDV_INS;
	event[3] = PAPI_REF_CYC;
	event[3] = PAPI_L1_LDM;
	event[3] = PAPI_L1_ICM;
	event[3] = PAPI_L2_TCA;
	event[3] = PAPI_L3_TCA;
	event[3] = PAPI_LD_INS;
	event[3] = PAPI_SR_INS;
	event[3] = PAPI_BR_INS;
	event[3] = PAPI_BR_MSP;
	event[3] = PAPI_L1_STM;
	event[3] = PAPI_L2_DCM;

	event[2] = PAPI_L1_DCM;
	event[3] = PAPI_L2_TCM;
	event[4] = PAPI_L3_TCM;
	num_events = 5;

/*
	event[2] = PAPI_BR_INS;
	event[3] = PAPI_BR_MSP;
	event[4] = PAPI_REF_CYC;

	num_events = 4;
	event[2] = PAPI_TLB_DM;
	event[3] = PAPI_FDV_INS;

	event[2] = PAPI_TLB_IM;
	event[3] = PAPI_LD_INS;
	//event[3] = PAPI_SR_INS;
*/

	num_events = 4;
	event[1] = PAPI_TLB_DM;
	event[2] = PAPI_L2_TCM;
	event[3] = PAPI_L3_TCM;

	num_events = 2;
	event[0] = PAPI_TOT_INS;
	event[1] = PAPI_TOT_CYC;

	char* env_events = getenv("PAPI_EVENTS");
	if (env_events) {
		char * e = strdup(env_events);
		char * ee = e;
		while (ee) {
			char * eee = strchr(ee, ' ');
			if (eee) {
				*eee = 0;
				++eee;
			}
			if ( ( retval = PAPI_add_named_event( EventSet, ee ) ) != PAPI_OK )
				throw PAPI_strerror(retval);
			ee = eee;
		}
		free(e);
	} else {
		PAPI_add_events(EventSet, event, num_events);
	}

	num_events = max_events;
	PAPI_list_events(EventSet, event, &num_events);
}

template <typename T>
papiwrapper<T>::~papiwrapper()
{
	PAPI_remove_events(EventSet, event, num_events);
	PAPI_destroy_eventset(&EventSet);
}

template <typename T>
void papiwrapper<T>::report(std::ostream &os, double steps, double bytes_load, double bytes_store)
{
	if (start_result != PAPI_OK)
		std::cerr << "PAPI_start failed: " << PAPI_strerror(start_result) << std::endl;
	if (stop_result != PAPI_OK)
		std::cerr << "PAPI_stop failed: " << PAPI_strerror(stop_result) << std::endl;
	double delta_wallclock = stop_time - start_time;
	os << "STEPS=" << steps << std::endl;
	os << "COUNTER(wallclock)=" << delta_wallclock;
	if (steps != 0.)
	{
		os << " per step: " << delta_wallclock * 1000000000. / steps << " ns";
	}
	os << std::endl;
	for (int i = 0; i < num_events; ++i) {
		PAPI_event_info_t info;
		if (PAPI_get_event_info( event[i], &info ) == PAPI_OK)
		{
			os << "COUNTER(" << info.symbol << ")=" << values[i];
			if (steps != 0.)
			{
				os << " per step: " << double(values[i]) / steps;
			}
			os << std::endl;
		}
	}
	if (num_events >= 2 && event[0] == PAPI_TOT_INS && event[1] == PAPI_TOT_CYC)
	{
		os << "CPI=" << double(values[1]) / double(values[0]) << std::endl;
	}
	if (event[1] == PAPI_TOT_CYC && (bytes_load + bytes_store) > 0 && steps != 0.) {
		double cyc_per_step = double(values[1]) / steps;
		os << "Throughput  load=" << bytes_load / cyc_per_step
			<< "  store=" << bytes_store / cyc_per_step
			<< "  total=" << (bytes_load + bytes_store) / cyc_per_step
			<< " bytes/cycle" << std::endl;
	}
}

#else

#include <iostream>

template <typename T>
void papiwrapper<T>::report(std::ostream &os, double steps, double, double)
{
	double delta_wallclock = stop_time - start_time;
	os << "STEPS=" << steps << std::endl;
	os << "COUNTER(wallclock)=" << delta_wallclock;
	if (steps != 0.)
		os << " per step: " << delta_wallclock * 1000000000. / steps << " ns";
	os << std::endl;
}

#endif

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
