/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_xyz_hpp
#define FMSOLVR_xyz_hpp

#include <ostream>
#include <cmath>

namespace fmsolvr
{

template <typename Real>
class XYZ
{
public:
    typedef Real value_type;

    XYZ() : x_(value_type()), y_(value_type()), z_(value_type())
    { }

    XYZ(value_type x, value_type y, value_type z) : x_(x), y_(y), z_(z)
    { }

    template <typename XYZCompat>
    explicit XYZ(const XYZCompat & something): x_(something.x()), y_(something.y()), z_(something.z())
    { }

    const value_type &x() const
    { return x_; }

    const value_type &y() const
    { return y_; }

    const value_type &z() const
    { return z_; }

    XYZ operator - ()
    {
        return XYZ(-x_, -y_, -z_);
    }

	XYZ operator + (const XYZ & a) const
	{
		return XYZ(x_ + a.x_, y_ + a.y_, z_ + a.z_);
	}

	XYZ operator - (const XYZ & a) const
	{
		return XYZ(x_ - a.x_, y_ - a.y_, z_ - a.z_);
	}

	XYZ operator * (const value_type s) const
	{
		return XYZ(x_ * s, y_ * s, z_ * s);
	}

	XYZ & operator += (const XYZ & a)
	{
		x_ += a.x_;
		y_ += a.y_;
		z_ += a.z_;
		return *this;
	}

	XYZ & operator -= (const XYZ & a)
	{
		x_ -= a.x_;
		y_ -= a.y_;
		z_ -= a.z_;
		return *this;
	}

	XYZ & operator *= (const value_type s)
	{
		x_ *= s;
		y_ *= s;
		z_ *= s;
		return *this;
	}

    bool operator ==(const XYZ &xyz) const
    {
        return x_ == xyz.x_ && y_ == xyz.y_ && z_ == xyz.z_;
    }

    bool operator !=(const XYZ &xyz) const
    {
        return !(*this == xyz);
    }

private:
    value_type x_, y_, z_;
};


// |a|
template <typename Real>
Real length(const XYZ<Real> & a)
{
	return std::sqrt(a.x() * a.x() + a.y() * a.y() + a.z() * a.z());
}


template <typename Real>
std::ostream & operator << (std::ostream & os, const XYZ<Real> & xyz)
{
    os << "(" << xyz.x() << ", " << xyz.y() << ", " << xyz.z() << ")";
    return os;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
