/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_xyzs_hpp
#define FMSOLVR_xyzs_hpp

#include <iostream>

namespace fmsolvr
{

// coordinates x, y, z and strength s (e.g. charge q or mass m)
template <typename Real>
class XYZS
{
public:
    typedef Real value_type;

    XYZS() : x_(value_type()), y_(value_type()), z_(value_type()), s_(value_type())
    { }

    XYZS(value_type x, value_type y, value_type z, value_type s) : x_(x), y_(y), z_(z), s_(s)
    { }

    template <typename XYZ>
    XYZS(const XYZ &xyz, value_type s) : x_(xyz.x()), y_(xyz.y()), z_(xyz.z()), s_(s)
    { }

    template <typename XYZ>
    void setCoordinates(const XYZ &xyz)
    {
        x_ = xyz.x();
        y_ = xyz.y();
        z_ = xyz.z();
    }

    const value_type &x() const
    { return x_; }

    const value_type &y() const
    { return y_; }

    const value_type &z() const
    { return z_; }

    const value_type &s() const
    { return s_; }

    const value_type &q() const
    { return s_; }

    int site() const
    { return 0; }

    int form() const
    { return 0; }

private:
    value_type x_, y_, z_, s_;
};

template <typename Real>
XYZS<Real> make_xyzs(const Real & x, const Real & y, const Real & z, const Real & s)
{
    return XYZS<Real>(x, y, z, s);
}

template <typename Real, typename XYZ>
XYZS<Real> make_xyzs(const XYZ & xyz, const Real & s)
{
    return XYZS<Real>(xyz.x(), xyz.y(), xyz.z(), s);
}

template <typename Real>
XYZS<Real> make_xyzq(const Real & x, const Real & y, const Real & z, const Real & s)
{
    return XYZS<Real>(x, y, z, s);
}

template <typename Real, typename XYZ>
XYZS<Real> make_xyzq(const XYZ & xyz, const Real & s)
{
    return XYZS<Real>(xyz.x(), xyz.y(), xyz.z(), s);
}

template <typename Real>
std::ostream & operator << (std::ostream & os, const XYZS<Real> & xyzs)
{
    os << "(" << xyzs.x() << ", " << xyzs.y() << ", " << xyzs.z() << "; " << xyzs.s() << ")";
    return os;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
