/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */
#include <iostream>
#include <chrono>
#include <vector>

typedef std::chrono::high_resolution_clock clk;
typedef clk::time_point time_point;

void bench(int n)
{
    // preheat memory
    {
        std::vector<time_point> tp(n, clk::now());
        if (std::chrono::nanoseconds(tp.back() - tp.front()).count() != 0)
            throw clk::now();
    }

    std::vector<time_point> tp;

    // push_back() w/o reserve()
    for (int i = 0; i < n; ++i)
        tp.push_back(clk::now());
    auto elapsed1 = tp.back() - tp.front();

    // push_back() w/ reserve()
    tp.resize(0);
    tp.reserve(n);
    for (int i = 0; i < n; ++i)
        tp.push_back(clk::now());
    auto elapsed2 = tp.back() - tp.front();

    // operator[] and value_type::operator=
    for (int i = 0; i < n; ++i)
        tp[i] = clk::now();
    auto elapsed3 = tp.back() - tp.front();

    auto ns1 = std::chrono::nanoseconds(elapsed1).count();
    auto ns2 = std::chrono::nanoseconds(elapsed2).count();
    auto ns3 = std::chrono::nanoseconds(elapsed3).count();

    std::cout << n << "\t" << double(ns1) / double(n - 1) << "\t" << double(ns2) / double(n - 1)
              << "\t" << double(ns3) / double(n - 1) << std::endl;
}

int main()
{
    bench(1000000);
    std::cout << "# sizeof(time_point)=" << sizeof(time_point) << std::endl;
    std::cout << "# n\tpush_back\treserve+pb\tassign" << std::endl;
    bench(1000);
    bench(10000);
    bench(100000);
    bench(1000000);
    bench(10000000);
    bench(100000000);
    bench(1000000000);
}
