/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */
#ifndef _AnBe_FMM_MICROBENCH_cubesum_hpp_
#define _AnBe_FMM_MICROBENCH_cubesum_hpp_

#include "xyzq.hpp"

template <typename T>
struct remove_complex
{
    typedef T type;
};

template <typename T>
struct remove_complex<std::complex<T> >
{
    typedef T type;
};


template <bool enable_x, bool enable_y, bool enable_z,
    typename Real33, typename Triangle,
    typename Operation>
size_t cubesum(
        const Real33 & abc,
        const size_t ws,
        Triangle & accu,
        const size_t p,
        Operation op)
{
    typedef typename remove_complex<typename Triangle::value_type>::type RealVec;
    typedef typename Vec_traits<RealVec>::scalar_type Real;

    const ssize_t ws_x = enable_x ? ws : 0;
    const ssize_t ws_y = enable_y ? ws : 0;
    const ssize_t ws_z = enable_z ? ws : 0;

    Vec_traits<RealVec> VT;
    RealVec rvi, rvj, rvk;
    size_t vi = 0;

    size_t cnt = 0;
    for (ssize_t i = -ws_x; i <= +ws_x; ++i) {
        Real ri(i);
        for (ssize_t j = -ws_y; j <= +ws_y; ++j) {
            Real rj(j);
            for (ssize_t k = -ws_z; k <= +ws_z; ++k) {
                Real rk(k);
                ++cnt;
                VT.element(rvi, vi) = ri;
                VT.element(rvj, vi) = rj;
                VT.element(rvk, vi) = rk;
                ++vi;
                if (vi == VT.VecSize)
                {
                    RealVec vx = VT.same(abc.a.x) * rvi + VT.same(abc.b.x) * rvj + VT.same(abc.c.x) * rvk;
                    RealVec vy = VT.same(abc.a.y) * rvi + VT.same(abc.b.y) * rvj + VT.same(abc.c.y) * rvk;
                    RealVec vz = VT.same(abc.a.z) * rvi + VT.same(abc.b.z) * rvj + VT.same(abc.c.z) * rvk;
                    op(make_xyzq(vx, vy, vz, VT.same(1.0)), accu, p);
                    vi = 0;
                }
            }
        }
    }
    if (vi > 0) {
        RealVec q = VT.same(1.0);
        for ( ; vi < VT.VecSize; ++vi) {
            VT.element(rvi, vi) = VT.element(rvj, vi) = VT.element(rvk, vi) = 1;
            VT.element(q, vi) = 0;
        }
        RealVec vx = VT.same(abc.a.x) * rvi + VT.same(abc.b.x) * rvj + VT.same(abc.c.x) * rvk;
        RealVec vy = VT.same(abc.a.y) * rvi + VT.same(abc.b.y) * rvj + VT.same(abc.c.y) * rvk;
        RealVec vz = VT.same(abc.a.z) * rvi + VT.same(abc.b.z) * rvj + VT.same(abc.c.z) * rvk;
        op(make_xyzq(vx, vy, vz, q), accu, p);
    }
    return cnt;
}

#endif
// vim: et:ts=4:sw=4
