/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */
#include <iostream>
#include <ctime>
#include <sys/time.h>
#include <typeinfo>

#include "wrpapi.hpp"
#include "Vec.h"

#include "multipole.hpp"
#include "abc.hpp"
#include "xyz.hpp"

#include "cubesum.hpp"
#include "triangle_accumulate.hpp"
#include "linear_accumulate.hpp"
#include "muladdmul.hpp"
#include "complex_mul_acc.hpp"


// Returns number of seconds since the epoch, high resolution.
inline double
timestamp()
{
    struct timeval tp;
    gettimeofday(&tp, NULL);
    return double(tp.tv_sec) + tp.tv_usec / 1000000.;
}

template <typename T>
T __attribute__((noinline)) bless(T val)
{
    return val;
}

template <typename CoefficientMatrix>
void dump(const CoefficientMatrix & a, size_t p)
{
    for (size_t m = p; m <= p; --m) {
        for (size_t l = 0; l <= p; ++l) {
            if (m <= l)
                std::cout << a(l, m) << " ";
            else
                std::cout << "() ";
        }
        std::cout << std::endl;
    }
}


int main()
{
#ifdef USE_FLOAT
    typedef float Real;
#else
    typedef double Real;
#endif
    typedef XYZ<Real> Real3;
    typedef ABC<Real3> Real33;
#ifndef USE_UPPERLOWER
    typedef MultipoleCoefficientsUpper<Real> CoeffMatrix;
#else
    typedef MultipoleCoefficientsUpperLower<Real> CoeffMatrix;
#endif

    typedef typename simd_engine_default<Real>::type simd_base_type;
#if 1
    typedef simd_base_type simd_type;
#else
    typedef SuperSIMD<simd_base_type, 2> super_simd_type;
    typedef super_simd_type simd_type;
#endif
    typedef Vec_traits<simd_type> simd_traits;
#if 1
    typedef std::complex<simd_type> complex_simd_type;
#else
    typedef simd_type complex_simd_type;
#endif
    typedef aligned_allocator<complex_simd_type> simd_allocator_type;
    size_t complex_size_factor = sizeof(complex_simd_type) / sizeof(simd_type);

    std::cout << "# FMM microbench"
        << " " << GITBRANCH << ":" << GITREVISION
        << " sizeof(Real)=" << sizeof(Real)
        << " sizeof(simd_type)=" << sizeof(simd_type)
        << " SIMDwidth=" << simd_traits::VecSize
        << " complex=" << complex_size_factor - 1
#ifdef USE_FLOAT
        << " FLOAT"
#else
        << " DOUBLE"
#endif
#ifdef USE_UPPERLOWER
        << " UPPERLOWER"
#else
        << " UPPER"
#endif
#ifdef NDEBUG
        << " NDEBUG"
#endif
#ifdef __FAST_MATH__
        << " __FAST_MATH__"
#endif
        << std::endl;

    papiwrapper pw;

    Real3 a(1, 0, 0);
    Real3 b(0, 1, 0);
    Real3 c(0, 0, 1);
    Real33 abc(a, b, c);

    if (1)
    {
//        size_t ps[] = {1, 2, 3, 4, 5, 7, 9, 10, 15, 20, 25, 30, 35, 40, 45, 50};
        size_t ps[] = {1, 2, 3, 4, 5, 7, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 25, 30, 35, 40, 45, 50};
        size_t wss[] = {15};

#if 0
        auto op = do_triangle_accumulate();
#elif 0
        auto op = do_linear_accumulate();
#elif 0
        auto op = do_linear_muladdmul();
#elif 1
        auto op = do_linear_complex_mul_acc();
#endif
        std::cout << "# cubesum op=" << typeid(__typeof__(op)).name() << std::endl;
        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t p = 2 * ps[ip];

                typedef UpperTriangularArrayRowMajor<complex_simd_type, void, simd_allocator_type> triangle_type;
                triangle_type accu(p);

                pw.start();
                size_t boxes = cubesum<true, true, true>(abc, ws, accu, p, op);
                pw.stop();
                size_t elements = (p + 1) * (p + 2) / 2;
                size_t mem = sizeof(typename triangle_type::value_type) * elements;
                std::cout << std::endl << "p=" << p << "  ws=" << ws << "  boxes=" << boxes << "  mem=" << double(mem) / 1024 << " Kb" << std::endl;
                size_t groupsize = simd_traits::VecSize;
                size_t elementsize = complex_size_factor * sizeof(Real);
                //pw.report(boxes * elements / groupsize, elementsize * groupsize, elementsize * groupsize);
                elements = elements / 2;
                pw.report(boxes * elements / groupsize / 3, elementsize * groupsize * 2 * 3);
            }
    }
}

// vim: et:ts=4:sw=4
