/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */
#include <iostream>
#include <iomanip>
#include <vector>

#define M2LP3_IMPL_DUMP

#include "fmsolvr/abc.hpp"
#include "fmsolvr/misc/energy_stress_triangular_array_group.hpp"
#include "fmsolvr/multipole.hpp"
#include "fmsolvr/operator/lattice/latticeoperator.hpp"
#include "fmsolvr/operator/p3/local2local_p3.hpp"
#include "fmsolvr/operator/p3/multipole2local_p3.hpp"
#include "fmsolvr/operator/p3/multipole2multipole_p3.hpp"
#include "fmsolvr/operator/p4/local2local.hpp"
#include "fmsolvr/operator/p4/local2local_derivative.hpp"
#include "fmsolvr/operator/p4/multipole2local.hpp"
#include "fmsolvr/operator/p4/multipole2local_optimized.hpp"
#include "fmsolvr/operator/p4/multipole2multipole.hpp"
#include "fmsolvr/operator/p4/multipole2multipole_derivative.hpp"
#include "fmsolvr/operator/particle2local.hpp"
#include "fmsolvr/operator/particle2multipole.hpp"
#include "fmsolvr/operator/rotation/rotation_types.hpp"
#include "fmsolvr/simd/simd_engine.hpp"
#include "fmsolvr/util/buildinfo.hpp"
#include "fmsolvr/util/envtools.hpp"
#include "fmsolvr/util/float128.hpp"
#include "fmsolvr/util/wrpapi.hpp"
#include "fmsolvr/xyz.hpp"

#include "make_grid_omega.hpp"
#include "homopole.hpp"
#include "Mlm400_64_iter64_1D.h"
#include "Mlm400_64_iter64_2D.h"
#include "Mlm400_64_iter64_3D.h"


namespace testdata {
//#include "qxyz1.hpp"
#include "qxyz17.hpp"
}

#define STRINGIFY(__x) _STRINGIFY(__x)
#define _STRINGIFY(__x) # __x

namespace fmsolvr
{

// relative error, using a as exact reference
template <typename CoefficientMatrix>
typename CoefficientMatrix::complex_type maxdelta(const CoefficientMatrix & a, const CoefficientMatrix & b, size_t p)
{
    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;
    scalar_type re(0), im(0);

    for (size_t l = 0; l <= p; ++l)
        for (size_t m = 0; m <= l; ++m) {
            complex_type alm = a.upper(l, m);
            complex_type blm = b.upper(l, m);
            typename complex_type::value_type almlen = fmm::abs(alm);
            if (almlen > 0) {
                re = std::max(re, std::abs((alm.real() - blm.real()) / almlen));
                im = std::max(im, std::abs((alm.imag() - blm.imag()) / almlen));
            }
        }
    return complex_type(re, im);
}

// relative error, using maximum value in current subspace
template <typename CoefficientMatrix>
typename CoefficientMatrix::complex_type maxdelta_col(const CoefficientMatrix & a, const CoefficientMatrix & b, size_t p)
{
    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;
    scalar_type re_max(0), im_max(0);

    for (size_t l = 0; l <= p; ++l)
        {

            scalar_type max_abs = 0;
            scalar_type re(0),im(0);

            for (size_t m = 0; m <= l; ++m) {


                complex_type alm = a.upper(l, m);
                complex_type blm = b.upper(l, m);

                max_abs = std::max(max_abs, fmm::abs(alm));
                re = std::max(re, std::abs((alm.real() - blm.real())));
                im = std::max(im, std::abs((alm.imag() - blm.imag())));
            }
            re_max = std::max(re_max,re/max_abs);
            im_max = std::max(re_max,re/max_abs);
        }
    return complex_type(re_max, im_max);
}


template <typename CoefficientMatrix>
void trianglediff(CoefficientMatrix & c, const CoefficientMatrix & a, const CoefficientMatrix & b, size_t p)
{
    for (size_t l = 0; l <= p; ++l)
        for (size_t m = 0; m <= l; ++m)
            c.upper(l, m) = a.upper(l, m) - b.upper(l, m);
}

template <typename T>
T __attribute__((noinline)) bless(T val)
{
    return val;
}

template <typename CoefficientMatrix>
void dump(const CoefficientMatrix & a, size_t p)
{
    std::cout << std::scientific;
    for (size_t m = p; m <= p; --m) {
        for (size_t l = 0; l <= p; ++l) {
            if (m <= l)
                std::cout << a.upper(l, m) << " ";
            else
                std::cout << "() ";
        }
        std::cout << std::endl;
    }
}

template <typename CoefficientMatrix>
void dump_lmreim(const CoefficientMatrix & a, size_t p, bool show_zeroes = true)
{
    typedef typename CoefficientMatrix::value_type complex_type;

    std::cout << std::scientific << std::setprecision(40);
    std::cout << "# l\tm\tre\tim" << std::endl;
    for (size_t l = 0; l <= p; ++l) {
        for (size_t m = 0; m <= l; ++m) {
            complex_type x = a.upper(l, m);
            if (show_zeroes || x != complex_type(0, 0))
                std::cout << l << "\t" << m << "\t" << x.real() << "\t" << x.imag() << std::endl;
        }
    }
}

template <typename CoefficientMatrix>
void dumpevenreal(const CoefficientMatrix & a, int p)
{
    for (int m = p - (p & 1); m >= 0; m -= 2) {
        for (int l = 0; l <= p; l += 2)
            if (m <= l)
                std::cout << l << "," << m << ":" << a.upper(l, m).real() << " ";
            else
                std::cout << "  ";
        std::cout << std::endl;
    }
}

template <typename CoefficientMatrix>
typename CoefficientMatrix::value_type multiply(const CoefficientMatrix & omega, const CoefficientMatrix & mu, size_t p)
{
    typename CoefficientMatrix::value_type res;
    for (size_t l = 0; l <= p; ++l)
        for (size_t m = 0; m <= l; ++m)
            res += omega.upper(l, m) * mu.upper(l, m);
    return res;
}

template <typename Complex>
typename Complex::value_type realonlycomplexmult(const Complex & a, const Complex & b)
{
    return a.real() * b.real() - a.imag() * b.imag();
}

template <typename Complex>
Complex quickcomplexmult(const Complex & a, const Complex & b)
{
    return Complex(a.real() * b.real() - a.imag() * b.imag(), a.real() * b.imag() + b.real() * a.imag());
}


int main(int argc, char **argv)
{
#ifdef USE_FLOAT
    typedef float Real;
#elif defined(USE_LONGDOUBLE)
    typedef long double Real;
#elif defined(USE_FLOAT128)
    typedef __float128 Real;
#else
    typedef double Real;
#endif
    typedef XYZ<Real> Real3;
    typedef ABC<Real3> Real33;
#ifdef USE_UPPER
    typedef MultipoleCoefficientsUpper<Real> CoeffMatrix;
#else
    typedef MultipoleCoefficientsUpperLower<Real> CoeffMatrix;
#endif

    typedef fmsolvr::RotationDataTypes<Real> RotationDataTypes;
    typedef typename CoeffMatrix::value_type coefficient_type;
    typedef typename RotationDataTypes::template ScratchTypeGenerator<coefficient_type>::Scratch scratch_type;

    std::string subtest = argc > 1 ? argv[1] : "";
    std::cout << "# FMM operator test"
              << " (" << subtest << ")"
              << " " << buildinfo<Real>() << std::endl;
    std::cout << "# " << compilerinfo() << std::endl;

    if (0)
    {
        typedef COMPLEX_GENERATOR<Real>::type cmplx;
        std::cout << "cmplx: " << get_name_of_type<cmplx>() << std::endl;
        std::cout << "-cmplx: " << get_name_of_type<decltype(-cmplx())>() << std::endl;
        std::cout << "cmplx+cmplx: " << get_name_of_type<decltype(cmplx() + cmplx())>()
                  << std::endl;
        std::cout << "cmplx*cmplx: " << get_name_of_type<decltype(cmplx() * cmplx())>()
                  << std::endl;
        std::cout << "cmplx*(cmplx+cmplx): "
                  << get_name_of_type<decltype(cmplx() * (cmplx() + cmplx()))>() << std::endl;
        std::cout << "(cmplx+cmplx)*cmplx: "
                  << get_name_of_type<decltype((cmplx() + cmplx()) * cmplx())>() << std::endl;
        std::cout << "(cmplx+cmplx)*(cmplx+cmplx): "
                  << get_name_of_type<decltype((cmplx() + cmplx()) * (cmplx() + cmplx()))>()
                  << std::endl;
    }

    papiwrapper<> pw;

    Real3 e1(1, 0, 0);
    Real3 e2(0, 1, 0);
    Real3 e3(0, 0, 1);

    if (0)
    {
#ifndef USE_FLOAT

    for (size_t p = 0; p < 51; ++p) {
        std::cout << " p = " << p << std::endl;

        Real3 loc1(0.5, 0.5, 0.5);
        Real3 loc2(2.25, 2.25, 2.25);
        scratch_type scratch(p);

        {
            CoeffMatrix mu1(p), mu2(p);
            P2L(23., Real3(0.1, 0.25, 0.33), mu1, Real3(0, 0, 0), p);
            P2L_old(make_xyzq<Real>(0.1, 0.25, 0.33, 23.), mu2, p);
            std::cout << "P2L vs. P2L_old: " << maxdelta(mu1, mu2, p) << std::endl;

        }

        std::cout << "P2M" << std::endl;
        CoeffMatrix omega(p);
        pw.start();
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega, p);
        pw.stop();
        pw.report(testdata::n);
        omega.populate_lower();

        std::cout << "P2L" << std::endl;
        CoeffMatrix mu(p);
        pw.start();
        P2L(testdata::x, testdata::y, testdata::z, testdata::q, loc2, 0, testdata::n, mu, p);
        pw.stop();
        pw.report(testdata::n);
        mu.populate_lower();

        std::cout << "M2L" << std::endl;
        CoeffMatrix mu2(p);
        pw.start();
        XM2L(omega, loc2 - loc1, mu2, p);
        pw.stop();
        pw.report(1);

        std::cout << "M2L_reference" << std::endl;
        CoeffMatrix mu2a(p);
        pw.start();
        M2L_reference(omega, loc1, mu2a, loc2, p);
        pw.stop();
        pw.report(1);

        std::cout << "M2M" << std::endl;
        CoeffMatrix omega2(p);
        pw.start();
        M2M(omega, loc1, omega2, loc2, p);
        pw.stop();
        pw.report(1);

        std::cout << "M2M_p3" << std::endl;
        CoeffMatrix omega2_p3(p);
        pw.start();
        M2Mp3(omega, loc1, omega2_p3, loc2, p, scratch);
        pw.stop();
        pw.report(1);

        std::cout << "L2L" << std::endl;
        CoeffMatrix mu3(p);
        pw.start();
        L2L(mu, loc1, mu3, loc2, p);
        pw.stop();
        pw.report(1);

        std::cout << "L2L_p3" << std::endl;
        pw.start();
        L2Lp3(mu, loc1, mu3, loc2, p, scratch);
        pw.stop();
        pw.report(1);

        //dump(omega, p);
        //dump(mu, p);
        //dump(mu2, p);
        //dump(mu2a, p);
        std::cout << mu2.upper(0,0) << mu2a.upper(0,0) << mu.upper(0,0) << std::endl;

        std::cout << "maxdelta " << maxdelta(mu, mu2, p) << maxdelta(mu2, mu2a, p) << std::endl;

        std::cout << omega.upper(p,0) + omega.upper(p,p)
            + mu.upper(p,0) + mu.upper(p,p)
            + mu2.upper(p,0) + mu2.upper(p,p)
            + mu2a.upper(p,0) + mu2a.upper(p,p)
            + omega2.upper(p,0) + omega2.upper(p,p)
            + mu3.upper(p,0) + mu3.upper(p,p)
            << std::endl;

        /*
        M2L_Operator<CoeffMatrix> Bforward(0.1, 0.25, 0.666, 7);
        M2L_Operator<CoeffMatrix> Bbackward(-0.1, -0.25, -0.666, 7);
        dump(Bforward, 14);
        dump(Bbackward, 14);
        */
    }

    }

//#define BENCH_M2L M2L
#define BENCH_M2L M2L_reference
//#define BENCH_M2L M2L_switch
//#define BENCH_M2L M2L_opt_m0
#define BENCH_M2L_PAIR M2L_pair_reference

    const int k868640 = 12;
    int reps868640[k868640] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 13};
    const int k108502 = 11;
    int reps108502[k108502] = {16, 20, 24, 25, 28, 33, 38, 43, 49, 60, 75};

    if (0)
    {

    {
        int reps = 868640;
        const size_t p = 10;
        std::cout << " p = " << p << std::endl;

        Real3 loc1(0.5, 0.5, 0.5);
        Real3 loc2(2.25, 2.25, 2.25);

        CoeffMatrix omega(p);
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega, p);
        omega.populate_lower();

        M2L_Operator<CoeffMatrix> B(loc1, loc2, p);

        CoeffMatrix mu2(p);
        pw.start();
        for (int rep = 0; rep < reps; ++rep)
            BENCH_M2L(omega, B, mu2, p);
        pw.stop();
        pw.report(reps);

        std::cout << mu2.upper(0,0) + mu2.upper(p,0) + mu2.upper(p,p) << std::endl;
    }

    std::cout << "BENCHMARK START" << std::endl;

    }

    if (0)
    {

        for (size_t p = 0; p <= 50 ; ++p)
        {
            std::cout << " p = " << p << std::endl;

            int reps = 2000000;
            if (p > 10)
                reps /= p;
            Real loc = 0.25;
            Real delta = Real(0.5) / Real(reps);
            CoeffMatrix::complex_type acc;

            pw.start();
            for (int rep = 0; rep < reps; ++rep) {
                M2L_Operator<CoeffMatrix> B(Real3(loc, loc, loc), p);
                loc += delta;
                acc += B.upper(0, 0) + B.upper(p, 0) + B.upper(p, p);
            }
            pw.stop();
            pw.report(reps);

            std::cout << acc << std::endl;
        }
    }
    else if (0)
    {

        for (size_t p = 0; p <= 50 ; ++p)
        {
            std::cout << " p = " << p << std::endl;

            int reps = 5000000;
            if (p > 10)
                reps /= p;
            Real locx = 0.25;
            Real locy = 0.80;
            Real locz = 0.33;
            Real delta = Real(0.5) / Real(reps);
            CoeffMatrix mu1(p);

            pw.start();
            for (int rep = 0; rep < reps; ++rep) {
                P2L_old(make_xyzq(locx, locy, locz, Real(42.)), mu1, p);
                locx += delta;
                locy -= delta;
                locz += delta;
            }
            pw.stop();
            pw.report(reps);

            locx = 0.25;
            locy = 0.80;
            locz = 0.33;
            CoeffMatrix mu2(p);

            pw.start();
            for (int rep = 0; rep < reps; ++rep) {
                P2L(42., Real3(locx, locy, locz), mu2, Real3(0, 0, 0), p);
                locx += delta;
                locy -= delta;
                locz += delta;
            }
            pw.stop();
            pw.report(reps);

            std::cout << maxdelta(mu1, mu2, p) << std::endl;
        }
    }
    else if (0)
    {
        for (int z = 0; z < k868640; ++z)
        {
            int reps = 868640 / 10;
            size_t p = reps868640[z];
            std::cout << " p = " << p << std::endl;

            Real3 loc1(0.5, 0.5, 0.5);
            Real3 loc2(2.25, 2.25, 2.25);

            CoeffMatrix omega1(p);
            P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega1, p);
            omega1.populate_lower();

            CoeffMatrix omega2(p);
            P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega2, p);
            omega2.populate_lower();

            CoeffMatrix mu2(p);
            CoeffMatrix mu1(p);
            pw.start();
            for (int rep = 0; rep < reps; ++rep)
                BENCH_M2L_PAIR(omega1, omega2, loc2 - loc1, mu2, mu1, p);
            pw.stop();
            pw.report(reps * 2);

            std::cout << mu2.upper(0,0) + mu2.upper(p,0) + mu2.upper(p,p) + mu1.upper(0,0) + mu1.upper(p,0) + mu1.upper(p,p) << std::endl;
        }

        for (int z = 0; z < k108502; ++z)
        {
            int reps = 108502 / 10 / 2;
            size_t p = reps108502[z];
            std::cout << " p = " << p << std::endl;

            Real3 loc1(0.5, 0.5, 0.5);
            Real3 loc2(2.25, 2.25, 2.25);

            CoeffMatrix omega1(p);
            P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega1, p);
            omega1.populate_lower();

            CoeffMatrix omega2(p);
            P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega2, p);
            omega2.populate_lower();

            CoeffMatrix mu2(p);
            CoeffMatrix mu1(p);
            pw.start();
            for (int rep = 0; rep < reps; ++rep)
                BENCH_M2L_PAIR(omega1, omega2, loc2 - loc1, mu2, mu1, p);
            pw.stop();
            pw.report(reps * 2);

            std::cout << mu2.upper(0,0) + mu2.upper(p,0) + mu2.upper(p,p) + mu1.upper(0,0) + mu1.upper(p,0) + mu1.upper(p,p) << std::endl;
        }
    }
    else if (0)
    {

#if 0  // needs M2L(source_omega, source_location, target_mu, target_location, p)
    for (int z = 0; z < k868640; ++z)
    {
        int reps = 868640 / 10;
        size_t p = reps868640[z];
        std::cout << " p = " << p << std::endl;

        Real3 loc1(0.5, 0.5, 0.5);
        Real3 loc2(2.25, 2.25, 2.25);

        CoeffMatrix omega(p);
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega, p);
        omega.populate_lower();

        CoeffMatrix mu2(p);
        pw.start();
        for (int rep = 0; rep < reps; ++rep)
            BENCH_M2L(omega, loc1, mu2, loc2, p);
        pw.stop();
        pw.report(reps);

        std::cout << mu2(0,0) + mu2(p,0) + mu2(p,p) << std::endl;
    }

    for (int z = 0; z < k108502; ++z)
    {
        int reps = 108502 / 10 / 2;
        size_t p = reps108502[z];
        std::cout << " p = " << p << std::endl;

        Real3 loc1(0.5, 0.5, 0.5);
        Real3 loc2(2.25, 2.25, 2.25);

        CoeffMatrix omega(p);
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega, p);
        omega.populate_lower();

        CoeffMatrix mu2(p);
        pw.start();
        for (int rep = 0; rep < reps; ++rep)
            BENCH_M2L(omega, loc1, mu2, loc2, p);
        pw.stop();
        pw.report(reps);

        std::cout << mu2(0,0) + mu2(p,0) + mu2(p,p) << std::endl;
    }
#endif

    }
    else if (0)
    {

    for (int z = 0; z < k868640; ++z)
    {
        int reps = 868640 / 10 ;
        size_t p = reps868640[z];
        std::cout << " p = " << p << std::endl;

        Real3 loc1(0.5, 0.5, 0.5);
        Real3 loc2(2.25, 2.25, 2.25);

        CoeffMatrix omega(p);
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega, p);
        omega.populate_lower();

        M2L_Operator<CoeffMatrix> B(loc1, loc2, p);

        CoeffMatrix mu2(p);
        pw.start();
        for (int rep = 0; rep < reps; ++rep)
            BENCH_M2L(omega, B, mu2, p);
        pw.stop();
        pw.report(reps);

        std::cout << mu2.upper(0,0) + mu2.upper(p,0) + mu2.upper(p,p) << std::endl;
    }

    for (int z = 0; z < k108502; ++z)
    {
        int reps = 108502 / 10 / 2 ;
        size_t p = reps108502[z];
        std::cout << " p = " << p << std::endl;

        Real3 loc1(0.5, 0.5, 0.5);
        Real3 loc2(2.25, 2.25, 2.25);

        CoeffMatrix omega(p);
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega, p);
        omega.populate_lower();

        M2L_Operator<CoeffMatrix> B(loc1, loc2, p);

        CoeffMatrix mu2(p);
        pw.start();
        for (int rep = 0; rep < reps; ++rep)
            BENCH_M2L(omega, B, mu2, p);
        pw.stop();
        pw.report(reps);

        std::cout << mu2.upper(0,0) + mu2.upper(p,0) + mu2.upper(p,p) << std::endl;
    }


    }
    else if (0)
    {

#define BENCH_ROTATE ApplyRotationMultipoleForward_folded<simd_engine_none>
//#define BENCH_ROTATE ApplyRotationLocalBackward_folded<simd_engine_none>

    typedef RotationMatrixUpper<Real> rotation_matrix;
    typedef RotationMatrixVector<rotation_matrix> rotation_matrix_vector;
    typedef RG::RotationGeneratorNoPrecomp<rotation_matrix_vector, Real3> rotation_generator;
    typedef typename rotation_generator::rotation_operator rotation_operator;

    auto bench_rotate = [&](size_t p, size_t reps) {
        std::cout << " algo = " << STRINGIFY(BENCH_ROTATE) << std::endl;
        std::cout << " p = " << p << std::endl;

        Real3 loc1(0.5, 0.5, 0.5);
        Real3 loc2(2.25, 2.25, 2.25);

        LinearArray<typename CoeffMatrix::value_type> scratcharray(p);

        CoeffMatrix omuga(p);
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omuga, p);
        omuga.populate_lower();

        rotation_generator G(p);
        rotation_operator R = G.generate_rotation(loc1 - loc2);

        CoeffMatrix omuga_rot(p);
        pw.start();
        for (size_t rep = 0; rep < reps; ++rep)
            BENCH_ROTATE(omuga, R, omuga_rot, p, scratcharray);
        pw.stop();
        pw.report(reps);

        std::cout << omuga_rot.upper(0,0) + omuga_rot.upper(p,0) + omuga_rot.upper(p,p) << std::endl;
    };

    for (int z = 0; z < k868640; ++z)
    {
        int reps = 868640 / 1 ;
        size_t p = reps868640[z];
        bench_rotate(p, reps);
    }

    for (int z = 0; z < k108502; ++z)
    {
        int reps = 108502 / 2 ;
        size_t p = reps108502[z];
        bench_rotate(p, reps);
    }

#endif
    }
    else if (0)
    {
        const size_t p = 10;
        std::cout << "P2L" << std::endl;
        CoeffMatrix mu(p);
        P2L(1., Real3(1., 1., 1.), mu, Real3(0, 0, 0), p);
        dump(mu, p);
    }
    else if (0)
    {
        const size_t p = 10;
        std::cout << "P2M" << std::endl;
        CoeffMatrix omega(p);
        P2M_nodivtable(1., Real3(1., 1., 1.), omega, Real3(0, 0, 0), p);
        dump(omega, p);
    }
    else if (0)
    {
        Real3 ptcl1(0.1, 0.3, 0.2);
        Real charge(1.2345);
        Real3 o1sml(1, 1, 1);
        Real3 o1big(2, 2, 2);
        Real3 ptcl2(11.4, 12.1, 13.7);
        Real3 o2sml(11, 11.5, 12.333);
        Real3 o2big(10, 11, 12);
        const size_t p = 50;
        const size_t pd = std::min(p, size_t(8));
        scratch_type scratch(p);

        CoeffMatrix mu(p);
        P2L(charge, ptcl1, mu, ptcl2, p);
        mu.populate_lower();
        std::cout << "P2L" << std::endl;
        dump(mu, pd);

        // @o1sml
        CoeffMatrix omega1(p);
        P2M_nodivtable(charge, ptcl1, omega1, o1sml, p);
        omega1.populate_lower();
        std::cout << "P2M" << std::endl;
        dump(omega1, pd);

        // @o1big
        CoeffMatrix omega2(p);
        M2M(omega1, o1sml, omega2, o1big, p);
        omega2.populate_lower();
        std::cout << "M2M" << std::endl;
        dump(omega2, pd);

        // @o1big
        CoeffMatrix omega2d(p);
        P2M_nodivtable(charge, ptcl1, omega2d, o1big, p);
        omega2d.populate_lower();
        std::cout << "P2M' " << maxdelta(omega2, omega2d, p) << std::endl;
        dump(omega2d, pd);

        // @o1big
        CoeffMatrix omega2p3(p);
        M2Mp3(omega1, o1sml, omega2p3, o1big, p, scratch);
        omega2p3.populate_lower();
        std::cout << "M2M_p3' " << maxdelta_col(omega2, omega2p3, p) << std::endl;
        dump(omega2p3, pd);

        // @o2big
        CoeffMatrix mu1(p);
        M2L_reference(omega2, o1big, mu1, o2big, p);
        mu1.populate_lower();
        CoeffMatrix mu1x(p);
        Real3 m2l = o1big - o2big;
        M2L_reference_DONOTUSE(omega2, m2l, mu1x, p);
        std::cout << "M2L " << maxdelta(mu1, mu1x, p) << std::endl;
        dump(mu1, pd);

        // @o2big
        CoeffMatrix mu1_p3(p);
        M2Lp3(omega2, o1big, mu1_p3, o2big, p, scratch);
        mu1_p3.populate_lower();
            std::cout << "M2L_p3 " << maxdelta_col(mu1, mu1_p3, p) << std::endl;
        dump(mu1_p3, pd);


        CoeffMatrix mu1d(p);
        P2L(charge, ptcl1, mu1d, o2big, p);
        mu1d.populate_lower();
        std::cout << "P2L' " << maxdelta(mu1, mu1d, p) << std::endl;
        dump(mu1d, pd);

        // @o2sml
        CoeffMatrix mu2(p);
        L2L(mu1, o2big, mu2, o2sml, p);
        mu2.populate_lower();
        std::cout << "L2L" << std::endl;
        dump(mu2, pd);

        CoeffMatrix mu2p3(p);
        L2Lp3(mu1, o2big, mu2p3, o2sml, p, scratch);
        mu2p3.populate_lower();
        std::cout << "L2L p^3 " << maxdelta_col(mu2, mu2p3, p) << std::endl;
        dump(mu2p3, pd);

        // @o2sml
        CoeffMatrix mu3(p);
        L2L(mu2, o2sml, mu3, ptcl2, p);
        mu3.populate_lower();
        std::cout << "L2L2L " << maxdelta(mu, mu3, p) << std::endl;
        dump(mu3, pd);

        CoeffMatrix mu3d(p);
        L2L(mu1, o2big, mu3d, ptcl2, p);
        mu3d.populate_lower();
        std::cout << "L2L' " << maxdelta(mu, mu3d, p) << " " << maxdelta(mu3, mu3d, p) << std::endl;
        dump(mu3d, pd);

    }
    else if (0)
    {
        const size_t p = 10;
        const size_t ws = 1;
        Real33 abc(e1, e2, e3);
        const ssize_t outer = ws * (2 * ws + 1) + ws;
        const ssize_t inner = ws;
        std::cout << "L_*(P2L)" << std::endl;
        CoeffMatrix l_star1(p);
        size_t lsboxes1 = L_star_layer_sum_reference<true, true, true>(abc, inner, outer, l_star1, p);
        std::cout << "l_star made from " << lsboxes1 << " boxes" << std::endl;
        dump(l_star1, p);
        std::cout << "L_*(Lattice3Dcubic)" << std::endl;
        CoeffMatrix l_star2(p);
        size_t lsboxes2 = Lattice3Dcubic_layer_sum<true, true, true>(abc, inner, outer, l_star2, p);
        std::cout << "l_star made from " << lsboxes2 << " boxes" << std::endl;
        dump(l_star2, p);
    }
    else if (0)
    {
        const size_t p = 10;
        const size_t ws = 1;
        const size_t maxlayers = 64;
        double maxerror = std::numeric_limits<double>::min();
        bool unitcube = true;
        Real33 abc(e1, e2, e3);
        size_t ly;

        std::cout << "p=" << p << "  ws=" << ws << "  maxlayers=" << maxlayers << std::endl;

        CoeffMatrix lattice1(2 * p);
        ly = LatticeOperator1D(abc, ws, lattice1, p, 1., maxlayers, maxerror);
        std::cout << "lattice1D(even,even) for p = " << p << "  layers=" << ly << std::endl;
        dumpevenreal(lattice1, 2 * p);
        if (unitcube && ws == 1) {
            CoeffMatrix L1Dtable(2 * p, LatticeSums_1D);
            std::cout << "maxreldelta: " << maxdelta(L1Dtable, lattice1, p) << std::endl;
        }
        if (ws == 2) {
            // add missing layer to turn lattice operator for ws=2 into lattice operator for ws=1
            L_star_layer_sum_reference<false, false, true>(abc, 1, 2, lattice1, 2 * p);
            dumpevenreal(lattice1, 2 * p);
        }

        CoeffMatrix lattice2(2 * p);
        ly = LatticeOperator2D(abc, ws, lattice2, p, 1., maxlayers, maxerror);
        std::cout << "lattice2D(even,even) for p = " << p << "  layers=" << ly << std::endl;
        dumpevenreal(lattice2, 2 * p);
        if (unitcube && ws == 1) {
            CoeffMatrix L2Dtable(2 * p, LatticeSums_2D);
            std::cout << "maxreldelta: " << maxdelta(L2Dtable, lattice2, p) << std::endl;
        }
        if (ws == 2) {
            // add missing layer to turn lattice operator for ws=2 into lattice operator for ws=1
            L_star_layer_sum_reference<true, true, false>(abc, 1, 2, lattice2, 2 * p);
            dumpevenreal(lattice2, 2 * p);
        }

        CoeffMatrix lattice3(2 * p);
        ly = LatticeOperator3D(abc, ws, lattice3, p, 1., maxlayers, maxerror);
        std::cout << "lattice3D(even,even) for p = " << p << "  layers=" << ly << std::endl;
        dumpevenreal(lattice3, 2 * p);
        if (unitcube && ws == 1) {
            CoeffMatrix L3Dtable(2 * p, LatticeSums_3D);
            std::cout << "maxreldelta: " << maxdelta(L3Dtable, lattice3, p) << std::endl;
        }
        if (ws == 2) {
            // add missing layer to turn lattice operator for ws=2 into lattice operator for ws=1
            L_star_layer_sum_reference<true, true, true>(abc, 1, 2, lattice3, 2 * p);
            dumpevenreal(lattice3, 2 * p);
        }
    }
    else if (0)
    {
        const size_t p = 85;
        const size_t ws = 2;
        const size_t maxlayers = 64;
        double maxerror = std::numeric_limits<double>::min();
        size_t convergeat = 4;
        Real33 abc(e1, e2, e3);
        size_t ly;

        CoeffMatrix lattice3d(2 * p);
        ly = LatticeOperatorInternal<true, true, true>(abc, ws, lattice3d, 2 * p, 1., maxlayers, convergeat, maxerror);
        std::cout << "# base=" << abc.a() << abc.b() << abc.c() << "  p=" << p << "  ws=" << ws << "  maxlayers=" << maxlayers << "  maxerror@(" << convergeat << ",0)=" << maxerror << "  efflayers=" << ly << std::endl;
        dump_lmreim(lattice3d, 2 * p);
    }
    else if (0)
    {
        const size_t maxlayers = bless(64);
        size_t convergeat = 4;
        double maxerror = std::numeric_limits<double>::min();
        Real33 abc(e1, e2, e3);

        if (1) {
            // CPU vorglühen
            const size_t ws = 2;
            size_t p = 50;
            double start = timestamp();
            do {
                CoeffMatrix lattice3(2 * p);
                LatticeOperator3D(abc, bless(ws), lattice3, p, 1., bless(64), bless(1e-16));
                ++p;
            } while ((timestamp() - start) < 4.0);
        }

        size_t ps[] = {5, 10, 20, 30, 40, 50, 60};
        size_t wss[] = {1, 2, 3, 4, 5, 10};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t p = ps[ip];

                std::cout << std::endl << "p=" << p << "  ws=" << ws << "  maxlayers=" << maxlayers << std::endl;

                std::cout << "O_*" << std::endl;
                CoeffMatrix o_star(2 * p);
                pw.start();
                size_t osboxes = O_star_operator<true, true, true>(abc, ws, o_star, 2 * p);
                pw.stop();
                pw.report(1);
                o_star.populate_lower();
                std::cout << "o_star made from " << osboxes << " boxes" << std::endl;

                std::cout << "L_*" << std::endl;
                CoeffMatrix l_star(2 * p);
                pw.start();
                size_t lsboxes = L_star_operator<true, true, true>(abc, ws, l_star, 2 * p);
                pw.stop();
                pw.report(1);
                l_star.populate_lower();
                std::cout << "l_star made from " << lsboxes << " boxes" << std::endl;

                std::cout << "S_L" << std::endl;
                CoeffMatrix scaled(2 * p);
                pw.start();
                ScaleL_operator(l_star, ws, scaled, 2 * p);
                pw.stop();
                pw.report(1);

                std::cout << "tL_i" << std::endl;
                CoeffMatrix Li(2 * p);
                pw.start();
                size_t ly = LatticeOperatorIterate(o_star, l_star, ws, Li, 2 * p, maxlayers, convergeat, maxerror);
                pw.stop();
                pw.report(1);
                std::cout << "effective layers=" << ly << "  [maxerror@(" << convergeat << ",0)=" << maxerror << "]" << std::endl;
                std::cout << "L_i" << std::endl;
                pw.report(ly);
            }
    }
    else if (0)
    {
        std::cout << std::scientific;
        const size_t maxlayers = bless(64);
        double maxerror = std::numeric_limits<double>::min();
        Real33 abc(e1, e2, e3);

        size_t ps[] = {5, 10, 20, 30, 40, 50, 60};
        size_t wss[] = {1, 2, 3, 4, 5};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t ws_outer = ws * (2 * ws + 1) + ws;
                const size_t p = ps[ip];

                std::cout << std::endl << "p=" << p << "  ws=" << ws << "  maxlayers=" << maxlayers << std::endl;

                CoeffMatrix o_star(2 * p);
                O_star_operator<true, true, true>(abc, ws, o_star, 2 * p);

                CoeffMatrix l_star_reference(2 * p);
                CoeffMatrix l_star_adaptive(2 * p);
                L_star_layer_sum_reference<true, true, true>(abc, ws, ws_outer, l_star_reference, 2 * p);
                size_t lsboxes = L_star_layer_sum_adaptive<true, true, true>(abc, ws, ws_outer, l_star_adaptive, 2 * p);
                std::cout << "l_star made from " << lsboxes << " boxes" << std::endl;

                CoeffMatrix Li_reference(2 * p);
                CoeffMatrix Li_adaptive(2 * p);
                size_t lyr = LatticeOperatorIterate(o_star, l_star_reference, ws, Li_reference, 2 * p, maxlayers, 4, maxerror);
                size_t lya = LatticeOperatorIterate(o_star, l_star_adaptive, ws, Li_adaptive, 2 * p, maxlayers, 4, maxerror);
                std::cout << "effective layers=" << lyr << "," << lya << "  [maxerror(4,0)=" << maxerror << "]"
                    << " l_star made from " << lsboxes << " boxes,"
                    << std::endl
                    << " maxdelta(p)=" << maxdelta(Li_reference, Li_adaptive, p)
                    << " maxdelta(p/2)=" << maxdelta(Li_reference, Li_adaptive, p / 2)
                    << " maxdelta(p/3)=" << maxdelta(Li_reference, Li_adaptive, p / 3)
                    << " maxdelta(p/4)=" << maxdelta(Li_reference, Li_adaptive, p / 4)
                    << " maxdelta(p/10)=" << maxdelta(Li_reference, Li_adaptive, p / 10)
                    << " delta(4,0)=" << l_star_reference.get(4, 0) - l_star_adaptive.get(4, 0) << std::endl;
            }
    }
    else if (0)
    {
        const size_t maxlayers = bless(64);
        size_t convergeat = 4;
        double maxerror = std::numeric_limits<double>::min();
        Real33 abc(e1, e2, e3);

        size_t ps[] = {/*5, 10, 20, 30, 40, 50, */60};
        size_t wss[] = {1, 2/*, 3, 4, 5, 10*/};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t p = ps[ip];
                const ssize_t outer = ws * (2 * ws + 1) + ws;
                const ssize_t inner = ws;

                std::cout << std::endl << "p=" << p << "  ws=" << ws << "  maxlayers=" << maxlayers << std::endl;

                std::cout << "O_*4" << std::endl;
                EnergyStressTriangularArrayGroup<CoeffMatrix> o_star(2 * p);
                pw.start();
                size_t osboxes = O_star_es_cube_sum_generic_rowstore<true, true, true>(abc, ws, o_star, 2 * p);
                pw.stop();
                pw.report(1);
                o_star.populate_lower();
                std::cout << "o_star made from " << osboxes << " boxes" << std::endl;

                std::cout << "L_*4" << std::endl;
                EnergyStressTriangularArrayGroup<CoeffMatrix> l_star(2 * p);
                pw.start();
                size_t lsboxes = L_star_es_layer_sum_generic_rowstore<true, true, true>(abc, inner, outer, l_star, 2 * p);
                pw.stop();
                pw.report(1);
                l_star.populate_lower();
                std::cout << "l_star made from " << lsboxes << " boxes" << std::endl;

                std::cout << "S_L" << std::endl;
                CoeffMatrix scaled(2 * p);
                pw.start();
                ScaleL_operator(l_star.energy, ws, scaled, 2 * p);
                pw.stop();
                pw.report(1);

                std::cout << "tL_i4" << std::endl;
                EnergyStressTriangularArrayGroup<CoeffMatrix> Li(2 * p);
                pw.start();
                size_t ly = LatticeOperatorESIterate(o_star, l_star, ws, Li, 2 * p, maxlayers, 4, maxerror);
                pw.stop();
                pw.report(1);
                std::cout << "effective layers=" << ly << "  [maxerror@(" << convergeat << ",0)=" << maxerror << "]" << std::endl;
                std::cout << "L_i4" << std::endl;
                pw.report(ly);

#if 0
                EnergyStressTriangularArrayGroup<CoeffMatrix> Lid(2 * p);
                for (size_t maxlys = 1; maxlys <= maxlayers; ++maxlys) {
                    Li.zero(p);
                    size_t lys = LatticeOperatorESIterate(o_star, l_star, ws, Li, 2 * p, maxlys, 4, maxerror);

                    std::cout << "L_" << maxlys << " via Lattice (" << lys << " iterations):" << std::endl;
                    dump(Li.energy, 11);
                    dump(Li.stress.a, 11);
                    dump(Li.stress.b, 11);
                    dump(Li.stress.c, 11);

                    Lid.zero(p);
                    size_t inner = ws;
                    // outer = ((2ws+1)^(i+2)-1)/2
                    size_t outer = 1;
                    for (size_t i = 0; i < maxlys+2; ++i)
                        outer *= 2 * ws + 1;
                    outer = (outer - 1) / 2;
                    size_t lsboxes = L_star_es_layer_sum_generic_rowstore_vec<true, true, true>(abc, inner, outer, Lid, 2 * p);

                    std::cout << "L_" << maxlys << " via sum(P2L) (" << lsboxes << " boxes):" << std::endl;
                    dump(Lid.energy, 11);
                    dump(Lid.stress.a, 11);
                    dump(Lid.stress.b, 11);
                    dump(Lid.stress.c, 11);

                    if (lys <= maxlys)
                        break;
                }
#endif

            }
    }
    else if (0)
    {
        Real33 abc(e1, e2, e3);

        size_t ps[] = {1, 2, 3, 4, 5, 10, 20, 30, 40, 50};
        size_t wss[] = {15};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t p = 2 * ps[ip];

                std::cout << std::endl << "p=" << p << "  ws=" << ws << std::endl;

                std::cout << "O_*" << std::endl;
                CoeffMatrix o_star(p);
                pw.start();
                size_t osboxes = O_star_operator<true, true, true>(abc, ws, o_star, p);
                pw.stop();
                pw.report(osboxes * (p - 1) * p / 2);
                std::cout << "o_star made from " << osboxes << " boxes" << std::endl;

            }
    }
    else if (0)
    {
        Real33 abc(e1, e2, e3);

        size_t p = 50;
        size_t ws = 1;
        size_t outer = ws * (2 * ws + 1) + ws;

        std::cout << std::endl << "# p=" << p << "  ws=" << ws << std::endl;
        std::cout << "# P2L dump" << std::endl;

        CoeffMatrix l_star(p);
        L_star_cube_sum_reference_dump<true, true, true>(abc, ws, outer, l_star, p);
    }
    else if (0)
    {
        Real33 abc(e1, e2, e3);

        size_t ps[] = {1, 2, 3, 4, 5, 10, 20, 30, 40, 50};
        size_t wss[] = {15};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t p = 2 * ps[ip];
                size_t osboxes;

                std::cout << std::endl << "p=" << p << "  ws=" << ws << std::endl;

                std::cout << "O_* (cube_sum_reference)" << std::endl;
                CoeffMatrix o_star(p);
                pw.start();
                osboxes = O_star_cube_sum_reference<true, true, true>(abc, ws, o_star, p);
                pw.stop();
                pw.report(osboxes * (p - 1) * p / 2);
                std::cout << "o_star made from " << osboxes << " boxes" << std::endl;

                CoeffMatrix o_star2(p);
                EnergyStressTriangularArrayGroup<CoeffMatrix> o_star_es(p);

                std::cout << "O_* (cube_sum_generic_colstore)" << std::endl;
                pw.start();
                osboxes = O_star_cube_sum_generic_colstore<true, true, true>(abc, ws, o_star2, p);
                pw.stop();
                pw.report(osboxes * (p - 1) * p / 2);
                std::cout << "o_star made from " << osboxes << " boxes " << maxdelta(o_star, o_star2, p) << std::endl;

                std::cout << "O_* (cube_sum_generic_rowstore)" << std::endl;
                pw.start();
                osboxes = O_star_cube_sum_generic_rowstore<true, true, true>(abc, ws, o_star2, p);
                pw.stop();
                pw.report(osboxes * (p - 1) * p / 2);
                std::cout << "o_star made from " << osboxes << " boxes " << maxdelta(o_star, o_star2, p) << std::endl;

                std::cout << "O_* (cube_sum_rowit)" << std::endl;
                pw.start();
                osboxes = O_star_cube_sum_rowit<true, true, true>(abc, ws, o_star2, p);
                pw.stop();
                pw.report(osboxes * (p - 1) * p / 2);
                std::cout << "o_star made from " << osboxes << " boxes " << maxdelta(o_star, o_star2, p) << std::endl;

                std::cout << "O_* (cube_sum_generic_colstore_scaled)" << std::endl;
                pw.start();
                osboxes = O_star_cube_sum_generic_colstore_scaled<true, true, true>(abc, ws, o_star2, p);
                pw.stop();
                pw.report(osboxes * (p - 1) * p / 2);
                std::cout << "o_star made from " << osboxes << " boxes " << maxdelta(o_star, o_star2, p) << std::endl;

                std::cout << "O_*4 (cube_sum_generic_colstore)" << std::endl;
                pw.start();
                osboxes = O_star_es_cube_sum_generic_colstore<true, true, true>(abc, ws, o_star_es, p);
                pw.stop();
                pw.report(osboxes * (p - 1) * p / 2);
                std::cout << "o_star made from " << osboxes << " boxes " << maxdelta(o_star, o_star_es.energy, p) << std::endl;

                std::cout << "O_*4 (cube_sum_generic_rowstore)" << std::endl;
                pw.start();
                osboxes = O_star_es_cube_sum_generic_rowstore<true, true, true>(abc, ws, o_star_es, p);
                pw.stop();
                pw.report(osboxes * (p - 1) * p / 2);
                std::cout << "o_star made from " << osboxes << " boxes " << maxdelta(o_star, o_star_es.energy, p) << std::endl;

            }
    }
    else if (0)
    {
        Real33 abc(e1, e2, e3);

        size_t ps[] = {1, 2, 3, 4, 5, 10, 20, 30, 40, 50};
        size_t wss[] = {3};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t p = 2 * ps[ip];

                std::cout << std::endl << "p=" << p << "  ws=" << ws << std::endl;

                std::cout << "L_*" << std::endl;
                CoeffMatrix l_star(p);
                pw.start();
                size_t lsboxes = L_star_operator<true, true, true>(abc, ws, l_star, p);
                pw.stop();
                pw.report(lsboxes * (p - 1) * p / 2);
                std::cout << "l_star made from " << lsboxes << " boxes" << std::endl;
            }
    }
    else if (0)
    {
        Real33 abc(e1, e2, e3);

        size_t ps[] = {1, 2, 3, 4, 5, 10, 20, 30, 40, 50};
        size_t wss[] = {3};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t p = 2 * ps[ip];
                const ssize_t outer = ws * (2 * ws + 1) + ws;
                const ssize_t inner = ws;
                size_t lsboxes;

                CoeffMatrix l_star(p);
                CoeffMatrix l_star2(p);
                EnergyStressTriangularArrayGroup<CoeffMatrix> l_star_es(p);

                std::cout << std::endl << "p=" << p << "  ws=" << ws << std::endl;

                std::cout << "L_*" << std::endl;
                pw.start();
                lsboxes = L_star_operator<true, true, true>(abc, ws, l_star, p);
                pw.stop();
                pw.report(lsboxes * (p - 1) * p / 2);
                std::cout << "l_star made from " << lsboxes << " boxes" << std::endl;

                std::cout << "L_* (layer_sum_reference)" << std::endl;
                pw.start();
                lsboxes = L_star_layer_sum_reference<true, true, true>(abc, inner, outer, l_star, p);
                pw.stop();
                pw.report(lsboxes * (p - 1) * p / 2);
                std::cout << "L_star made from " << lsboxes << " boxes" << std::endl;

                std::cout << "L_* (layer_sum_generic_colstore)" << std::endl;
                pw.start();
                lsboxes = L_star_layer_sum_generic_colstore<true, true, true>(abc, inner, outer, l_star2, p);
                pw.stop();
                pw.report(lsboxes * (p - 1) * p / 2);
                std::cout << "l_star made from " << lsboxes << " boxes " << maxdelta(l_star, l_star2, p) << std::endl;

                std::cout << "L_* (layer_sum_generic_rowstore)" << std::endl;
                pw.start();
                lsboxes = L_star_layer_sum_generic_rowstore<true, true, true>(abc, inner, outer, l_star2, p);
                pw.stop();
                pw.report(lsboxes * (p - 1) * p / 2);
                std::cout << "l_star made from " << lsboxes << " boxes " << maxdelta(l_star, l_star2, p) << std::endl;

                std::cout << "L_*4 (layer_sum_generic_colstore)" << std::endl;
                pw.start();
                lsboxes = L_star_es_layer_sum_generic_colstore<true, true, true>(abc, inner, outer, l_star_es, p);
                pw.stop();
                pw.report(lsboxes * (p - 1) * p / 2);
                std::cout << "l_star made from " << lsboxes << " boxes " << maxdelta(l_star, l_star_es.energy, p) << std::endl;

                std::cout << "L_*4 (layer_sum_generic_rowstore)" << std::endl;
                pw.start();
                lsboxes = L_star_es_layer_sum_generic_rowstore<true, true, true>(abc, inner, outer, l_star_es, p);
                pw.stop();
                pw.report(lsboxes * (p - 1) * p / 2);
                std::cout << "l_star made from " << lsboxes << " boxes " << maxdelta(l_star, l_star_es.energy, p) << std::endl;

            }
    }
    else if (0)
    {
        Real33 abc(e1, e2, e3);

        size_t ps[] = {1, 2, 3, 4, 5, 10, 20, 30, 40, 50};
        size_t wss[] = {1, 2, 3};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t ws_outer = ws * (2 * ws + 1) + ws;
                const size_t p = 2 * ps[ip];

                std::cout << std::endl << "p=" << p << "  ws=" << ws << std::endl;

                std::cout << "L_*" << std::endl;
                CoeffMatrix l_star_reference(p);
                pw.start();
                size_t lsboxes = L_star_layer_sum_reference<true, true, true>(abc, ws, ws_outer, l_star_reference, p);
                pw.stop();
                pw.report(lsboxes * (p - 1) * p / 2);
                CoeffMatrix l_star_adaptive(p);
                pw.start();
                lsboxes = L_star_layer_sum_adaptive<true, true, true>(abc, ws, ws_outer, l_star_adaptive, p);
                pw.stop();
                pw.report(lsboxes * (p - 1) * p / 2);
                std::cout << std::scientific;
                std::cout << "l_star made from " << lsboxes << " boxes,"
                    << " maxdelta(p)=" << maxdelta(l_star_reference, l_star_adaptive, p)
                    << " maxdelta(p/2)=" << maxdelta(l_star_reference, l_star_adaptive, p / 2)
                    << " maxdelta(p/3)=" << maxdelta(l_star_reference, l_star_adaptive, p / 3)
                    << " maxdelta(p/4)=" << maxdelta(l_star_reference, l_star_adaptive, p / 4)
                    << " delta(4,0)=" << l_star_reference.get(4, 0) - l_star_adaptive.get(4, 0) << std::endl;
            }
    }
    else if (0)
    {
        Real33 abc(e1, e2, e3);

        size_t ps[] = {1, 2, 3, 4, 5, 10, 20, 30, 40, 50};
        size_t wss[] = {1};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t p = 2 * ps[ip];
                const size_t reps = 100;

                CoeffMatrix o_star(p);
                O_star_operator<true, true, true>(abc, ws, o_star, p);
                CoeffMatrix l_star(p);
                L_star_operator<true, true, true>(abc, ws, l_star, p);

                std::cout << std::endl << "p=" << p << "  ws=" << ws << std::endl;

                std::cout << "rawM2L(omega(p),B(p))" << std::endl;
                CoeffMatrix Li(p);
                pw.start();
                for (size_t i = 0; i < reps; ++i) {
                    single_rawM2L(o_star, l_star, Li, p);
                }
                pw.stop();
                pw.report(reps * (p + 3) * (p + 1) * (p + 2) * (p + 2) / 12);
            }
    }
    else if (0)
    {
        Real33 abc(e1, e2, e3);

        size_t ps[] = {1, 2, 3, 4, 5, 10, 20, 30, 40, 50};
        size_t wss[] = {1};

        for (size_t iws = 0; iws < sizeof(wss) / sizeof(wss[0]); ++iws)
            for (size_t ip = 0; ip < sizeof(ps) / sizeof(ps[0]); ++ip)
            {
                const size_t ws = wss[iws];
                const size_t p = 2 * ps[ip];
                const size_t reps = 100;

                CoeffMatrix o_star(p);
                O_star_operator<true, true, true>(abc, ws, o_star, p);
                CoeffMatrix l_star(2 * p);
                L_star_operator<true, true, true>(abc, ws, l_star, 2 * p);

                std::cout << std::endl << "p=" << p << "  ws=" << ws << std::endl;

                std::cout << "rawM2L(omega(p),B(2*p))" << std::endl;
                CoeffMatrix Li(p);
                pw.start();
                for (size_t i = 0; i < reps; ++i) {
                    single_rawM2L(o_star, l_star, Li, p);
                }
                pw.stop();
                pw.report(reps * (p + 1) * (p + 1) * (p + 1) * (p + 2) / 2);
            }
    }
    else if (0)
    {
        size_t p = 6;
        Real3 loc1(0.5, 0.5, 0.5);
        Real3 loc2(2.25, 2.25, 2.25);

        CoeffMatrix omega(p);
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc1, 0, testdata::n, omega, p);
        omega.populate_lower();

        CoeffMatrix mu(p);
        P2L(testdata::x, testdata::y, testdata::z, testdata::q, loc2, 0, testdata::n, mu, p);
        mu.populate_lower();

        CoeffMatrix domegax(p);
        CoeffMatrix domegay(p);
        CoeffMatrix domegaz(p);
        CoeffMatrix dmux(p);
        CoeffMatrix dmuy(p);
        CoeffMatrix dmuz(p);

        pw.start();
        DxL2L_reference(mu, dmux, p);
        DyL2L_reference(mu, dmuy, p);
        DzL2L_reference(mu, dmuz, p);
        DxM2M_reference(omega, domegax, p);
        DyM2M_reference(omega, domegay, p);
        DzM2M_reference(omega, domegaz, p);
        pw.stop();
        pw.report(1);

        p = 6;
        std::cout << "mu" << std::endl;
        dump(mu, p);
        dump(dmux, p);
        dump(dmuy, p);
        dump(dmuz, p);
        std::cout << "omega" << std::endl;
        dump(omega, p);
        dump(domegax, p);
        dump(domegay, p);
        dump(domegaz, p);
    }
    else if (0)
    {
        std::vector<Real3> Fdm(testdata::n), Fdm2(testdata::n), Fdo(testdata::n), Fdo2(testdata::n), Fs(testdata::n);

        ssize_t p = bless(50);
        Real3 loc2(2.25, 2.25, 2.25);
        Real3 ref_point(0, 0, 0);
        P2M_divtable<Real> div_lut_p(p);

        CoeffMatrix mu(p);
        P2L(testdata::x, testdata::y, testdata::z, testdata::q, loc2, 0, testdata::n, mu, p);
        mu.populate_lower();

        CoeffMatrix dmux(p);
        CoeffMatrix dmuy(p);
        CoeffMatrix dmuz(p);

        DxL2L_reference(mu, dmux, p);
        DyL2L_reference(mu, dmuy, p);
        DzL2L_reference(mu, dmuz, p);
        dmux.populate_lower();
        dmuy.populate_lower();
        dmuz.populate_lower();

        std::cout << "F: d(mu)/dxyz, n=" << testdata::n << " p=" << p << std::endl;
        pw.start();
        for (size_t i = 0; i < testdata::n; ++i) {
            Real3 xyz(testdata::x[i], testdata::y[i], testdata::z[i]);
            Real q = testdata::q[i];
            CoeffMatrix omega(p);
			P2M(q, xyz, omega, ref_point, div_lut_p);
			omega.populate_lower();

            CoeffMatrix::complex_type Fx(0.);
            CoeffMatrix::complex_type Fy(0.);
            CoeffMatrix::complex_type Fz(0.);

			for (ssize_t l = 0; l <= p; ++l) {
				for (ssize_t m = -l; m <= l; ++m) {
                    CoeffMatrix::complex_type omega_lm = omega.get(l, m);
					Fx += quickcomplexmult(dmux.get(l, m), omega_lm);
					Fy += quickcomplexmult(dmuy.get(l, m), omega_lm);
					Fz += quickcomplexmult(dmuz.get(l, m), omega_lm);
				}
			}

			Fdm[i] = Real3(Fx.real(), Fy.real(), Fz.real());
        }
        pw.stop();
        pw.report(testdata::n);

        std::cout << "F: d(mu)/dxyz, upper, n=" << testdata::n << " p=" << p << std::endl;
        pw.start();
        for (size_t i = 0; i < testdata::n; ++i) {
            Real3 xyz(testdata::x[i], testdata::y[i], testdata::z[i]);
            Real q = testdata::q[i];
            CoeffMatrix omega(p);
			P2M(q, xyz, omega, ref_point, div_lut_p);
			//omega.populate_lower();

            Real Fx(0.);
            Real Fy(0.);
            Real Fz(0.);

			for (ssize_t l = 0; l <= p; ++l) {
                CoeffMatrix::complex_type omega_l0 = omega.upper(l, 0);
                Fx += realonlycomplexmult(dmux.upper(l, 0), omega_l0);
                Fy += realonlycomplexmult(dmuy.upper(l, 0), omega_l0);
                Fz += realonlycomplexmult(dmuz.upper(l, 0), omega_l0);
				for (ssize_t m = 1; m <= l; ++m) {
                    //CoeffMatrix::complex_type omega_lm = omega.upper(l, m) * 2.;
                    CoeffMatrix::complex_type omega_lm = omega.upper(l, m);
                    omega_lm += omega_lm;
					Fx += realonlycomplexmult(dmux.upper(l, m), omega_lm);
					Fy += realonlycomplexmult(dmuy.upper(l, m), omega_lm);
					Fz += realonlycomplexmult(dmuz.upper(l, m), omega_lm);
				}
			}

			Fdm2[i] = Real3(Fx, Fy, Fz);
        }
        pw.stop();
        pw.report(testdata::n);

        CoeffMatrix omega(p);
        CoeffMatrix domegax(p);
        CoeffMatrix domegay(p);
        CoeffMatrix domegaz(p);
        std::cout << "F: d(omega)/dxyz, n=" << testdata::n << " p=" << p << std::endl;
        pw.start();
        for (size_t i = 0; i < testdata::n; ++i) {
            Real3 xyz(testdata::x[i], testdata::y[i], testdata::z[i]);
            Real q = testdata::q[i];

            omega.zero();
			P2M(q, xyz, omega, ref_point, div_lut_p);
			omega.populate_lower();

			DxM2M_reference(omega, domegax, p);
			DyM2M_reference(omega, domegay, p);
			DzM2M_reference(omega, domegaz, p);
			domegax.populate_lower();
			domegay.populate_lower();
			domegaz.populate_lower();

            CoeffMatrix::complex_type Fx(0.);
            CoeffMatrix::complex_type Fy(0.);
            CoeffMatrix::complex_type Fz(0.);

			for (ssize_t l = 0; l <= p; ++l) {
				for (ssize_t m = -l; m <= l; ++m) {
                    CoeffMatrix::complex_type mu_lm = mu.get(l, m);
					Fx += quickcomplexmult(domegax.get(l, m), mu_lm);
					Fy += quickcomplexmult(domegay.get(l, m), mu_lm);
					Fz += quickcomplexmult(domegaz.get(l, m), mu_lm);
				}
			}

			Fdo[i] = Real3(-Fx.real(), -Fy.real(), -Fz.real());
        }
        pw.stop();
        pw.report(testdata::n);

        std::cout << "F: d(omega)/dxyz, upper, n=" << testdata::n << " p=" << p << std::endl;
        pw.start();
        for (size_t i = 0; i < testdata::n; ++i) {
            Real3 xyz(testdata::x[i], testdata::y[i], testdata::z[i]);
            Real q = testdata::q[i];

            omega.zero();
			P2M(q, xyz, omega, ref_point, div_lut_p);
			omega.populate_lower();

			DxM2M_reference(omega, domegax, p);
			DyM2M_reference(omega, domegay, p);
			DzM2M_reference(omega, domegaz, p);

            CoeffMatrix::complex_type Fx(0.);
            CoeffMatrix::complex_type Fy(0.);
            CoeffMatrix::complex_type Fz(0.);

			for (ssize_t l = 0; l <= p; ++l) {
                CoeffMatrix::complex_type mu_l0 = mu.get(l, 0);
                Fx += quickcomplexmult(domegax.upper(l, 0), mu_l0);
                Fy += quickcomplexmult(domegay.upper(l, 0), mu_l0);
                Fz += quickcomplexmult(domegaz.upper(l, 0), mu_l0);
				for (ssize_t m = 1; m <= l; ++m) {
                    CoeffMatrix::complex_type mu_lm = mu.upper(l, m) * Real(2.);
					Fx += quickcomplexmult(domegax.upper(l, m), mu_lm);
					Fy += quickcomplexmult(domegay.upper(l, m), mu_lm);
					Fz += quickcomplexmult(domegaz.upper(l, m), mu_lm);
				}
			}

			Fdo2[i] = Real3(-Fx.real(), -Fy.real(), -Fz.real());
        }
        pw.stop();
        pw.report(testdata::n);

        std::cout << "F: shift mu, n=" << testdata::n << " p=" << p << std::endl;
        pw.start();
        CoeffMatrix mu_shift(1);
        for (size_t i = 0; i < testdata::n; ++i) {
            Real3 xyz(testdata::x[i], testdata::y[i], testdata::z[i]);
            Real q = testdata::q[i];

            mu_shift.zero();
            // shift mu from (0,0,0) to (x,y,z)
            L2L(mu, ref_point, mu_shift, xyz, p, 1);

            Real Fx = mu_shift.upper(1, 1).real();
            Real Fy = mu_shift.upper(1, 1).imag();
            Real Fz = mu_shift.upper(1, 0).real();
			Fs[i] = Real3(Fx, Fy, Fz) * q;
        }
        pw.stop();
        pw.report(testdata::n);

        Real3 sFdm, sFdm2, sFdo, sFdo2, sFs;
        for (size_t i = 0; i < testdata::n; ++i) {
            sFdm += Fdm[i];
            sFdm2 += Fdm2[i];
            sFdo += Fdo[i];
            sFdo2 += Fdo2[i];
            sFs += Fs[i];
        }
        std::cout << sFdm-sFdm2 << sFdo-sFdo2 << sFdm-sFdo << sFdo-sFs << sFs-sFdm << std::endl;
    }
    else if (0)
    {
        const size_t maxlayers = bless(64);
        size_t convergeat = 4;
        double maxerror = std::numeric_limits<double>::min();
        const size_t ws = bless(1);
        const size_t p = bless(50);

        Real q1(.150078125e2);
        Real3 xyz1(.965, .222, .344);
        Real q2(-.17666015625e2);
        Real3 xyz2(.325, .767, .277);
        Real q3(.2658203125e1);
        Real3 xyz3(.965, .346, .889);
        Real33 abc(e1, e2, e3);
        Real3 ref_point = abc.center();

        std::cout << std::endl << "REF3PARTICLES  p=" << p << "  ws=" << ws << "  maxlayers=" << maxlayers << std::endl;
        std::cout << std::setprecision(12);
        std::cout << xyz1 << " " << q1 << std::endl;
        std::cout << xyz2 << " " << q2 << std::endl;
        std::cout << xyz3 << " " << q3 << std::endl;

        CoeffMatrix omega(p);
        P2M_nodivtable(q1, xyz1, omega, ref_point, p);
        P2M_nodivtable(q2, xyz2, omega, ref_point, p);
        P2M_nodivtable(q3, xyz3, omega, ref_point, p);
        omega.populate_lower();

        std::cout << "omega" << std::endl;
        dump_lmreim(omega, p);

        const ssize_t outer = ws * (2 * ws + 1) + ws;
        const ssize_t inner = ws;

        EnergyStressTriangularArrayGroup<CoeffMatrix> o_star(2 * p);
        size_t osboxes = O_star_es_cube_sum_generic_rowstore<true, true, true>(abc, ws, o_star, 2 * p);
        o_star.populate_lower();
        std::cout << "o_star made from " << osboxes << " boxes" << std::endl;

        EnergyStressTriangularArrayGroup<CoeffMatrix> l_star(2 * p);
        size_t lsboxes = L_star_es_layer_sum_generic_rowstore<true, true, true>(abc, inner, outer, l_star, 2 * p);
        l_star.populate_lower();
        std::cout << "l_star made from " << lsboxes << " boxes" << std::endl;

        EnergyStressTriangularArrayGroup<CoeffMatrix> Li(2 * p);
        size_t ly = LatticeOperatorESIterate(o_star, l_star, ws, Li, 2 * p, maxlayers, 4, maxerror);
        Li.energy.zero(3);
        Li.stress.a.zero(2);
        Li.stress.a.upper(3, 0) = 0;
        Li.stress.b.zero(2);
        Li.stress.b.upper(3, 0) = 0;
        Li.stress.c.zero(2);
        Li.stress.c.upper(3, 0) = 0;
        Li.populate_lower();
        std::cout << "effective layers=" << ly << "  [maxerror@(" << convergeat << ",0)=" << maxerror << "]" << std::endl;

        EnergyStressTriangularArrayGroup<CoeffMatrix> mu(p);
        single_M2L(omega, Li.energy, mu.energy, p);
        single_M2L(omega, Li.stress.a, mu.stress.a, p);
        single_M2L(omega, Li.stress.b, mu.stress.b, p);
        single_M2L(omega, Li.stress.c, mu.stress.c, p);

        std::cout << "mu.energy" << std::endl;
        dump_lmreim(mu.energy, p);
        std::cout << "mu.stress.a" << std::endl;
        dump_lmreim(mu.stress.a, p);
        std::cout << "mu.stress.b" << std::endl;
        dump_lmreim(mu.stress.b, p);
        std::cout << "mu.stress.c" << std::endl;
        dump_lmreim(mu.stress.c, p);

        std::cout << "omega * mu.energy " << multiply(omega, mu.energy, p) << std::endl;

        CoeffMatrix L(2 * p);
        LatticeOperator3D(abc, ws, L, p, 1., 42);
        CoeffMatrix dL(2 * p);
        trianglediff(dL, Li.energy, L, 2 * p);
        dump_lmreim(dL, 10);
        dump_lmreim(L, 4);

        L2L_Operator<CoeffMatrix> C1(ref_point, xyz1, p);
        std::cout << "L2L Operator #1 " << xyz1 << std::endl;
        dump_lmreim(C1, p);
        L2L_Operator<CoeffMatrix> C2(ref_point, xyz2, p);
        std::cout << "L2L Operator #2 " << xyz2 << std::endl;
        dump_lmreim(C2, p);
        L2L_Operator<CoeffMatrix> C3(ref_point, xyz3, p);
        std::cout << "L2L Operator #3 " << xyz3 << std::endl;
        dump_lmreim(C3, p);
    }
    else if (0)
    {
        Real3 ptcl(0.1, 0.3, 0.2);
        Real charge(1.2345);
        Real3 loc1(10, 11, 12);
        Real3 loc2 = loc1 + Real3(0, 0, 2.3456);
        const size_t p = 20;
        const size_t pd = std::min(p, size_t(8));
        scratch_type scratch(p);

        CoeffMatrix mu1(p);
        P2L(charge, ptcl, mu1, loc1, p);
        mu1.populate_lower();

        CoeffMatrix mu2p4(p);
        L2L(mu1, loc1, mu2p4, loc2, p);
        std::cout << "L2L p^4" << std::endl;
        dump(mu2p4, pd);

        CoeffMatrix mu2p3(p);
        L2Lp3(mu1, loc1, mu2p3, loc2, p, scratch);
        std::cout << "L2L p^3 " << maxdelta(mu2p4, mu2p3, p) << std::endl;
        dump(mu2p3, pd);
    }
    else if (0)
    {
        typedef RotationMatrixUpper<Real> rotation_matrix;
        typedef RotationMatrixVector<rotation_matrix> rotation_matrix_vector;
        typedef RG::RotationGeneratorNoPrecomp<rotation_matrix_vector, Real3> rotation_generator;
        typedef typename rotation_generator::rotation_operator rotation_operator;

        const size_t p = 7;
        CoeffMatrix omega(p);
        Real3 loc(0, 0, 0);
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, loc, 0, testdata::n, omega, p);
        omega.populate_lower();
        std::cout << "omega(" << testdata::n << " particles) @ " << loc << std::endl;
        dump(omega, p);

        Real3 there(1, 2, 3);
        rotation_generator RGen(p);
        rotation_operator R = RGen.generate_rotation(there);
        rotation_operator mR = RGen.generate_rotation(-there);
        CoeffMatrix omega_R(p);
        CoeffMatrix omega_mR(p);
        ApplyRotationMultipoleForward(omega, R, omega_R, p);
        ApplyRotationMultipoleForward(omega, mR, omega_mR, p);

        std::cout << "omega', rotated towards " << there << std::endl;
        dump(omega_R, p);
        std::cout << "omega', rotated towards " << -there << std::endl;
        dump(omega_mR, p);
    }
    else if (subtest == "M2Lp3_impl_folded_dump")
    {
        typedef RotationMatrixUpper<Real> rotation_matrix;
        typedef RotationMatrixVector<rotation_matrix> rotation_matrix_vector;
        typedef RG::RotationGeneratorNoPrecomp<rotation_matrix_vector, Real3> rotation_generator;
        typedef typename rotation_generator::rotation_operator rotation_operator;

        const size_t p = 50;
        CoeffMatrix omega(p);
        Real3 cntr(0.5, 0.5, 0.5);
        P2M(testdata::x, testdata::y, testdata::z, testdata::q, cntr, 0, testdata::n, omega, p);
        omega.populate_lower();

        Real3 loc(0, 0, 0);
        Real3 there(1, 2, 3);
        rotation_generator RGen(p);
        rotation_operator R = RGen.generate_rotation(there);
        Real shift_distance = length(there - loc);
        M2Lp3_Operator<Real> B(shift_distance, p);
        CoeffMatrix mu(p);
        scratch_type scratch(p);

        std::cout << "# omega(" << testdata::n << " particles) @ " << loc << "  ==> mu @ " << there << std::endl;
        std::cout << std::showpos << std::scientific
                  << std::setprecision(std::numeric_limits<Real>::max_digits10 + 1);
        M2Lp3_impl_folded_dump<simd_engine_default>(omega, R, B, mu, p, scratch);
    }
    else if (0)
    {
        size_t p = get_env_int("MULTIPOLEORDER", 6);
        size_t igrid = get_env_int("IGRID", 33);
        CoeffMatrix omega_grid(p);
        Real boxlength = 1.0;
        Real charge = 1.0;

        fmsolvr::make_grid_omega(igrid, boxlength, charge, omega_grid, p);
        std::cout << "# ω(grid(" << igrid << "))" << std::endl;
        dump_lmreim(omega_grid, p);
    }
    else if (0)
    {
        size_t p = get_env_int("MULTIPOLEORDER", 20);
        CoeffMatrix omega_homogeneous(p);

        fmsolvr::make_unitbox_homogeneous_charge_distribution_multipole_expansion(
                omega_homogeneous, p);
        std::cout << "# ω(homogeneous)" << std::endl;
        dump_lmreim(omega_homogeneous, p, false);
    }

    return 0;
}

}  // namespace fmsolvr

int main(int argc, char **argv)
{
    return fmsolvr::main(argc, argv);
}

// vim: et:ts=4:sw=4
