#ifndef _DaHa_DependencyCounter_
#define _DaHa_DependencyCounter_

#include "DependencyCounterDetails.hpp"
#include "EventDispatcher/sdfd.hpp"
#include "fmsolvr/tree/denseoctree.hpp"

#include <atomic>
#include <cstdint>
#include <tuple>

template <typename event_dispatcher, typename fmm_handle_type>
class DependencyCounters
{
public:
    using boxid_type = typename fmm_handle_type::boxid_type;

    using counter_tuple = typename event_dispatcher::dependency_counter_tuple_type;
    using DependencyCounterTreeType =
            fmsolvr::OctreeStorageDense<counter_tuple_wrapper<counter_tuple>>;

    explicit DependencyCounters(fmm_handle_type &fmm_handle)
            : force_computation_counter(fmm_handle.particle_handle().n_active() * 2),
              fmm_handle_(fmm_handle),
              m2l_interactions(fmm_handle_.m2l_interaction_count()),
              dependency_counter_tree_(fmm_handle.depth())

    {
        Reset();
    }

    void Init()
    {
        if (fmm_handle_.simulationbox().open_boundary_conditions())
        {
            InitOpenBoundary();
        }
        else
        {
            InitPeriodic();
        }
    }

    void InitOpenBoundary()
    {
        // If depth is smaller 2, there is no far field force
        if (fmm_handle_.depth() < 2)
        {
            force_computation_counter = fmm_handle_.particle_handle().n_active();
        }
        else
        {
            // for d=2 there is no l2l from parent in open boundary
            for (unsigned bid = 0; bid < size_t(1) << (3 * 2); ++bid)
            {
                auto &tuple = dependency_counter_tree_[boxid_type(bid, 2)];
                dependency_counter::details::get_counter<MU_Source>(tuple).DecrAndCheck();
                dependency_counter::details::get_counter<MU_Lowest>(tuple).DecrAndCheck();
            }
        }
    }
    void InitPeriodic()
    {
    }

    void Reset()
    {
        // reset all values to default
        force_computation_counter = fmm_handle_.particle_handle().n_active() * 2;
        for (unsigned d = 0; d <= fmm_handle_.depth(); ++d)
        {
            for (unsigned bid = 0; bid < size_t(1) << (3 * d); ++bid)
            {
                dependency_counter_tree_[boxid_type(bid, d)].Reset();
            }
        }

        // for d=0 lattice triggers partial mu only once
        // for d=0 there is no l2l from parent
        auto &tuple = dependency_counter_tree_[boxid_type(0, 0)];
        dependency_counter::details::get_counter<MU_Source>(tuple).counter = 1;
        dependency_counter::details::get_counter<MU_Lowest>(tuple).counter = 1;

        Init();
    }

    std::atomic<size_t> force_computation_counter;

    template <Events triggered_event, typename... Args>
    void dispatch(size_t bid, size_t d, Args &&... args)
    {
        auto &tuple = dependency_counter_tree_[boxid_type(bid, d)];
        auto &counter = dependency_counter::details::get_counter<triggered_event>(tuple);

        if (counter.DecrAndCheck())
        {
            event_dispatcher::template dispatch<BOX, triggered_event>(
                    bid, d, std::forward<Args>(args)...);
        }
    }


private:
    fmm_handle_type &fmm_handle_;
    const size_t m2l_interactions;
    DependencyCounterTreeType dependency_counter_tree_;
};


#endif
