#ifndef _sdfd_listener_
#define _sdfd_listener_


#include <cstddef>
#include <type_traits>
#include <utility>

#include "../EventDispatcherEnumerations.hpp"
#include "../CallbackImpl.hpp"

/**
 * The Listener consists of a callback and a quantity.
 * It is registered to a certain Event in the EventListener.
 * Calls to the Listeners dispatch method are filtered by the quantity.
 * Listeners are registered for a certain quantity and only executed if the configured quantity is
 * equal to the triggered.
 */
template <HierarchyLevels Quantity, EventHandlers Callback>
struct Listener
{
    /**
     * This dispatch method uses SFINAE.
     * With SFINAE we decide if the current quantity is equal to the configured quantity.
     * The following function is executed if the quantity is equal and action should follow.
     */

    template <HierarchyLevels triggered_hierarchy_level, Events triggered_event, typename... Args>
    static typename std::enable_if<triggered_hierarchy_level == Quantity, void>::type
    dispatch(Args &&... args)
    {
        CallbackImpl<Callback, Quantity, triggered_event>::template Call(std::forward<Args>(args)...);
    }

    /**
     * This function is executed if quantities are different and no action follows (NOOP).
     */
    template <HierarchyLevels triggered_hierarchy_level, Events triggered_event, typename... Args>
    static typename std::enable_if<triggered_hierarchy_level != Quantity, void>::type
    dispatch(Args &&... /*unused*/)
    {
        // This is a NOOP because the quantities are not the same and therefore no callback should
        // be called
    }
};


#endif
