#ifndef _DaHa_IntraNodeContainer_
#define _DaHa_IntraNodeContainer_

#include <cmath>
#include <vector>

#include "DependencyManager.hpp"
#include "Executor.hpp"
#include "LoadBalancer.hpp"
#include "MutexContainer.hpp"
#include "ThreadDormitory.hpp"

#include "TaskFactory.hpp"
#include "ThreadingWrapper.hpp"
#include "uninitialized_vector.hpp"

#include "ThreadIdentifier.hpp"

template <
        typename fmm_handle_type,
        typename mutex_container_type,
        typename event_dispatcher>
class IntraNodeContainer
{
public:
    using FMMHandle = fmm_handle_type;
    using EventDispatcher = event_dispatcher;
    using MutexContainerType = mutex_container_type;

    using DependencyManagerType = DependencyManager<EventDispatcher, FMMHandle>;
    using TaskFactoryType = TaskFactory<FMMHandle, MutexContainerType, DependencyManagerType>;


    using MultiQueueType = typename TaskFactoryType::MultiQueueType;
    using ExecutorType = Executor<MultiQueueType>;

    using LoadBalancerType = typename TaskFactoryType::LoadBalancerType;

    using ThreadingWrapperType = ThreadingWrapper<ExecutorType>;

    IntraNodeContainer(FMMHandle &fmm_handle, unsigned int num_threads)
            : fmm_handle_(fmm_handle),
              num_threads_(num_threads),
              dependency_manager_(fmm_handle),
              mutex_container_(fmm_handle_),
              queues_(num_threads_),
              load_balancer_(fmm_handle_, queues_),
              task_factory_(fmm_handle_, mutex_container_, load_balancer_, dependency_manager_),
              executors_(num_threads_),
              threading_wrapper_(num_threads_)
    {

        thread_id_rank_mapping_[ThreadingWrapperType::GetThisId()] = 0;
        executors_.emplace_back(0, queues_, thread_dormitory_);
        threading_wrapper_.emplace_back(executors_[0], true);

        // We need to initialize the ThreadLocal Variables of the Main thread here, because we never
        // call the executor.Run() method
        executors_[0].InitializeThreadLocal();

        // TODO(david): rethink main thread solution template sepcialization for threading wrapper?
        for (unsigned int i = 1; i < num_threads_; ++i)
        {
            executors_.emplace_back(i, queues_, thread_dormitory_);
            threading_wrapper_.emplace_back(executors_[i]);
            thread_id_rank_mapping_[threading_wrapper_[i].GetId()] = i;
        }
    }

    ~IntraNodeContainer()
    {
        thread_dormitory_.wakeUpAll();
    }

    TaskFactoryType &task_factory() { return task_factory_; }
    unsigned int num_threads() const { return num_threads_; }

    void FinishAndSoftJoin()
    {
        if (!IsPrimary())
        {
            return;
        }

        while (dependency_manager_.dc().force_computation_counter != 0)
        {
            executors_[0].RunAll();
        }

        if (dependency_manager_.dc().force_computation_counter == 0)
        {
            thread_dormitory_.putToSleepAll();
        }
    }

    void prepareForComputation()
    {
        DisableWorkStealing();
        thread_dormitory_.wakeUpAll();
        dependency_manager_.ResetDepsCounter();
    }

    bool IsPrimary() { return thread_id_rank_mapping_[ThreadingWrapperType::GetThisId()] == 0; }

    void EnableWorkStealing()
    {
        for (auto &exec : executors_)
        {
            exec.EnableWorkStealing();
        }
    }

    void DisableWorkStealing()
    {
        for (auto &exec : executors_)
        {
            exec.DisableWorkStealing();
        }
    }

private:
    FMMHandle &fmm_handle_;
    unsigned int num_threads_;

    DependencyManagerType dependency_manager_;

    MutexContainerType mutex_container_;

    fmsolvr::simple_vector<MultiQueueType> queues_;

    // multi thread scheduler
    LoadBalancerType load_balancer_;

    TaskFactoryType task_factory_;

    ThreadDormitory thread_dormitory_;

    // single thread scheduler
    uninitialized_vector<ExecutorType> executors_;

    // single thread wrapper
    uninitialized_vector<ThreadingWrapperType> threading_wrapper_;

    std::map<typename ThreadingWrapperType::id_type, int> thread_id_rank_mapping_;
};

#endif
