#ifndef _DaHa_Processor_
#define _DaHa_Processor_

#include "fmsolvr/tasks/pass1.hpp"
#include "fmsolvr/tasks/pass2.hpp"
#include "fmsolvr/sequential/pass2dl.hpp"
#include "fmsolvr/tasks/pass3.hpp"
#include "fmsolvr/tasks/pass4.hpp"
#include "fmsolvr/tasks/pass5.hpp"

#include <tuple>


template <typename FMMHandle, typename MutexContainer, typename dependency_manager>
class TaskFactory;

template <
        template <typename...> class ConcreteProcessor,
        typename fmm_handle_type,
        typename mutex_container_type,
        typename dependency_manager>
class AbstractProcessor
{
public:
    using FMMHandle = fmm_handle_type;
    using DependencyManagerType = dependency_manager;
    using MutexContainerType = mutex_container_type;
    using TaskFactoryType = TaskFactory<FMMHandle, MutexContainerType, DependencyManagerType>;

    using Processor = ConcreteProcessor<FMMHandle, MutexContainerType, DependencyManagerType>;

    AbstractProcessor(
            FMMHandle &fmm_handle,
            MutexContainerType &mutex_container,
            TaskFactoryType &task_factory,
            DependencyManagerType &dm)
            : fmm_handle_(fmm_handle),
              mutex_container_(mutex_container),
              task_factory_(task_factory),
              dm_(dm){}
    template <typename TreePartitionType>
    void execute(TreePartitionType tree_partition)
    {
        concrete_processor().pre_processing(tree_partition);
        concrete_processor().run_computation(tree_partition);
        concrete_processor().post_processing(tree_partition);
    }

    /**
     * All method should be empty by default and hidden with an implementation in child
     * class/concrete processor if needed.
     */
    template <typename TreePartitionType>
    void pre_processing(TreePartitionType /*unused*/)
    {
    }
    template <typename TreePartitionType>
    void run_computation(TreePartitionType /*unused*/)
    {
    }
    template <typename TreePartitionType>
    void post_processing(TreePartitionType /*unused*/)
    {
    }

protected:
    FMMHandle &fmm_handle_;
    MutexContainerType &mutex_container_;
    TaskFactoryType &task_factory_;
    DependencyManagerType &dm_;

    Processor &concrete_processor()
    {
        return *static_cast<Processor *>(this);
    }
};

template <typename... Args>
struct P2MProcessor : public AbstractProcessor<P2MProcessor, Args...>
{
    using P2MProcessor::AbstractProcessor::AbstractProcessor;
    template <typename TreePartitionType>
    void run_computation(TreePartitionType tree_partition)
    {
        fmsolvr::pass1_P2M(this->fmm_handle_, tree_partition, this->task_factory_, this->dm_);
    }
};

template <typename... Args>
struct M2MProcessor : public AbstractProcessor<M2MProcessor, Args...>
{
    using M2MProcessor::AbstractProcessor::AbstractProcessor;
    template <typename TreePartitionType>
    void run_computation(TreePartitionType tree_partition)
    {
        fmsolvr::pass1_M2M(
                this->fmm_handle_,
                tree_partition.first().sfcindex(),
                tree_partition.first().d(),
                this->task_factory_,
                this->dm_);
    }
};

template <typename... Args>
struct M2LProcessor : public AbstractProcessor<M2LProcessor, Args...>
{
    using M2LProcessor::AbstractProcessor::AbstractProcessor;
    template <typename TreePartitionType>
    void run_computation(TreePartitionType tree_partition)
    {
        fmsolvr::pass2_M2LSource(
                this->fmm_handle_,
                tree_partition.first().sfcindex(),
                tree_partition.first().d(),
                this->task_factory_,
                this->mutex_container_,
                this->dm_);
    }
};

template <typename... Args>
struct L2LProcessor : public AbstractProcessor<L2LProcessor, Args...>
{
    using L2LProcessor::AbstractProcessor::AbstractProcessor;
    template <typename TreePartitionType>
    void run_computation(TreePartitionType tree_partition)
    {
        fmsolvr::pass3_L2L(
                this->fmm_handle_,
                tree_partition.first().sfcindex(),
                tree_partition.first().d(),
                this->task_factory_,
                this->mutex_container_,
                this->dm_);
    }
};

template <typename... Args>
struct L2PProcessor : public AbstractProcessor<L2PProcessor, Args...>
{
    using L2PProcessor::AbstractProcessor::AbstractProcessor;
    template <typename TreePartitionType>
    void run_computation(TreePartitionType tree_partition)
    {
        fmsolvr::pass4_one_box(
                this->fmm_handle_,
                tree_partition.first().sfcindex(),
                this->mutex_container_,
                this->dm_);
    }
};

template <typename... Args>
struct P2PProcessor : public AbstractProcessor<P2PProcessor, Args...>
{
    using P2PProcessor::AbstractProcessor::AbstractProcessor;
    template <typename TreePartitionType>
    void run_computation(TreePartitionType tree_partition)
    {
        fmsolvr::pass5_P2P(this->fmm_handle_, tree_partition, this->mutex_container_, this->dm_);
    }
};

template <typename... Args>
struct LatticeProcessor : public AbstractProcessor<LatticeProcessor, Args...>
{
    using LatticeProcessor::AbstractProcessor::AbstractProcessor;

    using FMMHandle = typename LatticeProcessor::AbstractProcessor::FMMHandle;
    using BoxID = typename FMMHandle::boxid_type;
    using DipoleCompensation = fmsolvr::dipole_compensation<FMMHandle>;
    using Real33 = typename FMMHandle::real33_type;
    using Real3 = typename FMMHandle::real3_type;
    using Real = typename FMMHandle::real_type;

    template <typename TreePartitionType>
    void run_computation(TreePartitionType tree_partition)
    {
        auto &tree_mu = this->fmm_handle_.tree_mu();
        if (!this->fmm_handle_.simulationbox().open_boundary_conditions())
        {
            std::cout << "# PASS 2: Lattice\n";
            fmsolvr::pass2_dipole_lattice(this->fmm_handle_);
        }
        BoxID bid_root(tree_partition.first().sfcindex(), tree_partition.first().d());
        auto &mu = tree_mu[bid_root];
        mu.populate_lower();

        this->dm_.ResolvePartialMu(bid_root, this->task_factory_, this->mutex_container_.tree_ow_);
    }
};


#endif
// vim: et:ts=4:sw=4
