/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "fmsolvr/abc.hpp"
#include "fmsolvr/default_alloc.hpp"
#include "fmsolvr/math/ipowlog.hpp"
#include "fmsolvr/multipole.hpp"
#include "fmsolvr/operator/p3/local2local_p3.hpp"
#include "fmsolvr/operator/p3/multipole2local_p3.hpp"
#include "fmsolvr/operator/p3/multipole2multipole_p3.hpp"
#include "fmsolvr/operator/p4/local2local.hpp"
#include "fmsolvr/operator/p4/multipole2local.hpp"
#include "fmsolvr/operator/p4/multipole2multipole.hpp"
#include "fmsolvr/operator/particle2local.hpp"
#include "fmsolvr/operator/particle2multipole.hpp"
#include "fmsolvr/operator/rotation/rotation_generator.hpp"
#include "fmsolvr/operator/rotation/rotation_types.hpp"
#include "fmsolvr/operatorid_m2l.hpp"
#include "fmsolvr/simd/simd_engine.hpp"
#include "fmsolvr/tree/boxid.hpp"
#include "fmsolvr/tree/boxid_4d.hpp"
#include "fmsolvr/tree/simulationbox.hpp"
#include "fmsolvr/util/buildinfo.hpp"
#include "fmsolvr/util/envtools.hpp"
#include "fmsolvr/util/float128.hpp"
#include "fmsolvr/util/wrpapi.hpp"
#include "fmsolvr/xyz.hpp"
#include "fmsolvr/xyzs.hpp"

#include "bulk_ops.hpp"

#include <algorithm>
#include <chrono>
#include <random>
#include <vector>


namespace fmsolvr
{

papiwrapper<> pw;

unsigned seed;

unsigned default_seed()
{
#if 0
    return std::chrono::system_clock::now().time_since_epoch().count();
#else
    std::random_device rd;
    return rd();
#endif
}

template < typename Real3,
           typename CoeffMatrixIterator>
void generate_expansion_multipole(
	size_t p,
	ssize_t dd,
	CoeffMatrixIterator omega_iterator)
{
    size_t dim = 1 << dd;
    std::mt19937 gen(seed + p + dd);
    std::uniform_real_distribution<> dis(-pow(2,-dd-1),pow(2,-dd-1));
	for (size_t i = 0; i < dim; ++i) {
		for (size_t j = 0; j < dim; ++j) {
			for (size_t k = 0; k < dim; ++k) {
				Real3 particle = Real3(dis(gen), dis(gen), dis(gen));
				P2M_nodivtable(1.0,particle,**omega_iterator,Real3(0,0,0),p);
                (*omega_iterator)->populate_lower();
                omega_iterator++;
			}
		}
	}
}

template <typename Real3, typename CoefficientMatrix>
void generate_expansion_local(
                              ssize_t p,
                              CoefficientMatrix & mu)
{
	typedef typename CoefficientMatrix::value_type::value_type scalar_type;
    std::mt19937 gen(seed + p);
    std::uniform_real_distribution<> theta_dis(0,M_PI);
    std::uniform_real_distribution<> phi_dis(0,2*M_PI);
    std::uniform_real_distribution<> r_dis(3,5);

    scalar_type theta = theta_dis(gen);
    scalar_type phi = phi_dis(gen);
    scalar_type r = r_dis(gen);

    Real3 particle(r*sin(theta)*cos(phi),
                   r*sin(theta)*sin(phi),
                   r*cos(theta));

    P2L(1.0,particle,mu,Real3(0,0,0),p);
    mu.populate_lower();
}

// relative error, using a as exact reference
template <typename CoefficientMatrix>
typename CoefficientMatrix::complex_type::value_type maxdelta(
                                                              const CoefficientMatrix & a,
                                                              const CoefficientMatrix & b,
                                                              size_t p,
                                                              size_t & l_max,
                                                              size_t & m_max,
                                                              typename CoefficientMatrix::complex_type::value_type & ref,
                                                              typename CoefficientMatrix::complex_type::value_type & err)
{
    typedef typename CoefficientMatrix::complex_type complex_type;
    typedef typename complex_type::value_type scalar_type;
    scalar_type err_rel_max(-0), err_rel(0);

    for (size_t l = 0; l <= p; ++l)
        for (size_t m = 0; m <= l; ++m) {
            complex_type alm = a.upper(l, m);
            complex_type blm = b.upper(l, m);
            typename complex_type::value_type almlen = fmm::abs(alm);
            if (almlen > 0) {
                err_rel = fmm::abs(alm - blm)/almlen;
                if (err_rel > err_rel_max) {
                    err_rel_max = err_rel;
                    err = fmm::abs(alm - blm);
                    ref = almlen;
                    l_max = l;
                    m_max = m;
                }
            }
        }
    return err_rel_max;
}

// relative error, using a as exact reference
template <typename CoefficientMatrix>
void maxdelta_report(const CoefficientMatrix & a, const CoefficientMatrix & b, size_t p)
{
    typedef typename CoefficientMatrix::value_type::value_type Real;

    size_t l_max(0), m_max(0);
    Real ref_abs(0), err_abs(0);
    Real err_rel = maxdelta(a, b, p, l_max, m_max, ref_abs, err_abs);
    std::cout << "maximum relative error: " << err_rel
              << " at l = " << l_max << ", m = " << m_max << ", p = " << p << std::endl
              << "absolute error: " << err_abs << ", " << "reference value: " << ref_abs << std::endl
              << std::endl;
}

// relative error, using a as exact reference
template <typename CoefficientMatrixIterator,
          typename Real>
          Real maxdelta(
                                                CoefficientMatrixIterator a,
                                                CoefficientMatrixIterator b,
                                                size_t n,
                                                size_t p,
                                                size_t & l_max,
                                                size_t & m_max,
                                                Real & ref,
                                                Real & err)
{
    Real err_rel_max(-0), err_rel(0);
    for (size_t i = 0; i < n; ++i) {

        for (size_t l = 0; l <= p; ++l) {
            for (size_t m = 0; m <= l; ++m) {
                auto alm = (**a).upper(l, m);
                auto blm = (**b).upper(l, m);
                Real almlen = fmm::abs(alm);

                if (almlen > 0) {
                    err_rel = fmm::abs(alm - blm)/almlen;
                    if (err_rel > err_rel_max) {
                        err_rel_max = err_rel;
                        err = fmm::abs(alm - blm);
                        ref = almlen;
                        l_max = l;
                        m_max = m;
                    }
                }
            }
        }
        a++;
        b++;
    }
    return err_rel_max;
}


// relative error, using a as exact reference
template <typename CoefficientMatrixIterator>
void maxdelta_report(
                                                CoefficientMatrixIterator a,
                                                CoefficientMatrixIterator b,
                                                size_t n,
                                                size_t p)
{
    typedef typename std::remove_reference<decltype(**a)>::type::value_type::value_type Real;

    size_t l_max(0), m_max(0);
    Real ref_abs(0), err_abs(0);
    Real err_rel = maxdelta(a, b, n, p, l_max, m_max, ref_abs, err_abs);
    std::cout << "maximum relative error: " << err_rel;
    std::cout << " at l = " << l_max << ", m = " << m_max << ", p = " << p << std::endl;
    std::cout << "absolute error: " << err_abs << ", " << "reference value: " << ref_abs << std::endl;
    std::cout << std::endl;
}


template <typename InputCoefficents,
          typename RotationOperator,
          typename ShiftOperator,
          typename OutputCoefficients>
class Bulk4 {
    const InputCoefficents * in_;
    const RotationOperator R_;
    std::shared_ptr<const ShiftOperator> Op_;
    OutputCoefficients * out_;

public:
    typedef InputCoefficents in_type;
    typedef RotationOperator R_type;
    typedef ShiftOperator Op_type;
    typedef OutputCoefficients out_type;

    Bulk4(const InputCoefficents & in, const RotationOperator & R, std::shared_ptr<const ShiftOperator> && Op, OutputCoefficients & out) :
        in_(&in), R_(R), Op_(Op), out_(&out)
    { }

    const InputCoefficents & in() const
    { return *in_; }

    const RotationOperator & R() const
    { return R_; }

    const ShiftOperator & Op() const
    { return *Op_; }

    OutputCoefficients & out() const
    { return *out_; }
};

template <typename Allocator, typename RAIterator, typename Scratch>
void bulk_M2Lp3(Allocator /*alloc*/, RAIterator first, RAIterator last, size_t p, Scratch &scratch)
{
    typedef typename RAIterator::value_type value_type;
    typedef typename value_type::in_type in_type;
    typedef typename value_type::R_type R_type;
    typedef typename value_type::Op_type Op_type;
    typedef typename value_type::out_type out_type;
    typedef fmsolvr::bulk::BulkRotOp<in_type, R_type, Op_type, out_type> bulk_type;

    std::vector<bulk_type, typename Allocator::template rebind<bulk_type>::other> bulk(first, last);
    fmsolvr::bulk::M2Lp3(&*bulk.cbegin(), &*bulk.cend(), p, scratch);
}

template <typename Real>
void preheat(size_t p, size_t, size_t ws)
{
	typedef XYZ<Real> Real3;
	typedef typename fmsolvr::default_allocator<Real>::type Alloc;
    typedef typename fmsolvr::RotationOMuga<Real, Alloc>::CoeffMatrix CoeffMatrix;
    typedef M2L_Operator<CoeffMatrix> M2L_Operator;

    pw.start();
    std::cout << std::endl;
    std::cout << "PREHEAT M2L_p^4_onthefly p=" << p << " ws=" << ws << std::endl;
    ssize_t half_n_dim = 2*ws+1;
    ssize_t inner = ws;

    auto omega = std::allocate_shared<CoeffMatrix>(Alloc(), p);
    {
        CoeffMatrix * pomega = omega.get();
        generate_expansion_multipole<Real3>(p, 0, &pomega);
    }
    omega->populate_lower();
    auto mu = std::allocate_shared<CoeffMatrix>(Alloc(), p);

    for (ssize_t i = -half_n_dim; i <= half_n_dim; ++i) {
        for (ssize_t j = -half_n_dim; j <= half_n_dim; ++j) {
            for (ssize_t k = -half_n_dim; k <= half_n_dim; ++k) {
                if (i >= -inner && i <= +inner && j >= -inner && j <= +inner && k >= -inner && k <= +inner)
                    continue;
                M2L_Operator B(Real3(0,0,0),Real3(i,j,k),p);
                M2L_reference(*omega,B,*mu,p);
            }
        }
    }
    pw.stop();
    pw.report(1);
}

template <typename Real, template <typename> class SIMDengine>
void run_test(size_t p, size_t dd, size_t ws)
{
    bool test_m2m = get_env_int("TEST_M2M", 1);
    bool test_m2l = get_env_int("TEST_M2L", 1);
    bool test_l2l = get_env_int("TEST_L2L", 1);
    bool test_bulk = get_env_int("TEST_BULK", 1);

	typedef XYZ<Real> Real3;
	typedef ABC<Real3> Real33;
    typedef fmsolvr::BoxID<fmsolvr::BoxID4D<uint32_t> > BoxID;  // with a 64-bit size_t this is sufficient upto depth 21
    typedef fmsolvr::M2L_OperatorID<BoxID> OperatorID_M2L;
    typedef typename OperatorID_M2L::ijk_type IJK;
	typedef typename fmsolvr::default_allocator<Real>::type Alloc;
    typedef typename fmsolvr::RotationOMuga<Real, Alloc>::CoeffMatrix CoeffMatrix;
    typedef fmsolvr::RotationDataTypes<Real, Alloc> RotationDataTypes;
    typedef typename RotationDataTypes::RotMatVec RotMatVec;
    typedef RG::RotationGeneratorX2X<RotMatVec, Real3, Real33, Alloc> RotationGeneratorX2X;
    typedef RG::RotationGeneratorM2L<RotMatVec, Real3, Real33, Alloc> RotationGeneratorM2L;
    typedef typename RotationDataTypes::Rot_Operator RotationOperator;
    typedef M2M_Operator<CoeffMatrix> M2M_Operator;
    typedef typename RotationDataTypes::M2M_Operator M2Mp3_Operator;
    typedef L2L_Operator<CoeffMatrix> L2L_Operator;
    typedef typename RotationDataTypes::L2L_Operator L2Lp3_Operator;
    typedef M2L_Operator<CoeffMatrix> M2L_Operator;
    typedef typename RotationDataTypes::M2L_Operator M2Lp3_Operator;
    typedef typename CoeffMatrix::value_type coefficient_type;
    typedef typename RotationDataTypes::template ScratchTypeGenerator<coefficient_type>::Scratch scratch_type;

    std::cout << std::endl << "p=" << p << "  ws=" << ws << "  deltadepth=" << dd << std::endl << std::endl;

    Real3 e1(1, 0, 0);
    Real3 e2(0, 1, 0);
    Real3 e3(0, 0, 1);
    Real33 abc(e1, e2, e3);
    fmsolvr::SimulationBox<Real33> simbox(abc, Real3(0, 0, 0), false, false, false);

    const size_t c_dim = 1 << dd;
    const size_t nchildren = icubed(c_dim);

    // M2M
  if (test_m2m) {
    // Generate multipole expansions of child boxes.
    std::vector<std::shared_ptr<CoeffMatrix>> omega_c;
    omega_c.reserve(nchildren);
    for (size_t i = 0; i < nchildren; ++i)
        omega_c.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
    generate_expansion_multipole<Real3>(p,dd,omega_c.begin());

    RotationGeneratorX2X G(p, dd, abc);

    // Apply M2M and M2M_p3 operator to obtain expansion in the parents frame and compare.

    Real3 expn_point_p = simbox.expansion_point(BoxID(0, 0, 0, 0));
    auto omega_p = std::allocate_shared<CoeffMatrix>(Alloc(), p);
    auto omega_p_p3 = std::allocate_shared<CoeffMatrix>(Alloc(), p);
    auto omega_p_p3_2 = std::allocate_shared<CoeffMatrix>(Alloc(), p);
    auto omega_p_p3_ref = std::allocate_shared<CoeffMatrix>(Alloc(), p);
    auto omega_p_p3_bulk = std::allocate_shared<CoeffMatrix>(Alloc(), p);

    // allocate scratch
    scratch_type scratch(p);

    // M2M_p3, without precomputation.
    std::cout << "M2M_p^3_onthefly:" << std::endl;
    pw.start();
    size_t ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                RotationOperator R = G.generate_rotation_operator(IJK(i, j, k), dd);
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                M2Mp3_Operator A_p3(expn_point_c, expn_point_p, p);
                M2Mp3<SIMDengine>(*omega_c[ind], R, A_p3, *omega_p_p3, p, scratch);
                ind++;
            }
        }
    }
    assert(ind == nchildren);
    pw.stop();
    pw.report(nchildren);
    std::cout << std::endl;

    // M2M_p3, use precomputed matrices from the previous pass.
    std::cout << "M2M_p^3_precomputed:" << std::endl;
    pw.start();
    ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                RotationOperator R = G.generate_rotation_operator(IJK(i, j, k), dd);
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                M2Mp3_Operator A_p3(expn_point_c, expn_point_p, p);
                M2Mp3<SIMDengine>(*omega_c[ind], R, A_p3, *omega_p_p3_2, p, scratch);
                ind++;
            }
        }
    }
    assert(ind == nchildren);
    pw.stop();
    pw.report(nchildren);
    std::cout << std::endl;

    // M2M_p3 reference implementation, use precomputed matrices from the previous pass.
    std::cout << "M2M_p^3_reference:" << std::endl;
    pw.start();
    ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                RotationOperator R = G.generate_rotation_operator(IJK(i, j, k), dd);
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                M2Mp3_Operator A_p3(expn_point_c, expn_point_p, p);
                M2Mp3_impl_reference(*omega_c[ind],R,A_p3,*omega_p_p3_ref,p);
                ind++;
            }
        }
    }
    assert(ind == nchildren);
    pw.stop();
    pw.report(nchildren);
    std::cout << std::endl;

    maxdelta_report(*omega_p_p3_ref, *omega_p_p3_2, p);

   if (test_bulk) {
    // M2M_p3, bulk, use precomputed matrices from the previous pass.
    std::cout << "M2M_p^3_bulk_setup:" << std::endl;
    pw.start();
    typedef Bulk4<CoeffMatrix, RotationOperator, M2Mp3_Operator, CoeffMatrix> BulkM2MTuple;
    std::vector<BulkM2MTuple, typename Alloc::template rebind<BulkM2MTuple>::other> bulkM2Mdata;
    bulkM2Mdata.reserve(nchildren);
    ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                RotationOperator R = G.generate_rotation_operator(IJK(i, j, k), dd);
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                std::shared_ptr<M2Mp3_Operator> A_p3 = std::allocate_shared<M2Mp3_Operator>(Alloc(), expn_point_c, expn_point_p, p);
                bulkM2Mdata.push_back(BulkM2MTuple(*omega_c[ind], R, std::move(A_p3), *omega_p_p3_bulk));
                ind++;
            }
        }
    }
    assert(ind == nchildren);
    assert(bulkM2Mdata.size() == nchildren);
    pw.stop();
    pw.report(nchildren);
    std::cout << "M2M_p^3_bulk_run:" << std::endl;
    pw.start();
    for (auto it = bulkM2Mdata.begin(); it != bulkM2Mdata.end(); ++it) {
        M2Mp3<SIMDengine>(it->in(), it->R(), it->Op(), it->out(), p, scratch);
    }
    assert(bulkM2Mdata.size() == nchildren);
    pw.stop();
    pw.report(nchildren);
    std::cout << std::endl;

    maxdelta_report(*omega_p_p3, *omega_p_p3_bulk, p);
   }

    // M2M_p4
    std::cout << "M2M_p^4_onthefly:" << std::endl;
    pw.start();
    ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                M2M_Operator A(expn_point_c, expn_point_p, p);
                M2M(*omega_c[ind],A,*omega_p,p);

                ind++;
            }
        }
    }
    pw.stop();
    pw.report(nchildren);
    std::cout << std::endl;

    maxdelta_report(*omega_p, *omega_p_p3, p);
    maxdelta_report(*omega_p, *omega_p_p3_ref, p);
  }


    // L2L
  if (test_l2l) {
    // Generate local expansion of randomly positioned particle about (0,0,0)
    Real3 expn_point_p = simbox.expansion_point(BoxID(0, 0, 0, 0));
    auto mu_p = std::allocate_shared<CoeffMatrix>(Alloc(), p);
    generate_expansion_local<Real3>(p, *mu_p);

    // Apply L2L and L2L_p3 operator to obtain expansion in the childrens frame and compare.
    std::vector<std::shared_ptr<CoeffMatrix>> mu_c;
    std::vector<std::shared_ptr<CoeffMatrix>> mu_c_p3;
    std::vector<std::shared_ptr<CoeffMatrix>> mu_c_p3_2;
    std::vector<std::shared_ptr<CoeffMatrix>> mu_c_p3_ref;
    std::vector<std::shared_ptr<CoeffMatrix>> mu_c_p3_bulk;
    mu_c.reserve(nchildren);
    mu_c_p3.reserve(nchildren);
    mu_c_p3_2.reserve(nchildren);
    mu_c_p3_ref.reserve(nchildren);
    mu_c_p3_bulk.reserve(nchildren);
    for (size_t i = 0; i < nchildren; ++i) {
        mu_c.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
        mu_c_p3.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
        mu_c_p3_2.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
        mu_c_p3_ref.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
        mu_c_p3_bulk.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
    }

    // allocate scratch
    scratch_type scratch(p);

    RotationGeneratorX2X G2(p, dd, abc);
    // L2L_p3, without precomputation
    std::cout << "L2L_p^3_onthefly:" << std::endl;
    pw.start();
    size_t ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                RotationOperator R = G2.generate_rotation_operator(IJK(i, j, k), dd);
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                L2Lp3_Operator C_p3(expn_point_p, expn_point_c, p);
                L2Lp3<SIMDengine>(*mu_p, R, C_p3, *mu_c_p3[ind], p, scratch);

                ind++;
            }
        }
    }
    assert(ind == nchildren);
    pw.stop();
    pw.report(nchildren);
    std::cout << std::endl;

    // L2L_p3, with precomputation
    std::cout << "L2L_p^3_precomputed:" << std::endl;
    pw.start();
    ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                RotationOperator R = G2.generate_rotation_operator(IJK(i, j, k), dd);
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                L2Lp3_Operator C_p3(expn_point_p, expn_point_c, p);
                L2Lp3<SIMDengine>(*mu_p, R, C_p3, *mu_c_p3_2[ind], p, scratch);

                ind++;
            }
        }
    }
    pw.stop();
    pw.report(nchildren);
    std::cout << std::endl;

    // L2L_p3 reference implementation, with precomputation
    std::cout << "L2L_p^3_reference:" << std::endl;
    pw.start();
    ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                RotationOperator R = G2.generate_rotation_operator(IJK(i, j, k), dd);
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                L2Lp3_Operator C_p3(expn_point_p, expn_point_c, p);
                L2Lp3_impl_reference(*mu_p,R,C_p3,*mu_c_p3_ref[ind],p);

                ind++;
            }
        }
    }
    pw.stop();
    pw.report(nchildren);
    std::cout << std::endl;

    maxdelta_report(mu_c_p3_ref.begin(), mu_c_p3_2.begin(), nchildren, p);

   if (test_bulk) {
    // L2L_p3, bulk, with precomputation
    std::cout << "L2L_p^3_bulk_setup:" << std::endl;
    
    typedef Bulk4<CoeffMatrix, RotationOperator, L2Lp3_Operator, CoeffMatrix> BulkL2LTuple;
    std::vector<BulkL2LTuple, typename Alloc::template rebind<BulkL2LTuple>::other> bulkL2Ldata;
    bulkL2Ldata.reserve(nchildren);
   
    pw.start();
    ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                RotationOperator R = G2.generate_rotation_operator(IJK(i, j, k), dd);
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                std::shared_ptr<L2Lp3_Operator> C_p3 = std::allocate_shared<L2Lp3_Operator>(Alloc(),expn_point_p, expn_point_c,p);
                bulkL2Ldata.push_back(BulkL2LTuple(*mu_p,R,std::move(C_p3),*mu_c_p3_bulk[ind]));
                ind++;
            }
        }
    }

    assert(ind == nchildren);
    assert(bulkL2Ldata.size() == nchildren);
    pw.stop();
    pw.report(nchildren);
    std::cout << "L2L_p^3_bulk_run:" << std::endl;
    pw.start();
    for (auto it = bulkL2Ldata.begin(); it != bulkL2Ldata.end(); ++it) {
        L2Lp3<SIMDengine>(it->in(), it->R(), it->Op(), it->out(), p, scratch);
    }
    assert(bulkL2Ldata.size() == nchildren);
    pw.stop();
    pw.report(nchildren);
    std::cout << std::endl;

    maxdelta_report(mu_c_p3_2.begin(), mu_c_p3_bulk.begin(), nchildren, p); 
   }

    // L2L_p4
    std::cout << "L2L_p^4_onthefly:" << std::endl;
    pw.start();
    ind = 0;
    for (size_t i = 0; i < c_dim; ++i) {
        for (size_t j = 0; j < c_dim; ++j) {
            for (size_t k = 0; k < c_dim; ++k) {
                Real3 expn_point_c = simbox.expansion_point(BoxID(i, j, k, dd));
                L2L_Operator C(expn_point_p, expn_point_c, p);
                L2L_impl_reference(*mu_p,C,*mu_c[ind],p,p);

                ind++;
            }
        }
    }
    pw.stop();
    pw.report(nchildren);
    std::cout <<  std::endl;

    maxdelta_report(mu_c.begin(), mu_c_p3.begin(), nchildren, p);
    if (p > 10)
        maxdelta_report(mu_c.begin(), mu_c_p3.begin(), nchildren, 10);
    maxdelta_report(mu_c.begin(), mu_c_p3_ref.begin(), nchildren, p);
  }


    // M2L
  if (test_m2l) {
    // Apply M2L and M2L_p3 operator to the parents multipole expansion to obtain
    // the local expansions in the neighboring boxes and compare.

    RotationGeneratorM2L GM2L(p, ws, abc);
    ssize_t half_n_dim = 2*ws+1;

    const size_t nneighbors = icubed(2 * (2 * ws + 1) + 1) - icubed(2 * ws + 1);
    ssize_t inner = ws;

    // generate omega
    auto omega = std::allocate_shared<CoeffMatrix>(Alloc(), p);
    {
        CoeffMatrix * pomega = omega.get();
        generate_expansion_multipole<Real3>(p, 0, &pomega);
    }
    omega->populate_lower();

    // precompute B for p^4
    std::vector<M2L_Operator *> B;
    for (ssize_t i = -half_n_dim; i <= half_n_dim; ++i) {
        for (ssize_t j = -half_n_dim; j <= half_n_dim; ++j) {
            for (ssize_t k = -half_n_dim; k <= half_n_dim; ++k) {
                if (i >= -inner && i <= +inner && j >= -inner && j <= +inner && k >= -inner && k <= +inner)
                    continue;
                B.push_back(new M2L_Operator(Real3(0, 0, 0), Real3(i, j, k), p));
            }
        }
    }
    assert(B.size() == nneighbors);

    // allocate mu
    std::vector<std::shared_ptr<CoeffMatrix>> mu_p4_ref;
    std::vector<std::shared_ptr<CoeffMatrix>> mu_p4_2;
    std::vector<std::shared_ptr<CoeffMatrix>> mu_p3;
    std::vector<std::shared_ptr<CoeffMatrix>> mu_p3_2;
    std::vector<std::shared_ptr<CoeffMatrix>> mu_p3_ref;
    std::vector<std::shared_ptr<CoeffMatrix>> mu_p3_bulk;
    mu_p4_ref.reserve(nneighbors);
    mu_p4_2.reserve(nneighbors);
    mu_p3.reserve(nneighbors);
    mu_p3_2.reserve(nneighbors);
    mu_p3_ref.reserve(nneighbors);
    mu_p3_bulk.reserve(nneighbors);
    for (size_t i = 0; i < nneighbors; ++i) {
        mu_p4_ref.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
        mu_p4_2.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
        mu_p3.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
        mu_p3_2.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
        mu_p3_ref.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
        mu_p3_bulk.push_back(std::allocate_shared<CoeffMatrix>(Alloc(), p));
    }
    std::shared_ptr<CoeffMatrix> single_mu = std::allocate_shared<CoeffMatrix>(Alloc(), p);

    // allocate scratch
    scratch_type scratch(p);

    // M2L_p3, no precomputation
    std::cout << "M2L_p^3_onthefly:" << std::endl;
    pw.start();
    size_t ind = 0;
    for (ssize_t i = -half_n_dim; i <= half_n_dim; ++i) {
        for (ssize_t j = -half_n_dim; j <= half_n_dim; ++j) {
            for (ssize_t k = -half_n_dim; k <= half_n_dim; ++k) {
                if (i >= -inner && i <= +inner && j >= -inner && j <= +inner && k >= -inner && k <= +inner)
                    continue;
                IJK op_omega_ijk(0, 0, 0);
                IJK op_mu_ijk(i, j, k);
                RotationOperator R = GM2L.generate_rotation_operator(OperatorID_M2L(op_omega_ijk, op_mu_ijk, -1));
                M2Lp3_Operator B_p3(Real3(0,0,0),Real3(i,j,k),p);
                M2Lp3<SIMDengine>(*omega, R, B_p3, *mu_p3[ind], p, scratch);

                ind++;
            }
        }
    }
    assert(ind == nneighbors);
    pw.stop();
    pw.report(nneighbors);
    std::cout << std::endl;

    // M2L_p3, with precomputation
    std::cout << "M2L_p^3_one_d_many_omega:" << std::endl;
    pw.start();
    ind = 0;
    for (ssize_t i = -half_n_dim; i <= half_n_dim; ++i) {
        for (ssize_t j = -half_n_dim; j <= half_n_dim; ++j) {
            for (ssize_t k = -half_n_dim; k <= half_n_dim; ++k) {
                if (i >= -inner && i <= +inner && j >= -inner && j <= +inner && k >= -inner && k <= +inner)
                    continue;
                IJK op_omega_ijk(0, 0, 0);
                IJK op_mu_ijk(half_n_dim, half_n_dim, half_n_dim);
                RotationOperator R = GM2L.generate_rotation_operator(OperatorID_M2L(op_omega_ijk, op_mu_ijk, -1));
                M2Lp3_Operator B_p3(Real3(0,0,0),Real3(i,j,k),p);
                M2Lp3<SIMDengine>(*mu_p3[ind], R, B_p3, *single_mu, p, scratch);

                ind++;
            }
        }
    }
    pw.stop();
    pw.report(nneighbors);
    std::cout << std::endl;

    // M2L_p3, with precomputation
    std::cout << "M2L_p^3_precomputed:" << std::endl;
    pw.start();
    ind = 0;
    for (ssize_t i = -half_n_dim; i <= half_n_dim; ++i) {
        for (ssize_t j = -half_n_dim; j <= half_n_dim; ++j) {
            for (ssize_t k = -half_n_dim; k <= half_n_dim; ++k) {
                if (i >= -inner && i <= +inner && j >= -inner && j <= +inner && k >= -inner && k <= +inner)
                    continue;
                IJK op_omega_ijk(0, 0, 0);
                IJK op_mu_ijk(i, j, k);
                RotationOperator R = GM2L.generate_rotation_operator(OperatorID_M2L(op_omega_ijk, op_mu_ijk, -1));
                M2Lp3_Operator B_p3(Real3(0,0,0),Real3(i,j,k),p);
                //M2Lp3<SIMDengine>(*omega,R,B_p3,*mu_p3_2[ind],p);
                M2Lp3<SIMDengine>(*omega, R, B_p3, *mu_p3_2[ind], p, scratch);

                ind++;
            }
        }
    }
    pw.stop();
    pw.report(nneighbors);
    std::cout << std::endl;

    // M2L_p3 reference implementation, with precomputation
    std::cout << "M2L_p^3_reference:" << std::endl;
    pw.start();
    ind = 0;
    for (ssize_t i = -half_n_dim; i <= half_n_dim; ++i) {
        for (ssize_t j = -half_n_dim; j <= half_n_dim; ++j) {
            for (ssize_t k = -half_n_dim; k <= half_n_dim; ++k) {
                if (i >= -inner && i <= +inner && j >= -inner && j <= +inner && k >= -inner && k <= +inner)
                    continue;
                IJK op_omega_ijk(0, 0, 0);
                IJK op_mu_ijk(i, j, k);
                RotationOperator R = GM2L.generate_rotation_operator(OperatorID_M2L(op_omega_ijk, op_mu_ijk, -1));
                M2Lp3_Operator B_p3(Real3(0,0,0),Real3(i,j,k),p);
                M2Lp3_impl_reference(*omega,R,B_p3,*mu_p3_ref[ind],p);

                ind++;
            }
        }
    }
    pw.stop();
    pw.report(nneighbors);
    std::cout << std::endl;

    maxdelta_report(mu_p3_ref.begin(), mu_p3_2.begin(), nneighbors, p);

   if (test_bulk) {
    // M2L_p3, with precomputation, bulk
    std::cout << "M2L_p^3_bulk_setup:" << std::endl;

    typedef Bulk4<CoeffMatrix, RotationOperator, M2Lp3_Operator, CoeffMatrix> BulkM2LTuple;
    std::vector<BulkM2LTuple, typename Alloc::template rebind<BulkM2LTuple>::other> bulkM2Ldata;
    bulkM2Ldata.reserve(nneighbors);

    pw.start();
    ind = 0;
    for (ssize_t i = -half_n_dim; i <= half_n_dim; ++i) {
        for (ssize_t j = -half_n_dim; j <= half_n_dim; ++j) {
            for (ssize_t k = -half_n_dim; k <= half_n_dim; ++k) {
                if (i >= -inner && i <= +inner && j >= -inner && j <= +inner && k >= -inner && k <= +inner)
                    continue;
                IJK op_omega_ijk(0, 0, 0);
                IJK op_mu_ijk(i, j, k);
                RotationOperator R = GM2L.generate_rotation_operator(OperatorID_M2L(op_omega_ijk, op_mu_ijk, -1));
                std::shared_ptr<M2Lp3_Operator> B_p3 = std::allocate_shared<M2Lp3_Operator>(Alloc(),Real3(0,0,0),Real3(i,j,k),p);
                bulkM2Ldata.push_back(BulkM2LTuple(*omega,R,std::move(B_p3),*mu_p3_bulk[ind]));
                ind++;
            }
        }
    }

    assert(ind == nneighbors);
    assert(bulkM2Ldata.size() == nneighbors);
    pw.stop();
    pw.report(nneighbors);
    std::cout << "M2L_p^3_bulk_run:" << std::endl;
    pw.start();
    assert(bulkM2Ldata.size() == nneighbors);
    bulk_M2Lp3(Alloc(), bulkM2Ldata.cbegin(), bulkM2Ldata.cend(), p, scratch);
    pw.stop();
    pw.report(nneighbors);
    std::cout << std::endl;
    
    maxdelta_report(mu_p3_2.begin(), mu_p3_bulk.begin(), nneighbors, p);
   }

    // M2L_p4
    std::cout << "M2L_p^4_onthefly:" << std::endl;
    pw.start();
    ind = 0;
    for (ssize_t i = -half_n_dim; i <= half_n_dim; ++i) {
        for (ssize_t j = -half_n_dim; j <= half_n_dim; ++j) {
            for (ssize_t k = -half_n_dim; k <= half_n_dim; ++k) {
                if (i >= -inner && i <= +inner && j >= -inner && j <= +inner && k >= -inner && k <= +inner)
                    continue;
                M2L_Operator B(Real3(0,0,0),Real3(i,j,k),p);
                M2L_reference(*omega,B,*mu_p4_ref[ind],p);

                ind++;
            }
        }
    }
    pw.stop();
    pw.report(nneighbors);
    std::cout << std::endl;

    maxdelta_report(mu_p4_ref.begin(), mu_p3.begin(), nneighbors, p);

    // M2L_p4 with precomputed B operator
    std::cout << "M2L_p^4_precomputed:" << std::endl;
    pw.start();
    ind = 0;
    for (ssize_t i = -half_n_dim; i <= half_n_dim; ++i) {
        for (ssize_t j = -half_n_dim; j <= half_n_dim; ++j) {
            for (ssize_t k = -half_n_dim; k <= half_n_dim; ++k) {
                if (i >= -inner && i <= +inner && j >= -inner && j <= +inner && k >= -inner && k <= +inner)
                    continue;
                M2L_reference(*omega, *B[ind], *mu_p4_2[ind], p);
                ++ind;
            }
        }
    }
    pw.stop();
    pw.report(nneighbors);
    std::cout << std::endl;

    maxdelta_report(mu_p4_ref.begin(), mu_p4_2.begin(), nneighbors, p);
    maxdelta_report(mu_p4_ref.begin(), mu_p3_ref.begin(), nneighbors, p);

    for (size_t i = 0; i < nneighbors; ++i) {
        delete B[i];
    }
  }
}

int main(int argc, char** argv)
{
#ifdef USE_FLOAT
    typedef float Real;
#elif defined(USE_LONGDOUBLE)
    typedef long double Real;
#elif defined(USE_FLOAT128)
    typedef __float128 Real;
#else
    typedef double Real;
#endif

    std::cout << "# FMM p^3 operator test"
        << " " << buildinfo<Real>()
        << std::endl;
    std::cout << "# " << compilerinfo() << std::endl;

    size_t p_min = 50, p_max = 50;
    size_t dd_min = 1, dd_max = 1;
    size_t ws_min = 1, ws_max = 1;

    if (argc > 1)
        p_min = p_max = atoi(argv[1]);
    if (argc > 2)
        p_max = atoi(argv[2]);
    if (argc > 3)
        dd_min = dd_max = atoi(argv[3]);
    if (argc > 4)
        dd_max = atoi(argv[4]);
    if (argc > 5)
        ws_min = ws_max = atoi(argv[5]);
    if (argc > 6)
        ws_max = atoi(argv[6]);

    seed = get_env_int("SEED", default_seed());
    std::cout << "# SEED=" << seed << std::endl;

    if (int prh = get_env_int("PREHEAT", 0))
        preheat<Real>(40, 1, prh);

    for (size_t p = p_min; p <= p_max; ++p)
        for (size_t dd = dd_min; dd <= dd_max; ++dd)
            for (size_t ws = ws_min; ws <= ws_max; ++ws) {
                run_test<Real, simd_engine_none>(p, dd, ws);
            }

    return 0;
}

}  // namespace fmsolvr

int main(int argc, char **argv)
{
    return fmsolvr::main(argc, argv);
}

// vim: et:ts=4:sw=4
