#!/usr/bin/python
#
# This file is part of FMSOLVR, a Fast Multipole Solver for MD.
#
# Copyright © 2019 Andreas Beckmann
#
# FMSOLVR is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# FMSOLVR is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with FMSOLVR.  If not, see
# <https://www.gnu.org/licenses/>.
#

import sys
import math

shift = [0.0, 0.0, 0.0]

box_a = [2.0, 0.0, 0.0]
box_b = [0.0, 2.0, 0.0]
box_c = [0.0, 0.0, 2.0]

#l = 2 / math.sqrt(2)
#box_a = [l, l, 0]
#box_b = [l, -l, 0]

#l = 2 / math.sqrt(3)

#box_a = [l, l, l]
#box_b = [l, -l, -l]
#box_c = [-l, l, -l]

#box_a = [4.0, 0.0, 0.0]
#box_b = [0.0, 2.0, 0.0]
#box_c = [2.0, 0.0, 2.0]

box_a = [1.0, 0.0, 0.0]
box_b = [0.5, 1.0, 0.0]
box_c = [0.5, 0.5, 1.0]

box_a = [2.0, 0.0, 0.0]
box_b = [0.3, 2.0, 0.0]
box_c = [0.0, 0.5, 2.0]

#box_a = [1.0, 0.0, 0.0]
#box_b = [0.0, 1.0, 0.0]
#box_c = [0.0, 0.0, 1.0]

args = [float(x) for x in sys.argv[1:]]

if len(args) >= 9:
	box_a = args[0:3]
	box_b = args[3:6]
	box_c = args[6:9]
if len(args) >= 12:
	shift = args[9:12]


def vect_len(a):
	return math.sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2])

def vect_scale(a, s):
	return [a[0] * s, a[1] * s, a[2] * s]

def vect_neg(a):
	return [-a[0], -a[1], -a[2]]

def vect_add(a, b):
	return [
			a[0] + b[0],
			a[1] + b[1],
			a[2] + b[2]
			]

def vect_sub(a, b):
	return [
			a[0] - b[0],
			a[1] - b[1],
			a[2] - b[2]
			]

def scalar_prod(a, b):
	return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]

def cross_prod(a, b):
	return [
			a[1] * b[2] - a[2] * b[1],
			a[2] * b[0] - a[0] * b[2],
			a[0] * b[1] - a[1] * b[0]
			]

def spat_prod(a, b, c):
	return scalar_prod(cross_prod(a, b), c)

def box_midpoint(a, b, c):
	return vect_scale(vect_add(vect_add(a, b), c), 0.5)

def box_umkugel_diameter(a, b, c):
	return max(
			vect_len(vect_add(vect_add(a, b),c)),
			vect_len(vect_sub(vect_add(b, c),a)),
			vect_len(vect_sub(vect_add(c, a),b)),
			vect_len(vect_sub(vect_add(a, b),c))
	)

# distance of point A from the plane defined by b,c
def box_height(a, b, c):
	cross_bc = cross_prod(b, c)
	return abs(scalar_prod(a, cross_bc) / vect_len(cross_bc))

# distance of point A from the line defined by b
def side_height(a, b):
	ortho = cross_prod(a, b)
	cross = cross_prod(b, ortho)
	return abs(scalar_prod(a, cross) / vect_len(cross))

def get_ws1(h, d):
	ws = 1
	magic = 2 * math.sqrt(3) / 4
	while d / ((ws + 1) * h) > magic:
		ws += 1
	return ws

def get_ws(lens, d):
	return [get_ws1(l, d) for l in lens]

def signum(x):
	return (x > 0) - (x < 0)


# is point p inside the parallelepiped(a, b, c) with a corner at origin?
# returns +1 if inside; 0 if at the border, -1 if outside
def inside_parallelepiped(origin, a, b, c, p):
	A = origin
	W = vect_add(A, vect_add(a, vect_add(b, c)))
	cross_ab = cross_prod(a, b)
	cross_bc = cross_prod(b, c)
	cross_ca = cross_prod(c, a)
	len_cross_ab = vect_len(cross_ab)
	len_cross_bc = vect_len(cross_bc)
	len_cross_ca = vect_len(cross_ca)
	AP = vect_sub(p, A)
	WP = vect_sub(p, W)
	hAab = scalar_prod(AP, cross_ab) / len_cross_ab
	hAbc = scalar_prod(AP, cross_bc) / len_cross_bc
	hAca = scalar_prod(AP, cross_ca) / len_cross_ca
	hWab = scalar_prod(WP, cross_ab) / len_cross_ab
	hWbc = scalar_prod(WP, cross_bc) / len_cross_bc
	hWca = scalar_prod(WP, cross_ca) / len_cross_ca
	ret = - signum(max(hAab * hWab, hAbc * hWbc, hAca * hWca))
	#print p, hAab, hAbc, hAca, hWab, hWbc, hWca, ret
	return ret

# ExpandedCoordinateToAlternatingSign
def ectas(x):
	return 1 - 2 * (((x - 1) / 2) & 1)

def enumerate_nacl(radius, shift, a, b, c):
	print "# NaCl shifted by ", shift
	fmt = "% d\t% .16g\t% .16g\t% .16g"
	totalcharge = 0
	particles = 0
	Onacl = [0., 0., 0.]
	Mnacl = vect_add(box_midpoint(a, b, c), shift)
	# two opposite corners of the bounding box of the sphere with given radius around
	# box with origin shift and midpoint M
	lower_corner = vect_sub(Mnacl, [radius, radius, radius])
	upper_corner = vect_add(Mnacl, [radius, radius, radius])
	imin = 4 * (int(math.floor(lower_corner[0])) - 1) - 1
	imax = 4 * (int(math.ceil(upper_corner[0])) + 1) + 1
	jmin = 4 * (int(math.floor(lower_corner[1])) - 1) - 1
	jmax = 4 * (int(math.ceil(upper_corner[1])) + 1) + 1
	kmin = 4 * (int(math.floor(lower_corner[2])) - 1) - 1
	kmax = 4 * (int(math.ceil(upper_corner[2])) + 1) + 1
	for i in range(imin, imax + 1, 2):
		for j in range(jmin, jmax + 1, 2):
			for k in range(kmin, kmax + 1, 2):
				charge = ectas(i) * ectas(j) * ectas(k)
				xyz_nacl = [i / 4., j / 4., k / 4.]
				inside = inside_parallelepiped(shift, a, b, c, xyz_nacl)
				xyz = vect_sub(xyz_nacl, Mnacl)
				if inside > 0:
					totalcharge += charge
					particles += 1
					print fmt % tuple([charge] + xyz)
				elif inside == 0:
					print "#" + fmt % tuple([charge] + xyz)
	# 2 * sqrt(3) probe point
	probe = [1./6., 1./6., 1./6.]
	print fmt % tuple([0] + vect_sub(probe, Mnacl))
	print "# particles: %d, total charge: %g" % (particles, totalcharge)


def cuboid_stats(a, b, c, lens):
	x = [a, 0, 0]
	y = [0, b, 0]
	z = [0, 0, c]
	v = abs(spat_prod(x, y, z))
	d = box_umkugel_diameter(x, y, z)
	ws_abc = get_ws(lens, d)
	print '# cuboid: ', [a, b, c], " v=%g d=%g ws=%g " % (v, d, max(ws_abc)), ws_abc

def box_stats(a, b, c):
	print "# box spanner: ", a, b, c
	print "# box volume: ", abs(spat_prod(a, b, c))
	d = box_umkugel_diameter(a, b, c)
	print "# umkugel diam: ", d
	print "# midpoint: ", box_midpoint(a,b,c)
	h_a = box_height(a, b, c)
	h_b = box_height(b, a, c)
	h_c = box_height(c, a, b)
	lens = [vect_len(a), vect_len(b), vect_len(c)]
	print "# heights: ", [h_a, h_b, h_c]
	ws_abc = get_ws(lens, d)
	print "# ws: ", ws_abc
	h_ab = side_height(a, b)
	h_ac = side_height(a, c)
	h_ba = side_height(b, a)
	h_bc = side_height(b, c)
	h_ca = side_height(c, a)
	h_cb = side_height(c, b)
	cuboid_stats(vect_len(a), h_ba, h_c, lens)
	cuboid_stats(vect_len(a), h_ca, h_b, lens)
	cuboid_stats(vect_len(b), h_ab, h_c, lens)
	cuboid_stats(vect_len(b), h_cb, h_a, lens)
	cuboid_stats(vect_len(c), h_ac, h_b, lens)
	cuboid_stats(vect_len(c), h_bc, h_a, lens)
	enumerate_nacl(d/2., shift, box_a, box_b, box_c)
	print '# export UNITBOX="%s %d"' % (" ".join(["%.16g" % x for x in a + b + c]), max(ws_abc))

def test_inside(a, b, c):
	inside_spat(a, b, c, [0, 0, 0])
	inside_spat(a, b, c, [0.1, 0.2, 0.3])
	inside_spat(a, b, c, [1, 1, 1])
	inside_spat(a, b, c, [3, 0, 0])
	inside_spat(a, b, c, [0, 3, 0])
	inside_spat(a, b, c, [0, 0, 3])
	inside_spat(a, b, c, [5, 5, 5])
	inside_spat(a, b, c, [-1, -1, -1])
	inside_spat(a, b, c, [-3, -0, -0])
	inside_spat(a, b, c, [-0, -3, -0])
	inside_spat(a, b, c, [-0, -0, -3])
	inside_spat(a, b, c, [-5, -5, -5])


box_stats(box_a, box_b, box_c)
