/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */
#ifndef _AnBe_FMM_SIMD_AVX_complex_double2_h_
#define _AnBe_FMM_SIMD_AVX_complex_double2_h_

#include <complex>
#include <immintrin.h>

class avx_complex_double2
{
    typedef __m256d vdouble4;
    vdouble4 vval;

    avx_complex_double2(const vdouble4 & a) : vval(a)
    { }

public:
    avx_complex_double2() = default;

    avx_complex_double2(double a, double b, double c, double d) : vval{a, b, c, d}
    { }

    avx_complex_double2(std::complex<double> ab, std::complex<double> cd) : vval{ab.real(), ab.imag(), cd.real(), cd.imag()}
    { }

    avx_complex_double2 operator + (const avx_complex_double2 & a) const
    {
        return vval + a.vval;
    }

    avx_complex_double2 operator - (const avx_complex_double2 & a) const
    {
        return vval - a.vval;
    }

    avx_complex_double2 operator * (const avx_complex_double2 & b) const
    {
        __m256d a3210 = vval;                           // im(a1),re(a1),im(a0),re(a0)
        __m256d b3210 = b.vval;                         // im(b1),re(b1),im(b0),re(b0)
        __m256d a2301 = _mm256_permute_pd(a3210, 0x5);  // swap real + imag
        __m256d b2200 = _mm256_permute_pd(b3210, 0x0);  // real parts
        __m256d b3311 = _mm256_permute_pd(b3210, 0xf);  // imag parts
        __m256d a3210b2200 = a3210 * b2200;             // im(a1)*re(b1),re(a1)*re(b1),im(a0)*re(b0),re(a0)*re(b0)
        __m256d a2301b3311 = a2301 * b3311;             // re(a1)*im(b1),im(a1)*im(b1),re(a0)*im(b0),im(a0)*im(b0)
        return _mm256_addsub_pd(a3210b2200, a2301b3311);
        // im(a1)*re(b1)+re(a1)*im(b1), re(a1)*re(b1)-im(a1)*im(b1),
        // im(a0)*re(b0)+re(a0)*im(b0), re(a0)*re(b0)-im(a0)*im(b0)
    }

    avx_complex_double2 operator - () const
    {
        return -vval;
    }

    avx_complex_double2 & operator += (const avx_complex_double2 & a)
    {
        vval += a.vval;
        return *this;
    }

    avx_complex_double2 & operator -= (const avx_complex_double2 & a)
    {
        vval -= a.vval;
        return *this;
    }

    avx_complex_double2 & acc_mul(const avx_complex_double2 & a, const avx_complex_double2 & b)
    {
        avx_complex_double2 c = a * b;
        *this += c;
        return *this;
    }

    __m256d get() const
    {
        return vval;
    }
};

#endif
// vim: et:ts=4:sw=4
