/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2014 Simon Pfreundschuh
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_OPERATOR_ROTATION_rotation_generator_hpp
#define FMSOLVR_OPERATOR_ROTATION_rotation_generator_hpp

#include "fmsolvr/math/angle.hpp"
#include "fmsolvr/math/ipowlog.hpp"
#include "fmsolvr/operator/rotation/eikphitable.hpp"
#include "fmsolvr/operator/rotation/rotation_constants.hpp"
#include "fmsolvr/operator/rotation/rotation_operator.hpp"

#include <vector>
#include <cassert>
#include <cmath>
#include <sys/types.h>

namespace fmsolvr
{

// RotationGenerators provide an abstraction layer for the computation
// of RotationOperators. They exploit structure of the M2M, M2L and L2L
// operators, to allow for reuse of RotationOperators. A RotationGenerator
// provides a member function generate_rotation that returns a suitable RotationOperator
// that can be used for both the hinwards and backwards rotation of the expansion.

namespace rotation_generator_utility
{

template <typename Real, template <typename> class Real3>
Angle<Real> compute_phi(Real3<Real> d)
{
    Real cos_phi, sin_phi;
    Real length_x_y = std::sqrt(d.x() * d.x() + d.y() * d.y());

    if (length_x_y > 0)
    {
        cos_phi = d.x() / length_x_y;
        sin_phi = d.y() / length_x_y;
    }
    else
    {
        cos_phi = 1;
        sin_phi = 0;
    }

    Angle<Real> phi(cos_phi, sin_phi);
    return phi;
}

template <typename Real, template <typename> class Real3>
Angle<Real> compute_theta(Real3<Real> d)
{
    Real cos_theta, sin_theta;
    Real length_x_y = std::sqrt(d.x() * d.x() + d.y() * d.y());
    Real length     = std::sqrt(d.x() * d.x() + d.y() * d.y() + d.z() * d.z());
    if (length > 0)
    {
        cos_theta = d.z() / length;
        sin_theta = length_x_y / length;
    }
    else
    {
        cos_theta = 1;
        sin_theta = 0;
    }

    Angle<Real> theta(cos_theta, sin_theta);
    return theta;
}
}

namespace RG
{

// RotationGenerator for rotation-based M2M and L2L operators. For a given
// relative depth dd, RotationGeneratorX2X stores the RotationMatrixVectors
// for each child after their first computation and thus allows for cheap
// generation of RotationGenerators.
// The indexing of the children is given by three integer coordinates (i,j,k).
// The relative position of the child in the parent coordinate system is given
// by (c * (i + 0.5 - 2^{d-1}),(c * (j + 0.5 - 2^{d-1}),(c * (k + 0.5 - 2^{d-1}))
// where c = 2^{-d}.

//TODO: Exploit reoccurences of rotational angles to reduce memory footprint..
template <typename RotationMatrixVector,
          typename Real3,
          typename Real33,
          typename Allocator = std::allocator<RotationMatrixVector> >
class RotationGeneratorX2X
{
    RotationGeneratorX2X(const RotationGeneratorX2X &)            = delete;
    RotationGeneratorX2X operator= (const RotationGeneratorX2X &) = delete;

    public:
        typedef typename Real3::value_type real_type;
private:
    typedef typename Allocator::template rebind<real_type>::other real_allocator_type;
    typedef fmsolvr::EikPhiTable<real_type, real_allocator_type> EikPhiTable;
    typedef RotationConstants<real_type, real_allocator_type> rotation_constants;
public:
    typedef RotationOperator<RotationMatrixVector, EikPhiTable, rotation_constants> rotation_operator;

        // RotationGeneratorX2X with uncomputed RotationMatrixVectors, allocation and computation
        // of the RotationMatrixVectors happens at the first request.
        RotationGeneratorX2X(
            ssize_t       p,
            ssize_t       ddmax,
            const Real33 &basis) : p_(p), ddmax_(ddmax), basis_(basis), rot_constants_(p)
        {
            array_.resize(size());
            eikphitable_array_.resize(size());
        }

        ssize_t p() const
        {
            return p_;
        }

    template <typename IJK>
    rotation_operator generate_rotation_operator(const IJK &ijk_c, size_t dd)
    {
        return generate_rotation_operator(ijk_c.i(), ijk_c.j(), ijk_c.k(), dd);
    }

private:
        rotation_operator generate_rotation_operator(
            ssize_t i,
            ssize_t j,
            ssize_t k,
            ssize_t dd)
        {
            using namespace rotation_generator_utility;

            real_type              half_dim = 1 << (dd - 1);
            Real3                  origin(half_dim, half_dim, half_dim);
            Real3                  child(real_type(i)+0.5, real_type(j)+0.5, real_type(k)+0.5);
            Real3                  d   = basis_ * (child-origin);

            auto & R = array_[offset(i, j, k, dd)];
            if (!R)
                R = std::allocate_shared<RotationMatrixVector>(Allocator(), p_);
            if (!R->is_computed())
            {
                Angle<real_type> theta = compute_theta(d);
                R->compute(theta, rot_constants_);
            }
            auto & Rphi = eikphitable_array_[offset(i, j, k, dd)];
            if (!Rphi) {
                Angle<real_type>       phi = compute_phi(d);
                Rphi = std::allocate_shared<EikPhiTable>(Allocator(), phi, p_);
            }
            return rotation_operator(*R, *Rphi, rot_constants_, offset(i, j, k, dd));
        }

    private:
        size_t                  p_;
        size_t                  ddmax_;
        const Real33            basis_;
        std::vector<std::shared_ptr<RotationMatrixVector> > array_;
        std::vector<std::shared_ptr<EikPhiTable> > eikphitable_array_;
        rotation_constants rot_constants_;

        size_t size() const
        {
            return 1 << (ddmax_ * 3);
        }

        size_t offset(
            size_t i,
            size_t j,
            size_t k,
            size_t dd)
        {
            // dd has to be smaller than ddmax_
            assert(dd <= ddmax_);

            size_t row_size = 1 << ddmax_;
            size_t dind     = (1 << (ddmax_-1)) - (1 << (dd -1));
            return ((k+dind)*row_size+(j+dind))*row_size+i+dind;
        }
};

// RotationGenerator for the M2L operator. Stores RotationMatrixVectors
// for all interacting boxes after their first generation. The
// interacting boxes are indexed by the triple (i,j,k) where is the shift
// vector in the normalized box-coordinate system.

//TODO: Exploit reoccurance of angles to reduce memory footprint.
template <typename RotationMatrixVector,
          typename Real3,
          typename Real33,
          typename Allocator = std::allocator<RotationMatrixVector> >
class RotationGeneratorM2L
{
    RotationGeneratorM2L(const RotationGeneratorM2L &)            = delete;
    RotationGeneratorM2L operator= (const RotationGeneratorM2L &) = delete;

    public:

        typedef typename Real3::value_type real_type;
private:
    typedef typename Allocator::template rebind<real_type>::other real_allocator_type;
    typedef fmsolvr::EikPhiTable<real_type, real_allocator_type> EikPhiTable;
    typedef RotationConstants<real_type, real_allocator_type> rotation_constants;
public:
    typedef RotationOperator<RotationMatrixVector, EikPhiTable, rotation_constants> rotation_operator;

        RotationGeneratorM2L(
            ssize_t      p,
            ssize_t      ws,
            const Real33 basis) : p_(p), ws_(ws), basis_(basis), rot_constants_(p)
        {
            array_.resize(size());
            eikphitable_array_.resize(size());
        }

        ssize_t p() const
        {
            return p_;
        }

    template <typename OpID>
    rotation_operator generate_rotation_operator(const OpID &opID)
    {
        auto ijk = opID.relative_to_source();
        return generate_rotation_operator(ijk.i(), ijk.j(), ijk.k());
    }

private:
        rotation_operator generate_rotation_operator(
            ssize_t i,
            ssize_t j,
            ssize_t k)
        {
            //std::cout << "gro(" << i << "," << j << "," << k << ") ofs=" << offset(i, j, k) << "/" << size() << std::endl;
            using namespace rotation_generator_utility;

            Real3                  d(basis_*Real3(-i, -j, -k));

            assert(offset(i, j, k) < size());
            auto & R = array_[offset(i, j, k)];
            if (!R)
                R = std::allocate_shared<RotationMatrixVector>(Allocator(), p_);
            if (!R->is_computed())
            {
                Angle<real_type> theta = compute_theta(d);
                R->compute(theta, rot_constants_);
            }
            auto & Rphi = eikphitable_array_[offset(i, j, k)];
            if (!Rphi)
            {
                Angle<real_type> phi = compute_phi(d);
                Rphi = std::allocate_shared<EikPhiTable>(Allocator(), phi, p_);
            }
            return rotation_operator(*R, *Rphi, rot_constants_, offset(i, j, k));
        }

    private:
        size_t                  p_;
        size_t                  ws_;
        const Real33            basis_;
        std::vector<std::shared_ptr<RotationMatrixVector> > array_;
        std::vector<std::shared_ptr<EikPhiTable> > eikphitable_array_;
        rotation_constants rot_constants_;

        size_t size() const
        {
            return icubed(4 * ws_ + 3);
        }

        size_t offset(
            ssize_t i,
            ssize_t j,
            ssize_t k)
        {
            size_t dim = 4*ws_+3;
            size_t i_normalized(i + (dim-1) / 2);
            size_t j_normalized(j + (dim-1) / 2);
            size_t k_normalized(k + (dim-1) / 2);
            return (k_normalized * dim + j_normalized) * dim + i_normalized;
        }
};

// RotationGenerator for arbitrary shift vectors. RotationMatrixVectors
// are computed on the fly, thus RotationOperator genereation is expensive.
template <typename RotationMatrixVector,
          typename Real3,
          typename Allocator = std::allocator<RotationMatrixVector> >
class RotationGeneratorNoPrecomp
{
    RotationGeneratorNoPrecomp(const RotationGeneratorNoPrecomp &)            = delete;
    RotationGeneratorNoPrecomp operator= (const RotationGeneratorNoPrecomp &) = delete;

    public:

        typedef typename RotationMatrixVector::size_type size_type;
        typedef typename RotationMatrixVector::index_type index_type;
        typedef typename RotationMatrixVector::value_type real_type;

private:
    typedef typename Allocator::template rebind<real_type>::other real_allocator_type;
    typedef fmsolvr::EikPhiTable<real_type, real_allocator_type> EikPhiTable;
    typedef RotationConstants<real_type, real_allocator_type> rotation_constants;

    typedef std::shared_ptr<RotationMatrixVector> shared_rmv;
    typedef std::shared_ptr<EikPhiTable> shared_EikPhiTable;
    typedef std::shared_ptr<rotation_constants> shared_rc;

    struct wrap_shared {
        shared_rmv rmv_;
        shared_EikPhiTable eikphitable_;
        shared_rc rc_;

        wrap_shared(shared_rmv rmv, shared_EikPhiTable eikphitable, shared_rc rc)
            : rmv_(rmv), eikphitable_(eikphitable), rc_(rc)
        { }
    };

public:
        typedef RotationOperator<RotationMatrixVector, EikPhiTable, rotation_constants, wrap_shared> rotation_operator;

        RotationGeneratorNoPrecomp( ssize_t p ) : p_(p) {};

        // Generate RotationOperator that aligns the given shift vector with
        // the z-axis of the reference frame.
        // The  shift vector for the different operators is defined as follows:
        // M2M: d = source - target
        // M2L: d = source - target
        // L2L: d = target - source
        rotation_operator generate_rotation(Real3 d)
        {

            using namespace rotation_generator_utility;

            Angle<real_type>      phi   = compute_phi(d);
            Angle<real_type>      theta = compute_theta(d);
            auto rc = std::allocate_shared<rotation_constants>(Allocator(), p_);
            auto R = std::allocate_shared<RotationMatrixVector>(Allocator(), p_, theta, *rc);
            auto Rphi = std::allocate_shared<EikPhiTable>(Allocator(), phi, p_);
            return rotation_operator(*R, *Rphi, *rc, -1, wrap_shared(R, Rphi, rc));
        }

    private:
        ssize_t p_;
};

}  // namespace RG

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
