/*
 * This file is part of FMSOLVR, a Fast Multipole Solver for MD.
 *
 * Copyright © 2019 Andreas Beckmann
 *
 * FMSOLVR is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * FMSOLVR is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FMSOLVR.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#ifndef FMSOLVR_SEQUENTIAL_pass3_hpp
#define FMSOLVR_SEQUENTIAL_pass3_hpp

#include "fmsolvr/ijk.hpp"
#include "fmsolvr/operator/p3/local2local_p3.hpp"
#include "fmsolvr/operator/p4/local2local.hpp"
#include "fmsolvr/util/envtools.hpp"


// Note: In contrast to P2M, P2L, M2M, and M2L that all use (source - target),
// for L2L it is required to use (target - source) instead. Really.

// The forward rotation of an M2M source (child) and the forward rotation of
// an L2L source (parent) are identical. The L2L rotation thereby seems off
// by 180°, but is intrinsically compensated by the shift vector pointing
// into opposite directions: M2M (source-target), L2L (target-source).
// The p3 shift operator consists only of the distance (non-negative,
// undirected), not the (target-source) vector itself. Hence the p4 vector
// property (different directions for M2M and L2L) is not visible and is
// handled via the seemingly wrong rotation.
// The backward rotation is handled accordingly. It is always defined
// (M2M, M2L, L2L) to "undo" the corresponding forward rotation.

namespace fmsolvr {

//Same as PASS 1:  just start at the root of the tree. We can also exploit more parallelism at the levels nearer the root
//by doing some redundant computations
// this routine processes all tree levels
template <typename FMMHandle>
size_t pass3_L2L_tree(FMMHandle & fmm_handle)
{
    const bool p4 = get_env_int("P4", 0);
    typedef typename FMMHandle::real3_type Real3;
    typedef typename FMMHandle::size_type size_type;
    typedef typename FMMHandle::boxid_type BoxID;
    typedef typename FMMHandle::coefficient_type coefficient_type;
    typedef typename FMMHandle::RotationDataTypes::Rot_Operator RotationOperator;
    typedef typename FMMHandle::RotationDataTypes::L2L_Operator L2Lp3_Operator;
    typedef typename FMMHandle::RotationDataTypes::template ScratchTypeGenerator<coefficient_type>::Scratch scratch_type;

    const auto &simulationbox = fmm_handle.simulationbox();
    const auto &tree_omega = fmm_handle.tree_omega();
    auto & tree_mu = fmm_handle.tree_mu();
    auto &G = fmm_handle.RotationGeneratorL2L();
    const size_type depth = fmm_handle.depth();
    const size_type p = fmm_handle.p();
    scratch_type scratch(p);

    // parent = source
    // child = target
    size_t l2lcount = 0;
    tree_mu[BoxID(0, 0, 0, 0)].populate_lower();
    for (size_t d_c = 1; d_c <= depth; ++d_c) {
        size_t d_p = d_c - 1;
        size_t dim_p = size_t(1) << d_p;
        size_t d_delta = d_c - d_p;
        size_t dim_op = size_t(1) << d_delta;
        const auto & l2l_operator = fmm_handle.l2l_operator_cube(d_c, d_delta);

        for (size_t i = 0; i < dim_p; ++i) {
            for (size_t j = 0; j < dim_p; ++j) {
                for (size_t k = 0; k < dim_p; ++k) {
                    BoxID bid_p(i, j, k, d_p);
                    Real3 expn_point_p = simulationbox.expansion_point(bid_p);
                    const auto & mu_p = tree_mu[bid_p];
                    for (size_t ii = 0; ii < dim_op; ++ii) {
                        for (size_t jj = 0; jj < dim_op; ++jj) {
                            for (size_t kk = 0; kk < dim_op; ++kk) {
                                BoxID bid_c = bid_p.descendant(ii, jj, kk, d_delta);
                                // no particles in the target box?
                                // (this HACK only works if source tree and target tree are
                                // identical)
                                if (tree_omega[bid_c].empty())
                                    continue;
                                IJK<size_t> ijk_c(ii, jj, kk);
                                Real3 expn_point_c = simulationbox.expansion_point(bid_c);
                                auto & mu_c = tree_mu[bid_c];
                                //L2L operator similar to M2M operator -> exploit parallelism to achieve good memory access pattern (see above)
                                if (p4) {
                                    L2L(mu_p, l2l_operator(ii, jj, kk), mu_c, p);
                                } else {
                                    RotationOperator R =
                                            G.generate_rotation_operator(ijk_c, d_delta);
                                    L2Lp3_Operator C_p3(expn_point_p, expn_point_c, p);
                                    L2Lp3<FMMHandle::template SIMDengine>(mu_p, R, C_p3, mu_c, p, scratch);
                                }
                                ++l2lcount;
                                mu_c.addSourceCounters(mu_p.sourceCounters());
                                mu_c.populate_lower();
                            }
                        }
                    }
                }
            }
        }
    }
    return l2lcount;
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
