#ifndef _AnBe_FMM_pass1_hpp_
#define _AnBe_FMM_pass1_hpp_

#include "fmsolvr/ijk.hpp"
#include "fmsolvr/operator/p3/multipole2multipole_p3.hpp"
#include "fmsolvr/operator/p4/multipole2multipole.hpp"
#include "fmsolvr/operator/particle2multipole.hpp"
#include "fmsolvr/util/envtools.hpp"

#include <cassert>

namespace fmsolvr {

// construct multipoles on lowest level
// this routine processes all particles
template <typename FMMHandle, typename TaskFactory, typename DependencyManager>
void pass1_P2M(
    FMMHandle &fmm_handle, typename FMMHandle::tree_partition_type tp, TaskFactory &task_factory, DependencyManager &dm)
{
    typedef typename FMMHandle::real3_type Real3;
    typedef typename FMMHandle::real_type Real;
    typedef typename FMMHandle::size_type size_type;

    const auto & particle_handle = fmm_handle.particle_handle();
    auto & tree_omega = fmm_handle.tree_omega();
    const size_type p = fmm_handle.p();
    const auto & div_lut_p = fmm_handle.div_lut_p(p);
    const auto & ordered_particles = particle_handle.ordered_particles();
    const auto & box_particle_offset = particle_handle.box_particle_offset();

    //loop over all boxes on the lowes level (neighbourships do not matter) num_boxes_lowest = dim*dim*dim
    //each box can be handled independtly (1 thread block per box? -> see below)
    for (auto bid : tp.box_range()) {
        size_t id = bid.sfcindex();
        // CoeffMatrix is a triangular array longest line has size 2*p need for every multipole
        Real3 expn_point = particle_handle.simulationbox().expansion_point(bid); //middle point of current box
        //all particles in the box can be handled independtly but this requires atomic writes to the CoeffMatrix
        //maybe one warp per box? This would avoid memory coalescing issues
        //also consider putting CoeffMatrix into shared memory for float p <= 30 => upper lower storage requires 30**2**8 =~ 7kb shared memory (upper bound)
        auto & omega = tree_omega[bid];
        omega.zero(p);
        omega.clearSourceCounters();
        for (size_t i = box_particle_offset[id]; i < box_particle_offset[id + 1]; ++i) {
            Real3 xyz = Real3(ordered_particles[i]);
            Real q = ordered_particles[i].s();
            //this writes to all elements of CoeffMatrix (p**2 operation)
            P2M(q, xyz, omega, expn_point, div_lut_p);
            omega.incSourceCounters();
            //within P2M some more parallelism could be exploited.
        }
        omega.populate_lower();
        dm.ResolveFinalOmega(bid, task_factory);
    }
}

template <typename FMMHandle, typename TaskFactory, typename DependencyManager>
void pass1_M2M(
        FMMHandle &fmm_handle,
        size_t id,
        size_t d,
        TaskFactory &task_factory,
        DependencyManager &dm)
{
    const bool p4 = get_env_int("P4", 0);
    typedef typename FMMHandle::real3_type Real3;
    typedef typename FMMHandle::size_type size_type;
    typedef typename FMMHandle::boxid_type BoxID;
    typedef typename FMMHandle::coefficient_type coefficient_type;
    typedef typename FMMHandle::RotationDataTypes::Rot_Operator RotationOperator;
    typedef typename FMMHandle::RotationDataTypes::M2M_Operator M2Mp3_Operator;
    typedef typename FMMHandle::RotationDataTypes::template ScratchTypeGenerator<coefficient_type>::Scratch scratch_type;

    const auto &simulationbox = fmm_handle.simulationbox();
    auto & tree_omega = fmm_handle.tree_omega();
    auto &G = fmm_handle.RotationGeneratorM2M();
    const size_type p = fmm_handle.p();
    scratch_type scratch(p);

    {
        size_t d_p = d;
        size_t d_c = d + 1;

        size_t d_delta = d_c - d_p;
        size_t dim_op = size_t(1) << d_delta;
        const auto & m2m_operator = fmm_handle.m2m_operator_cube(d_c, d_delta);

        {
            {
                {
                    BoxID bid_p(id, d);
                    Real3 expn_point_p = simulationbox.expansion_point(bid_p);
                    auto & omega_p = tree_omega[bid_p];
                    omega_p.zero(p);
                    omega_p.clearSourceCounters();
                    for (size_t ii = 0; ii < dim_op; ++ii) {
                        for (size_t jj = 0; jj < dim_op; ++jj) {
                            for (size_t kk = 0; kk < dim_op; ++kk) {
                                BoxID bid_c = bid_p.descendant(ii, jj, kk, d_delta);
                                IJK<size_t> ijk_c(ii, jj, kk);
                                Real3 expn_point_c = simulationbox.expansion_point(bid_c);
                                const auto & omega_c = tree_omega[bid_c];
                                if (omega_c.empty())
                                    continue;
                                //8 lower level CoeffMatrices are combined to a new CoeffMatrix similar to a series of complex matrix vector multiplications
                                //contains parallelism we should exploit it to get good memory access patterns
                                if (p4) {
                                    M2M(omega_c, m2m_operator(ii, jj, kk), omega_p, p);
                                } else {
                                    RotationOperator R =
                                            G.generate_rotation_operator(ijk_c, d_delta);
                                    M2Mp3_Operator A_p3(expn_point_c, expn_point_p, p);
                                    M2Mp3<FMMHandle::template SIMDengine>(omega_c, R, A_p3, omega_p, p, scratch);
                                }
                                omega_p.addSourceCounters(omega_c.sourceCounters());
                            }
                        }
                    }
                    omega_p.populate_lower();
                    dm.ResolveFinalOmega(bid_p, task_factory);
                }
            }
        }
    }
}

}  // namespace fmsolvr

#endif
// vim: et:ts=4:sw=4
